----------------------------------------------------------------------
-- Module Auxiliary.hs                                              --
--                                                                  --
-- This module contains various auxiliary functions for different   --
-- purposes.							    --
--                                                                  --
-- 12.11.96  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Auxiliary (
  -- functions on formulae
  depth, size,     		-- depth and size of a formula
  newVar, newVar',   		-- generate next variable
  newCon,			-- generate next constant
  subst, substLit,      	-- variable substitution functions
  exchange, exchangeLit,	-- variable exchange functions
  fdelete, fdelete2,		-- variable deletion functions
  getPredicates,		-- get formula's predicates
  -- tuple selectors
  fst3, snd3, thd3,
  fst5, snd5, thrd5, frth5, fifth5,
  fst6, snd6, thrd6, frth6, fifth6, sixth6,
  -- misc
  fix,       			-- fix-point operator
  takeOneOfEach,		-- build combinations
  insertNew,			-- insert into list, if not present
  quickSort,			-- quick sort
  quickSortBy,			-- quick sort with spec. ordering
  newLine			-- for IO monad
) where


-------------------
----- imports -----
-------------------

import Formula
import List ((\\), nub)
import Set


------------------------
-- exported functions --
------------------------

-- "y = fix f x" computes the fixpoint of function f with initial
-- value x if it exists. Otherwise the function does not terminate.

fix :: Eq a => (a -> a) -> a -> a
fix f x = let y = f x in if x == y then x else fix f y


-- compute the depth of a formula

depth :: Formula -> Int
depth f
  | isVerum f || isFalsum f || isLiteral f = 1
  | otherwise = maximum (map depth (getSubFormulae f)) + 1


-- compute the size of a formula

size :: Formula -> Int
size f
  | isVerum f || isFalsum f || isLiteral f = 1
  | otherwise = sum (map size (getSubFormulae f)) + 1


-- "mv = newVar m f" generates a new variable depending on a formula.
-- m is an integer denoting the highest admissible variable index, f is
-- a formula. mv is the first admissible variable not free in f,
-- if it exists, Nothing otherwise.

newVar :: Int -> Formula -> Maybe Variable
newVar maxv f = if null nvs then Nothing else Just (head nvs)
  where nvs = [mkVar 0..mkVar maxv] \\ setToOrdList (vars f)


-- "mv = newVar' m vs" generates a new variable depending on a variable set.
-- m is an integer denoting the highest admissible variable index, vs
-- is a set of variables. mv is the first admissible variable not in vs,
-- if it exists, Nothing otherwise.

newVar' :: Int -> Set Variable -> Maybe Variable
newVar' maxv vs = if null nvs then Nothing else Just (head nvs)
  where nvs = [mkVar 0..mkVar maxv] \\ setToOrdList vs


-- "c = newCon f" generates a new constant depending on a formula f.
-- The new constant c is the constant with the lowest index not occurring
-- in f.

newCon :: Formula -> Constant
newCon f = head ([mkCon 0..] \\ setToOrdList (consts f))


-- "l' = substLit v t l" substitutes variable v by term t in literal l,
-- thus resulting in literal l'.

substLit :: Variable -> Term -> Literal -> Literal
substLit x t l = mkLit p (map sub ts)
  where p = getLitPred l; ts = getLitTerms l
	sub a | termIsVar a = if getTermVar a == x then t else a
              | otherwise   = a


-- "l' = exchangeLit v v' l" replaces any occurrence of variable v
-- by variable v' in literal l, and vice versa. The resulting literal is l'.

exchangeLit :: Variable -> Variable -> Literal -> Literal
exchangeLit x y l = mkLit p (map exc ts)
  where p = getLitPred l; ts = getLitTerms l
        exc a | termIsVar a = if getTermVar a == x then termFromVar y
                              else if getTermVar a == y then termFromVar x
                              else a
              | otherwise   = a


-- "f' = subst v t f" substitutes variable v by term t in formula f,
-- resulting in formula f'.

subst :: Variable -> Term -> Formula -> Formula
subst x t f
  | isPosLit f  = literal (substLit x t (getLiteral f))
  | isExists f = if qv == x then f else 
      if termIsCon t then f'
      else -- termIsVar
        if qv == getTermVar t && x `isElement` free f then
          exists x (exchange x qv sf) else f'
  | isForAll f = if qv == x then f else
      if termIsCon t then f'
      else -- termIsVar
        if qv == getTermVar t && x `isElement` free f then
          forall x (exchange x qv sf) else f'
  | otherwise = f'
  where qv = getQuantVar f
        sf = getSubFormula f
        f' = setSubFormulae (map (subst x t) (getSubFormulae f)) f


-- "f' = exchange v v' f" replaces any occurrence of variable v
-- by variable v' in formula f, and vice versa. The resulting formula
-- is f'.

exchange :: Variable -> Variable -> Formula -> Formula
exchange x y f
  | isPosLit f = literal (exchangeLit x y (getLiteral f))
  | isExists f = setQuantVar qv' f'
  | isForAll f = setQuantVar qv' f'
  | otherwise = f'
  where qv  = getQuantVar f
        qv' = if qv == x then y else if qv == y then x else qv
        sf  = getSubFormula f
        f'  = setSubFormulae (map (exchange x y) (getSubFormulae f)) f


-- "f' = fdelete v f" removes any free occurrence of variable v in formula f
-- by replacing the corresponding literal by falsum.

fdelete :: Variable -> Formula -> Formula
fdelete v f
  | isLiteral f = if v `isElement` free f then falsum else f
  | isExists f = if getQuantVar f == v then f else
      setSubFormula (fdelete v (getSubFormula f)) f
  | isForAll f = if getQuantVar f == v then f else
      setSubFormula (fdelete v (getSubFormula f)) f
  | otherwise = setSubFormulae (map (fdelete v) (getSubFormulae f)) f


-- "f' = fdelete2 x y f" removes any free occurrence of variable y from formula f
-- by replacing the corresponding literal by falsum. This replacement is only
-- performed if x /= y, otherwise f' = f.

fdelete2 :: Variable -> Variable -> Formula -> Formula
fdelete2 x y f = if x == y then f else fdelete y f


-- "l' = takeOneOfEach l" builds all combinations of elements of list l,
-- i.e. if l = [[a_{1,1},...a_{1,n_1}],...,[a_{m,1},...,a_{m,n_m}]], then
-- the list l' contains as elements all lists of the form
-- [a_{1,i_1},...,a_{m,i_m}] where 1 <= i_m <= n_m.

takeOneOfEach :: [[a]] -> [[a]]
takeOneOfEach [] = []
takeOneOfEach (x:[]) = [[c] | c <- x]
takeOneOfEach (x:xs) = [(i:r) | i <- x, r <- takeOneOfEach xs]


-- "l' = insertNew a l" adds a in front of l, if a is not already contained
-- in l, otherwise l' = l.

insertNew :: Eq a => a -> [a] -> [a]
insertNew a l = if a `elem` l then l else a:l

-- "l' = quickSort l" sorts list l in non-descending order using the
-- quicksort algorithm. The standard ordering on the elements is used.

quickSort :: Ord a => [a] -> [a]
quickSort [] = []
quickSort (x:xs) = quickSort [u | u <- xs, u < x] ++ [x] ++
                   quickSort [u | u <- xs, u >= x]


-- "l' = quickSortBy o l" sorts list l in non-descending order using
-- the ordering o.

quickSortBy :: (a -> a -> Ordering) -> [a] -> [a]
quickSortBy _ [] = []
quickSortBy o (x:xs) =
  quickSortBy o [u | u <- xs, o u x == LT]  ++ [x] ++
  quickSortBy o [u | u <- xs, o u x == GT || o u x == EQ]


-- tuple selector functions:

fst3 :: (a,b,c) -> a
fst3 (x,_,_) = x

snd3 :: (a,b,c) -> b
snd3 (_,x,_) = x

thd3 :: (a,b,c) -> c
thd3 (_,_,x) = x

fst5 :: (a,b,c,d,e) -> a
fst5 (x,_,_,_,_) = x

snd5 :: (a,b,c,d,e) -> b
snd5 (_,x,_,_,_) = x

thrd5 :: (a,b,c,d,e) -> c
thrd5 (_,_,x,_,_) = x

frth5 :: (a,b,c,d,e) -> d
frth5 (_,_,_,x,_) = x

fifth5 :: (a,b,c,d,e) -> e
fifth5 (_,_,_,_,x) = x

fst6 :: (a,b,c,d,e,f) -> a
fst6 (x,_,_,_,_,_) = x

snd6 :: (a,b,c,d,e,f) -> b
snd6 (_,x,_,_,_,_) = x

thrd6 :: (a,b,c,d,e,f) -> c
thrd6 (_,_,x,_,_,_) = x

frth6 :: (a,b,c,d,e,f) -> d
frth6 (_,_,_,x,_,_) = x

fifth6 :: (a,b,c,d,e,f) -> e
fifth6 (_,_,_,_,x,_) = x

sixth6 :: (a,b,c,d,e,f) -> f
sixth6 (_,_,_,_,_,x) = x


-- "(p,n) = getPredicates f" extracts all predicates of formula f,
-- and separates these into a list p of predicates occurring in positive
-- literals in f, and a list n of predicates from negative literals.

getPredicates :: Formula -> ([Predicate],[Predicate])
getPredicates f
  | isPosLit f = ([getLitPred (getLiteral f)], [])
  | isNegLit f = ([], [getLitPred (getLiteral f)])
  | otherwise = foldr (concTup . getPredicates) ([],[]) sfs
  where sfs = getSubFormulae f
        concTup (xs,ys) (xs',ys') = (nub (xs ++ xs'), nub (ys ++ ys'))


-- print newline

newLine :: IO ()
newLine = putChar '\n'
