----------------------------------------------------------------------
-- Module Formula.hs                                                --
--                                                                  --
-- Implementation module of some basic abstract data types:         --
-- Variables, constants, terms, predicates, literals, formulae.     --
--                                                                  --
-- 26.09.96  initial revision                                       --
----------------------------------------------------------------------


--------------------
----- comments -----
--------------------

-- Formulae are represented as trees, where the nodes of a formula tree
-- consist of a constructor and additional information needed by
-- different reduction strategies.
-- Descendents of a node in a formula tree represent direct subformulae.
-- Constructors are: true, false, literal, not, or, and, exists and forall.
-- true, false and literal nodes have no descendents; not, exists and forall
-- have one descendent; or and and have at least two descendents.
-- The node information consists either of a strategy info field or is nothing.
-- The strategy information field contains generation-term pairs, where
-- for each pair the term may also be omitted.
-- Strategy information is used for conjunction and quantifier nodes.
-- In the first case (conjunction node) there is only one generation-term
-- pair, where the term is set to Nothing. The generation indicates
-- the number of the reduction step, where reduction at this node was
-- possible for the first time.
-- For quantifier nodes the generation has the same semantics, but
-- it is stored for each term independently. Thus, the meaning of a
-- pair (g,t) in the strategy information field is that reduction with
-- term t was at first possible after reduction step g.
-- For existential quantifier nodes, strategy information for a specific
-- term is not re-initialized, i.e. a reduction with term t is performed
-- only once. It may be the case, however, that introduction of new variables
-- by means of a universal rule application expands the strategy information
-- field of an existential quantifier by another generation-term pair.
-- Strategy information for existential quantifer nodes contains at most
-- those terms which are present in the formula.
-- For universal quantifier nodes strategy information is only available
-- for variable terms, but not for constant terms. Moreover, strategy info
-- fields are updated for all variable terms, i.e. reduction with some 
-- variable, e.g. x, may be performed multiply.
-- Strategy information for universal quantifier nodes contains not only
-- those variables occurring in the formula, but also another fresh
-- variable not used so far, if the variable restriction of the calculus
-- admits this.


----------------------------
----- module interface -----
----------------------------

module Formula (
  -- variables
  Variable, mkVar, getVarIndex,
  -- constants
  Constant, mkCon, getConIndex,
  -- terms
  Term, termFromVar, termFromCon, termIsVar, termIsCon, getTermVar,
        getTermCon,
  -- predicates
  Predicate, mkPred, getPredIndex, getPredRank,
  -- literals
  Literal, mkLit, getLitPred, getLitTerms, getLitVars, getLitCons,
  -- generations
  Generation, mkGen, initialGen, getGen, nextGen,
  -- node and strategy information
  NodeInfo, emptyNodeInfo, mkNodeInfo, nodeInfoIsEmpty, getStrategyInfo,
  StrategyInfo, mergeNodeInfo, getSITerms,
  -- formulae
  Formula, verum, falsum, literal, neg, disj, conj, exists, forall,
           isVerum, isFalsum, isLiteral, isPosLit, isNegLit, isNeg, isDisj, isConj,
           isExists, isForAll, isQuant, getLiteral, getQuantVar, getSubFormula,
           getSubFormulae, getNodeInfo, (@@), setQuantVar, setSubFormula,
           setSubFormulae, setNodeInfo, addSubFormulae, setSubFormulaAtPos,
           vars, free, bound, consts,
  -- positions
  Position, lambda, mkPos, concPos, posToList, listToPos, commonPrefix,
  -- paths
  Path, emptyPath, mkPath, getPath, filterPath, reversePath, pathToList,
        pathIsEmpty, pathHead, pathRest, pathLength, decomposePath
) where


-------------------
----- imports -----
-------------------

import Set     -- for variable sets
import List (nub)


--------------------
----- notation -----
--------------------

-- To get the subformula at a specified position use the infix operator @@:
-- f@@p delivers the subformula of f at position p.

infix 8 @@


-------------------------------
----- abstract data types -----
-------------------------------

---------------------
----- variables -----
---------------------

-- variable data type:
--   data type name: Variable
--   methods:        mkVar, getVarIndex
--   instance of:    Eq, Ord, Enum, Read, Show

newtype Variable = MkVar Int
  deriving (Eq, Ord)

-- constructor

mkVar :: Int -> Variable
mkVar x = MkVar x

-- selector

getVarIndex :: Variable -> Int
getVarIndex (MkVar x) = x

-- instance declarations

instance Enum Variable where
  toEnum = mkVar
  fromEnum = getVarIndex
  enumFrom (MkVar x) = map MkVar [x..]
  enumFromThen (MkVar x) (MkVar x') = map MkVar [x,x'..]

-- instance Read: in module FormulaIO
-- instance Show: in module FormulaIO


---------------------
----- constants -----
---------------------

-- constant data type:
--   data type name: Contant
--   methods:        mkCon, getConIndex
--   instance of:    Eq, Ord, Enum, Read, Show

newtype Constant = MkCon Int
  deriving (Eq, Ord)

-- constructor

mkCon :: Int -> Constant
mkCon x = MkCon x

-- selector

getConIndex :: Constant -> Int
getConIndex (MkCon x) = x

-- instance declarations

instance Enum Constant where
  toEnum = mkCon
  fromEnum = getConIndex
  enumFrom (MkCon x) = map MkCon [x..]
  enumFromThen (MkCon x) (MkCon x') = map MkCon [x,x'..]

-- instance Read: in module FormulaIO
-- instance Show: in module FormulaIO


-----------------
----- terms -----
-----------------

-- term data type:
--   data type name: Term
--   methods:        termFromVar, termFromCon, termIsVar, termIsCon,
--                   getTermVar, getTermCon
--   instance of:    Eq, Ord, Read, Show

data Term = Var Variable | Con Constant
  deriving (Eq, Ord)

-- constructors

termFromVar :: Variable -> Term
termFromVar v = Var v

termFromCon :: Constant -> Term
termFromCon c = Con c

-- selectors

termIsVar :: Term -> Bool
termIsVar (Var _) = True
termIsVar _       = False

termIsCon :: Term -> Bool
termIsCon (Con _) = True
termIsCon _       = False

getTermVar :: Term -> Variable
getTermVar (Var v) = v
getTermVar _       = error "wrong argument in getTermVar"

getTermCon :: Term -> Constant
getTermCon (Con c) = c
getTermCon _       = error "wrong argument in getTermCon"

-- instance declarations

-- instance Read: in module FormulaIO
-- instance Show: in module FormulaIO


----------------
-- predicates --
----------------

-- predicate data type:
--   data type name: Predicate
--   methods:        mkPred, getPredIndex, getPredRank
--   instance of:    Eq, Ord, Show, Enum
--   comment: The first argument passed to the constructor encodes a
--            predicate P, the second is the rank of P.

newtype Predicate = MkPred (Int, Int)

-- constructor

mkPred :: Int -> Int -> Predicate
mkPred x r = MkPred (x,r)

-- selectors

getPredIndex :: Predicate -> Int
getPredIndex (MkPred (x,_)) = x

getPredRank :: Predicate -> Int
getPredRank (MkPred (_,r)) = r

-- instance declarations

instance Eq Predicate where
 MkPred (x,_) == MkPred (y,_) = x == y

instance Ord Predicate where
 compare (MkPred (x,_)) (MkPred (y,_)) = compare x y

instance Enum Predicate where
  toEnum i = MkPred (i,0)
  fromEnum = getPredIndex
  enumFrom (MkPred (x,y)) = [MkPred (i,y) | i <- [x..]]
  enumFromThen (MkPred (x,y)) (MkPred (x',_)) = [MkPred (i,y) | i <- [x,x'..]]

-- instance Show: in module FormulaIO


--------------
-- literals --
--------------

-- literal data types:
--   data type name: Literal
--   methods:        mkLit, getLitPred, getLitTerms, getLitVars, getLitCons
--   instance of:    Eq, Ord, Show

newtype Literal = MkLit (Predicate, [Term])
  deriving (Eq, Ord)

-- constructor

mkLit :: Predicate -> [Term] -> Literal
mkLit p ts = MkLit (p, ts)

-- selectors

getLitPred :: Literal -> Predicate
getLitPred (MkLit (p, _)) = p

getLitTerms :: Literal -> [Term]
getLitTerms (MkLit (_, ts)) = ts

getLitVars :: Literal -> Set Variable
getLitVars (MkLit (_, ts)) = listToSet (map getTermVar (filter termIsVar ts))

getLitCons :: Literal -> Set Constant
getLitCons (MkLit (_, ts)) = listToSet (map getTermCon (filter termIsCon ts))

-- instance declarations

-- instance Show: in module FormulaIO


-----------------------
----- generations -----
-----------------------

-- generation data type:
--   data type name: Generation
--   methods:        mkGen, getGen
--   instance of:    Eq, Ord, Show

newtype Generation = MkGen Int
  deriving (Eq, Ord, Show)

-- constructors

initialGen :: Generation
initialGen = MkGen 0

mkGen :: Int -> Generation
mkGen x = MkGen x

-- selector

getGen :: Generation -> Int
getGen (MkGen x) = x

-- operations

nextGen :: Generation -> Generation
nextGen (MkGen x) = MkGen (x+1)


----------------------------
----- node information -----
----------------------------

-- strategy and node information data type:
--   data type names: StrategyInfo, NodeInfo
--   methods:         emptyNodeInfo, mkNodeInfo, nodeInfoIsEmpty,
--                    getStrategyInfo, mergeNodeInfo
--   instance of:     Eq, Show

type StrategyInfo = [(Generation, Maybe Term)]

newtype NodeInfo = MkNodeInfo (Maybe StrategyInfo)
  deriving Eq


-- constructors

emptyNodeInfo :: NodeInfo
emptyNodeInfo = MkNodeInfo Nothing

mkNodeInfo :: StrategyInfo -> NodeInfo
mkNodeInfo gt = MkNodeInfo (Just gt)

-- selectors

nodeInfoIsEmpty :: NodeInfo -> Bool
nodeInfoIsEmpty (MkNodeInfo Nothing) = True
nodeInfoIsEmpty _                    = False

getStrategyInfo :: NodeInfo -> StrategyInfo
getStrategyInfo (MkNodeInfo (Just gt)) = gt
getStrategyInfo (MkNodeInfo Nothing) = []

-- operations

mergeNodeInfo :: NodeInfo -> NodeInfo -> NodeInfo
mergeNodeInfo (MkNodeInfo Nothing) ni = ni
mergeNodeInfo ni (MkNodeInfo Nothing) = ni
mergeNodeInfo (MkNodeInfo (Just gt)) (MkNodeInfo (Just gt')) =
  MkNodeInfo (Just (nub (gt ++ gt')))

getSITerms :: StrategyInfo -> [Term]
getSITerms [] = []
getSITerms ((g,mt):si) =
  let ts = getSITerms si
      Just t = mt
  in if mt == Nothing then ts else t:ts

-- instance declarations

-- instance Show: in module FormulaIO


--------------------
----- formulae -----
--------------------

-- formula data type:
--   data type name: Formula
--   methods:        verum, falsum, literal, neg, disj, conj, exists, forall,
--                   isVerum, isFalsum, isLiteral, isPosLit, isNegLit, isNeg,
--                   isDisj, isConj, isExists, isForAll, isQuant, getLiteral,
--                   getQuantVar, getSubFormula, getSubFormulae, getNodeInfo,
--                   @@, setQuantVar, setSubFormula, setSubFormulae,
--                   setNodeInfo, addSubFormulae, setSubFormulaAtPos,
--                   vars, free, bound, consts
--   instance of:    Eq, Show, Read

type FormulaTree = Tree (ForConstructor, NodeInfo)
newtype Formula = MkFor FormulaTree

-- auxiliary data types (not abstract!)

data ForConstructor = 
  T | F |             -- true and false
  Lit Literal |       -- literal
  Not | Or | And |    -- negation, disjunction, conjunction
  Exists Variable |   -- existential quantification
  ForAll Variable     -- universal quantification
 deriving Eq

data Tree a = Node a (Forest a)
  deriving (Eq, Show)
type Forest a = [Tree a]

-- constructors

verum, falsum :: Formula
verum = MkFor (Node (T, MkNodeInfo Nothing) [])
falsum = MkFor (Node (F, MkNodeInfo Nothing) [])

literal :: Literal -> Formula
literal l = MkFor (Node (Lit l, MkNodeInfo Nothing) [])

neg :: Formula -> Formula
neg (MkFor f) = MkFor (Node (Not, MkNodeInfo Nothing) [f])

-- If formula list is [f] then just disj f = f,
-- if the formula list is empty, then disj f = false
-- For conj the same applies mutatis mutandis (duality!)
disj, conj :: [Formula] -> Formula
disj [] = MkFor (Node (F, MkNodeInfo Nothing) [])
disj [f] = f
disj fs = MkFor (Node (Or, MkNodeInfo Nothing) (map strip fs))
conj [] = MkFor (Node (T, MkNodeInfo Nothing) [])
conj [f] = f
conj fs = MkFor (Node (And, MkNodeInfo Nothing) (map strip fs))

exists, forall :: Variable -> Formula -> Formula
exists v (MkFor f) = MkFor (Node (Exists v, MkNodeInfo (Just [])) [f])
forall v (MkFor f) = MkFor (Node (ForAll v, MkNodeInfo (Just [])) [f])

-- selectors

isVerum, isFalsum, isPosLit, isNegLit, isLiteral, isNeg,
  isDisj, isConj, isExists, isForAll, isQuant :: Formula -> Bool

isVerum (MkFor (Node (T, _) _)) = True
isVerum _ = False

isFalsum (MkFor (Node (F, _) _)) = True
isFalsum _ = False

isPosLit (MkFor (Node (Lit _, _) _)) = True
isPosLit _ = False

isNegLit (MkFor (Node (Not, _) [Node (Lit _, _) _])) = True
isNegLit _ = False

isLiteral f = isPosLit f || isNegLit f

isNeg (MkFor (Node (Not, _) _)) = True
isNeg _ = False

isDisj (MkFor (Node (Or, _) _)) = True
isDisj _ = False

isConj (MkFor (Node (And, _) _)) = True
isConj _ = False

isExists (MkFor (Node (Exists _, _) _)) = True
isExists _ = False

isForAll (MkFor (Node (ForAll _, _) _)) = True
isForAll _ = False

isQuant (MkFor (Node (Exists _, _) _)) = True
isQuant (MkFor (Node (ForAll _, _) _)) = True
isQuant _ = False

-- getLiteral is applicable for positive and negative literal formulae
getLiteral :: Formula -> Literal
getLiteral (MkFor (Node (Lit l, _) _)) = l
getLiteral (MkFor (Node (Not,_) [Node (Lit l, _) _])) = l
getLiteral _ = error "wrong argument in getLiteral"

getQuantVar :: Formula -> Variable
getQuantVar (MkFor (Node (c, _) ~(ft:_))) = case c of
  Exists v -> v
  ForAll v -> v
  otherwise -> error "wrong argument in getQuantVar"
 where f = MkFor ft

getSubFormula :: Formula -> Formula
getSubFormula (MkFor (Node (c, _) ~(ft:_))) = case c of
  Not -> f
  Exists _ -> f
  ForAll _ -> f
  otherwise -> error "wrong argument in getSubFormula"
 where f = MkFor ft

getSubFormulae :: Formula -> [Formula]
getSubFormulae (MkFor (Node _ fts)) = map MkFor fts

getNodeInfo :: Formula -> NodeInfo
getNodeInfo (MkFor (Node (_,ni) _)) = ni

(@@) :: Formula -> Position -> Formula
(MkFor ft)@@ p = MkFor (getAtPos p ft)
  where getAtPos (Pos []) ft = ft
        getAtPos (Pos (i:is)) (Node _ fts) = getAtPos (Pos is) (ith fts i)
        ith [] _ = error "inaccessible position in @@"
        ith (x:_) 0 = x
        ith (x:xs) n = ith xs (n-1)

setSubFormula :: Formula -> Formula -> Formula
setSubFormula (MkFor sft) (MkFor (Node n@(c,_) _)) = case c of
  Not -> MkFor (Node n [sft])
  Exists _ -> MkFor (Node n [sft])
  ForAll _ -> MkFor (Node n [sft])
  otherwise -> error "wrong argument in setSubFormula"

setSubFormulae :: [Formula] -> Formula -> Formula
setSubFormulae sfs (MkFor (Node n _)) =
  MkFor (Node n (map strip sfs))

setNodeInfo :: NodeInfo -> Formula -> Formula
setNodeInfo ni (MkFor (Node (c,_) fts)) = MkFor (Node (c, ni) fts)

setQuantVar :: Variable -> Formula -> Formula
setQuantVar v (MkFor (Node (c, ni) fts)) = case c of
  Exists _ -> MkFor (Node (Exists v, ni) fts)
  ForAll _ -> MkFor (Node (ForAll v, ni) fts)
  otherwise -> error "wrong argument in setQuantVar"

addSubFormulae :: [Formula] -> Formula -> Formula
addSubFormulae sfs (MkFor (Node n@(c,_) fts)) = case c of
  Or -> MkFor (Node n (fts ++ map strip sfs))
  And -> MkFor (Node n (fts ++ map strip sfs))
  otherwise -> error "wrong argument in addSubFormulae"

setSubFormulaAtPos :: Position -> Formula -> Formula -> Formula
setSubFormulaAtPos (Pos p) (MkFor sf) (MkFor f) = MkFor (setSF p sf f)
  where setSF [] sf _ = sf
        setSF (i:is) sf (Node n fts) =
            Node n (fs ++ [setSF is sf (fts!!i)] ++ gs)
          where fs = take i fts
                gs = drop (i+1) fts

-- operations

vars :: Formula -> Set Variable
vars (MkFor ft) = vars' ft
  where vars' (Node (c, _) fts) = case c of
          Lit l -> getLitVars l
          Exists v -> add v (vars' (head fts))
          ForAll v -> add v (vars' (head fts))
          otherwise -> foldr (sunion . vars') empty fts

free :: Formula -> Set Variable
free (MkFor ft) = free' ft
  where free' (Node (c, _) fts) = case c of
          Lit l -> getLitVars l
          Exists v -> free' (head fts) `minus` singleton v
          ForAll v -> free' (head fts) `minus` singleton v
          otherwise -> foldr (sunion . free') empty fts

bound :: Formula -> Set Variable
bound (MkFor ft) = bound' ft
  where bound' (Node (c, _) fts) = case c of
          Exists v -> add v (bound' (head fts))
          ForAll v -> add v (bound' (head fts))
          otherwise -> foldr (sunion . bound') empty fts

consts :: Formula -> Set Constant
consts (MkFor ft) = consts' ft
  where consts' (Node (c, _) fts) = case c of
          Lit l -> getLitCons l
          otherwise -> foldr (sunion . consts') empty fts

-- auxiliary functions 

strip :: Formula -> FormulaTree
strip (MkFor ft) = ft

-- instance declarations

instance Eq Formula where
  (MkFor ft) == (MkFor ft') = forEqual (ft, ft')
    where forEqual ((Node (c,_) fts), (Node (c',_) fts')) =
            c == c' && all forEqual (zip fts fts')

-- instance Show: in module FormulaIO


---------------------
----- positions -----
---------------------

-- formula position data type:
--   data type name: Position
--   methods:        lambda, mkPos, concPos, posToList, listToPos
--   intance of:     Eq, Ord, Show
-- NOTE: The ordering is not total! Comparisons using <= or >= are not allowed!
--       To perform a comparision as e.g. p <= p' use p < p' || p = p'.

newtype Position = Pos [Int]
  deriving Eq

-- constructors

lambda :: Position
lambda = Pos []

mkPos :: Position -> Int -> Position
mkPos (Pos p) i = Pos (p ++ [i])

-- selectors

posToList :: Position -> [Int]
posToList (Pos p) = p

-- operations

concPos :: Position -> Position -> Position
concPos (Pos p) (Pos p') = Pos (p ++ p')

listToPos :: [Int] -> Position
listToPos p = (Pos p)

-- "cp = commonPrefix p p'" returns the longest common prefix of
-- positions p and p', i.e. if p = [i_0,...,i_n] and p' = [j_0,...,j_m]
-- then cp = [i_0,...,i_k] where i_l = j_l for 0 <= l <= k, k <= n,m and
-- i_{l+1} /= j_{l+1}.

commonPrefix :: Position -> Position -> Position
commonPrefix (Pos []) _ = Pos []
commonPrefix _ (Pos []) = Pos []
commonPrefix (Pos (i:p)) (Pos (i':p')) =
  let Pos c = commonPrefix (Pos p) (Pos p')
  in if i==i' then Pos (i:c) else Pos []

-- instance declarations

-- NOTE: This is no total ordering! Therefore a comparision using
-- operators <= and >= is not allowed! (It returns always equal!)
-- A test whether p <= q should be replaced by p < q || p == q.

instance Ord Position where
  compare (Pos []) (Pos []) = EQ
  compare (Pos []) _ = GT
  compare _ (Pos []) = LT
  compare (Pos (i:p)) (Pos (i':p')) =
    if i == i' then compare (Pos p) (Pos p')
    else EQ

instance Show Position where
  showsPrec _ (Pos []) = showString "lambda"
  showsPrec _ (Pos p) = showChar '[' . showPos p . showChar ']'
    where showPos [] = id
          showPos (i:[]) = shows i
          showPos (i:is) = shows i . showChar '.' . showPos is

-----------------
----- paths -----
-----------------

-- path data type:
--   data type name: Path
--   methods:        emptyPath, mkPath, getPath, pathLength, reversePath,
--                   pathIsEmpty, pathToList, pathHead, pathRest,
--                   decomposePath
--   instance of:    Eq, Show

newtype Path = MkPath [Formula]
  deriving Eq

-- constructors

emptyPath :: Path
emptyPath = MkPath []

mkPath :: Formula -> Path -> Path
mkPath f (MkPath pt) = MkPath (f:pt)

-- operations

getPath :: Formula -> Position -> Position -> Path
getPath f (Pos p) (Pos p') = MkPath (getPath' (f@@(Pos p)) (drop (length p) p'))
  where getPath' sf [] = [sf]
        getPath' sf@(MkFor (Node _ sfs)) (i:ps) =
                           sf : getPath' (MkFor (sfs!!i)) ps

pathLength :: Path -> Int
pathLength (MkPath pt) = length pt

filterPath :: (Formula -> Bool) -> Path -> Path
filterPath f (MkPath pt) = MkPath (filter f pt)

reversePath :: Path -> Path
reversePath (MkPath pt) = MkPath (reverse pt)

pathIsEmpty :: Path -> Bool
pathIsEmpty (MkPath []) = True
pathIsEmpty _ = False

pathToList :: Path -> [Formula]
pathToList (MkPath pt) = pt

pathHead :: Path -> Formula
pathHead (MkPath []) = error "wrong argument in pathHead"
pathHead (MkPath (f:_)) = f

pathRest :: Path -> Path
pathRest (MkPath []) = error "wrong argument in pathRest"
pathRest (MkPath (_:fs)) = MkPath fs

-- assertion: decomposePath (mkPath f p) = (f,p)

decomposePath :: Path -> (Formula, Path)
decomposePath (MkPath []) = error "wrong argument in decomposePath"
decomposePath (MkPath (f:pt)) = (f, MkPath pt)

-- instance declarations

-- instance Show: in module FormulaIO
