----------------------------------------------------------------------
-- Module FormulaIO.hs                                              --
--                                                                  --
-- Routines for reading and writing formulae, variables, constants, --
-- terms, etc.							    --
-- This file also includes the main formula parser.		    --
--                                                                  --
-- 22.10.96  initial revision                                       --
-- 12.01.00  RA extensions					    --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module FormulaIO (
  -- data types
  FormulaSyntax(
    PL1, RA
  ),
  -- input functions
  readFormulaRA,	-- read a formula in an IO state monad (RA syntax)
  readFormula, 		-- read a formula in an IO state monad (PL1 syntax)
  readTerm,		-- read a term in an IO state monad
  readVar,		-- read a variable in an IO state monad
  readCon,		-- read a constant in an IO state monad
  checkLitRank,		-- assures that all lit. with same predicates have the same rank
  rReadLn,		-- robust read line for IO state monad
  rReadVar,		-- robust read variable for IO state monad
  rReadCon,		-- robust read constant for IO state monad
  rReadTerm,		-- robust read term for IO state monad
  rReadWith,		-- robust generic read function for IO state monad
  rReadCond,		-- robust generic read function with test of read value

  -- output functions
  showFor,		-- print a PL1 formula in an IO state monad
  draw,			-- draw a PL1 formula in an IO state monad
  drawWithNodeInfo,	-- '' with node info
  drawWithNodeInfoAndLitPositions, -- '' with literal positons
  showUResult,		-- print unification result

  -- interaction functions
  query			-- interactive choice

  -- NOTE: Instances of class Read are declared and exported for the following
  --       data types: Formula, Term, Variable, Constant.
  --       Instances of class Show are declared and exported for the following
  --       data types: Variable, Constant, Term, Predicate, Literal, Formula,
  --                   NodeInfo, Path
) where


----------------------------
----- imported modules -----
----------------------------

import Char
import Formula
import Auxiliary
import Scanner
import IOStateMonad
import GlobalState
import Unification
import Monad
import Simplify (collectLiterals, simplify, nnf)
import List ((\\), partition)

--------------
-- notation --
--------------

-- alternatives in parser rules are separated by `+++'.
infixr 8 +++


----------------------
----- data types -----
----------------------

-- parser data type

-- The PFail data type is provided in order to get acceptable error
-- messages. The integer denotes the token position of the error,
-- the string is an error message, and the list is the list of remaining
-- tokens when the error occurred.

data PFail t = PFail Int String [t]
  deriving Show

-- A ParseResult v t is either a (partially) successful parse or a failure
-- (where t is the token type, v the resulting value type).
-- In the first case (v,r)-pairs and the best possibility to fail
-- are returned, where v is the value of the parsed token stream and r
-- is the rest of tokens not parsed; in the second case a failure
-- possibility is returned.

data ParseResult v t = Parse [(v,Int,[t])] (PFail t)
                     | None (PFail t)
  deriving Show

-- A parser takes a list of tokens of type t and a token position and
-- returns a parse result.

newtype Parser t v = Parser ([t] -> Int -> ParseResult v t)
--  deriving Show


-- Formula syntax: PL1 or RA 

data FormulaSyntax = PL1 | RA
  deriving (Show, Eq)

---------------------------------
----- instance declarations -----
---------------------------------

-- instances for parser type

instance Functor (Parser t) where
  -- map :: (v -> w) -> Parser t v -> Parser t w
  fmap f (Parser p) =
    Parser (\ts n -> let res = p ts n
                   in case res of
                     Parse l fa -> Parse [(f v,k, r) | (v,k,r) <- l] fa
                     None fa -> None fa)

instance Monad (Parser t) where
  -- return :: v -> Parser t v
  return v = Parser (\ts n -> Parse [(v,n,ts)] noFail)
  -- >>= :: Parser t v -> (v -> Parser t w) -> Parser t w
  (Parser p) >>= f = Parser (\ts n ->
    case p ts n of
      Parse l fa -> processAlts fa [] [fv r k | (v,k,r) <- l, let Parser fv = f v]
      None fa -> None fa)

instance MonadPlus (Parser t) where
  -- mzero :: Parser t v
  mzero = Parser (\ts n -> None (PFail n "error in last token." ts))
  -- mplus :: Parser t v -> Parser t v -> Parser t v
  mplus (Parser p) (Parser q) = Parser (\ts n ->
    let rp = p ts n; rq = q ts n
    in case (rp,rq) of
      (Parse l fa, Parse l' fa') -> Parse (l ++ l') (bestFail fa fa')
      (Parse l fa, None fa') -> Parse l (bestFail fa fa')
      (None fa, _) -> updateFail fa rq)


-- instances of class Show (no symbol table support)

instance Show Variable where
  showsPrec _ v = showVar emptySymbolTable v

instance Show Constant where
  showsPrec _ c = showCon emptySymbolTable c

instance Show Term where
  showsPrec _ t = showTerm emptySymbolTable t

instance Show Predicate where
  showsPrec _ p = showPred emptySymbolTable p

instance Show Literal where
  showsPrec _ l = showLit emptySymbolTable l

instance Show Formula where
  showsPrec p f = showFormula emptySymbolTable p f

instance Show NodeInfo where
  showsPrec _ ni
   | nodeInfoIsEmpty ni = id
   | otherwise = let gts = getStrategyInfo ni in showNIList "" gts
   where showNIList _ [] = id
         showNIList s ((g,mt):gts) = showString s . showChar '(' . shows g .
          showChar ',' . case mt of
           { Nothing -> showString "||"; Just t -> shows t } .
           showChar ')' . showNIList "," gts

instance Show Path where
  showsPrec _ pa
    | pathIsEmpty pa = showString "<< >>"
    | otherwise = let fs = pathToList pa
                  in showString "<< " . showPath fs . showString " >>"
    where showPath [] = id
          showPath (f:[]) = showPathE f
          showPath (f:fs) = showPathE f . showString ", " . showPath fs
          showPathE f
            | isVerum f = showString "true"
            | isFalsum f = showString "false"
            | isPosLit f = shows (getLiteral f)
            | isNegLit f = showChar '~' . shows (getLiteral (getSubFormula f))
            | isConj f = showString "/\\"
            | isDisj f = showString "\\/"
            | isExists f = showString "(exists " . shows (getQuantVar f) .
                showChar ')'
            | isForAll f = showString "(forall " . shows (getQuantVar f) .
                showChar ')'


-- instances of class Read (no symbol table support)

-- read, but no symbol table support

-- Input string should contain a semicolor to mark the end of the
-- formula.

instance Read Formula where
  readsPrec _ s = if null toks then [] 
                  else either (const []) (\((f,_,_):_) -> [(f,[])]) parse
    where (toks,_) = forScannerPL emptySymbolTable s
          parse = papply mainForParserPL toks

-- input following term is discarded (usually up to a newline)
instance Read Term where
  readsPrec _ s = readTerm' (dropWhile isSpace s)
    where readTerm' ('v':'_':r) = let n = read r
                                  in [(termFromVar (mkVar n), [])]
          readTerm' ('c':'_':r) = let n = read r
                                  in [(termFromCon (mkCon n), [])]
          readVar' _ = []

-- input following variable is discarded (usually up to a newline)
instance Read Variable where
  readsPrec _ s = readVar' (dropWhile isSpace s)
    where readVar' ('v':'_':r) = let n = read r in [(mkVar n, [])]
          readVar' _ = []

-- input following constant is discarded (usually up to a newline)
instance Read Constant where
  readsPrec _ s = readCon' (dropWhile isSpace s)
    where readCon' ('c':'_':r) = let n = read r in [(mkCon n, [])]
          readCon' _ = []


------------------------------
----- exported functions -----
------------------------------

-----------------------
-- parsing functions --
-----------------------

-- Error codes raised by these functions (these relate to IO state monad errors):
--  10: variable parse failed
--  11: constant parse failed
--  12: term parse failed
--  13: formula parse failed
--  20: literal rank check failed

-- Read a formula into an IO state monad (PL1 syntax).
-- The formula parser expects a semicolon at the end of the input string.

readFormula :: String -> IOState Formula
readFormula s = do
  let (toks,symTab) = forScannerPL emptySymbolTable s -- convert string to tokens
      parse = papply mainForParserPL toks             -- parse token string
  setSymbolTable symTab				      -- update symbol table
  case parse of
    Left errMsg -> failS (iosError 13 errMsg)
    Right ((f,_,_):_) -> return f


{-
igs = MkGlobalState {
  totalProofSteps = 0,
  totalProofGenerations = 0,
  totalProofTime = 0,
  subProofTime = 0,
  proofOptions = MkProofOptions {
    splitting = NoSplitting,
    equalModRenSteps = 10,
    unusedLitSteps = 20,
    outerAQuants = True,
    printing = OMVerbose,
    debugging = False
  },
  formula = falsum,
  generation = initialGen,
  step = 0,
  maxVar = 0,
  posLitPosition = lambda,
  negLitPosition = lambda,
  unificationMap = [],
  unificationResult = Fail,
  symbolTable = emptySymbolTable,
  randomNumbers = []
}

ttt s = do { f <- readFormulaRA s; showFor f; draw f } `startingWith` igs
-}


-- Read a formula in RA syntax into an IO state monad 
  
readFormulaRA :: String -> IOState Formula
readFormulaRA s = do
  let (toks,symTab) = forScannerRA initialRASymbolTable s
      Just x = stLookupVarByName "x" symTab
      Just y = stLookupVarByName "y" symTab
      Just z = stLookupVarByName "z" symTab
      xyz = (x,y,z)
      parse = papply (mainParserRA xyz) toks          -- parse token string
  setSymbolTable symTab			 	      -- update symbol table
  case parse of
    Left errMsg -> failS (iosError 13 errMsg)
    Right ((f,_,_):_) -> do draw (reduceEqRAFor $ nnf f)
                            return (extendRAFor $ reduceEqRAFor $ nnf f)

-- RA equality predicate i
eqPredRA :: Predicate
eqPredRA = mkPred 0 2

-- remove occurrances of the equality predicate by using the following
-- two reductions:
-- (exists x)(x = y /\ F(x))  -->  F(y)
-- (forall x)(x = y -> F(x))  -->  F(y)

reduceEqRAFor :: Formula -> Formula
reduceEqRAFor f
  | isVerum f = f
  | isFalsum f = f
  | isExists f' && isConj sf && mt /= Nothing = simplify $ subst x t (conj r)
  | isForAll f' && isDisj sf && mt /= Nothing = simplify $ subst x t (disj r)
  | otherwise = f'
  where f' = simplify $ setSubFormulae (map reduceEqRAFor (getSubFormulae f)) f
        sf = getSubFormula f'
	x = getQuantVar f
        (mt,r) = searchEqPred x (isExists f) (getSubFormulae sf)
	Just t = mt

-- Search positive resp. negative equality predicate containing
-- variable x in list of formulae.
-- searchEqPred x p fs = (t', fs')  searches for a predicate x=t if p is
-- true (or x/=t if p is false) for some term t in the formula list fs.
-- If such a term t is found, it is returned in t' together with all other
-- subformulae in fs'. Otherwise t' is Nothing, and fs' equals fs.

searchEqPred :: Variable -> Bool -> [Formula] -> (Maybe Term, [Formula])
searchEqPred x p [] = (Nothing, [])
searchEqPred x p (f:fs) =
  if p && isPosLit f || not p && isNegLit f then
    if mt' /= Nothing then (mt', f:fs') else
    if mt  /= Nothing then (mt, fs) else (Nothing,f:fs)
  else (mt',f:fs')
  where mt = check x f
        (mt', fs') = searchEqPred x p fs
        check :: Variable -> Formula -> Maybe Term
        check v g = let l = getLiteral g
                        pr = getLitPred l
                        [x,y] = getLitTerms l
                    in if pr == eqPredRA then
                         if v == getTermVar x then Just y
                         else if v == getTermVar y then Just x
                         else Nothing
                       else Nothing

-- extendRAFor: add negated equality axioms if formula contains
-- equality predicate i

extendRAFor :: Formula -> Formula
extendRAFor f =
  let preds = map fst3 (collectLiterals f)
      containsEq = eqPredRA `elem` preds
      x = mkVar 0; y = mkVar 1; z = mkVar 2
      eqLit v1 v2 = literal (mkLit eqPredRA [termFromVar v1, termFromVar v2])
      mLit p v1 v2 = literal (mkLit p [termFromVar v1, termFromVar v2])
      eqForRA1 = forall x (eqLit x x)
      -- eqForRA2 = forall x (forall y (disj [neg (eqLit x y), eqLit y x]))
      eqForRA3 p = forall x (forall y (forall z
                    (conj [disj [neg (mLit p x z), neg (eqLit z y), mLit p x y],
			   disj [neg (mLit p z y), neg (eqLit x z), mLit p x y]])))
      eqForRA = neg (conj (eqForRA1:(map eqForRA3 preds))) -- \\ [eqPredRA]))))
  in if containsEq then disj [eqForRA, f] else f


{-
extendRAFor :: Formula -> Formula
extendRAFor f =
  let preds = map fst3 (collectLiterals f)
      containsEq = eqPredRA `elem` preds
      x = mkVar 0; y = mkVar 1; z = mkVar 2
      eqLit v1 v2 = literal (mkLit eqPredRA [termFromVar v1, termFromVar v2])
      mLit p v1 v2 = literal (mkLit p [termFromVar v1, termFromVar v2])
      eqForP p = forall x (forall y (
                   conj [disj [(exists z)(conj [mLit p x z, eqLit z y]), neg (mLit p x y)],
		         disj [mLit p x y, (forall z)(disj [neg (mLit p x z), neg (eqLit z y)])]]));
      eqFors = map (neg . eqForP) preds
  in if containsEq then disj (f:eqFors) else f
-}

-- Read a term in an IO state monad.

readTerm :: String -> IOState Term
readTerm s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mt = stLookupTermByName name symTab; Just t = mt
      -- default format
      rdf ('v':'_':is) = Just (termFromVar (mkVar (read is)))
      rdf ('c':'_':is) = Just (termFromCon (mkCon (read is)))
      rdf _ = Nothing
      mdt = rdf name; Just dt = mdt
  if mt == Nothing then
    if mdt == Nothing then failS (iosError 12 "no term parse")
     else return dt
   else return t
  
-- Read a variable in an IO state monad

readVar :: String -> IOState Variable
readVar s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mv = stLookupVarByName name symTab; Just v = mv
      -- default format
      rdf ('v':'_':is) = Just (mkVar (read is))
      rdf _ = Nothing
      mdv = rdf name; Just dv = mdv
  if mv == Nothing then
    if mdv == Nothing then failS (iosError 10 "no variable parse")
     else return dv
   else return v

-- Read a constant in an IO state monad

readCon :: String -> IOState Constant
readCon s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mc = stLookupConByName name symTab; Just c = mc
      -- default format
      rdf ('c':'_':is) = Just (mkCon (read is))
      rdf _ = Nothing
      mdc = rdf name; Just dc = mdc
  if mc == Nothing then
    if mdc == Nothing then failS (iosError 11 "no constant parse")
     else return dc
   else return c


-----------------------
-- semantical checks --
-----------------------

-- "checkLitRank f" raises an exception, if there are literals with the same
-- predicate in formula f, otherwise it returns formula f with literal ranks
-- updated in the symbol table of the global state and in all predicates. 

checkLitRank :: Formula -> IOState Formula
checkLitRank f = do
  st <- getSymbolTable
  if isLiteral f then do
    let l = getLiteral f; p = getLitPred l; ts = getLitTerms l
        i = getPredIndex p; r' = length ts; r = stLookupPredRank st p
    if r == -1 then do
      setSymbolTable (stSetPredRank st p r')
      return (if isPosLit f then literal (mkLit (mkPred i r') ts)
              else neg (literal (mkLit (mkPred i r') ts)))
     else if r == r' then return f
     else failS (iosError 20 ("Semantic error: different ranks " ++ show r' ++ 
      " resp. " ++ show r ++ " assigned to predicate " ++ showPred' st p ++ "."))
   else do
    let fs = getSubFormulae f
        checkLitRank' fs = foldR (\f ff -> do f' <- checkLitRank f
                                              return (f':ff)) [] fs
    fs' <- checkLitRank' fs
    return (setSubFormulae fs' f)


---------------------------
-- robust read functions --
---------------------------

-- (For IO global state monad.)

-- Robust functions never raise an IO exception. Instead they repeatedly
-- read input, until a correct parse is achieved.

-- read line
rReadLn :: Read a => IOState a
rReadLn = rReadWith (readWith . readIO)

-- read variable
rReadVar :: IOState Variable
rReadVar = rReadWith readVar

-- read constant
rReadCon :: IOState Constant
rReadCon = rReadWith readCon

-- read term
rReadTerm :: IOState Term
rReadTerm = rReadWith readTerm

-- Generic robust read function; promotes a read function to a robust
-- version.

rReadWith :: (String -> IOState a) -> IOState a
rReadWith rf = catchS $$ do s <- readWith getLine
                            rf s
                      $$ \_ -> do putStrLn' "Error in input."
                                  rReadWith rf

-- Generic robust read function with test of input value.
-- The condition specified as second parameter must be fulfilled in order
-- to return from this function. Otherwise the input process is repeated.

rReadCond :: IOState a -> (a -> Bool) -> IOState a
rReadCond reader cond = do
  x <- reader
  if cond x then return x
   else do putStrLn' "Input not admitted."
           rReadCond reader cond


----------------------
-- output functions --
----------------------

-- (Working on IO global state monads)

-- show formula (with symbol table support)
 
showFor :: Formula -> IOState ()
showFor f = do
  st <- getSymbolTable
  putStr' (showFormula st 0 f "")

-- Draw formula as a tree (with symbol table support).
draw :: Formula -> IOState ()
draw f = do
  st <- getSymbolTable
  putStr' (drawWO st f)

-- Draw formula as a tree with adjoined node information (and symbol table support).

drawWithNodeInfo :: Formula -> IOState ()
drawWithNodeInfo f = do
  st <- getSymbolTable
  putStr' (drawWNI st f)

-- Draw formula as a tree with adjoined node information and literal positions
-- (with symbol table support).

drawWithNodeInfoAndLitPositions :: Formula -> IOState ()
drawWithNodeInfoAndLitPositions f = do
  st <- getSymbolTable
  putStr' (drawWNILP st f)


----------------------------
-- generic query function --
----------------------------

-- "c = query msg cs" writes the string msg on the screen, and reads
-- characters from standard input until a character matches any of those
-- contained in cs. This character is returned.

query :: String -> [Char] -> IOState Char
query msg cs = do
  printWith putStr msg 	 	-- print message
  doIO (readAnswer cs)		-- read input
 where readAnswer :: [Char] -> IO Char
       readAnswer cs = do
         cl <- getLine
         let cl' = dropWhile (\c -> c `notElem` cs) cl
         if null cl' then do putStrLn "Input error."
                             readAnswer cs
                     else return (head cl')



------------------------------
----- internal functions -----
------------------------------

---------------------
-- parser routines --
---------------------

-- At first some common combinators:

-- item parser: returns the next token as value

item :: Parser t t
item = Parser (\ts n -> case ts of
   [] -> None (PFail n "unexpected end of file." [])
   (t:r) -> Parse [(t,n+1,r)] noFail)

-- generic formula token parser with error messages

ftoken :: (Show t, Eq t) => t -> Parser (Token t) (Token t)
ftoken ftok = Parser (\ts n -> 
  case ts of 
    [] -> None (PFail n "unexpected end of file" [])
    (t@(ft,(l,c)):r) -> if ft == ftok then Parse [(t,n+1,r)] noFail
                        else  None (PFail n ("unexpected `" ++ show ft ++
                                            "'.") ts))

-- Formula token parser with error messages and checking predicate.
-- The parser "ftokenp checkf msg" reads the next token from the input stream
-- and checks whether it fulfills predicate checkf or not. If the check fails,
-- an error message is reported, where msg is the name of the expected token;
-- otherwise the formula token with its position in the input is returned.

ftokenp :: Show t => (t -> Bool) -> String -> Parser (Token t) (Token t)
ftokenp p expStr = Parser (\ts n -> 
  case ts of [] -> None (PFail 0 "unexpected end of file" [])
             (t@(ft,(l,c)):r) -> if p ft then Parse [(t,n+1,r)] noFail
                                 else  None (PFail n ("expected `" ++ expStr ++
                                                   "', found `" ++ show ft ++
                                                   "'.") ts))

-- some predicates useful in function ftokenp:

-- ftoken is a variable token
--isTokVar :: FToken -> Bool
isTokVar (TokVar _) = True
isTokVar _ = False

-- ftoken is a constant token
--isTokCon :: FToken -> Bool
isTokCon (TokCon _) = True
isTokCon _ = False

-- ftoken is a predicate token
--isTokPred :: FToken -> Bool
isTokPred (TokPred _) = True
isTokPred _ = False

isRATokPred (RATokPred _) = True
isRATokPred _ = False


-- "first" returns the first alternative of a parse result

first :: Parser t v -> Parser t v
first (Parser p) = Parser (\ts n -> case p ts n of
                                      None err -> None err
                                      Parse (x:xs) fa -> Parse [x] fa)


-- force parser application

force :: Parser t v -> Parser t v
force (Parser p)        =
  Parser (\ts n -> case p ts n of
    Parse xs fa -> Parse ((fst3 (head xs), snd3 (head xs), thd3 (head xs)) : tail xs) fa
    None fa -> None fa)


-- alternative parser: accepts input matching first or second parser

(+++) :: Parser t v -> Parser t v -> Parser t v
p +++ q = mplus p q

-- repeated application of parser, zero times or more

many :: Parser t v -> Parser t [v]
many p = force (many1 p +++ return [])

-- repeated application of parser, at least once

many1 :: Parser t v -> Parser t [v]
many1 p = do { x <- p; xs <- many p; return (x:xs) }

-- repeated application of first parser (zero or more times),
-- with second parser interspersed

sepby :: Parser t v -> Parser t w -> Parser t [v]
p `sepby` sep = (p `sepby1` sep) +++ return []

-- repeated application of first parser (at least once)
-- with second parser interspersed

sepby1 :: Parser t v -> Parser t w -> Parser t [v]
p `sepby1` sep = do {x <- p; xs <- many (do {sep; p}); return (x:xs)}


--------------------------------
-- auxiliary parser functions --
--------------------------------

-- compute parse result (and best failure) from alternatives

processAlts :: PFail t -> [(v,Int,[t])] -> [ParseResult v t]
  -> ParseResult v t
processAlts f [] [] = None f
processAlts f rs [] = Parse rs f
processAlts f rs (None f':rws) = processAlts (bestFail f f') rs rws
processAlts f rs (Parse rs' f':rws) = processAlts (bestFail f f') (rs++rs') rws

-- "noFail" indicates that no parse error has occurred so far.

noFail :: PFail t
noFail = PFail (-1) "" []

-- "updateFail" computes the best failure when an error has occurred.

updateFail :: PFail t -> ParseResult v t -> ParseResult v t
updateFail f (Parse cas f') = let r = bestFail f f' in Parse cas r
updateFail f (None f') = None (bestFail f f')

-- "bestFail" selects one out of two failure possibilities, such that the
-- following rule holds: the more tokens parsed, the better.

bestFail :: PFail t -> PFail t -> PFail t
bestFail f@(PFail n _ _) f'@(PFail n' _ _) =
  if n > n' then f
  else if n' > n then f'
  else if n == -1 then noFail
  else f



-------------------------
-- PL1 formula parsers --
-------------------------

-- The following parsers represent a grammar for formulae with precedences
-- as usually defined.

-- formula parser (PL1 syntax)

mainForParserPL :: Parser TokenPL Formula
mainForParserPL = do
  f <- forParser
  ftoken TokSemicolon
  return f

-- PL1 formula parser

forParser :: Parser TokenPL Formula
forParser = do
  f <- impParser
  g' <- forParser'
  case g' of
    Nothing -> return f
    Just g  -> return (conj [disj [neg f,g], disj [neg g, f]])

forParser' :: Parser TokenPL (Maybe Formula)
forParser' = do { ftoken TokEquiv; f <- forParser; return (Just f) }
  +++ return Nothing

-- implication parsers

impParser :: Parser TokenPL Formula
impParser = do
  f <- disjParser
  g' <- impParser'
  case g' of
    Nothing -> return f
    Just g  -> return (disj [neg f, g])

impParser' :: Parser TokenPL (Maybe Formula)
impParser' = do { ftoken TokImplies; f <- impParser; return (Just f) }
  +++ return Nothing

-- disjunction parsers

disjParser :: Parser TokenPL Formula
disjParser = do
  f <- conjParser
  g' <- disjParser'
  case g' of
    Nothing -> return f
    Just g  -> return (disj [f, g])

disjParser' :: Parser TokenPL (Maybe Formula)
disjParser' = do { ftoken TokOr; f <- disjParser; return (Just f) }
  +++ return Nothing

-- conjunction parsers

conjParser :: Parser TokenPL Formula
conjParser = do
  f <- quantParser
  g' <- conjParser'
  case g' of
    Nothing -> return f
    Just g  -> return (conj [f, g])

conjParser' :: Parser TokenPL (Maybe Formula)
conjParser' = do { ftoken TokAnd; f <-conjParser; return (Just f) }
  +++ return Nothing

-- quantified formula parsers

quantParser :: Parser TokenPL Formula
quantParser =
  do { ftoken TokOpenPar; ftoken TokExists;
       vs <- varParser `sepby1` commaParser;
       ftoken TokClosePar; f <- quantParser; return (foldr exists f vs) }
  +++ do ftoken TokOpenPar; ftoken TokForAll
         vs <- varParser `sepby1` commaParser
         ftoken TokClosePar; f <- quantParser; return (foldr forall f vs)
  +++ negParser

-- negation parser

negParser :: Parser TokenPL Formula
negParser =
  do { ftoken TokNot; f <- quantParser; return (neg f) }
  +++ primForParser

-- primitive formula parser
-- A primitive formula is either an atomic formula or any formula
-- within parenthesis.

primForParser :: Parser TokenPL Formula
primForParser =
  do { ftoken TokOpenPar; f <- forParser; ftoken TokClosePar; return f }
  +++ atomParser

-- atomic formula parser

atomParser :: Parser TokenPL Formula
atomParser = do { ftoken TokTrue; return verum }
  +++ do { ftoken TokFalse; return falsum }
  +++ do (TokPred p,_) <- ftokenp isTokPred "predicate"
         ftoken TokOpenPar
         ts <- termParser `sepby` commaParser
         ftoken TokClosePar
         return (literal (mkLit p ts))
  +++ do (TokPred p,_) <- ftokenp isTokPred "predicate"
         ts <- many termParser
         return (literal (mkLit p ts))
  +++ do t1 <- termParser
         (TokPred p,_) <- ftokenp isTokPred "predicate"
         t2 <- termParser
         return (literal (mkLit p [t1,t2]))
  +++ do t1 <- termParser
         ftoken TokNot
         (TokPred p,_) <- ftokenp isTokPred "predicate"
         t2 <- termParser
         return (neg (literal (mkLit p [t1,t2])))

-- auxiliary parsers

commaParser :: Parser TokenPL TokenPL
commaParser = do { t <- ftoken TokComma; return t }

termParser :: Parser TokenPL Term
termParser = do { (TokVar v,_) <- ftokenp isTokVar "term"; return (termFromVar v) }
  +++ do { (TokCon c,_) <- ftokenp isTokCon "term"; return (termFromCon c) }

varParser :: Parser TokenPL Variable
varParser = do (TokVar v,_) <- ftokenp isTokVar "variable"
               return v



----------------
-- RA parsers --
----------------

-- variable triple (x,y,z)
-- relations R are interpreted as xRy,
-- last variable z of var-triple is free for use

type VarTriple = (Variable, Variable, Variable)


-- parse RA expression:
-- {E_1,...,E_n} |- E
-- where E_i and E are RA equations

mainParserRA :: VarTriple -> Parser TokenRA Formula
mainParserRA (xyz@(x,y,_)) = do
  fs <- eqListParser xyz
  ftoken RATokTurnstile
  f <- eqParser xyz
  ftoken RATokSemicolon
  let res = if null fs then (forall x (forall y f))
            else (disj [exists x (exists y (disj (map neg fs))),
                        forall x (forall y f)])
  return res


-- parse an equation (one of three kinds):
-- L=R   - equality           (forall x,y)([L]xy <=> [R]xy)
-- L<R   - inclusion          (forall x,y)([L]xy -> [R]xy)
-- L     - equivalent to L=1  (forall x,y)([L]xy)

eqParser :: VarTriple -> Parser TokenRA Formula
eqParser xyz = do
  f <- rf1Parser xyz
  g' <- eqParser' xyz
  case g' of
    Nothing -> return f
    Just (RATokEquals, g) -> return (conj [disj [neg f, g], disj [neg g, f]])
    Just (RATokIncludes, g) -> return (disj [neg f, g])

eqParser' :: VarTriple -> Parser TokenRA (Maybe (FTokenRA, Formula))
eqParser' xyz =
  do (t,_) <- ftokenp (\t -> t == RATokEquals || t == RATokIncludes) "= or <"
     f <- rf1Parser xyz
     return (Just (t, f))
  +++ return Nothing

eqListParser :: VarTriple -> Parser TokenRA [Formula]
eqListParser xyz = eqParser xyz `sepby` (ftoken RATokComma)


-- parse a RA formula, priority level 1: relative sum, #
-- (# is right associative)

rf1Parser :: VarTriple -> Parser TokenRA Formula
rf1Parser (xyz@(x,y,z)) = do
  f <- rf2Parser xyz
  g' <- rf1Parser' xyz
  case g' of
    Nothing -> return f
    Just g  -> return (forall z (disj [exchange y z f, exchange x z g]))

rf1Parser' :: VarTriple -> Parser TokenRA (Maybe Formula)
rf1Parser' xyz = do { ftoken RATokRelSum; f <- rf1Parser xyz; return (Just f)}
  +++ return Nothing


-- parse a RA formula, priority level 2: absolute sum, +
-- (+ is right associative)

rf2Parser :: VarTriple -> Parser TokenRA Formula
rf2Parser xyz = do
  f <- rf3Parser xyz
  g' <- rf2Parser' xyz
  case g' of
    Nothing -> return f
    Just g  -> return (disj [f, g])

rf2Parser' :: VarTriple -> Parser TokenRA (Maybe Formula)
rf2Parser' xyz = do { ftoken RATokAbsSum; f <- rf2Parser xyz; return (Just f)}
  +++ return Nothing


-- parse a RA formula, priority level 3: relative product, @
-- (@ is right associative)

rf3Parser :: VarTriple -> Parser TokenRA Formula
rf3Parser (xyz@(x,y,z)) = do
  f <- rf4Parser xyz
  g' <- rf3Parser' xyz
  case g' of
    Nothing -> return f
    Just g  -> return (exists z (conj [exchange y z f, exchange x z g]))

rf3Parser' :: VarTriple -> Parser TokenRA (Maybe Formula)
rf3Parser' xyz = do { ftoken RATokRelProd; f <- rf3Parser xyz; return(Just f)}
  +++ return Nothing


-- parse a RA formula, priority level 4: absolute product, *
-- (* is right associative)

rf4Parser :: VarTriple -> Parser TokenRA Formula
rf4Parser xyz = do
  f <- rf5Parser xyz
  g' <- rf4Parser' xyz
  case g' of
    Nothing -> return f
    Just g  -> return (conj [f, g])

rf4Parser' :: VarTriple -> Parser TokenRA (Maybe Formula)
rf4Parser' xyz = do { ftoken RATokAbsProd; f <- rf4Parser xyz; return(Just f)}
  +++ return Nothing


-- parse a RA formula, priority level 5: negation (~) and conversion (^)

rf5Parser :: VarTriple -> Parser TokenRA Formula
rf5Parser (xyz@(x,y,_)) = do 
  let negParser = do { ftoken RATokNegation; return (\f -> neg f) }
  negtrans <- many negParser
  f <- rf6Parser xyz
  let convParser = do { ftoken RATokConverse; return (\f -> exchange x y f)}
  convtrans <- many convParser
  return (foldr id f (convtrans ++ negtrans))



-- parse a RA formula, priority level 6: atoms
-- (abs./rel. one, abs. zero, predicates, parenthesized expr.)

rf6Parser :: VarTriple -> Parser TokenRA Formula
rf6Parser (xyz@(x,y,_)) =
      do { ftoken RATokAbsZero; return falsum }
  +++ do { ftoken RATokAbsOne; return verum }
  +++ do { ftoken RATokRelOne; return (literal (mkLit eqPredRA [termFromVar x, termFromVar y])) }
  +++ do (RATokPred p,_) <- ftokenp isRATokPred "predicate"
         return (literal (mkLit p [termFromVar x, termFromVar y]))
  +++ do ftoken RATokOpenPar
         f <- rf1Parser xyz
         ftoken RATokClosePar
         return f



----------
-- misc --
----------

-- Apply a parser to a token stream, and return either the parse result
-- or a string denoting an error message.

papply :: Parser (Token t) v -> [Token t] -> Either String [(v,Int,[Token t])]
papply (Parser p) ts =
  let res = p ts 0
  in case res of
    None (PFail _ msg ts) ->
      Left ("Error at line " ++ show l ++ ", column " ++ show c ++ 
            ": " ++ msg)
         where ((_,(l,c)):_) = ts
    Parse l _ -> Right l


------------------------
-- printing functions --
------------------------

-- print a literal (with symbol table support)

showLit :: SymbolTable -> Literal -> ShowS
showLit st l =
  let p = getLitPred l; ts = getLitTerms l
      showTerms = showTL id '('
      showTL e _ [] = e
      showTL e c (t:ts) = showChar c . showTerm st t . 
                          showTL (showChar ')') ',' ts
  in showPred st p . showTerms ts

-- print a predicate (with symbol table support)

showPred :: SymbolTable -> Predicate -> ShowS
showPred st p =
  let mps = stLookupByPred p st; Just ps = mps
      pi = getPredIndex p
  in if mps == Nothing then showString "P_" . shows pi
     else showString ps

-- write predicate representation to a string (with symbol table support)

showPred' :: SymbolTable -> Predicate -> String
showPred' st p = showPred st p ""

-- show a constant (with symbol table support)

showCon :: SymbolTable -> Constant -> ShowS
showCon st c =
  let mcs = stLookupByCon c st; Just cs = mcs
      ci = getConIndex c
  in if mcs == Nothing then showString "c_" . shows ci
     else showString cs

-- show a variable (with symbol table support)

showVar :: SymbolTable -> Variable -> ShowS
showVar st v =
  let mvs = stLookupByVar v st; Just vs = mvs
      vi = getVarIndex v
  in if mvs == Nothing then showString "v_" . shows vi
     else showString vs

-- show a term (with symbol table support)

showTerm :: SymbolTable -> Term -> ShowS
showTerm st t
  | termIsVar t = let v = getTermVar t in showVar st v
  | termIsCon t = let c = getTermCon t in showCon st c

-- Show a formula (with symbol table support); the second parameter is
-- the precedence level, which should be set to zero on entry.

showFormula :: SymbolTable -> Int -> Formula -> ShowS
showFormula st p f
  | isVerum f = showString "true"
  | isFalsum f = showString "false"
  | isPosLit f = showLit st (getLiteral f)
  | isNeg f = showChar '~' . showFormula st 2 sf
  | isDisj f = showParen (p > 0) (showForList st 0 sfs " \\/ ")
  | isConj f = showParen (p > 1) (showForList st 1 sfs " /\\ ")
  | isExists f = showString "(exists " . showVar st x . showChar ')' .
                 showFormula st 2 sf
  | isForAll f = showString "(forall " . showVar st x . showChar ')' .
                 showFormula st 2 sf
 where sfs@(sf:_) = getSubFormulae f
       x = getQuantVar f

-- Show a list of formulae with a interspersed separator string.
-- The second parameter is used as above, the fourth is the separator
-- string printed between individual formulae.

showForList :: SymbolTable -> Int -> [Formula] -> String -> ShowS
showForList _ _ [] _ = error "empty formula list in showForList"
showForList st p (f:[]) _ = showFormula st p f
showForList st p (f:fs) s = showFormula st p f . showString s .
                            showForList st p fs s

-- print node info (with symbol table support for variables and constants)

showNodeInfo :: SymbolTable -> NodeInfo -> ShowS
showNodeInfo st ni
  | nodeInfoIsEmpty ni = id
  | otherwise = let gts = getStrategyInfo ni in showNIList "" gts
  where showNIList _ [] = id
        showNIList s ((g,mt):gts) = showString s . showChar '(' .
         shows (getGen g) . showChar ',' . case mt of
          { Nothing -> showString "||"; Just t -> showTerm st t } .
          showChar ')' . showNIList "," gts


-------------------------------
-- unification result output --
-------------------------------

-- (The following functions operate on a global IO state monad and are
-- mainly used for debugging purposes.)

-- show unification result (i.e. unification mapping and result)

showUResult :: IOState ()
showUResult = do
  (um,ur) <- getUnification
  showUR ur um
 where
   -- sub-routine
   showUR :: UResult -> UMap -> IOState ()
   showUR Fail _ = putStrLn' "<no solution>"
   showUR (Solution es) um = do
     putStr' "{ "
     showEQList "" es um
     putStr' " }"
   -- show list of unification equations
   showEQList :: String -> [UEquation] -> UMap -> IOState ()
   showEQList _ [] _ = done
   showEQList s (e:es) um = do
     putStr' s
     showEQ e um
     showEQList ", " es um
   -- show unification equation
   showEQ :: (UTerm, UTerm) -> UMap -> IOState ()
   showEQ (l,r) um = do
     showUTerm l um
     putStr' " -> "
     showUTerm r um
   -- show unification term
   showUTerm :: UTerm -> UMap -> IOState ()
   showUTerm ut um =
     let mti = uMapLookupByUTerm ut um; Just (t,pi) = mti
     in if mti == Nothing then putStr' "<undefd.>" else do
       st <- getSymbolTable
       putStr' (showTerm st t "")
       case pi of RootPath -> putStr' "#R"
                  PosLitPath -> putStr' "#P"
                  NegLitPath -> putStr' "#N"
                  Unknown -> done


----------------------------
-- formula draw functions --
----------------------------

-- write a formula as a tree to a string (with symbol table support)

drawWO :: SymbolTable -> Formula -> String
drawWO st f = draw' st 0 f ""
  where draw' st i f = showString (replicate i ' ') . draw'' st i f
        draw'' st i f
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf
          | isDisj f = showString "\\/\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isConj f = showString "/\\\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isExists f = showString "exists " . showVar st qv . showChar '\n' .
                         draw' st (i+3) sf
          | isForAll f = showString "forall " . showVar st qv . showChar '\n' .
                         draw' st (i+3) sf
          where qv = getQuantVar f
                l  = getLiteral f
                sfs = getSubFormulae f
                sf:_ = sfs


-- write a formula as a tree to a string, adjoin node information

drawWNI :: SymbolTable -> Formula -> String
drawWNI st f = draw' st 0 f ""
  where draw' st i f = showString (replicate i ' ') . draw'' st i f
        draw'' st i f
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showString sni . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showString sni . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf
          | isDisj f = showString "\\/\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isConj f = showString "/\\" . showString sni . showChar '\n' .
              foldr ((.) . draw' st (i+3)) id sfs
          | isExists f = showString "exists " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf
          | isForAll f = showString "forall " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf
          where qv = getQuantVar f
                l  = getLiteral f
                ni = getNodeInfo f
                sni= if nodeInfoIsEmpty ni then ""
                     else "   <" ++ showNodeInfo st ni "" ++ ">"
                sfs = getSubFormulae f
                sf:_ = sfs


-- write a formula to a string, adjoin node information and literal positions

drawWNILP :: SymbolTable -> Formula -> String
drawWNILP st f = draw' st 0 f lambda ""
  where draw' st i f p = showString (replicate i ' ') . draw'' st i f p
        draw'' st i f p
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showString sp . showString sni . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showString sp .
              showString sni . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf (mkPos p 0)
          | isDisj f = showString "\\/\n" .
              foldr (\(f,j) s -> draw' st (i+3) f (mkPos p j) . s) id sfs'
          | isConj f = showString "/\\" . showString sni . showChar '\n' .
              foldr (\(f,j) s -> draw' st (i+3) f (mkPos p j) . s) id sfs'
          | isExists f = showString "exists " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf (mkPos p 0)
          | isForAll f = showString "forall " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf (mkPos p 0)
          where qv = getQuantVar f
                l  = getLiteral f
                ni = getNodeInfo f
                sni= if nodeInfoIsEmpty ni then ""
                     else "   <" ++ showNodeInfo st ni "" ++ ">"
                sp = "  " ++ show p
                sfs = getSubFormulae f
                sfs' = zip sfs ([0..length sfs-1])
                sf:_ = sfs
