----------------------------------------------------------------------
-- Module GlobalState.hs                                            --
--                                                                  --
-- This module implements an instance of the IO state monad         --
-- representing the global state of the RPC prover.     	    --
-- Moreover, it provides several data types to encode proof         --
-- options.							    --
--                                                                  --
-- 11.03.97  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module GlobalState (
  -- data types
  Step,			-- proof step number
  SplitType (		-- subgoal generation type
    NoSplitting,
    RootSplitting,
    CompleteSplitting,
    RecursiveSplitting
  ),
  OutputMode (		-- output mode specifier
    OMTerse,
    OMStandard,
    OMVerbose
  ),
  ProofOptions (	-- proof option data structure
    MkProofOptions,
    printing, debugging, splitting,
    equalModRenSteps,
    unusedLitSteps,
    outerAQuants
  ),
  PreProcessOptions (	-- options for formula pre-processing
    MkPPOptions,
    ppSkolemization,
    ppSplitting
  ),
  GlobalState (		-- the global state data structure
    MkGlobalState,
    currentProofStack,
    totalProofSteps,
    totalProofGenerations,
    totalProofTime,
    subProofTime,
    proofOptions,
    maxSteps,
    formula,
    generation,
    step,
    maxVar,
    posLitPosition,
    negLitPosition,
    unificationMap,
    unificationResult,
    symbolTable,
    randomNumbers,
    lastAK3info
  ),
  IOState,		-- specialization of IO state monad to global state

  -- functions to extract parts of the global state
  getProofOptions, getWorkingFormula, getGeneration,
  getStep, getMaxVar, getMaxSteps, getSubProofTime,
  getLiteralPair, getUnification, getWorkingState, getLitPairState,
  getProofStack, getTotalProofSteps, getTotalProofGenerations,
  getTotalProofTime, getSymbolTable, getLastAK3Info,

  -- functions to manipulate the global state
  setProofStack, updateProofStack, setTotalProofSteps, setTotalProofGenerations,
  setTotalProofTime, setSubProofTime, setMaxSteps, setProofOptions,
  setWorkingFormula, setGeneration, setStep, setMaxVar, setLiteralPair,
  setUnification, setWorkingState, setLitPairState, nextGeneration,
  nextStep, updateWorkingFormula, increaseTotalProofGenerationsBy,
  increaseTotalProofStepsBy, increaseTotalProofTimeBy, increaseSubProofTimeBy,
  updateMaxSteps, setSymbolTable, setLastAK3Info,

  -- auxiliary functions for I/O on global state monad
  putStr', putStrLn', readLn', newLine'
) where


-------------------
----- imports -----
-------------------

import Formula
import ProofStack
import IOStateMonad
import Unification
import Auxiliary
import Scanner


----------------------
----- data types -----
----------------------

-- number of proof step

type Step = Int


-- SplitType indicates how subgoals are created:
-- o NoSplitting: No generation of subgoals
-- o RootSplitting: Whenever the top operator of the formula to prove is
--   a conjunction splitting of the formula into the subformulae of the
--   disjunction is performed.
-- o CompleteSplitting: Same as RootSplitting, after having applied
--   a maximal series of distribution rule applications to the formula.
--   (I.e. A \/ (B /\ C) --> (A \/ B) /\ (A \/ C)).
-- o RecursiveSplitting: Only sensible in conjunction with partial
--   skolemization. In this case splitting steps followed by a skolemization
--   steps are performed until no further subgoals can be generated.

data SplitType = NoSplitting
               | RootSplitting
               | CompleteSplitting
               | RecursiveSplitting
  deriving (Eq, Ord, Show)


-- output mode for proofs

data OutputMode = OMTerse | OMStandard | OMVerbose
  deriving (Eq, Ord, Show)


-- Proof options consist of:
-- o splitting: a splitting type, as described above,
-- o equalModRenSteps: the number of steps, where-after subformulae
--   equal modulo renaming of bound variables are deleted.
-- o unusedLitSteps: the number of steps, where-after unused literals
--   are removed.
-- o outerAQuants: whether or not to delete outermost universal quantifiers
-- o printing: an output mode for proofs
-- o debugging: whether or not to print debugging messages

data ProofOptions = MkProofOptions {
  splitting :: SplitType,
  equalModRenSteps :: Int,
  unusedLitSteps :: Int,
  outerAQuants :: Bool,
  printing :: OutputMode,
  debugging :: Bool
}
  deriving (Eq, Show)


-- Pre-process optinos consist of:
-- o ppSkolemization: whether or not to perform a partial skolemization
--   before starting the proof
-- o ppSplitting: a splitting type, as indicated above

data PreProcessOptions = MkPPOptions {
  ppSkolemization :: Bool,
  ppSplitting :: SplitType
}
  deriving (Eq, Show)


-- The global state of the prover:

data GlobalState = MkGlobalState {
  -- proof management variables:
  currentProofStack :: ProofStack (Generation, Step),	-- subgoal stack
  totalProofSteps :: Int,	  			-- # of major proof steps
  totalProofGenerations :: Int,				-- # of RPC reductions
  totalProofTime :: Integer,				-- total proof time
  subProofTime :: Integer,				-- total time for sub-proof
  proofOptions :: ProofOptions,				-- current proof options
  maxSteps :: Int,					-- maximal # of steps for proof

  -- variables for different reduction strategies
  formula :: Formula,					-- current formula to prove
  generation :: Generation,				-- current generation
  step :: Step,						-- current major proof step
  maxVar :: Int,  					-- highest usable variable index

  -- variables for literal pair reduction strategy
  posLitPosition :: Position,				-- pos. of positive literal
  negLitPosition :: Position,				-- pos. of negative literal
  unificationMap :: UMap,				-- mapping UResult -> Vars,Cons
  unificationResult :: UResult,				-- unification result

  -- symbol table for parser
  symbolTable :: SymbolTable,				-- parser symbol table
  
  -- random number sequence
  randomNumbers :: [Int],				-- used by some reduction strat.

  -- information about last AK3 rule
  lastAK3info :: Maybe (Position, Term)
}


-- IO state monad where the state variable is substitued by global state.

type IOState a = IOStateMonad GlobalState a


------------------------------------------------
-- functions to extract parts of global state --
------------------------------------------------

getProofOptions :: IOState ProofOptions
getProofOptions = fetchWith (\state -> proofOptions state)

getWorkingFormula :: IOState Formula
getWorkingFormula = fetchWith (\state -> formula state)

getGeneration :: IOState Generation
getGeneration = fetchWith (\state -> generation state)

getStep :: IOState Step
getStep = fetchWith (\state -> step state)

getMaxVar :: IOState Int
getMaxVar = fetchWith (\state -> maxVar state)

getMaxSteps :: IOState Int
getMaxSteps = fetchWith (\state -> maxSteps state)

getSubProofTime :: IOState Integer
getSubProofTime = fetchWith (\state -> subProofTime state)

getLiteralPair :: IOState (Position, Position)
getLiteralPair = fetchWith (\state -> (posLitPosition state, negLitPosition state))

getUnification :: IOState (UMap, UResult)
getUnification = fetchWith (\state -> (unificationMap state, unificationResult state))

getWorkingState :: IOState (Formula, Int, Generation, Step)
getWorkingState = fetchWith (\state -> (formula state, maxVar state,
                                         generation state, step state))

getLitPairState :: IOState (Formula, Int, Generation, Step, Position, Position,
                            UMap, UResult)
getLitPairState = do
  fetchWith (\state -> (formula state, maxVar state, generation state, step state,
                        posLitPosition state, negLitPosition state,
                   	unificationMap state, unificationResult state))

getProofStack :: IOState (ProofStack (Generation, Step))
getProofStack = fetchWith (\state -> currentProofStack state)

getTotalProofSteps :: IOState Step
getTotalProofSteps = fetchWith (\state -> totalProofSteps state)

getTotalProofGenerations :: IOState Int
getTotalProofGenerations = fetchWith (\state -> totalProofGenerations state)

getTotalProofTime :: IOState Integer
getTotalProofTime = fetchWith (\state -> totalProofTime state)

getSymbolTable :: IOState SymbolTable
getSymbolTable = fetchWith (\state -> symbolTable state)

getLastAK3Info :: IOState (Maybe (Position, Term))
getLastAK3Info = fetchWith (\state -> lastAK3info state)


---------------------------------------------------
-- functions to manipulate parts of global state --
---------------------------------------------------

setProofStack :: ProofStack (Generation, Step) -> IOState ()
setProofStack ps = update (\state -> state { currentProofStack = ps })

updateProofStack :: (ProofStack (Generation, Step) -> ProofStack (Generation, Step))
  -> IOState ()
updateProofStack f = update (\state -> state { currentProofStack =
					 f (currentProofStack state) })

setTotalProofSteps :: Int -> IOState ()
setTotalProofSteps s = update (\state -> state { totalProofSteps = s} )

setTotalProofGenerations :: Int -> IOState ()
setTotalProofGenerations g = update (\state -> state { totalProofGenerations = g })

setTotalProofTime :: Integer -> IOState ()
setTotalProofTime t = update (\state -> state { totalProofTime = t })

setSubProofTime :: Integer -> IOState ()
setSubProofTime t = update (\state -> state { subProofTime = t })

setMaxSteps :: Int -> IOState ()
setMaxSteps ms = update (\state -> state { maxSteps = ms })

setProofOptions :: ProofOptions -> IOState ()
setProofOptions po = update (\state -> state { proofOptions = po })

setWorkingFormula :: Formula -> IOState ()
setWorkingFormula f = update (\state -> state { formula = f })

setGeneration :: Generation -> IOState ()
setGeneration g = update (\state -> state { generation = g })

setStep :: Step -> IOState ()
setStep s = update (\state -> state { step = s })

setMaxVar :: Int -> IOState ()
setMaxVar mv = update (\state -> state { maxVar = mv })

setLiteralPair :: Position -> Position -> IOState ()
setLiteralPair p p' =
  update (\state -> state { posLitPosition = p, negLitPosition = p' })

setUnification :: UMap -> UResult -> IOState ()
setUnification um ur =
  update (\state -> state { unificationMap = um, unificationResult = ur })

setWorkingState :: Formula -> Int -> Generation -> Step -> IOState ()
setWorkingState f mv g s =
  update (\state -> state { formula = f, maxVar = mv, generation = g, step = s })

setLitPairState :: Formula -> Int -> Generation -> Step -> Position -> Position
  -> UMap -> UResult -> IOState ()
setLitPairState f v g s p p' um ur =
  update (\state -> state { formula = f, maxVar = v,
                            generation = g, step = s,
                            posLitPosition = p, negLitPosition = p',
                            unificationMap = um, unificationResult = ur })

nextGeneration :: IOState ()
nextGeneration = update (\state -> state { generation =
                                      nextGen (generation state) })

nextStep :: IOState ()
nextStep = update (\state -> state { step = step state + 1 })

updateWorkingFormula :: (Formula -> Formula) -> IOState ()
updateWorkingFormula f =
  update (\state -> state { formula = f (formula state) })

increaseTotalProofGenerationsBy :: Int -> IOState ()
increaseTotalProofGenerationsBy x =
  update (\state -> state { totalProofGenerations = totalProofGenerations state + x })

increaseTotalProofStepsBy :: Int -> IOState ()
increaseTotalProofStepsBy x =
  update (\state -> state { totalProofSteps = totalProofSteps state + x })

increaseTotalProofTimeBy :: Integer -> IOState ()
increaseTotalProofTimeBy x =
  update (\state -> state { totalProofTime = totalProofTime state + x })

increaseSubProofTimeBy :: Integer -> IOState ()
increaseSubProofTimeBy x =
  update (\state -> state { subProofTime = subProofTime state + x })

updateMaxSteps :: (Int -> Int) -> IOState ()
updateMaxSteps f =
  update (\state -> state { maxSteps = f (maxSteps state) })

setSymbolTable :: SymbolTable -> IOState ()
setSymbolTable st =
  update (\state -> state { symbolTable = st })

setLastAK3Info :: Maybe (Position, Term) -> IOState ()
setLastAK3Info mpt =
  update (\state -> state { lastAK3info = mpt })


-------------------------
-- auxiliary functions --
-------------------------

-- similar to those defined in the standard prelude, but for IO state monad

putStr' :: String -> IOState ()
putStr' = printWith putStr

putStrLn' :: String -> IOState ()
putStrLn' = printWith putStrLn

readLn' :: Read a => IOState a
readLn' = readWith readLn

newLine' :: IOState ()
newLine' = doIO newLine
