----------------------------------------------------------------------
-- Module IOStateMonad.hs                                           --
--                                                                  --
-- This module implements a combined state/IO monad with            --
-- built-in exception handling.                                     --
--                                                                  --
-- 24.01.97  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module IOStateMonad (
  -- data types
  IOSErrCode, IOSError, -- IO state monad error type
  IOStateMonad,         -- main IO state monad type
  -- functions on IOStateMonads
  startingWith,       	-- initialization routine for monad
  fetch,		-- get current state
  fetchWith,		-- get current state and apply function to it
  update,		-- update current state by function
  set,			-- set current state to constant value
  printWith,		-- promote output function from IO monad
  readWith,		-- promote input function from IO monad
  doIO,			-- promote IO function from IO monad
  doIOWithErrCode,	--  '' with specific error code on failure
  -- exception handling
  failS,		-- raise an exception
  catchS,		-- catch an exception
  iosError,		-- error constructor
  getIOSErrCode,	-- extract error code from caught exception
  getIOSErrMsg,		-- extract error message from caught exception
  ($$),			-- apply function to avoid parentheses
  -- generic monad functions
  done, fold1, foldR, foldL
--  unless, when
) where


-------------------
----- imports -----
-------------------

import IO

--------------------
----- notation -----
--------------------

-- left associative apply function with low priority

infixl 0 $$


----------------------
----- data types -----
----------------------

-- An exception for the IO/state monad consists of an error code
-- and an error message.

type IOSErrCode = Int
data IOSError = IOSError IOSErrCode String
  deriving (Eq, Show)

-- The main IO/state monad type:

data IOSException a = Return a | Raise IOSError
  deriving (Eq, Show)
newtype IOStateMonad s a = IOSt (s -> IO (IOSException a, s))


---------------------------------
----- instance declarations -----
---------------------------------

instance Functor (IOStateMonad s) where
  fmap f (IOSt m) = IOSt $ \s -> do (x,s') <- m s
                                    case x of
                                      Raise e  -> return (Raise e, s')
                                      Return y -> return (Return (f y), s')

instance Monad (IOStateMonad s) where
  return x   = IOSt (\s -> return (Return x, s))
  IOSt m >>= f = IOSt $ \s -> do (x,s') <- m s
                                 case x of
                                   Raise e  -> return (Raise e, s')
                                   Return y -> do let IOSt f' = f y
                                                  f' s'

------------------------------
----- exported functions -----
------------------------------

-- functions for IOStateMonads

-- initialization of state monad 

startingWith :: IOStateMonad s a -> s -> IO a
(IOSt m) `startingWith` v = do (x,s') <- m v
                               case x of
                                 Raise e  -> do let IOSError _ msg = e
                                                fail msg
-- Haskell 1.4 :                                fail (userError msg)
                                 Return y -> return y

-- fetch current state

fetch :: IOStateMonad s s
fetch = IOSt (\s -> return (Return s,s))

-- fetch current state and apply function to result

fetchWith :: (s -> a) -> IOStateMonad s a
fetchWith f = IOSt (\s -> return (Return (f s), s))

-- update current state by function

update :: (s -> s) -> IOStateMonad s ()
update f = IOSt (\s -> return (Return (), f s))

-- set current state to constant value

set :: s -> IOStateMonad s ()
set s' = IOSt (\s -> return (Return (), s'))


-- IO functions

printWith :: Show a => (a -> IO ()) -> a -> IOStateMonad s ()
printWith pf x = doIO (pf x)

readWith :: Read a => IO a -> IOStateMonad s a
readWith = doIO

doIO :: IO a -> IOStateMonad s a
doIO = doIOWithErrCode 0

doIOWithErrCode :: IOSErrCode -> IO a -> IOStateMonad s a
doIOWithErrCode ec f =
  IOSt $ \s -> catch $$ do x <- f
                           return (Return x, s)
                     $$ \e -> let msg = ioeGetErrorString e
                                  err = IOSError ec msg
                              in return (Raise err, s)

------------------------
-- exception handling --
------------------------

-- raise an exception

failS :: IOSError -> IOStateMonad s a
failS e = IOSt (\s -> return (Raise e, s))

-- catch an exception

catchS :: IOStateMonad s a -> (IOSError -> IOStateMonad s a) -> IOStateMonad s a
catchS (IOSt m) h = IOSt $ \s -> do (x,s') <- m s
                                    case x of
                                      Raise e  -> let IOSt y = h e in y s
                                      Return y -> return (x, s')

-- error (exception) constructor for IO state monads

iosError :: IOSErrCode -> String -> IOSError
iosError c msg = IOSError c msg

-- extraction of error code and message from exception

getIOSErrCode :: IOSError -> IOSErrCode
getIOSErrCode (IOSError c _) = c

getIOSErrMsg :: IOSError -> String
getIOSErrMsg (IOSError _ m) = m

-- apply function

($$) :: (a -> b) -> a -> b
f $$ x = f x


-----------------------------
-- generic monad functions --
-----------------------------

fold1 :: Monad m => (a -> m b) -> [a] -> m b
fold1 f [x] = f x
fold1 f (x:xs) = f x >> fold1 f xs

foldL :: Monad m => (a -> b -> m a) -> a -> [b] -> m a
foldL f a [] = return a
foldL f a (x:xs) = do y <- f a x
                      foldL f y xs

foldR :: Monad m => (a -> b -> m b) -> b -> [a] -> m b
foldR f a [] = return a
foldR f a (x:xs) = do y <- foldR f a xs
                      f x y

done :: Monad m => m ()
done = return ()

-- unless :: Monad m => Bool -> m () -> m ()
-- unless p s = if p then done else s

-- when :: Monad m => Bool -> m () -> m ()
-- when p s = if p then s else done
