----------------------------------------------------------------------
-- Module LiteralPair.hs					    --
--                                                                  --
-- Implements functions to perform reductions that make a unifiable --
-- literal pair vanish.						    --
--                                                                  --
-- 16.01.97  initial revision                                       --
----------------------------------------------------------------------

--------------------
----- comments -----
--------------------

-- A complete literal pair reduction makes a specific literal pair
-- vanish. This is accomplished by a series of appropriate reductions.
-- New subformulae (and thus literals) may arise during this process, however.
-- The entry point and process manager is the function `processLitPath' (resp.
-- `preProcessRootPath' and `postProcessRootPath'), which repeatedly calls
-- one of the following functions:
-- o litDFlatten (flattening of disjunctions)
-- o litCReduce (distributive law application, conjunction handling)
-- o litAReduce (universal quantifier (forall) rule application)
-- o litEReduce (existential quantifier rule application)
-- These functions also determine suitable reduction terms and variables;
-- they call in turn functions of the next lower level:
-- `applyDRuleX', `applyARuleX' and `applyERuleX'.
-- The functions on this level are responsible for updates of the literals'
-- positions and of the unification result and mapping.
-- These extended reduction functions (apply?RuleX) call the corresponding
-- reduction functions on the underlying level (in module `Reduction'),
-- which perform the reduction process proper.
-- Thus, the whole process incorporates a four-level model:
-- 1. manager (processLitPath,...)
-- 2. main reduction functions (litDFlatten, litCReduce,...)
-- 3. extended reduction rule application (applyDRuleX,...)
-- 4. normal reduction rule application (applyRule)
-- A complete literal pair unification step can be performed by successively
-- calling `preProcessRootPath', `processLitPath' and `postProcessRootPath'.


----------------------------
----- module interface -----
----------------------------

module LiteralPair (
  -- data types
  LitPairSelector,	-- select a unifiable literal pair for unification
  TermSelector,		-- choose a term as reduction parameter for unification
  TermSelType (
    AnyVar, AnyTerm, AnyOf
  ),
  DisjSelectType (	-- which part of a disjunction takes part in a reduction?
    MaxDSel, MinDSel, NormDSel
  ),

  -- operations
  processLitPath,	-- main unification process (works on literal paths)
  preProcessRootQuant,	-- unification pre-processing (works on root path)
  postProcessRootQuant,	-- unification post-processing (works on root path)

  -- auxiliary functions
  getFormulaeAndPositions
) where


----------------------------
----- imported modules -----
----------------------------

import Set
import Auxiliary
import Formula
import List
import IOStateMonad
import Reduction
import Simplify
import Unification
import GlobalState
import Monad


----------------------
----- data types -----
----------------------

-- A literal pair selector chooses a literal pair for unification.
-- Literal pairs are searched for in the current working formula stored in
-- the global state monad. The positions (p,p') of the positive and negative
-- literal, respectively, are returned.

type LitPairSelector = IOState (Position, Position)

-- A term selector is a function that takes a formula position p and a
-- term selector type (see below), and returns a term that may be used as
-- a reduction parameter for reduction at position p, while fulfilling
-- the conditions imposed by the term selector type.

type TermSelector = Position -> TermSelType -> IOState Term

-- The term selector type distinguishes three alternatives:
-- o AnyVar:  any variable is suitable as a reduction term (but no constant)
-- o AnyTerm: any variable or constant may be used as a reduction term
-- o AnyOf ts: any term of term list ts may be used as a reduction term

data TermSelType = AnyVar | AnyTerm | AnyOf [Term]

-- The disjunction selector type is used for similar purposes as a formula
-- list partitioner, but it does not allow that degree of freedom.
-- Three pre-defined disjunction selectors are available:
-- o MaxDSel (maximal disjunction selector): take the whole disjunction for
--   reduction
-- o MinDSel (minimal disjunction selector): take only that part of the
--   disjunction containing a literal of the current literal pair for
--   reduction
-- o NormDSel (normal disjunction selector): include those formulae of the
--   disjunction into the redex, which fulfill some literal condition.
--   This condition is specified below.

data DisjSelectType = MaxDSel | MinDSel | NormDSel
  deriving (Eq, Show)

-- The following two auxiliary data types are used to determine which literal
-- is relevant in a reduced formula, where two or more possibilities arise.
-- This is the case for existential and universal quantifier reductions.

-- In case of existential quantifier reduction there are two alternatives:
-- [reduction rule is, say:  (exists x)A  -->  (exists x)A \/ A']
-- o NewE: relevant position is in A'
-- o OldE: relevant position is in A

data LitFollowE = NewE | OldE
  deriving (Eq, Show)

-- In case of universal quantifier reduction there are (at most) three
-- alternatives:
-- [reduction rule is one of, say:
--  (UQR1)  (forall x)A \/ B  -->  (forall x)A \/ A'' \/ (forall y)((forall y)B' \/ A') \/ B
--  (UQR2)  (forall x)A       -->  (forall x)A \/ A'
--  (UQR3)  [not considered, as there is only one position alternative]]
-- o OldA: relevant position is in A or B
-- o NewA: relevant positon is in A' or B'
-- o DelA: relevant position is in A''

data LitFollowA = OldA | DelA | NewA
  deriving (Eq, Show)



------------------------------
----- exported functions -----
------------------------------

-- "processLitPath ts ds" performs reduction steps in order to unify two literals.
-- ts is a term selector, ds a disjunction selector type.
-- Reduction steps are performed on the current working formula. The formula, as
-- well as the literal pair to be unified, is stored in the global state monad.
-- The literal pair stored in the state monad must be unifiable and fulfill two
-- additional conditions, which hold after having called `preProcessRootQuant':
-- 1. The unification solution must not contain a substitution (resp. equation)
--    of the form x = f(x') where x and x' are complementary variables.
--    (Complementary variables are unification variables which differ only by sign.)
-- 2. The unification solution must not contaion a pair of substitutions of the
--    form x = t, x' = t' with t /= t', where x and x' are complementary variables.
-- This algorithm performs several reduction steps such that the literals
-- are moved from their original positions to neighbour positions (i.e. they
-- occur in a disjunction as L \/ L').
-- It may be the case, however, that ~L /= L'. This is only allowed if
-- by a series of existential quantifier reductions on the literal pair's
-- root path the literals can be unified (This is performed by algorithm
-- `postProcessRootQuant').
-- The original literal positions as well as the literal positions of each intermediate
-- step are marked with the current step number.
-- This is made in order to avoid multiple unifications with the same literal
-- pair.
-- On termination of this algorithm, the working formula is the reduced formula
-- as described above, and the positions of the literal pair are neighbour
-- positions.
--
-- NOTES:
-- 1. Reductions are performed outermost first with priority to universal quantifiers.
--    Further information on which reducion path (positive or negative literal)
--    is chosen, should be extracted from the algorithm below.
-- 2. The working formula produced by this algorithm is neither simplified nor
--    flat.
-- 3. The unification result and unification map of the state monad may be modified
--    in order to reflect changes induced by existential quantifier reductions.
-- 4. Literal postions and the generation stored in the global state monad are
--    updated.

processLitPath :: TermSelector -> DisjSelectType -> IOState ()
processLitPath tSel dSel =
  do (f,_,_,_,p,p',_,_) <- getLitPairState
     opt <- getProofOptions
     -- mark literals
     markLiterals
     -- find outermost reduction positions on pos. and neg. literal path
     let rootp = commonPrefix p p'; rl = length (posToList rootp)
         rp1 = listToPos (take (rl+1) (posToList p))   -- red. position on pos. lit. path
         rp2 = listToPos (take (rl+1) (posToList p'))  -- red. position on neg. lit. path
     -- select next reduction step 
         sf1 = f@@rp1; sf2 = f@@rp2
     -- select position and reducer
     if isLiteral sf1 && isLiteral sf2 then done  -- finished
      else do
       -- consider disjunctions and conjunctions first
       if isDisj sf1 then litDFlatten rp1
        else if isDisj sf2 then litDFlatten rp2
        else if isConj sf1 then litCReduce rp1 dSel
        else if isConj sf2 then litCReduce rp2 dSel
        -- only one quantifier or at least one universal
        else if isForAll sf1 then litAReduce rp1 PosLitPath tSel dSel
        else if isForAll sf2 then litAReduce rp2 NegLitPath tSel dSel
        else if isExists sf1 && not (isExists sf2) then
          litEReduce rp1 PosLitPath tSel
        else if isExists sf2 && not (isExists sf1) then
          litEReduce rp2 NegLitPath tSel
        else do
          -- two existential quantifiers
          let -- skipping of unused quantifiers
              rp1' = listToPos (take (rl+2) (posToList p))  -- subfor. positions of
              rp2' = listToPos (take (rl+2) (posToList p')) -- exist. quantifiers
              plv = getLitVars (getLiteral (f@@p))	    -- pos. lit.'s variables
              nlv = getLitVars (getLiteral (f@@p'))	    -- neg. lit.'s variables
              pbv = quantVarsOnPath f rp1' p	-- bound variables on paths to pos.
              nbv = quantVarsOnPath f rp2' p'   -- resp. neg. literal
              px = getQuantVar (f@@rp1)		-- bound vars at exist. quant. reduction
              nx = getQuantVar (f@@rp2)		-- positions
              -- skipping is allowed if no literal variable is affected
              pskip = not (px `isElement` plv) || px `isElement` pbv
              nskip = not (nx `isElement` nlv) || nx `isElement` nbv
              -- reveal universal quantifiers: puq resp. nuq are true iff. there
              -- is a universal quantifier on the path to the literal
              puq = not (pathIsEmpty (filterPath isForAll (getPath f rp1' p)))
              nuq = not (pathIsEmpty (filterPath isForAll (getPath f rp2' p')))
              -- choose reduction position, criteria for selection:
              crp = if pskip then rp1       -- 1. skipping on pos. lit. path
                    else if nskip then rp2  --    skipping on neg. lit. path
                    else if nuq then rp2    -- 2. revelation of universal
                    else if puq then rp1    --    quantifiers
                    else rp1		    -- 3. arbitrary, here: pos. lit. path
              pi = if crp == rp1 then PosLitPath else NegLitPath -- path info of red.
          litEReduce crp pi tSel
       -- perform next reduction step
       processLitPath tSel dSel


-- "preProcessRootQuant ts" performs reduction steps on the root path of
-- a unifiable literal. ts is a term selector.
-- Pre-processing on the root path is the first step in a literal unification
-- process. This step is needed to handle some special cases (depending on the
-- unification equation system solution) which are not considered in the
-- (supposedly) following main unification process. These cases are:
-- 1. The unification solution contains a substitution of the form x = f(x')
--    where x and x' are complementary variables.
-- 2. The unification solution contains a pair of substitutions of the form
--    x = t, x' = t' with t /= t', where x and x' are complementary variables.
-- To circumvent these cases a suitable existential quantifier reduction
-- is performed on the root path.
-- Moreover, literal positions, the unification result and mapping, and the
-- working formula are updated. The resulting working formula is not necessarily
-- flat or simplified. The literal positions of the original working formula
-- are marked with the current step number (stored in the state monad)
-- in order to avoid multiple unifications with the same literal pair.
-- When neither of the aforementioned two cases holds, no reduction is performed
-- and no updates are made.

preProcessRootQuant :: TermSelector -> IOState ()
preProcessRootQuant termSelector =
  do (f,_,_,_,p,p',umap,ur) <- getLitPairState
     let rootp = commonPrefix p p'  -- position of the connecting disjunction
         -- reduction necessary? if so, rrp is a pair (p,ts) containing the
         -- (existential quantifier) reduction position p and the list of
         -- admissable reduction terms ts. ts == Nothing indicates arbitrary
         -- reduction term.
         rrp = findRootReductionPosition f rootp ur umap
     if rrp == Nothing then done	        -- no reduction necessary
      else do  -- perform reduction no root path
        -- mark literals
        markLiterals
        -- prepare reduction step
        let Just (rp,ts) = rrp 			-- get red. position and term
            pVars = quantVarsOnPath f rootp p	-- get bound variables on pos.
            nVars = quantVarsOnPath f rootp p'	-- and neg. literal path
            x = getQuantVar (f@@rp)		-- get bound var. x at reduction pos.
            vfp = vars (f@@p)		-- get positive and negative literals'
            vfp' = vars (f@@p')		-- variables
            -- compute red. variable information:
            -- xin[p/n]: variable x occurs in pos./neg. literal
            -- xb[p/n]: variable x is bound on pos./neg. literal path
            -- NOTE: rp is the innermost root position binding variable x
            xinp = x `isElement` vfp; xbp = x `isElement` pVars
            xinn = x `isElement` vfp'; xbn = x `isElement` nVars
        -- compute reduction term
            -- search unification term for variable x
            -- NOTE: 1. this never fails, due to `findRootReductionPosition'
            --       2. the result is not unique, the sign is a degree of freedom
            Just xut = uMapLookupByTermAndPathInfo (termFromVar x)
                                                   RootPath umap
            -- unification term may be positive or negative, so we differentiate:
            pxut = absUTerm xut; nxut = negUTerm pxut
            -- lookup substitution terms for x in UResult (for variable doublet case)
            Just stp = uResultLookup pxut ur
            Just stn = uResultLookup nxut ur
            -- compute critical terms, i.e. dependent terms of unification
            -- operator (resp. universal quantifier)
            ctp = map absUTerm (getUOpVars stp)
            ctn = map absUTerm (getUOpVars stn)
            -- is x (as unification var.) in critical terms? if not, reduction term is ok
            t1ok = pxut `notElem` ctp
            t2ok = pxut `notElem` ctn  -- pxut used because ctn terms have pos. sign
            -- compute reduction term list
            ts' = if ts == Nothing then AnyTerm  -- occur check of compl. var
                  else let Just (t1,t2) = ts     -- variable doublet case
                       in if t1ok && t2ok then AnyOf [t1,t2]
                          else if t1ok then AnyOf [t1] else if t2ok then AnyOf [t2]
                          else error ("no appropriate reduction term in " ++
                                      "preProcessRootQuant")
        -- choose reduction term
        rt <- termSelector rp ts'
        -- which literals are relevant in new formula?
        let posLitNew =
              if ts == Nothing then  -- occur check case:
                -- relevant positive literal position is in new formula if variable x
                -- occurs free in negative literal
                if xinp && not xbp then False
                else if xinn && not xbn then True
                else error "reduce error in preProcessRootQuant"
              else  -- variable doublet case
                let Just (frt,_) = ts in t1ok && rt == frt
            (pFollow, nFollow) = if posLitNew then (NewE, OldE) else (OldE, NewE)
        -- compute whether to update UResult or not: if occurrence x is innermost
        -- bound occurrence then we have to update the uResult.
            updURes = if pFollow == NewE then not xbp else not xbn
        -- perform reduction
        opt <- getProofOptions
        let om = printing opt
        if om == OMVerbose then printWith putChar 'E' else done
        applyERuleX rp rt pFollow nFollow updURes


-- "postProcessRootQuant" performs existential quantifier reductions on the root
-- path. Post-processing is the third step in a literal unification process.
-- On entry, the literal pair's positions (p,p') stored in the state monad
-- must be neighbour positions, i.e. init p = init p'.
-- Moreover, the literals (l,l') at positions (p,p') must be convertible
-- to l \/ l' where ~l = l' by a series of existential quantifier reductions
-- on the root path.
-- The working formula, unification map and generation information are updated,
-- literals before each reduction are marked.
-- The resulting working formula (which is also stored in the global state monad)
-- is neither flat nor simplified.

postProcessRootQuant :: IOState ()
postProcessRootQuant =
  do (f,_,_,_,p,p',umap,_) <- getLitPairState
     -- get literals
     let fp = f@@p; fp' = f@@p'
     if nnf (neg fp) == fp' then done  -- (L,~L) pair already here
      else do -- prepare reductions
        let -- position of connecting disjunction
            rootPos = listToPos (init (posToList p))
            -- get formulae on root path with their positions
            rootFors = getFormulaeAndPositions f lambda rootPos
            -- only existential quantifiers are relevant
            redPosEQ = filter (isExists . fst) rootFors
            -- delete outermost multiple occurrences of same variable
            redPosL = map snd (remOMVar redPosEQ)
        -- perform reductions (innermost first)
        foldR ppReducer () redPosL
  where ppReducer :: Position -> () -> IOState ()
        ppReducer redPos () = do
          (f,_,_,_,p,p',umap,ur) <- getLitPairState
          if neg (f@@p) == f@@p' then done  -- (L,~L) pair reached
           else do
            -- perform existential quantifier reduction
            markLiterals
            opt <- getProofOptions
            let -- compute reduction term
                x = getQuantVar (f@@redPos)
                mrt = computeERedTerm x RootPath False ur umap; Just rt = mrt
                om = printing opt
            when (om == OMVerbose) $ printWith putChar 'E'
            applyERuleX redPos rt NewE NewE False
        -- remove outermost multiple occurrences of same variable
        -- input list is sorted outermost - innermost, only existential
        -- quantifier formulae are present in input list
        remOMVar :: [(Formula, Position)] -> [(Formula, Position)]
        remOMVar fpl =
          let -- input list to remOMVar' is sorted innermost - outermost
              remOMVar' _ [] = []
              remOMVar' vs ((f,p):r) =
               let v = getQuantVar f
               in if v `isElement` vs then remOMVar' vs r
                  else (f,p):(remOMVar' (add v vs) r)
          in reverse (remOMVar' empty (reverse fpl))



------------------------------
----- internal functions -----
------------------------------

---------------------------------------------------
-- positive and negative literal path reductions --
---------------------------------------------------

-- The following functions are used to perform reductions needed to unify
-- a literal pair. These reductions include quantifier reductions, distributive
-- law application and flattening of disjunctions.
-- All of these reductions keep track of the literal pair that is to be unified
-- and adapt the unification result and mapping.
-- They do not mark the literal pair with the current step number.

-- "litDFlatten p" performs a flattening step at position p on the working
-- formula f stored in the global state monad.
-- Formula f must contain a disjunction symbol at position p. Moreover, the
-- direct superformula (at position p' where p = p'.i for some i) must
-- also be a disjunction symbol in f.
-- The resulting formula is f with the disjunction at position p' (partially)
-- flattened. I.e. the disjunctions at positions p and p' are merged.
-- The literal pair positions of the global state monad are updated,
-- the generation field is not increased (as this reduction step is not
-- considered a `real' reduction, but merely a simplification).

litDFlatten :: Position -> IOState ()
litDFlatten redPos =
  do -- extract information from global state monad
     f <- getWorkingFormula
     (p,p') <- getLiteralPair
     -- compute new formula f'
     let rp = posToList redPos				-- inner \/-pos.
         rootPos = listToPos (init rp); j = last rp	-- outer \/-pos.
         mainDisjFs = getSubFormulae (f@@rootPos)	-- get outer \/'s subfors,
         bs = take j mainDisjFs ++ drop (j+1) mainDisjFs -- but not inner \/
         as = getSubFormulae (mainDisjFs!!j) 		-- inner \/'s subfors
         l = length bs					-- # of outer subfors w/o inner \/
         newDisjFs = bs ++ as				-- merged subformulae
         f' = setSubFormulaAtPos rootPos (disj newDisjFs) f
     -- update literal positions
         (np,np') = updatePosD rootPos redPos p p' j l
     -- return new formula and literal pair positions
     setWorkingFormula f'
     setLiteralPair np np'
 where
   -- "updatePosD odp idp p p' j l" updates literal positions p and p' after
   -- a flattening step at position odp. odp and idp are (old) \/-positions,
   -- which have been merged (in the new formula odp is still a \/-position).
   -- l is the number of \/-subformulae at position idp, j is the subformula-index of
   -- the (old) inner \/-disjunction. In the new formula old outer subformulae occur
   -- before old inner subformulae as subformulae of the new disjunction at position
   -- odp.
   updatePosD :: Position -> Position -> Position -> Position -> Int -> Int
     -> (Position, Position)
   updatePosD rootp redp p p' j l =
     let r = posToList redp; rl = length r
         -- tpa (tpa') is position of p (p') in inner \/-subformula
         tpa = drop rl (posToList p); tpa' = drop rl (posToList p')
         -- ia (ia') is index of subformula containing p (p') w.r.t. inner \/-subformula
         ia = head tpa; ia' = head tpa'
         r' = posToList rootp; rl' = length r'
         -- tpb (tpb') is position of p (p') in outer \/-subformula
         tpb = drop rl' (posToList p); tpb' = drop rl' (posToList p')
         -- ib (ib') is index of subformula containing p (p') w.r.t. outer \/-subformula
         ib = head tpb; ib' = head tpb'
         np = if p < redp then
               -- old position is subposition of inner \/
               concPos rootp (listToPos (l+ia:tail tpa))
              else if p < rootp then  -- subposition of outer \/
                if ib < j then p  -- subposition of outer \/, before (old) inner \/
                else  -- subposition of outer \/, after (old) inner \/
                 concPos rootp (listToPos (ib-1:tail tpb))
              else p  -- p is no subposition of outer \/
         -- for comments see def. of np above
         np' = if p' < redp then concPos rootp (listToPos (l+ia':tail tpa'))
               else if p' < rootp then
                 if ib' < j then p'
                 else concPos rootp (listToPos (ib'-1:tail tpb'))
               else p'
     in (np,np')

-- "litAReduce p i ts ds" performs a universal quantifier reduction step
-- at position p (in the sense that the performed reduction is a p-reduction)
-- on the working formula f stored in the global state monad.
-- Formula f must contain a universal quantifier symbol at position p.
-- Reduction has to take place either on the positive or the negative
-- literal path (of the current literal pair stored in the state monad);
-- i is the corresponding path information.
-- ts is a term selector function, ds a disjunction selector type, as described
-- above.
-- Then the working formula is updated (reflecting the quantifier reduction)
-- as well as the literal pair's positions and the unification mapping.
-- (See algorithm applyARuleX for more accurate information on how the unification
-- mapping is changed.)

litAReduce :: Position -> PathInfo -> TermSelector -> DisjSelectType
  -> IOState ()
litAReduce redPos pi termSelector dSel =
  do -- get printing options and write information when requested
     opt <- getProofOptions
     let om = printing opt
     when (om == OMVerbose) $ printWith putChar 'A'
     -- get current state
     (f,maxVar,_,_,p,p',_,_) <- getLitPairState
     -- The next steps are used to compute a list of admissible reduction terms.
     -- get variable x in (forall x)A
     let x = getQuantVar (f@@redPos)
     -- compute position of direct superformula (to test whether the symbol at
     -- this position is a \/), and index of (forall x)A in disjunction
         r = posToList redPos; l = length r
         rootPos = listToPos (init r); j = last r
     -- find literal position within scope of universal quantifer (ap) and other (ap')
         (ap,ap') = if pi == PosLitPath then (p,p') else (p',p)
     -- get variables free in A and B
         dfs = getSubFormulae (f@@rootPos)  -- dfs is: (forall A) \/ B
         a = getSubFormula (dfs!!j); b = disj (take j dfs ++ drop (j+1) dfs)
         favs = free a  -- extract free variables of A
         fbvs = free b  -- and B
     -- get literal variables
         lvs = getLitVars (getLiteral (f@@ap))    -- literal in scope of univ. quant.
         lvs' = getLitVars (getLiteral (f@@ap'))  -- other literal
     -- get bound variables on literal path
         sx = singleton x   -- set containing just variable x
         -- bound variables on universally quantified literal path
         bvs = quantVarsOnPath f (listToPos (take (l + 1) (posToList ap))) ap
         -- bound variables on other literal's path
         bvs' = quantVarsOnPath f rootPos ap'
     -- arbitrary reduction variable? (yes, if x is not free in literal)
         arv = not (x `isElement` (lvs `minus` bvs))
     -- build admissible variable sets (only used if x occurs in literal ap)
         vs = listToSet [mkVar 0..mkVar maxVar]  -- all variables used in curr. calculus
         -- vs1: variables not free in A and B, with exception of x in A
         -- vs2: variables not free in literal in B, and not free in A (if not x)
         -- vs3: variables not free in literal in A (if not x) and not free in literal in B
         -- Variables from set vs_i should be preferably used before variables
         -- from a set vs_j with j>i.
         vs1 = (vs `minus` fbvs) `minus` (favs `minus` sx)
         vs2 = (vs `minus` (lvs' `minus` bvs')) `minus` (favs `minus` sx)
         vs3 = (vs `minus` (lvs' `minus` bvs')) `minus` ((lvs `minus` bvs)
               `minus` sx)
     -- choose first non-empty set (of vs{1..3}, priority for lower index)
         vs' = if isEmpty vs1 then
                 if isEmpty vs2 then vs3 else vs2
               else vs1
         -- compute term selector argument
         tselt = if arv then AnyVar
                        else AnyOf (map termFromVar (setToOrdList vs'))
     when (not arv && isEmpty vs') $ error "not enough variables in LitAReduce"
     -- choose reduction term (term is variable!)
     rt <- termSelector redPos tselt
     let rv = getTermVar rt
     -- determine relevant literals depending on path information:
     let (pFollow, nFollow) = if arv then if pi == PosLitPath
                                          then (DelA, OldA)
                                          else (OldA, DelA)
                              else (NewA, NewA)
     -- perform reduction (and update positions and unification map)
     applyARuleX redPos rv pFollow nFollow dSel


-- "litEReduce p i ts" performs a existential quantifier reduction step at
-- position p (in the sense that the performed reduction is a p-reduction)
-- on the working formula f stored in the global state monad.
-- Formula f must contain a existential quantifier symbol at position p.
-- Reduction has to take place either on the positive or the negative literal
-- path (of the current literal pair stored in the state monad); i is the
-- corresponding path information.
-- ts is a term selector function as described above.
-- Then the working formula is updated (reflecting the quantifier reduction)
-- as well as the literal pair's positions, unification mapping and unification
-- result (all stored in the state monad).
-- (See algorithm applyERuleX for more accurate information on how the unification
-- mapping and result are changed.)

litEReduce :: Position -> PathInfo -> TermSelector -> IOState ()
litEReduce redPos i termSelector =
  do -- get printing options and write information when requested
     opts <- getProofOptions
     let om = printing opts
     when (om == OMVerbose) $ printWith putChar 'E'
     -- get current state
     (f,_,_,_,p,p',umap,ur) <- getLitPairState
{-
     draw f
     putStrLn' ("positions: " ++ show (posToList p) ++ " " ++ show (posToList p'))
     draw (f@@p)
     draw (f@@p')
     putStrLn' ("redpos: " ++ show redPos)
-}
     -- Is this the innermost existential quantifier binding variable x?
     -- If not, reduction term is arbitrary.
     let x = getQuantVar (f@@redPos)
         -- find literal position pp within scope of existential quantifier
         pp = if i == PosLitPath then p else p'
         -- get variables (existentially) bound on matching path
         eqpath = filterPath isExists (pathRest (getPath f redPos pp))
         pathVars = foldr (add . getQuantVar) empty (pathToList eqpath)
         -- innermost exist. quant. for variable x?
         innermost = not (x `isElement` pathVars)
     -- compute substitution term from unification result and mapping
         mrt = computeERedTerm x i True ur umap; Just crt = mrt
     rt <- if not innermost || mrt == Nothing
           then -- mrt = Nothing inidicates arbitrary reduction term
             termSelector redPos AnyTerm
           else return crt -- use computed reduction term
     -- determine relevant literals depending on path information:
     let (pFollow,nFollow) = if i == PosLitPath then (NewE, OldE)
                             else (OldE, NewE)
     -- perform reduction (and update positions, unification map and result)
     f <- getWorkingFormula
{-
     putStrLn' ("mrt = " ++ show mrt ++ ", innermost = " ++ show innermost)
     putStrLn' ("redPos = " ++ show redPos)
     putStrLn' ("before apply, redterm is" ++ show rt)
--     draw f  -- f ok
-}
     applyERuleX redPos rt pFollow nFollow (mrt == Nothing && innermost)
{-
     f <- getWorkingFormula
     putStrLn' ("almost done")
     draw f  -- f corrupted
     putStrLn' ("E rule application done.")
-}

-- "litCReduce p ds" performs a distributive law reduction step at position p
-- (in the sense that the performed reduction is a p-reduction) on the working
-- formula f stored in the global state monad.
-- Formula f must contain a conjunction symbol at position p. Reduction has to
-- take place either on the positive or negative literal path (of the current
-- literal pair stored in the state monad). ds is a disjunction selector type,
-- as described above (which is used to build the disjunction part incorporated
-- in the reduction step).
-- Then the working formula is updated (reflecting the distributive law reduction)
-- as well as the literal pair's positions.

litCReduce :: Position -> DisjSelectType -> IOState ()
litCReduce rp dSel =
  do -- get printing options and write information when requested
     opt <- getProofOptions
     let om = printing opt
     when (om == OMVerbose) $ printWith putChar 'C'
     -- perform reduction (and update positons)
     applyDRuleX rp dSel


-----------------------------------
-- rule application with updates --
-----------------------------------

-- The following functions apply a reduction rule and update literal pairs'
-- positions, unification mappings and unification results.

-- "applyDRuleX p ds" applies a distributive law rule to the working formula
-- f stored in the global state monad.
-- p is the reduction position (in the sense that the performed reduction is
-- a p-reduction), ds is a disjunction selector type used to determine
-- the disjunctions used for the reduction step.
-- Literal positions are updated, changes on unification mapping and result
-- are not necessary, and are thus not performed.
-- The generation is increased by one.
-- Not both literals (positive and negative) are allowed to occur at subpositions
-- of p.

applyDRuleX :: Position -> DisjSelectType -> IOState ()
applyDRuleX redPos dSelType =
  do -- extract formula and literal positions from state monad
     f <- getWorkingFormula
     (p,p') <- getLiteralPair
     -- build conjunction selector
     let posLitInConj = p < redPos  -- does pos. lit. occur in conjunction?
         rp = posToList redPos; rl = length rp
         -- tp (tp') is the position of the matching literal rel. to /\
         tp = drop rl (posToList p); tp' = drop rl (posToList p')
         -- i is the index of the subfor. containing the relevant literal
         -- in the conjunction
         i = if posLitInConj then head tp else head tp'
         -- now, this is the conjunction selector for reduction:
         conjSel fs = return ([fs!!i], take i fs ++ drop (i+1) fs)
     -- build disjunction selector
         rootPos = listToPos (init rp)  -- rootPos: position of \/
         j = last rp  -- index of /\ in disjunction
         -- tpa (tpa') is the positive (negative) literal position relative
         -- to the disjunction
         tpa = drop (rl-1) (posToList p); tpa' = drop (rl-1) (posToList p')
         -- k1 (k2) is the index of the subformula containing the positive
         -- (negative) literal in the (old) disjunction
         k1 = head tpa; k2 = head tpa'
         -- k is the index of the non-conjunction literal in the disjunction
         k | posLitInConj = if k2 < j then k2 else k2 - 1
           | otherwise    = if k1 < j then k1 else k1 - 1
         -- now, this is the disjunction selector (for different disjunction
         -- selector types):
         -- NOTE: the first selected function contains one relevant literal
         disjSel fs =
           -- f' is the disjunction containing the non-conjunction literal, fs' is
           -- the rest
           let f' = fs!!k; fs' = take k fs ++ drop (k+1) fs
           in if dSelType == MaxDSel then return (f':fs', [])     -- max. disj.
              else if dSelType == MinDSel then return ([f'],fs')  -- min. disj.
              else -- dSelType == NormDSel,  disjunction parts selected acc. to literals
                let -- get predicates of B_i
                    b = f@@redPos    -- b is the conjunction
                    (bpls,bnls) = getPredicates b
                    -- partition formula list: first list contains formulae containing
                    -- predicates also occurring in b, second list all other
                    pfl [] = ([],[])
                    pfl (f:fs) =
                      let (ps,ns) = getPredicates f
                          -- build cross-section with (pos. and neg.) predicates
                          ok = not (null (intersect bnls ps ++ intersect bpls ns))
                          (fs',nfs') = pfl fs
                      in if ok then (f:fs',nfs') else (fs',f:nfs')
                    (xf,nxf) = pfl fs'  -- this is the partition of disjunction w/o f'
                in return (f':xf,nxf) -- f' is known to be ok, so add to first list
     -- reduce with distributive law rule using computed selector functions (implicitly
     -- updates working formula in state monad and increases generation by one)
     applyRule DistributionRule redPos Nothing (Just disjSel) (Just conjSel) Nothing
     -- compute new literal positions
     let -- `simulate' disjunction selector function
         b = f@@redPos  -- b is the main (relevant) conjunction
         -- as are all subformulae of the main disjunction
         as = getSubFormulae (f@@rootPos)
         -- as' is the main disjunction without the relevant conjunction,
         -- as'' is as' without the subformula containing the other literal 
         as' = take j as ++ drop (j+1) as; as'' = take k as' ++ drop (k+1) as'
         (bpls,bnls) = getPredicates b  -- predicates of conjunction
         -- "adfl fs" simulates the normDSel disjunction selector, but returns
         -- only the second formula list, not incorporated in the reduction
         adfl :: [Formula] -> [Formula]
         adfl [] = []
         adfl (f:fs) =      -- see disjSel definition above
           let (ps,ns) = getPredicates f
               notok = null (intersect bnls ps ++ intersect bpls ns)
               nfs' = adfl fs
           in if notok then f:nfs' else nfs'  -- add to list, if not used for reduction
         -- noDANormDSel is true iff the whole disjunction was used for reduction
         -- (in case of NormDSel was the disjunction selector used for reduction)
         noDANormDSel = null (adfl as'')
         -- l is the # of subformulae of the main disjunction
         l = length as
         -- noDA is true, iff the whole disjunction was used for reduction
         noDA = dSelType == MaxDSel ||              -- obviously
                (dSelType == MinDSel && l <= 2) ||  -- only subfors cont. rel. literals
                (dSelType == NormDSel && noDANormDSel)  -- see above
         -- now, we can compute the new literal positions
         (np,np') = updatePosD rootPos redPos p p' noDA
     -- store result in state monad: working formula and literal
     -- positions are updated (working formula, see above)
     setLiteralPair np np'
  where
    -- "(np,np') = updatePosD dp cp p p' w" updates the literal positions
    -- after a distributive law rule application. dp, cp, p and p' are
    -- the disjunction position, the conjunction position, the positive
    -- and negative literal positions, respectively. w is true, iff the
    -- whole disjunction was incorporated in the reduction process.
    updatePosD :: Position -> Position -> Position -> Position -> Bool
      -> (Position, Position)
    updatePosD rootp redp p p' noDA =
      let r = posToList redp; rl = length r
          -- tpb (tpb') is position of p (p') w.r.t. main /\
          tpb = drop rl (posToList p); tpb' = drop rl (posToList p')
          r' = posToList rootp; rl' = length r'
          -- tpa (tpa') is position of p (p') w.r.t. main \/
          tpa = drop rl' (posToList p); tpa' = drop rl' (posToList p')
          -- literal update function for positive and negative literal
          posUD :: Position -> [Int] -> [Int] -> Position
          posUD lp pb pa =
            if noDA then
              -- no additional disjunction (whole disj. incorp. in reduction)
              if lp < redp then        -- lp is (was) position in conjunction
                listToPos (r' ++ [0,0] ++ tail pb)
              else if lp < rootp then  -- lp is (was) position in disjunction
                listToPos (r' ++ [0,1] ++ tail pa)
              else lp  -- no subposition of main disjunction
            else -- additional disjunction
              if lp < redp then listToPos (r' ++ [0,0,0] ++ tail pb)
              else if lp < rootp then listToPos (r' ++ [0,0,1] ++ tail pa)
              else lp
      -- here are the new positions:
      in (posUD p tpb tpa, posUD p' tpb' tpa')
    

-- "applyERuleX p t pf nf b" applies an existential quantifier rule to the
-- working formula f stored in the global state monad.
-- p is the reduction position (in the sense that the performed reduction is
-- a p-reduction), t is a term, pf and nf indicate which literal positions
-- in the new formula are relevant (for positive and negative literal, resp.),
-- b is a boolean value, which determines whether or not the unification result
-- is to be updated.
-- There are only two admissible combinations for the (pf,nf) pair:
-- (NewE,OldE) is used, if the positive literal is in the newly created subformula,
-- the negative literal is part of the old subformula.
-- (OldE,NewE) is used similarly, with `positive' and `negative' changed in the
-- last sentence. (See also the definition of LitFollowE, above.)
-- Literal positions are updated, as well as the unification mapping. The unification
-- result is only updated when b is true.
-- The generation (also stored in the global state monad) is increased by one.
-- Not both literals (positive and negative) are allowed to occur at subpositions
-- of p.
-- The new unification mapping and result are determined as follows:
-- ?????

applyERuleX :: Position -> Term -> LitFollowE -> LitFollowE -> Bool -> IOState ()
applyERuleX redPos redTerm pFollow nFollow updateUResult =
  do -- extract formula, literal positions and unification information from state monad
     (f,maxVar,g,s,p,p',umap,ur) <- getLitPairState
     -- reduce with existential rule (implicitly updates working formula of state monad
     -- and increases generation by one)
{-
     putStrLn' ("before ERule-apply, redTerm is " ++ show redTerm)
-}
     applyRule ExistentialRule redPos Nothing Nothing Nothing (Just redTerm)
{-
     putStrLn' ("ERule applied!")
     fff <- getWorkingFormula
     draw fff
-}
     -- compute new literal positions considering relevant literal information
     let (np,np') = updatePosE redPos p p' pFollow nFollow
     -- update unification mapping
         -- compute position of disjunction linking literals
         rootPos = commonPrefix p p'
         -- determine on which literal path reduction has taken place (see module
         -- `Unification' for information about literal paths)
         i | rootPos < redPos = RootPath
           | p < redPos = PosLitPath
           | p' < redPos = NegLitPath
           | otherwise = error "wrong reduction position in applyERuleX"
         -- update unification map
         umap' = updateUMapE f redTerm redPos rootPos i p p' umap
     -- Update unification result, if demanded (i.e. if updateUResult = True; this
     -- occurs during root path pre-processing or during the main unification process,
     -- when a reduction with a freely chosen term is performed).
     -- This is only necessary when a unification equation of kind UVar x == UVar y
     -- is handled.
         -- find relevant literal path of literal under existential quantifier;
         -- in case of a reduction on the root path, where both literals occur
         -- in the scope of this quantifier, the relevant literal is the one
         -- which is not in the scope of the quantifier after application of
         -- the reduction.
         i' = if pFollow == NewE then PosLitPath else NegLitPath
         -- get relevant literal ap in scope of existential quantifier used for
         -- reduction (see also comment above)
         ap = if i' == PosLitPath then p else p'
         -- compute new (post-reduction) name v' of variable bound by existential
         -- quantifier (???)
         v' = let -- get quantified variables between literal and quantifier
                  qvop = quantVarsOnPath f redPos ap
                  -- rv is chosen reduction variable, qv is bound variable
                  rv = getTermVar redTerm; qv = getQuantVar (f@@redPos)
              in -- ???? <<< NEEDS REWORKING !!! >>>
                 if termIsVar redTerm && (rv `isElement` qvop ||
                    qv `isElement` qvop) then redTerm
                 else termFromVar qv
         mxut = uMapLookupByTermAndPathInfo v' i' umap'; Just xut = mxut
         (ur',um') = if updateUResult && mxut /= Nothing && isUVar xut then
                       updateVarEq xut ur umap' redTerm
                     else (ur,umap')
     f' <- getWorkingFormula
     setLitPairState f' maxVar (nextGen g) s np np' um' ur'
 where
   -- update literal positions
   updatePosE :: Position -> Position -> Position -> LitFollowE -> LitFollowE
     -> (Position, Position)
   updatePosE rp p p' pFollow nFollow =
     let r = posToList rp; rl = length r
         tp = drop rl (posToList p); tp' = drop rl (posToList p')
         np = if pFollow == NewE then
                -- new position is in newly created formula A[x/a]
                if p < rp then listToPos (r ++ [1] ++ tail tp) else p
              else
                -- new position is in old formula (exists x)A
                if p < rp then listToPos (r ++ [0] ++ tp) else p
         np' = if nFollow == NewE then
                 if p' < rp then listToPos (r ++ [1] ++ tail tp') else p'
               else
                 if p' < rp then listToPos (r ++ [0] ++ tp') else p'
     in (np,np')
   -- update UnificationMap
   updateUMapE :: Formula -> Term -> Position -> Position -> PathInfo
     -> Position -> Position -> UMap -> UMap
   updateUMapE f redTerm redPos rootPos redPosInfo p p' umap =
     let rootVars = quantVarsOnPath f redPos rootPos
         posVars = quantVarsOnPath f redPos p
         negVars = quantVarsOnPath f redPos p'
         x = getQuantVar (f@@redPos)
         redTermIsVar = termIsVar redTerm
         redVar = getTermVar redTerm
         -- process root path variables (change path info)
         rumap = processRootVars rootVars
         -- process pos path variables (rename, if needed)
         pumap = if pFollow == NewE && redTermIsVar
           then exchangePathVars x redVar posVars PosLitPath
           else id
         -- process neg path variables (rename, if needed)
         numap = if nFollow == NewE && redTermIsVar
           then exchangePathVars x redVar negVars NegLitPath
           else id
     in (numap . pumap . rumap) umap
   -- adapt unification result and umap, when an equation UVar x = UVar y
   -- is handled by the reduction step
   updateVarEq :: UTerm -> UResult -> UMap -> Term
     -> (UResult, UMap)
   updateVarEq xut ur um rt =
     -- lookup equation in unification result
     let -- "(es,es') = lueq ut ur": Find all equations in ur which contain ut
         -- and another variable, orient these equations such that ut
         -- is on the left. These equations are es, all other make up es'.
         lueq :: UTerm -> UResult -> ([UEquation], [UEquation])
         lueq _ (Solution []) = ([],[])
         lueq ut (Solution (e@(l,r):er)) =
           let (es,es') = lueq ut (Solution er)
           in if ut == l then if isUVar r then (e:es, es')
                              else (es, e:es')
              else if ut == r then if isUVar l then ((r,l):es, es')
                              else (es, e:es')
              else (es, e:es')
         (eqs, ur') = lueq xut ur
     -- update equation and umap: constant case
         cupd e um =
           let rci = getConIndex (getTermCon rt)
               ceq = let (_,r) = e in (r, UCon rci)
               crtlu = uMapLookupByTermAndPathInfo rt Unknown um
               cum = if crtlu == Nothing then (UCon rci, rt, Unknown):um
                                         else um
           in (ceq, cum)
     -- update equation and umap: variable case
         vupd e um =
           let rvi = getVarIndex (getTermVar rt); (UVar v,r) = e
               veq = (r, UOp v [])
               umud _ [] = []
               umud v (e@(UVar v',_,pi):r) =
                 if v == v' then (UOp v [], rt, pi):r else e : umud v r
               umud v (e:r) = e : umud v r
           in (veq, umud v um)
     -- apply update functions
         new udf = foldr (\e (es,um) -> let (ve,vum) = udf e um
                                        in (ve:es,vum))
                         (ur',um) eqs
         (ceqs, cum) = new cupd; (veqs, vum) = new vupd
     in if termIsCon rt then (solve ceqs, cum)
                        else (solve veqs, vum)


-- "um' = processRootVars vs um" 

processRootVars :: Set Variable -> UMap -> UMap
processRootVars vs umap = foldl procRV umap (setToOrdList vs)
  where procRV :: UMap -> Variable -> UMap
        procRV [] _ = []
        procRV (e@(ut@(UVar x),t,_):r) v =
          if termFromVar v == t then
            if x < 0 then (ut, t, NegLitPath):procRV r v
            else (ut,t, PosLitPath):procRV r v
          else e : procRV r v
        procRV (e@(ut@(UOp x _),t,_):r) v =
          if termFromVar v == t then
            if x < 0 then (ut,t,NegLitPath):procRV r v
            else (ut,t,PosLitPath):procRV r v
          else e : procRV r v
        procRV (e:r) v = e : procRV r v

-- "um' = exchangePathVars x y vs i um" changes entries of the unification mapping um
-- by replacing variable x by variable y in the term part and vice versa.
-- This is only performed for those entries with path information part equal to i.
-- Moreover entries with term part x are only changed, if x is not contained in
-- the variable set vs. (The same holds, mutatis mutandis, for variable y).
-- All other entries, as well as the unfication term parts of all entries, are not
-- affected.
-- The changed unification mapping um' is returned.

exchangePathVars :: Variable -> Variable -> Set Variable -> PathInfo -> UMap -> UMap
exchangePathVars _ _ _ _ [] = []
exchangePathVars x y vs i um =
  let tx = termFromVar x    -- build corresponding terms
      ty = termFromVar y
      -- alteration functions
      renX e@(ut,t,pi) | t == tx && pi == i = (ut, ty, pi)
                       | otherwise          = e
      renY e@(ut,t,pi) | t == ty && pi == i = (ut, tx, pi)
                       | otherwise          = e
      exc  e@(ut,t,pi) | t == tx && pi == i = (ut, ty, pi)
                       | t == ty && pi == i = (ut, tx, pi)
                       | otherwise          = e
  -- changes depend on variable set vs
  in if x `isElement` vs then
       if y `isElement` vs then map exc um
       else map renX um
     else if y `isElement` vs then map renY um
     else um


------------
-- A Rule --
------------

-- admitted possibilities for (pFollow, nFollow):
-- (OldA,DelA), (DelA,OldA), (NewA,NewA)

applyARuleX :: Position -> Variable -> LitFollowA -> LitFollowA
  -> DisjSelectType -> IOState ()
applyARuleX redPos redVar pFollow nFollow dSelType =
  do (f,maxVar,g,s,p,p',umap,ur) <- getLitPairState
     let -- get relevant positions to build disjunction selector
         r = posToList redPos; rootPos = listToPos (init r)
         rl = length r; j = last r
         tpb = drop (rl-1) (posToList p); tpb' = drop (rl-1) (posToList p')
         i1 = head tpb; i2 = head tpb'
         i = if p < redPos then if  i2 < j then i2 else i2 - 1
             else if i1 < j then i1 else i1 - 1
         -- build disjunction selector: first formula in selected disjuction
         -- is relevant formula (i.e formula containing p or p')
         disjSel fs =
           let f' = fs!!i; fs' = take i fs ++ drop (i+1) fs
           in if dSelType == MaxDSel then return (f':fs', [])
              else if dSelType == MinDSel then
                -- apply A rule kind 3?
                let freea = free (getSubFormula (f@@redPos))
                    freeb = free f'
                    rvFree = not (redVar `isElement` (freea `sunion` freeb))
                in if pFollow == NewA || rvFree then
                     -- p < redPos or p' < redPos
                     return ([f'],fs')
                   else return ([], fs) -- one of pFollow, nFollow is DelA
              else -- dSelType == NormDSel
                let -- get subformula A
                    fa = getSubFormula (f@@redPos)
                    -- get literals
                    (fapls,fanls) = getPredicates fa
                    -- partition formula list
                    pfl [] = ([],[])
                    pfl (f:fs) =
                      let (ps,ns) = getPredicates f
                          ok = not (null (intersect fanls ps ++ intersect fapls ns))
                          (fs',nfs') = pfl fs
                      in if ok then (f:fs',nfs') else (fs',f:nfs')
                    (xf,nxf) = pfl fs'
                    -- apply A rule kind 3?
                    freea = free fa
                    freeb = free (disj xf)
                    rvFree = not (redVar `isElement` (freea `sunion` freeb))
                in if pFollow == NewA || rvFree then return (f':xf, nxf)
                   else return ([], fs)
     -- reduce
     applyRule UniversalRule redPos (Just disjSel) Nothing Nothing (Just (termFromVar redVar))
     f' <- getWorkingFormula
     -- compute new positions
     let rootPos = listToPos (init r)
         -- compute part B when using NormDSel as disjunction selector
         aAndB = getSubFormulae (f@@rootPos)
         partA = aAndB!!j
         partB = take j aAndB ++ drop (j+1) aAndB
         (fapls,fanls) = getPredicates partA
         bfl [] = []
         bfl (f:fs) = let (ps,ns) = getPredicates f
                          ok = not (null (intersect fanls ps ++ intersect fapls ns))
                          fs' = bfl fs
                      in if ok then f:fs' else fs'
         -- assertion: bfl partB has length >= 1
         joBNormDSel = null (tail (bfl partB))
         -- no part B in reduction formula
         rf = f@@rootPos
         noB = null r || not (isDisj rf) || 
               (dSelType == MinDSel && pFollow /= NewA) ||
               (dSelType == NormDSel && pFollow /= NewA)
         -- part B is just one formula (only valid if noB == False)
         joB = length partB < 2 ||
               (dSelType == MinDSel && pFollow == NewA) ||
               (dSelType == NormDSel && pFollow == NewA && joBNormDSel)
         -- redVar is not free in A and B
         freeA = free (getSubFormula partA)
         freeB = case dSelType of
                   MinDSel -> free (partB!!i)
                   NormDSel -> free (disj (bfl partB))
                   MaxDSel -> free (disj partB)
         rvNotFree = not (redVar `isElement` (freeA `sunion` freeB))
         -- no rest in case of ARule kind 3 application
         noR = case dSelType of
                 MinDSel -> length partB < 2
                 NormDSel -> length (bfl partB) == length partB
                 MaxDSel -> True
         (np,np') = updatePosA redPos p p' pFollow nFollow noB joB rvNotFree noR
-----------------
{-
     putStrLn' ("\nredVar = " ++ show redVar)
     putStrLn' ("rvNotFree = " ++ show rvNotFree)
     putStrLn' ("noR = " ++ show noR)
     putStrLn' ("noB = " ++ show noB)
     putStrLn' ("new positions: " ++ show (np,np'))
     putStrLn' ("old umap: <<" ++ show umap ++ ">>")
     putStrLn' "f' = "
     draw f'
     showUResult
     let
-}
-----------------
     -- update UMap in case (NewA,NewA), or if rvNotFree == True
--         umap' = if pFollow == NewA && nFollow == NewA || rvNotFree then
         umap' = if pFollow == NewA && nFollow == NewA then
                   updateUMapA f redVar redPos p p' umap
                 else umap
----------------
{-
     putStrLn' ("new umap: <<" ++ show umap' ++ ">>")
     showUResult
-}
----------------
     setLitPairState f' maxVar (nextGen g) s np np' umap' ur
 where
   -- update literal positions
   updatePosA :: Position -> Position -> Position -> LitFollowA -> LitFollowA
     -> Bool -> Bool -> Bool -> Bool -> (Position, Position)
   updatePosA rp p p' pFollow nFollow noB joB rvNotFree noR =
     let r = posToList rp; rl = length r
         root = init r; j = last r; rootp = listToPos root
         tpa = drop rl (posToList p); tpa' = drop rl (posToList p')
         tpb = drop (rl-1) (posToList p); tpb' = drop (rl-1) (posToList p')
         ib = head tpb; ib' = head tpb'
         np = if rvNotFree then
                -- part B exists, redVar not free in A or B, ARule kind 3
                if noR then
                  -- simple case, complete disjunction selected, no rest
                  if p < rp then listToPos (root ++ [0,0] ++ tail tpa)
                  else if p < rootp then listToPos (root ++ [0,1] ++ tail tpb)
                  else error "illegal case in updatePosA"
                else
                  -- not complete disjunction selected
                  if p < rp then listToPos (root ++ [0,0,0] ++ tail tpa)
                  else if p < rootp then listToPos (root ++ [0,0,1] ++ tail tpb)
                  else error "illegal case in updatePosA"
              else if noB then
                -- no part B in reduction, ARule kind 2
                if pFollow == OldA then
                  -- new position is in old formula (forall x)A
                  if p < rp then listToPos (r ++ [0] ++ tpa) else p
                else if pFollow == DelA then
                  -- new position is in A[-x]
                  if p < rp then listToPos (r ++ [1] ++ tail tpa)
                  else error "illegal case in updatePosA"
                else error "illegal case in updatePosA"
              else
                -- part B exists, ARule kind 1
                if pFollow == OldA then
                  -- new position is in old formula (forall x)A
                  if p < rp then listToPos (r ++ [0] ++ tpa)
                  else if p < rootp then
                    if ib < j then listToPos (root ++ [ib+3] ++ tail tpb)
                    else listToPos (root ++ [ib+2] ++ tail tpb)
                  else p
                else if pFollow == DelA then
                  -- new position in A[-x]
                  if p < rp then listToPos (root ++ [1] ++ tail tpa)
                  else error "illegal case in updatePosA"
                else
                  -- new position in (forall y)((forall y)B \/ A[x-y][x/y])
                  if p < rp then listToPos (root ++ [2,0,1] ++ tail tpa)
                  else if p < rootp then
                    if joB then listToPos (root ++ [2,0,0,0] ++ tail tpb)
                    else listToPos (root ++ [2,0,0,0,0] ++ tail tpb)
                  else p
         np' = if rvNotFree then
                 -- part B exists, redVar not free in A or B, ARule kind 3
                 if noR then
                   -- simple case, complete disjunction selected, no rest
                   if p' < rp then listToPos (root ++ [0,0] ++ tail tpa')
                   else if p' < rootp then listToPos (root ++ [0,1] ++ tail tpb')
                   else error "illegal case in updatePosA"
                 else
                   -- not complete disjunction selected
                   if p' < rp then listToPos (root ++ [0,0,0] ++ tail tpa')
                   else if p' < rootp then listToPos (root ++ [0,0,1] ++ tail tpb')
                   else error "illegal case in updatePosA"
               else if noB then
                 -- no part B in reduction, ARule kind 2
                 if nFollow == OldA then
                   -- new position is in old formula (forall x)A
                   if p' < rp then listToPos (r ++ [0] ++ tpa') else p'
                 else if nFollow == DelA then
                   -- new position is in A[-x]
                   if p' < rp then listToPos (r ++ [1] ++ tail tpa')
                   else error "illegal case in updatePosA"
                 else error "illegal case in updatePosA"
               else
                 -- part B exists
                 if nFollow == OldA then
                   -- new position is in old formula (forall x)A
                   if p' < rp then listToPos (r ++ [0] ++ tpa')
                   else if p' < rootp then
                     if ib' < j then listToPos (root ++ [ib'+3] ++ tail tpb')
                     else listToPos (root ++ [ib'+2] ++ tail tpb')
                   else p
                 else if nFollow == DelA then
                   -- new position in A[-x]
                   if p' < rp then listToPos (root ++ [1] ++ tail tpa')
                   else error "illegal case in updatePosA"
                 else
                   -- new position in (forall y)((forall y)B \/ A[x-y][x/y])
                   if p' < rp then listToPos (root ++ [2,0,1] ++ tail tpa')
                   else if p' < rootp then
                     if joB then listToPos (root ++ [2,0,0,0] ++ tail tpb')
                     else listToPos (root ++ [2,0,0,0,0] ++ tail tpb')
                   else p'
     in (np,np')
   updateUMapA :: Formula -> Variable -> Position -> Position -> Position
     -> UMap -> UMap
   updateUMapA f redVar redPos p p' umap =
     let x = getQuantVar (f@@redPos)
         redPosInfo | p < redPos = PosLitPath
                    | p' < redPos = NegLitPath
                    | otherwise = error "illegal case in updateUMapA"
         cp = if p < redPos then p else p'
         vs = quantVarsOnPath f redPos cp
     in exchangePathVars x redVar vs redPosInfo umap
{-
         mrOp = uMapLookupByTermAndPathInfo (termFromVar x) redPosInfo umap
         Just rOp = mrOp
         cp = if p < redPos then p else p'
         vs = quantVarsOnPath f redPos cp
     in if mrOp == Nothing then error ("UOp lookup failed in updateUMapA: " ++ show x ++ show redPosInfo )
        else exchangePathVars x redVar vs redPosInfo umap
-}



-- "findRootReductionPosition f p ur um": Determines, if a reduction on the
-- root path is necessary to unify a literal pair (p,p'). The literal pair
-- (p,p') must be unifiable, where ur is the result of the unification
-- and um the corresponding unification map.
-- Reduction is necessary, if the unification solution contains an equation
-- of the form x = f(...x'...) where x' is the complementary variable of x,
-- or if the solution contains an equation pair x = t_1, x = t_2 with
-- t1 /= t2 (x is understood as either x or the complementary term of x).
-- The function returns the position where reduction has to take place
-- to resolve the aforementioned cases together with a pair of corresponding
-- reduction terms (or Nothing of the reduction term is arbitrary).
-- The first reduction term may be used for substitution in the positive
-- literal, the second in the negative literal.
-- Nothing is returned if no reduction on the root path is needed.

findRootReductionPosition :: Formula -> Position -> UResult -> UMap
  -> Maybe (Position, Maybe (Term, Term))
findRootReductionPosition f rootPos (Solution es) umap =
  let -- find critical equations, i.e. equations of the form
      -- v = f(...v'...) where v and v' are complementary variables
      critEqFilter (v, UOp _ rts) = occurCheck (negUTerm v) rts
      critEqFilter _ = False
      -- Get left hand sides of critical equations, i.e. variables
      -- and find their original variable names.
      -- Assertion: All critical variables lie on the root path!
      crit1VarsAndRTerms = map (\(v,_) -> (v, Nothing)) (filter critEqFilter es)
      crit2VarsAndRTerms = map (\(v,tp) -> (v, Just tp))
                               (doubletVars es umap)
      criticalVarsAndRTerms = map
        (\(v,ts) -> let Just (v',_) = uMapLookupByUTerm v umap
                    in (v', ts))
        (crit1VarsAndRTerms ++ crit2VarsAndRTerms)
      -- find positions of critical variables
      critPosAndRTerms =
        map ((\(v,mtp) -> (findVarPos f (getTermVar v) rootPos, mtp)))
            criticalVarsAndRTerms
  in if null critPosAndRTerms then Nothing
     else Just (foldl1 (\(p,mtp) (p',mtp') -> 
                        if p > p' then (p,mtp) else (p',mtp'))
                critPosAndRTerms)

-- "doubletVars es" finds all unification variables x, such that x' = t1 and
-- x' = t2 with t1 /= t2 is in the equation system es and not both t1 and
-- t2 are unification variables.
-- x' is either x or the complementary variable of x.
-- It returns a list of UVars for which the aforementioned criterion holds
-- together with a pair (t1,t2) of terms, which are candidates for a reduction
-- term. t1 corresponds with the positive literal, t2 with the negative.
-- I.e. substitution of x by t1 in the subformula containing the positive literal p
-- is a sensible reduction step avoiding the doublet case.
-- NOTE: The unification equation system must be interreduced.
 
doubletVars :: [UEquation] -> UMap -> [(UTerm, (Term, Term))]
doubletVars es umap =
  let -- add all equations (v,l) where (l,v) is in es and v is a variable
      rVarFilter (_, UVar _) = True; rVarFilter _ = False
      eqsWithVarOnRight = filter rVarFilter es
      uvars = es ++ map (\(l,v) -> (v,l)) eqsWithVarOnRight
      -- convert neg. uvars -> pos. uvars, and remember original sign:
      -- true: positive literal, false: negative literal
      absUTerm' (UVar x) = (UVar (abs x), if x > 0 then True else False)
      absUTerm' (UOp x ts) = (UOp (abs x) ts, if x > 0 then True else False)
      absUTerm' t = (t, False)
      puvars = map (\(l,r) -> (absUTerm' l, absUTerm' r)) uvars
      -- sort uvar list
      spuvars = quickSort puvars
      -- doublet finder for sorted lists
      df [] = []; df [e] = []
      df (((l,pol),(r,por)):e@((l',pol'),(r',por')):es) =
        if l `equalsModSubTerms` l' && pol /= pol' &&
-----
           not (r `equalsModSubTerms` r') && not (isUVar r && isUVar r') then
-----
          let -- reconstruct sign
              sr = if por then r else negUTerm r
              sr' = if por' then r' else negUTerm r'
              -- get positive and negative right hand sides
              (pr,nr) = if pol then (sr,sr') else (sr',sr)
              Just (opr,_) = uMapLookupByUTerm pr umap
              Just (onr,_) = uMapLookupByUTerm nr umap
          in (l, (opr,onr)) : df (e:es)
        else df (e:es)
  in df spuvars
      

-------------------------
-- auxiliary functions --
-------------------------

-- assertion: p > p' (if not, the result is the empty set)
quantVarsOnPath :: Formula -> Position -> Position -> Set Variable
quantVarsOnPath f p p' =
  let path = getPath f p p'
      qpath = filterPath isQuant path
  in if p < p' then empty
     else foldr (add . getQuantVar) empty (pathToList qpath)

-- "findVarPos f v rp": Find innermost occurrence of bound variable v on
-- path lambda -- rp in formula f.
-- It is assumed that such a positon exists!

findVarPos :: Formula -> Variable -> Position -> Position
findVarPos f v rp =
  let rootPathP = getFormulaeAndPositions f lambda rp
      -- get existential quantifiers on root path in reverse order
      pathP = reverse (filter (isExists . fst) rootPathP)
      -- filter bound variable v
      vpathP = filter (\(f,_) -> getQuantVar f == v) pathP
  in if null vpathP then
       error "variable not found in findVarPos"
     else snd (head vpathP)

-- "getFormulaeAndPositions f p p'" returns all subformulae on the path
-- p -- p' in formula f together with their positions in f.
-- It is assumed that p is a superposition of p'.

getFormulaeAndPositions :: Formula -> Position -> Position
   -> [(Formula, Position)]
getFormulaeAndPositions f p p' =
  map (\(f,pl) -> (f, concPos p (listToPos pl))) (getfp (f@@p)
       (drop (length (posToList p)) (posToList p')))
  where getfp :: Formula -> [Int] -> [(Formula, [Int])]
        getfp f [] = [(f, [])]
        getfp f p@(i:ps) = (f,[]) : map (\(f,pos) -> (f,i:pos))
                                       (getfp (getSubFormulae f !! i) ps)



-- "markLiterals" marks the current literal pair in the working formula
-- stored in the global state monad with the current step number. This is
-- needed by the literal pair unification strategy to avoid multiple
-- unifications with the same literal pair.

markLiterals :: IOState ()
markLiterals = do
  -- get literal pair
  (f,_,_,s,p,p',_,_) <- getLitPairState
  -- get node information and add new step number
  let fp = f@@p; fp' = f@@p'
      sip = getStrategyInfo (getNodeInfo fp)
      sip' = getStrategyInfo (getNodeInfo fp')
      nip = mkNodeInfo (insertNew (mkGen s, Nothing) sip)
      nip' = mkNodeInfo (insertNew (mkGen s, Nothing) sip')
      nfp = setNodeInfo nip fp; nfp' = setNodeInfo nip' fp'
  -- adapt working formula
  setWorkingFormula (setSubFormulaAtPos p nfp (setSubFormulaAtPos p' nfp' f))
