module Main where

import Set
import Formula
import Scanner
import FormulaIO
import Simplify
import Reduction
import IO
import Unification
import Auxiliary
import LiteralPair
import Subsumption
import ProofStack
import Strategy
import ProofManager
import IOStateMonad
import GlobalState

main = do
  showWelcomeMessage
  mainLoop `startingWith` initialGlobalState

showWelcomeMessage :: IO ()
showWelcomeMessage = do
  putStrLn (replicate 80 '-')
  putStr "---  Welcome to the ARA prover, version 0.9, (c) 1996-2000 "
  putStrLn "by Carsten Sinz.  ---"
  putStrLn (replicate 80 '-')
  newLine

initialGlobalState :: GlobalState
initialGlobalState = MkGlobalState {
  currentProofStack = emptyPS,
  totalProofSteps = 0,
  totalProofGenerations = 0,
  totalProofTime = 0,
  subProofTime = 0,
  proofOptions = MkProofOptions {
    splitting = NoSplitting,
    equalModRenSteps = 10,
    unusedLitSteps = 20,
    outerAQuants = True,
    printing = OMVerbose,
    debugging = False
  },
  formula = falsum,
  generation = initialGen,
  step = 0,
  maxVar = 0,
  posLitPosition = lambda,
  negLitPosition = lambda,
  unificationMap = [],
  unificationResult = Fail,
  symbolTable = emptySymbolTable,
  randomNumbers = [],
  lastAK3info = Nothing
}


mainLoop :: IOState ()
mainLoop =
  do printMenu
     c <- query "Selection (a/r/f/o/c/g/s/p/q)? " "arfocgspq"
     newLine'
     case c of
       'a' -> do mf <- readInputFormula RA
                 if mf == Nothing then do
                   putStrLn' "Formula input aborted."
                   newLine'
                   mainLoop
                  else do
                   newLine'
                   putStrLn' ("Input formula read successfully, converted" ++
                              " to PL1 negation normal form,")
                   putStrLn' ("simplified and added to subgoal list.\n")
                   let Just f = mf; f' = simplify (nnf f)
                   setTotalProofSteps 0
                   setTotalProofGenerations 0
		   setProofStack emptyPS
		   setCurrentSubGoal f' initialGen 0   -- pushes f' onto the proof stack
                   setTotalProofGenerations 0
                   setTotalProofTime 0
                   setSubProofTime 0
                   mainLoop
       'r' -> do mf <- readInputFormula PL1
                 if mf == Nothing then do
                   putStrLn' "Formula input aborted."
                   newLine'
                   mainLoop
                  else do
                   newLine'
                   putStrLn' ("Input formula read successfully, converted" ++
                              " to negation normal form,")
                   putStrLn' ("simplified and added to subgoal list.\n")
                   let Just f = mf; f' = simplify (nnf f)
                   setTotalProofSteps 0
                   setTotalProofGenerations 0
		   setProofStack emptyPS
		   setCurrentSubGoal f' initialGen 0   -- pushes f' onto the proof stack
                   setTotalProofGenerations 0
                   setTotalProofTime 0
                   setSubProofTime 0
                   mainLoop
       'f' -> do mf <- getCurrentSubGoal
                 let Just (f,g,s) = mf
                 if mf == Nothing then do
                   putStrLn' "No current formula."
                   newLine'
                   mainLoop
                  else do
                   ps <- getProofStack
                   let size = psSize ps
                   c' <- if size > 1 then
                           query ("Attention: All subgoals will be " ++
                                  "deleted! Continue or abort (c/a)? ") "ac"
                          else return 'c'
                   if c' == 'a' then do
                     newLine'
                     mainLoop
                    else do
                     ppOpts <- readPreProcessOptions
                     preProcess ppOpts
                     newProofStack <- getProofStack
                     putStrLn' ("Generated " ++ show (psSize newProofStack) ++ 
                                " subgoal(s).")
                     newLine'
                     setProofStack newProofStack
                     mainLoop
       'o' -> do changeProofOptions
                 newLine'
                 mainLoop
       'c' -> do mf <- getCurrentSubGoal
                 let Just (f,_,_) = mf
                 if mf == Nothing then putStrLn' "No current formula."
                  else do c' <- query "Simple output or layout (s/l)? " "sl"
                          newLine'
                          if c' == 's' then do { showFor f; newLine' }
                           else draw f
                 newLine'
                 mainLoop
       'g' -> do printProofStack
                 newLine'
                 mainLoop
       's' -> do mf <- getCurrentSubGoal
                 let Just (f,_,_) = mf
                 if mf == Nothing then putStrLn' "No current formula."
                  else showFormulaStatistics f
                 newLine'
                 mainLoop
       'p' -> do mf <- getCurrentSubGoal
                 if mf == Nothing then do
                   putStrLn' "No current formula.\n"
                   mainLoop
                  else do
                   proveIt
                   newLine'
                   mainLoop
       'q' -> do putStrLn' "Good bye!"
                 newLine'


proveIt :: IOState ()
proveIt = do
  proofResult <- tryProof
  if proofResult == ProofSucceeded then do
    ts <- getTotalProofSteps
    tg <- getTotalProofGenerations
    tt <- getTotalProofTime
    newLine'
    putStrLn' (replicate 80 '-')
    putStrLn' "*****  P R O O F  C O M P L E T E D  *****"
    newLine'
    printProofStatistics False tg ts tt
    putStrLn' (replicate 80 '-')
    return ()
   else if proofResult == PartialProof then do
    -- only one formula had to be proved
    newLine'
    putStrLn' (replicate 80 '-')
    putStrLn' "***** PARTIAL PROOF SUCCEEDED *****"
    putStrLn' (replicate 80 '-')
    return ()    
   else if proofResult == MaxStepReached then do
    -- maximal number of proof steps reached
    newLine'
    putStrLn' (replicate 80 '-')
    putStrLn' "The maximal number of proof steps was reached."
    c <- query "Continue or abort (c/a)? " "ca"
    if c == 'c' then do
      putStrLn' "Resuming...\n"
      proveIt
     else do
      putStrLn' "Proof search aborted."
      return ()
   else do  -- proof failed
    newLine'
    putStrLn' (replicate 80 '-')
    putStrLn' "***** PROOF FAILED *****"
    putStrLn' (replicate 80 '-')
    return ()


-- assertion: proof stack not empty

tryProof :: IOState ProofResult
tryProof = do
  (cfonly, maxVar, maxSteps, redStrat) <- queryProofSetup
  setMaxVar maxVar
  setMaxSteps maxSteps
  if cfonly then do
    -- proof only one subgoal
    proveSubGoal redStrat
   else do
    -- proof all subgoals
    proveAll redStrat


printMenu :: IOState ()
printMenu = do
  printWith putStrLn "You have the following choice:"
  printWith putStrLn "  (a) read formula (RA syntax)"
  printWith putStrLn "  (r) read formula (PL1 syntax)"
  printWith putStrLn "  (f) preprocess current formula"
  printWith putStrLn "  (o) print / change proof options"
  printWith putStrLn "  (c) show current formula"
  printWith putStrLn "  (g) show generated subgoals"
  printWith putStrLn "  (s) print formula statistics"
  printWith putStrLn "  (p) start proof / next subgoal proof"
  printWith putStrLn "  (q) quit"
     
-- result:
--  1. only first subgoal (<-> all subgoals)
--  2. max var
--  3. max step
queryProofSetup :: IOState (Bool, Int, Int, ReductionStrategy)
queryProofSetup = do
  ps <- getProofStack
  let size = psSize ps
  c' <- if size > 1 then
          query "Prove current formula or all subgoals (f/a)? " "fa"
         else return 'a'
  putStr' "Enter maximal number of variables: "
  maxVar <- rReadLn
  putStr' "Enter maximal number of proof steps: "
  maxStep <- rReadLn
  redStrat <- queryRedStrat
  return (c' == 'f', maxVar-1, maxStep, redStrat)
          

queryRedStrat :: IOState ReductionStrategy
queryRedStrat = do
  putStrLn' "Choose a major reduction strategy:"
  putStrLn' "  (l) literal pair reduction strategy"
  putStrLn' "  (o) oldest reduction possibility strategy"
--  putStrLn' "  (m) mix both strategies"
  putStrLn' "  (a) LP strategy with priority to non-expanding quantifier rule"
--  c <- query "Selection (l/o/m/a)? " "loma"
  c <- query "Selection (l/o/a)? " "loa"
  case c of 'l' -> initLPReducer
            'o' -> initORPReducer
            'a' -> initAK3Reducer
            otherwise -> failS (iosError 111 "not yet implemented")


readPreProcessOptions :: IOState PreProcessOptions
readPreProcessOptions = do
  c <- query "Perform partial skolemization (y/n)? " "yn"
  c' <- query "Subgoal generation (y/n)? " "yn"
  c'' <- if c' == 'y' then
           query "  Partial, complete or recursive subgoal generation (p/c/r)? " "cpr"
         else return '#'
  let sp = case (c',c'') of
             ('y','c') -> CompleteSplitting
             ('y','p') -> RootSplitting
             ('y','r') -> RecursiveSplitting
             ('n','#') -> NoSplitting
      sk = c == 'y'
  return (MkPPOptions { ppSkolemization = sk, ppSplitting = sp })

    

changeProofOptions :: IOState ()
changeProofOptions = do
  showProofOptions
  c <- query "Enter proof option to change, or 'e' to exit (o/g/d/a/r/l/e): "
             "osgderal"
  opts <- getProofOptions
  case c of
    'o' -> do putStrLn' "Possible output modes:"
              putStrLn' "  (t) terse"
              putStrLn' "  (s) standard"
              putStrLn' "  (v) verbose"
              c' <- query "New output mode (t/s/v)? " "tsv"
              let om = case c' of
                         't' -> OMTerse
                         's' -> OMStandard
                         'v' -> OMVerbose
              setProofOptions (opts { printing = om })
    'g' -> do putStrLn' "Possible values for subgoal generation:"
              putStrLn' "  (o) on"
              putStrLn' "  (p) partial"
              putStrLn' "  (f) off"
              c' <- query "New value (o/p/f)? " "opf"
              let sp = case c' of
                         'o' -> CompleteSplitting
                         'p' -> RootSplitting
                         'f' -> NoSplitting
              setProofOptions (opts { splitting = sp })
    'd' -> do let d' = not (debugging opts)
              setProofOptions (opts { debugging = d' })
    'r' -> do putStrLn' "Enter new value: "
              r <- rReadCond rReadLn (>= 1)
              setProofOptions (opts { equalModRenSteps = r })
    'l' -> do putStrLn' "Enter new value: "
              l <- rReadCond rReadLn (>= 1)
              setProofOptions (opts { unusedLitSteps = l })
    'a' -> do let a' = not (outerAQuants opts)
              setProofOptions (opts { outerAQuants = a' })
    'e' -> done
  if c == 'e' then done else
    do newLine'
       changeProofOptions
  

showProofOptions :: IOState ()
showProofOptions = do
  opts <- getProofOptions
  let sp = splitting opts; p = printing opts; d = debugging opts
      r = equalModRenSteps opts; l = unusedLitSteps opts
      a = outerAQuants opts
      pp = case p of OMTerse -> "terse"
                     OMStandard -> "standard"
                     OMVerbose -> "verbose"
      psp = case sp of NoSplitting -> "off"
                       RootSplitting -> "partial"
                       CompleteSplitting -> "on"
      pd = if d then "on" else "off"
      pa = if a then "on" else "off"
  putStrLn' "Global proof options:"
  putStrLn' ("  (o) output mode:        " ++ pp)
  putStrLn' ("  (g) subgoal generation: " ++ psp)
  putStrLn' ("  (d) debug mode:         " ++ pd)
  putStrLn' ("  (a) remove outermost univ. quants.:  " ++ pa)
  putStrLn' "   Steps before removing..."
  putStrLn' ("  (r) subformulae equal mod. renaming: " ++ show r)
  putStrLn' ("  (l) unused literals:                 " ++ show l)


-- print non-empty proof-stack
printProofStack :: IOState ()
printProofStack = do
  ps <- getProofStack
  if psIsEmpty ps then putStrLn' "No subgoals." else pps 1 ps
 where pps n ps = if psIsEmpty ps then return () else
         if n > 10 then do
            putStrLn' ("...and " ++ show (psSize ps) ++ " further subgoals.")
          else do
            let (f,(g,s),r) = topPopPS ps
            putStrLn' (show n ++ ") origin: generation " ++ show g ++ ", step " ++
              show s)
            putStr' "   "
            showFor f
            newLine'
            pps (n+1) r


readInputFormula :: FormulaSyntax -> IOState (Maybe Formula)
readInputFormula syntax = do
  c <- query "Read input formula from file or standard input (f/s)? " "fs"
  if c == 'f' then readInputFormulaFromFile syntax
    else readInputFormulaFromStdIn syntax


readInputFormulaFromFile :: FormulaSyntax -> IOState (Maybe Formula)
readInputFormulaFromFile syntax = do
  putStrLn' "Enter file name: "
  filename <- doIO getLine
  if null filename then return Nothing else
    catchS $$ do fstring <- doIO (readFile filename)
                 let reader | syntax == PL1 = readFormula
                            | syntax == RA  = readFormulaRA
                 catchS $$ do f <- reader fstring
                              f' <- checkLitRank f
                              return (Just f')
                        $$ \e -> do let msg = getIOSErrMsg e
                                    putStrLn' msg
				    newLine'
                                    readInputFormulaFromFile syntax
            $$ \e -> do putStrLn' ("Can't open file \"" ++ filename ++ "\".")
                        readInputFormulaFromFile syntax

readInputFormulaFromStdIn :: FormulaSyntax -> IOState (Maybe Formula)
readInputFormulaFromStdIn syntax = do
  putStrLn' "Enter formula:"
  catchS $$ do s <- doIO getSemLines
               let reader | syntax == PL1 = readFormula
                          | syntax == RA  = readFormulaRA
               f <- reader s
               f' <- checkLitRank f
               return (Just f')
         $$ \e -> do let msg = getIOSErrMsg e
                     putStrLn' msg
                     c <- query "\nRetry or abort (r/a)? "
                                "ra"
                     if c == 'r' then readInputFormulaFromStdIn syntax
                                 else return Nothing

getSemLines :: IO String
getSemLines = gsl False where
  gsl sc = do c <- getChar
              if c == ';' then do cs <- gsl True
                                  return (c:cs)
                   else if c == '\n' && sc then return []
                   else do cs <- gsl sc
                           return (c:cs)


readLnF :: Read a => IO a
readLnF = do l <- getSemLines
             r <- readIO l
             return r


showFormulaStatistics :: Formula -> IOState ()
showFormulaStatistics f = doIO (do
  putStrLn "Statistics of current formula:"
  putStrLn ("  size:                      " ++ show (size f))
  putStrLn ("  depth:                     " ++ show (depth f))
  putStrLn ("  # variables:               " ++ show (cardinality (vars f)))
  putStrLn ("  # free variables:          " ++ show (cardinality (free f)))
  putStrLn ("  # constants:               " ++ show (cardinality (consts f)))
  putStrLn ("  # predicates:              " ++ show (length (collectLiterals f)))
  putStrLn ("  # existential quantifiers: " ++ show (quantCount isExists f))
  putStrLn ("  # universal quantifiers:   " ++ show (quantCount isForAll f))
 )
 where quantCount q f =
         let n = sum (map (quantCount q) (getSubFormulae f))
         in if q f then n+1 else n
