module ProofManager where

import IOStateMonad
import Formula
import Strategy
import Simplify
import LiteralPair
import ProofStack
import Subsumption
import Reduction
import Auxiliary
import Timing  -- hugs
--import CPUTime   -- hbc/ghc
import GlobalState
import Monad (when)

data ProofResult = ProofSucceeded | ProofFailed | PartialProof | MaxStepReached
  deriving Eq


-- set current formula and adapt proof stack (current formula is the top of the
-- proof stack)

setCurrentSubGoal :: Formula -> Generation -> Step -> IOState ()
setCurrentSubGoal f g s = updateProofStack (\stack -> pushPS f (g,s) stack)

getCurrentSubGoal :: IOState (Maybe (Formula, Generation, Step))
getCurrentSubGoal = do
  ps <- getProofStack
  let (f,(g,s)) = topPS ps
  if psIsEmpty ps then return Nothing else return (Just (f,g,s))

delCurrentSubGoal :: IOState ()
delCurrentSubGoal = do
  ps <- getProofStack
  if psIsEmpty ps then done else setProofStack (popPS ps)


-- pre-process current formula, i.e. top of proof stack
-- result is put on the proof stack
-- proof stack must not be empty
preProcess :: PreProcessOptions -> IOState ()
preProcess opt = do
  ~(Just (f,g,s)) <- getCurrentSubGoal
  let skol = minimizeVarAndConstDomains . partialSkolemization . simplify .
             deleteOuterUniversalQuantifiers . minimizeQuantifierScopes
      spl t = case t of
                RootSplitting -> topSplit
                CompleteSplitting -> split
                otherwise -> return
      simp = minimizeVarAndConstDomains . simplify
      -- first skolemization step
      f' = if ppSkolemization opt then skol f else simp f
      -- splitting step, if ppSplitting /= NoSplitting
      fs = spl (ppSplitting opt) f'
      -- skolemization after complete splitting
      fs' = if ppSkolemization opt then map skol fs else map simp fs
      -- skolemization and recursive splitting
      recSplit :: [Formula] -> [Formula]
      recSplit fs =
        let fs' = concat (map split fs)
            sfs' = map skol fs'
        in if length fs == length sfs' then sfs' else recSplit sfs'
      fs'' = if ppSkolemization opt then recSplit [f'] else map simp fs
      -- preliminary result, depending on splitting options
      pr = case ppSplitting opt of
             NoSplitting -> [f']
             RootSplitting -> fs
             CompleteSplitting -> fs'
             RecursiveSplitting -> fs''
      -- filter out true formulae
      res = let pr' = filter (not . isVerum) pr in if null pr' then [verum] else pr'
  -- push result on proof stack
  setProofStack (pushManyPS res (g,s) emptyPS)


-- "res = proveAll redStrat" tries to prove all subgoals on the proof stack.
-- Therefore reduction strategy redStrat is applied at most maxStep times.
-- Node information is not required to be set up correctly for the input
-- formulae on the proof stack.
-- Total proof information is updated.
-- On result three cases can hold:
--   1. res == ProofSucceeded: All subgoals could be proved. The subgoal
--      stack returned in the state monad is empty in this case.
--   2. res == ProofFailed: One or more subgoals could not be proved.
--      The subgoal stack contains the first formula which could not be proved
--      as well as all further formulae not considered so far.
--   3. res == MaxStepReached: The proof was suspened due to the max step
--      option. In this case the proof stack contains all currently unproved
--      formulae, with the last considered formula on top.

proveAll :: ReductionStrategy -> IOState ProofResult
proveAll redStrat = do
  -- get proof options, maxSteps and current subgoal stack
  opts <- getProofOptions
  let dbg = debugging opts
  maxSteps <- getMaxSteps
  proofStack <- getProofStack
  -- reduction step limit reached?
  if maxSteps <= 0 then do
    if dbg then do
      putStrLn' "\nMaximal number of proof steps reached, aborting."
     else done
    return MaxStepReached
  -- proof complete?
   else if psIsEmpty proofStack then do
    if dbg then putStrLn' "\nNo further subgoals, aborting." else done
    return ProofSucceeded
   else do
   -- proof stack is not empty, prove next subgoal
    s <- getStep
    res <- proveSubGoal redStrat
    s' <- getStep
    -- subgoal proved?
    if res == PartialProof then do
      -- subgoal proved, next subgoal
      newProofStack <- getProofStack
      if printing opts == OMVerbose then putStrLn' ("\n" ++
        show (psSize newProofStack) ++ " subgoal(s) left to prove.")
       else done
      -- adapt maxSteps
      updateMaxSteps (\ms -> ms - s' + s)
      -- next subgoal proof
      proveAll redStrat
     else
      -- subgoal not (yet) proved, return proof stack
      return res


-- "res = proveSubGoal redStrat" tries to prove one subgoal. The subgoal f to
-- prove is the top of the current proof stack stored in the state monad, so
-- the current proof stack must not be empty.
-- The subgoal f is removed from the proof stack, and reduction strategy
-- redStrat is applied at most maxSteps times, where maxSteps is read from the
-- state monad.
-- The input formula need not have node information set up correctly.
-- proveSubGoal updates generation and prove step, and puts
-- further generated subgoals on the proof stack of the state monad.
-- Total proof information is updated as well.
-- res is one of PartialProof, MaxStepReached or ProofFailed.
-- The resulting formula is stored under workingFormula in the state monad.
-- If the formula f could not be proved because of MaxStepReached,
-- f is pushed back onto the proof stack.

proveSubGoal :: ReductionStrategy -> IOState ProofResult
proveSubGoal redStrat = do
  -- extract proof information from the state monad
  maxVar <- getMaxVar
  maxSteps <- getMaxSteps
  -- initialize proof formula
  ~(Just (f,g,s)) <- getCurrentSubGoal
  setWorkingState f maxVar g s
  updateWorkingFormula (initStrategyInfoGen g maxVar)
  -- delete current subgoal
  delCurrentSubGoal
  -- enter reduction loop
  putStrLn' (replicate 80 '-')
  putStrLn' ("Starting RPC-proof with " ++ show (maxVar+1) ++ " variable(s).")
  t <- doIO getCPUTime
  res <- reductionLoop (maxSteps+s) redStrat
  t' <- doIO getCPUTime
  -- subgoal proved?
  if res == PartialProof then do
    -- compute subproof time
    spt <- getSubProofTime
    let spt' = spt + t' - t
    -- update total proof information
    g' <- getGeneration
    s' <- getStep
    increaseTotalProofGenerationsBy (getGen g' - getGen g)
    increaseTotalProofStepsBy (s'-s)
    increaseTotalProofTimeBy spt'
    -- print message
    putStrLn' "\n***** SUBGOAL  P R O V E D *****\n"
    printProofStatistics True (getGen g') s' spt'
    -- reset subproof time
    setSubProofTime 0
    return PartialProof
   else if res == MaxStepReached then do
    -- adapt subproof time
    increaseSubProofTimeBy (t'-t)
    -- push subgoal back onto proof stack
    (f',_,g',s') <- getWorkingState
    setCurrentSubGoal f' g' s'
    return MaxStepReached
   else return ProofFailed
    


-- "res = reductionLoop maxStep redStrat" applies reduction strategy redStrat
-- to the working formula stored in the state monad until either the step
-- number of the current formula has reached maxStep or the formula is
-- irreducible.
-- In the first case MaxStepReached is returned, in the second case the result
-- depends on the current formula f. If f = true, then PartialProof is returned,
-- otherwise the result is ProofFailed.
-- The input formula must have node information set up correctly.
-- The resulting formula, generation, step and furhter generated
-- subgoals are returned in the state monad.

reductionLoop :: Int -> ReductionStrategy -> IOState ProofResult
reductionLoop maxStep redStrat = do
  -- get output options
  opts <- getProofOptions
  let dbg = debugging opts
      om = printing opts
  -- get current step number and working formula
  s <- getStep
  f <- getWorkingFormula
  -- finished? (i.e. formula irreducible or maxStep prove steps occurred)
  if s >= maxStep then do
    if dbg then
      putStrLn' "\nMaximal number of proof steps reached, aborting."
     else done
    return MaxStepReached
   else if irreducible f then do
     if dbg then putStrLn' "\nFormula is irreducible, aborting."
            else done
     if isVerum f then return PartialProof else return ProofFailed
   else catchS
     -- perform a reduction step
     $$ do -- apply reduction strategy to working formula
           redStrat
           -- simplify result and set up splitting reduction stack
           ~(nf:r) <- simp
           when (not (null r)) $ do
             when (om == OMVerbose) $
               putStrLn' ("\nGenerated " ++ show (length r) ++
                          " further subgoal(s).")
             g' <- getGeneration
             s' <- getStep
             updateProofStack (\stack -> pushManyPS r (g',s') stack)
           -- update current working formula and proof stack
           setWorkingFormula nf
           -- next reduction step
           reductionLoop maxStep redStrat
      $$ \e ->
           let c = getIOSErrCode e
               m = getIOSErrMsg e
           in case c of
             1 -> do putStrLn' ("Reduction strategy terminated " ++
                       "because no further reductions possible!")
                     putStrLn' ("Error message: " ++ m)
                     return ProofFailed
             otherwise -> do
               putStrLn' "Unknown error in reduction strategy!"
               return ProofFailed
 where
   -- simplification of working formula, depending on options
   simp :: IOState [Formula]
   simp = do
     opts <- getProofOptions
     let sp = splitting opts; r = equalModRenSteps opts
         l = unusedLitSteps opts; a = outerAQuants opts
     s <- getStep
     wf <- getWorkingFormula
     let f = if a then dropOuterUniversalQuantifiers wf else wf
     if s `mod` r == 0 then putStrLn' "\n>> Removed subformulae equal modulo variable renaming." else done
     if s `mod` l == 0 then putStrLn' "\n>> Removed unused literals." else done
     let f' = if s `mod` r == 0 then simplify (remEqSubFors (simplify f))
              else simplify f
         simp' = if s `mod` l == 0 then simplify . delNotUnifiableLiterals
                 else simplify
     case sp of
       NoSplitting -> return [f']
       RootSplitting -> return (map simp' (topSplit f'))
       CompleteSplitting -> do updateWorkingFormula (\_ -> f')
                               distribOut
                               f'' <- getWorkingFormula
                               return (map simp' (topSplit f''))


printProofStatistics :: Bool -> Int -> Int -> Integer -> IOState ()
printProofStatistics sg dg ds dt =
  do let header = if sg then "Statistics of subgoal proof:"
                  else "Statistics of complete proof:"
         timemsg = if sg then "  CPU time used for subgoal proof:  "
                   else "  CPU time used for complete proof: "
     putStrLn' header
     putStrLn' ("  RPC-reductions performed:         " ++ show dg)
     putStrLn' ("  major reduction steps:            " ++ show ds)
     putStrLn' (timemsg ++ show (dt `div` 1000000000) ++ " ms")
