----------------------------------------------------------------------
-- Module ProofStack.hs                                             --
--                                                                  --
-- Implements the abstract data type proof stack, which is used     --
-- to handle subgoal storage.					    --
--                                                                  --
-- 03.12.96  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module ProofStack (
  -- data type
  ProofStack,
  -- constructors
  emptyPS,		-- empty proof stack
  pushPS,		-- push a formula onto the proof stack
  -- operators
  pushManyPS,		-- push a series of formulae onto the proof stack
  topPS,		-- get top element of proof stack
  psIsEmpty,		-- is the proof stack empty?
  popPS,		-- pop top element from proof stack
  topPopPS,		-- combination of top and pop
  setTopPS,		-- change top element of proof stack
  psSize,		-- determine the number of formulae on the proof stack
  mergePS		-- merge two proof stacks into one
) where


----------------------------
----- imported modules -----
----------------------------

import Formula


----------------------
----- data types -----
----------------------

-- This is the proof stack data type:
-- The parameter a is used to store additional information for
-- each formula.

newtype ProofStack a = MkPS [(Formula, a)]
  deriving Eq


------------------------------
----- exported functions -----
------------------------------

-- make an empty proof stack

emptyPS :: ProofStack a
emptyPS = MkPS []

-- Push a formula and associated formula information onto the
-- proof stack.

pushPS :: Formula -> a -> ProofStack a -> ProofStack a
pushPS f i (MkPS ps) = MkPS ((f,i):ps)

-- Push a list of formulae with common associated information onto
-- the proof stack.

pushManyPS :: [Formula] -> a -> ProofStack a -> ProofStack a
pushManyPS [] _ s = s
pushManyPS (f:fs) i s =
  let n = pushManyPS fs i s
  in pushPS f i n

-- get the top element (formula and information) of the proof stack

topPS :: ProofStack a -> (Formula, a)
topPS (MkPS []) = error "empty proof stack in topPS"
topPS (MkPS (t:_)) = t

-- Test whether the proof stack is empty or not.

psIsEmpty :: ProofStack a -> Bool
psIsEmpty (MkPS []) = True
psIsEmpty _ = False

-- Drop the top element of the proof stack.

popPS :: ProofStack a -> ProofStack a
popPS (MkPS []) = error "empty proof stack in popPS"
popPS (MkPS (_:r)) = MkPS r

-- Get the top element (formula and information) of the proof stack.
-- The top element is deleted from the returned proof stack.

topPopPS :: ProofStack a -> (Formula, a, ProofStack a)
topPopPS (MkPS []) = error "empty proof stack in topPopPS"
topPopPS (MkPS ((f,i):r)) = (f,i, MkPS r)

-- Change the top element of the proof stack.

setTopPS :: Formula -> a -> ProofStack a -> ProofStack a
setTopPS _ _ (MkPS []) = error "empty proof stack in setTopPS"
setTopPS f i (MkPS (_:r)) = MkPS ((f, i):r)

-- Determine the size of the proof stack.

psSize :: ProofStack a -> Int
psSize (MkPS s) = length s

-- Merge two proof stacks. The formulae of the first proof
-- stack lie on top of the formulae of the second proof stack.

mergePS :: ProofStack a -> ProofStack a -> ProofStack a
mergePS (MkPS ps) (MkPS ps') = MkPS (ps ++ ps')
