module ProofManager where

import IOStateMonad
import Formula
import Strategy
import Simplify
import LiteralPair
import ProofStack
import RedStratTypes
import Subsumption
import Reduction
import Auxiliary


-- input formula must be in nnf and simplified
preProcess :: Formula -> Generation -> Step -> PreProcessOptions
  -> IO (ProofStack (Generation, Step))
preProcess f g s opt = do
  let skol = minimizeVarAndConstDomains . partialSkolemization . simplify .
             deleteOuterUniversalQuantifiers . minimizeQuantifierDomains
      spl t = if t == RootSplitting then topSplit else split
      -- first skolemization step
      f' = if ppSkolemization opt then skol f else f
      -- splitting step, if ppSplitting /= NoSplitting
      fs = spl (ppSplitting opt) f'
      -- skolemization after complete splitting
      fs' = map skol fs
  if ppSplitting opt == NoSplitting then return (pushPS f' (g,s) emptyPS)
   else -- splitting demanded
     if ppSplitting opt == RootSplitting then
        return (pushManyPS fs (g,s) emptyPS)
      else return (pushManyPS fs' (g,s) emptyPS)

-- proveFor tries to prove one formula without splitting and skolem.

simpleProve :: Step -> Int -> Generation -> Step -> ReductionStrategy
  -> ProveOptions -> Formula -> IO Formula
simpleProve maxStep maxVar g s redStrat opt f = do
  putStrLn (replicate 80 '-')
  putStrLn ("Strarting proof with " ++ show maxVar ++ " variable(s).")
  let initf = initStrategyInfo g maxVar f
  f' <- initForGenIOSt (simpleProveStep maxStep redStrat opt) initf g s
  return f'


simpleProveStep :: Step -> ReductionStrategy -> ProveOptions
  -> ForGenIOSt Formula
simpleProveStep maxStep redStrat opt = do
  (f,g,s) <- getForGenStep
  -- proof finished?
  if s >= maxStep then do
    printWith putStrLn "Maximal number of proof steps reached, aborting."
    printWith putStrLn (replicate 80 '-')
    return f
   else if irreducible f then do
     printWith putStrLn "Formula is irreducible, aborting."
     printWith putStrLn (replicate 80 '-')
     return f
   else catchS
    -- perform reduction step
      $$ do let om = printing opt
            redStrat opt
            (f',g',s') <- getForGenStep
            if om >= OMStandard then 
              printWith putStrLn ("reduction step #" ++ show s' ++ " completed, "
                ++ show (g'-g) ++ " RPC-reduction(s), " ++ show g' ++ " in total.")
             else return ()
            -- simplify result
            let nf = (simplify . remEqSubFors . simplify) f'
            setFormula nf
            simpleProveStep maxStep redStrat opt
      $$ \e ->
           let c = getIOSErrCode e
               m = getIOSErrMsg e
           in case c of
             1 -> do printWith putStrLn ("Reduction strategy terminated " ++
                       "because no further reductions possible!")
                     printWith putStrLn ("Error message: " ++ m)
                     return f
             otherwise -> do
               printWith putStrLn "Unknown error in reduction strategy!"
               return f
