----------------------------------------------------------------------
-- Module Reduction.hs                                              --
--                                                                  --
-- Reductions, strategy info maintenance algorithms, computation    --
-- of all possible reductions, irreducibility test.                 --
--                                                                  --
-- 11.12.96  initial revision                                       --
----------------------------------------------------------------------


--------------------
----- comments -----
--------------------

-- Not all reductions of the RPC calculus are handled within this module.
-- The simpler boolean reduction rules, with the exception of the distributive
-- law, are handled in module `Simplify' directly.
-- Thus, only three kinds of reduction rules are managed here:
-- o distributive law rule:     A \/ (B /\ C)    --> (A \/ B) /\ (A \/ C)
-- o existential quant. rule:   (exists x)A      --> (exists x)A \/ A[x/y]
-- o universal quant. rule (UQR):
--     (forall x)A \/ B --> (forall x)A \/ A[-x]		(UQR 1)
--                          \/ (forall y)((forall y)B \/ A[x-y][x/y]) \/ B
--     (forall x)A      --> (forall x)A \/ A[-x]		(UQR 2)
--     (forall x)A \/ B --> (forall y)(A[x/y] \/ B),            (UQR 3)
--                          if y is not free in A or B
-- Which of the rules (UQR 1-3) is applied depends on context:
-- When part B is not present, (UQR 2) is applied, otherwise selection depends
-- on the condition mentioned after (UQR 3): if variable y does not
-- occur free in A as well as B then (UQR 3) is chosen, otherwise (UQR 1).
--
-- A reduction rule application consists of the following steps:
--   1. Take an input formula F with correctly set-up strategy info.
--   2. Perform a reduction using `applyRule'. This step consists of
--      three sub-steps:
--      a) Perform the reduction step proper.
--      b) Set up node information for newly created subformulae
--         (this is done by `initSubStrategyInfo').
--      c) Update strategy info field used for the current reduction,
--         i.e. the appropriate component of the strategy info is erased.
--   3. Update all strategy info fields using `updateStrategyInfo'.
--      This exclusively alters conjunction marks and those quantifier marks
--      which incorporate the same term as used for the reduction.
-- After these three steps the precondition for step 1 is fulfilled again.
-- The output formula F' obtained after applying the beforementioned three
-- steps is neither flat nor simplified.


----------------------------
----- module interface -----
----------------------------

module Reduction (
  -- data types
  ForListPartitioner,	-- partitions a list of formulae in a global IO state monad
  ReductionSelector,	-- chooses one reduction possibility in an IO state monad
  RuleType (		-- indicates the kind of rule for reduction:
    DistributionRule,	-- o distributive law
    ExistentialRule,	-- o existential quantifier reduction
    UniversalRule	-- o universal quantifier reduction
  ),			-- other rules are not handled explicitly, but in mod. Simplify
  ReductionInfo,	-- information about a reduction possibility
  -- operations
  resetStrategyInfo,	-- reset all strategy information fields to initial value
  initStrategyInfo,	-- initialize all strategy info fields
  initStrategyInfoGen,	-- init. all strat. info fields for a spec. generation
  initSubStrategyInfo,	-- initialize strategy info fields for a subformula
  updateStrategyInfo,	-- update strategy info fields after a reduction step
  findReductions,	-- searches for admissable reduction possibilities
  irreducible,		-- indicates whether or not a formula is irreducible
  applyRule,		-- performs a reduction step

  applyDRule 	-- @!!!!!! only used by Subsumption:distribOut
) where


----------------------------
----- imported modules -----
----------------------------

import List ( (\\) )
import Set
import Formula
import Auxiliary
import IOStateMonad
import GlobalState


----------------------
----- data types -----
----------------------

-- A formula list partitioner splits a list into two lists of formulae.
-- This is done within the global IO state monad in order to give access
-- to global parameters and to perform IO operations.
-- The partitioner is used to select disjunctions and conjunctions
-- used for a reduction rule and those not incorporated in that rule.
-- The distributive and universal quantifier rule allow different disjunctions
-- in the redex. Moreover, in the case of the distributive rule, the
-- conjunction has to be split into two.
-- Different formula list partitioners are specified in the module `Strategy'.

type ForListPartitioner = [Formula] -> IOState ([Formula],[Formula])

-- Reduction information identifies a reduction possibility.
-- A reduction is uniquely determined by the reduction position and
-- the term to be used as a reduction parameter. To simplify application
-- of a reduction rule, the rule type which is to be applied at the
-- reduction position is also provided.
-- Node information at the reduction position may be used by reduction
-- strategies to select a specific reduction, though this information
-- has to be extracted from the working formula stored in the global
-- state monad.

type ReductionInfo = (Position, RuleType, [Term])

-- A reduction selector chooses one reduction out of all possible ones.
-- It works within the global IO state monad to allow access to the
-- global parameters and to perform IO.
-- Different reduction selectors are specified in the module `Strategy'.

type ReductionSelector = IOState ReductionInfo


-- The rule type specifies the kind of rule to be applied. Simpler
-- reductions (boolean part) are not handled here, but in module `Simplify'.

data RuleType = DistributionRule | ExistentialRule | UniversalRule
  deriving Eq


---------------------------------
----- instance declarations -----
---------------------------------

-- to show rule types (for debugging purposes)

instance Show RuleType where
  showsPrec _ DistributionRule = showString "/\\-rule"
  showsPrec _ ExistentialRule = showString "exists-rule"
  showsPrec _ UniversalRule = showString "forall-rule"


------------------------------
----- exported functions -----
------------------------------

---------------------------------------------------
-- functions for strategy information management --
---------------------------------------------------

-- "f' = resetStrategyInfo f" resets all node information fields in formula f
-- to their initial values, resulting in formula f'. f' is unchanged in other
-- respects.
-- More specifically, cunjunction and quantifier nodes get an empty strategy
-- information list, all other nodes are assigned node information Nothing.

resetStrategyInfo :: Formula -> Formula
resetStrategyInfo f
  | isConj f   = setNodeInfo (mkNodeInfo []) f'
  | isExists f = setNodeInfo (mkNodeInfo []) f'
  | isForAll f = setNodeInfo (mkNodeInfo []) f'
  | otherwise  = setNodeInfo emptyNodeInfo f'
  where f' = setSubFormulae (map resetStrategyInfo (getSubFormulae f)) f

-- "f' = initStrategyInfo m f" initializes the strategy information fields
-- of formula f as it is needed by all reduction strategies.
-- m is the highest variable index to be used in the calculus.
-- All conjunction nodes are marked with the empty term (and generation 0),
-- all universal quantifier nodes are marked with all variables plus one
-- new variable not occurring in formula f (all with generation 0),
-- existential quantifier nodes are marked with all terms occurring in
-- formula f.

initStrategyInfoGen :: Generation -> Int -> Formula -> Formula
initStrategyInfoGen g mv f =
  initSubStrategyInfo g mv (vars f) (consts f) False f

initStrategyInfo :: Int -> Formula -> Formula
initStrategyInfo = initStrategyInfoGen initialGen


-- "initSubStrategyInfo" initializes the node info fields of a formula
-- which is needed by all reduction strategies.
-- In "f' = initSubStrategyInfo g mv vs cs d f" g denotes the generation of the
-- formula, mv is the highest variable index to be used, vs and cs are
-- the sets of variables of a formula f* and f is a subformula of f*.
-- (f* is the main formula which is to be proved, f is a subformula incorporated
-- in the reduction step.)
-- d is true if the direct superformula of f is a disjunction, otherwise false.
-- The result f' is f with the appropriate node information inserted.

initSubStrategyInfo :: Generation -> Int ->
  Set Variable -> Set Constant -> Bool -> Formula -> Formula
initSubStrategyInfo n maxv vs cs d f = initSI n d f
  where tvs = map termFromVar (setToOrdList vs)  -- list of variables in main formula
        tcs = map termFromCon (setToOrdList cs)	 -- list of constants in main formula
        mnv = newVar' maxv vs	-- fresh variable
        tvs' = if mnv == Nothing then tvs	 -- variable list expanded by new variable
               else ((\(Just x) -> termFromVar x)mnv):tvs
        initSI :: Generation -> Bool -> Formula -> Formula
        initSI n sfd f
         | isDisj f = setSubFormulae (map (initSI n True) (getSubFormulae f)) f
         | isConj f =
             if sfd then -- superformula is disjunction
               setNodeInfo (mkNodeInfo [(n, Nothing)]) (setSubFormulae
                 (map (initSI n False) (getSubFormulae f)) f)
             else -- superformula is no disjunction
               setNodeInfo emptyNodeInfo (setSubFormulae (map (initSI n False)
                 (getSubFormulae f)) f)
         | isExists f = setNodeInfo (mkNodeInfo
            (map (\t -> (n, Just t)) (tvs ++ tcs)))
            (setSubFormula ((initSI n False) (getSubFormula f)) f)
         | isForAll f =
             if sfd then -- superformula is disjunction
               setNodeInfo (mkNodeInfo (map (\v -> (n, Just v)) tvs'))
                 (setSubFormula ((initSI n False) (getSubFormula f)) f)
             else -- superformula is no disjunction
               setNodeInfo (mkNodeInfo [(n,Nothing)])
                 (setSubFormula ((initSI n False) (getSubFormula f)) f)
         | otherwise = setSubFormulae (map (initSI n False)(getSubFormulae f)) f
         where nni si = if sfd then mkNodeInfo si else emptyNodeInfo

-- "f' = updateStrategyInfo g m nv t f" marks new reduction possibilities in
-- formula f.
-- g is the generation of the last prove step, m is the highest variable index
-- to be used, nv indicates whether or not a new variable not occurring in
-- formula f was used as reduction parameter in the last reduction,
-- t is the reduction term used for the last reduction (if any).
-- Then all nodes of formula f are updated, such that a new reduction step
-- can be applied afterwards. f must be the whole formula to prove, not a
-- subformula (e.g. the reduced subformula).
-- No strategy information is deleted, only new information inserted. This
-- comprises:
-- o conjunctions with no mark are re-marked
-- o universal quantifiers are re-marked with all variables occurring in formula
--   f plus one fresh variable, if the direct superformula is a disjunction;
--   otherwise the quantifer is necessarily marked with the empty term
-- o existential quantifiers are re-marked with term t if it was a newly generated
--   variable

updateStrategyInfo :: Generation -> Int -> Bool -> Maybe Term -> Formula
  -> Formula
updateStrategyInfo g maxv newV mt f = updateSI g False f
  where Just t = mt
        vs = map termFromVar (setToOrdList (vars f))	-- variable list of formula f
        cs = map termFromCon (setToOrdList (consts f))  -- constant list of formula f
        mnv = newVar maxv f    -- fresh variable
	-- expanded list of variables (by new variable)
        vs' = if mnv == Nothing then vs else ((\(Just x) -> termFromVar x)mnv):vs
        -- update information
        updateSI :: Generation -> Bool -> Formula -> Formula
        updateSI g sfd f     -- sfd: superformula is disj.
          | isDisj f = nf True f -- nothing to do for disjunctions
          | isConj f =
              if sfd then	 -- superformula is disjunction
                if null si	 -- re-mark, if it is not marked already
                  then setNodeInfo (mkNodeInfo [(g, Nothing)]) (nf False f)
                else nf False f	 -- no change required, already marked
              else setNodeInfo (mkNodeInfo []) (nf False f) -- no reduction possib.
          | isExists f = if newV then setNodeInfo -- re-mark, if new variable inserted
             (mkNodeInfo (addNI (g,mt) si)) (nf False f) else nf False f
          | isForAll f =
              if sfd then        -- direct superformula is disjunction
                -- not yet marked or just marked with empty term?
                if null si || snd (head si) == Nothing then
                  -- all terms have to be added
                  setNodeInfo (mkNodeInfo [(g,Just t) | t <- vs']) (nf False f)
                else
                  -- was already marked, just update missing terms
                  setNodeInfo (mkNodeInfo ([(g,Just t) | t <- vs' \\ si'] ++ si))
                              (nf False f)
              else -- direct superformula is no disjunction
                -- not yet marked or just marked with empty term?
                if null si || snd (head si) == Nothing then
                  -- update node info
                  setNodeInfo (mkNodeInfo [(g,Nothing)]) (nf False f)
                -- keep reduction terms unchanged for further usage
                else nf False f
          | otherwise = nf False f
          where si = getStrategyInfo (getNodeInfo f)
                si' = getSITerms si
        -- update subformulae: first parameter indicates whether the
        -- direct superformula is a disjunction
        nf :: Bool -> Formula -> Formula
        nf d f = setSubFormulae (map (updateSI g d) (getSubFormulae f)) f
        -- add node information for a term, if not already present
        addNI :: (Generation, Maybe Term) -> StrategyInfo -> StrategyInfo
        addNI (g,mt) [] = [(g,mt)]
        addNI (g,mt) x@((g',mt'):xs) = if mt == mt' then x
          else (g', mt') : addNI (g,mt) xs


------------------------
-- reducibility tests --
------------------------

-- "r = findReductions f" collects all reduction possibilities in formula
-- f and gathers them in a reduction information list.
-- Only reduction possibilities for distributive and quantifier rules
-- is collected, as the other (boolean) reductions are handled in module
-- `Simplify'.
-- Strategy information has to be set-up correctly.

findReductions :: Formula -> [ReductionInfo]
findReductions f = findrpos lambda f where
  findrpos :: Position -> Formula -> [ReductionInfo]
  findrpos p f
    | isConj f && nnsi = (p, DistributionRule, getSITerms si) : prs
    | isExists f && nnsi = (p, ExistentialRule, getSITerms si) : prs
    | isForAll f && nnsi = (p, UniversalRule, getSITerms si) : prs
    | otherwise = prs
    where sf  = getSubFormulae f
          l   = length sf
          -- reduction possibilities in subformulae
          prs = [pr | i <- [0..l-1], pr <- findrpos (mkPos p i) (sf!!i)]
          -- mark reduction possibility only if there is any term
          -- in the strategy information
          si = getStrategyInfo (getNodeInfo f)
          nnsi = not (null si)

-- "t = irreducible f" returns True if formula f is irreducible, False
-- otherwise. Strategy information must be set-up correctly to deliver
-- sensible results.

irreducible :: Formula -> Bool
irreducible = null . findReductions


--------------------------------
-- application of RPC_n rules --
--------------------------------

-- "applyRule r p udp ddp dcp t" applies a reduction rule of calculus
-- RPC to the working formula stored in the global state monad.
-- r is the kind of rule to apply, p is the reduction position
-- (in the sense that the performed reduction is a p-reduction),
-- udp, ddp and dcp are formula list partitioners, where udp is used
-- to select the appropriate disjunction for universal quantifier rule
-- application, ddp is used to select the appropriate disjunction for
-- distribution rule application, and dcp is used to partition the
-- conjunction for the same rule.
-- t may be a term used as reduction parameter. t should be set to Nothing
-- when a distribution rule is to be applied, in case of a universal
-- quantifier rule it has to denote a variable, for existential quantifier
-- rule applications it may be any term.
-- The formula stored in the state monad must be in negation normal form and
-- should be flat in order to make the formula list partitioners to work as
-- expected.
-- The resulting formula is in negation normal form, but neither flat
-- nor simplified.
-- Strategy information is updated, such that before the next reduction
-- step no further change on strategy information is required.
-- The generation (stored in the global state monad) is increased by one.

applyRule :: RuleType -> Position -> Maybe ForListPartitioner -> Maybe ForListPartitioner
  -> Maybe ForListPartitioner -> Maybe Term -> IOState ()
applyRule r p mDisjSelA mDisjSelD mConjSelD mt =
  do let Just t = mt	-- extract reduction parameters and partitioners
         Just disjSelA = mDisjSelA
         Just disjSelD = mDisjSelD
         Just conjSelD = mConjSelD
         v = getTermVar t
     -- get formula before reduction step
     f <- getWorkingFormula
     case r of		-- apply rule, depending on type (performed on state monad)
       DistributionRule -> applyDRule p disjSelD conjSelD
       ExistentialRule -> applyERule p t
       UniversalRule -> applyARule p disjSelA v
     -- get formula after reduction step
     (f',maxVar,g,_) <- getWorkingState
     -- was a new variable introduced in the reduced formula?
     let nv = mt /= Nothing && termIsVar t && v `isElement` vars f'
                            && not (v `isElement` vars f)
     -- update strategy information and working formula, increase generation
     setWorkingFormula (updateStrategyInfo (nextGen g) maxVar nv mt f')
     nextGeneration

-- "applyDRule p ddp dcp" applies a distribution rule to the working
-- formula stored in the global state monad. p is the reduction position
-- in the sense that the performed recuction is a p-reduction, ddp and
-- dcp are formula list partitioners, where ddp is used to determine
-- the disjunctions used for the reduction step (part A, below),
-- dcp is used to split the conjunction (parts B and C, below).
-- The rule performed is:
--   A \/ (B /\ C) --> (B \/ A) /\ (C  \/ A)
-- The input formula must be in negation normal form and should be flat
-- in order to make the partitioners work as expected.
-- The resulting formula is in negation normal form, but not necessarily
-- flat nor simplified; strategy information at the /\ - node is deleted,
-- other strategy information is not updated, so this has to be done
-- explicitly afterwards, as is e.g. done by `applyRule'.
-- The generation field in the state monad is not updated.

applyDRule :: Position -> ForListPartitioner -> ForListPartitioner -> IOState ()
applyDRule p disjSel conjSel =
  do f <- getWorkingFormula
     let posList = posToList p                -- p is the position of /\
         disjPos = listToPos (init posList)   -- disjPos is the position of \/
         i = last posList                     -- conj. index in disj.
         disjFs = getSubFormulae (f@@disjPos) -- the subformulae of the disj.
         disjRest = take i disjFs ++ drop (i+1) disjFs  -- `` without conj.
         conjFs = getSubFormulae (disjFs!!i)  -- get subformulae of conj.
     (a,na) <- disjSel disjRest    -- select part A of disj. (na is the rest)
     (b,c) <- conjSel conjFs       -- select parts B and C of conj.
     let ni = getNodeInfo (f@@p)   -- get node info of /\
         b' = if not (null (tail b)) then setNodeInfo ni (conj b) else head b
         c' = if not (null (tail c)) then setNodeInfo ni (conj c) else head c
         -- g' is the new, reduced subformula. Note that if na is empty,
         -- the disj.-constructor doesn't build a disjunction.
         g' = disj (conj [disj (b' : a), disj (c' : a)] : na)
     setWorkingFormula (setSubFormulaAtPos disjPos g' f)

-- "applyERule p t" applies an existential quantifier rule to the working
-- formula stored in the global state monad. p is the reduciton position
-- in the sense that the performed reduction is a p-reduction, t
-- is the term used as reduction parameter.
-- The rule performed is:
--   (exists x)A --> (exists x)A \/ A[x/t]
-- The resulting formula is neither flat nor simplified.
-- Strategy information for the existential quantifier is adapted, but
-- no further changes on the information is made. So this has to be done
-- explicitly afterwards.
-- The generation field in the state monad is not updated.

applyERule :: Position -> Term -> IOState ()
applyERule p t =
  do (f,maxVar,g,_) <- getWorkingState
     let exa = f@@p            -- get subfor. to be reduced, i.e. (exists x) A
         a = getSubFormula exa -- get subformula A
         x = getQuantVar exa   -- get bound variable x
         si = getStrategyInfo (getNodeInfo exa) -- get strat. i. of ex. quant.
         -- remove those parts of strategy info which correspond with term t
         siFilter p = snd p /= Just t         -- set up strategy info filter
         newSI = setNodeInfo (mkNodeInfo (filter siFilter si))
         -- build A[x/a] and set new strategy info for this subformula
         a' = initSubStrategyInfo (nextGen g) maxVar (vars f) (consts f) True (subst x t a)
         -- exa' is the new, reduced subformula
         exa' = disj [newSI exa, a']
     setWorkingFormula (setSubFormulaAtPos p exa' f)

-- "applyARule p udp y" applies a universal quantifier rule to the
-- working formula stored in the global state monad. p is the reduction
-- position in the sense that the performed reduction is a p-reduction.
-- udp is a formula list partitioner, which is used to select part B of
-- the formula (see below). y is the variable used as reduction parameter,
-- when required.
-- The rule performed is one of the following three, depending on context:
--   (forall x)A \/ B --> (forall x)A \/ A[-x]		(UQR 1)
--                         \/ (forall y)((forall y)B \/ A[x-y][x/y]) \/ B
--    or 
--   (forall x)A --> (forall x)A \/ A[-x]		(UQR 2)
--    or						(UQR 3)
--   (forall x)A \/ B --> (forall y)(A[x/y] \/ B), if y is not free in A and B
-- Rule (UQR 2) is applied if no part B is present, otherwise rule (UQR 3)
-- is performed, whenever the mentioned condition holds. When neither
-- (UQR 2) nor (UQR 3) can be used, a reduction with (UQR 1) is made.
-- The input formula stored in the global state monad must be in negation normal
-- form and should be flat in order to make the formula list partitioner
-- to work as expected.
-- The resulting formula is in negation normal form, but neither flat
-- nor simplified. Node information for the universal quantifier is updated,
-- all other strategy information is kept unchanged.
-- The generation field in the state monad is not updated.

applyARule :: Position -> ForListPartitioner -> Variable -> IOState ()
applyARule p disjSel y =
  do f <- getWorkingFormula
     let -- determine which kind of rule to apply
         posList = posToList p  -- p is the position of forall
         disjPos = listToPos (init posList)
         kind1 = not (null posList) && isDisj (f@@disjPos)
     if kind1 then applyARuleKind1 p disjSel y
              else applyARuleKind2 p


-- Apply universal quantifier rule (UQR 1). Specification as denoted above.
-- The performed reduction is:
--   (forall x)A \/ B --> (forall x)A \/ A[-x]
--                        \/ (forall y)((forall y)B \/ A[x-y][x/y]) \/ B

applyARuleKind1 :: Position -> ForListPartitioner -> Variable -> IOState ()
applyARuleKind1 p disjSel y =
  do (f,maxVar,g,_) <- getWorkingState
     let posList = posToList p                -- p is the position of forall
         disjPos = listToPos (init posList)   -- disjPos is the position of \/
         i = last posList                     -- forall index in disj.
         disjFs = getSubFormulae (f@@disjPos) -- all subformulae of the disj.
         disjRest = take i disjFs ++ drop (i+1) disjFs  -- `` without forall
         faa = disjFs!!i                      -- subformula (forall x)A
         a = getSubFormula faa                -- subformula A
         x = getQuantVar faa                  -- bound variable x
     (b,_) <- disjSel disjRest  -- select part B of disj. (b is a list of fors!)
     if null b then applyARuleKind2 p -- no part B due to selector function
      else if not (y `isElement` (free a `sunion` free (disj b))) then
       applyARuleKind3 p disjSel y
      else do
       let si = getStrategyInfo (getNodeInfo faa) -- get strat.info of fa. quant.
           ty = termFromVar y                     -- variable y as term
           -- remove those parts of strategy info which corresponds with term t
           siFilter p = snd p /= Just ty          -- set up strategy info filter
           newSI = setNodeInfo (mkNodeInfo (filter siFilter si))
           -- build A[-x] and set new strategy info for this subformula
           a' = initSubStrategyInfo (nextGen g) maxVar (vars f) (consts f) True
                  (fdelete x a)
           -- build A[x-y][x/y] and (forall y)B
           a'' = subst x ty (fdelete2 x y a)
           b' = forall y (disj b)
           -- build (forall y)((forall y)B \/ A[x-y][x/y]) and set up strat. info
           r = initSubStrategyInfo (nextGen g) maxVar (vars f) (consts f) True
                 (forall y (disj [b', a'']))
           -- faa' is the new, reduced subformula
           faa' = disj (newSI faa : a' : r : disjRest)
       setWorkingFormula (setSubFormulaAtPos disjPos faa' f)


-- Apply universal quantifier rule (UQR 2). Specification as denoted above.
-- The reduction performed is:
--   (forall x)A --> (forall x)A \/ A[-x]

applyARuleKind2 :: Position -> IOState ()
applyARuleKind2 p =
  do (f,maxVar,g,_) <- getWorkingState
     let -- get subformulae and bound variable
         faa = f@@p                   -- subformula (forall x)A
         a = getSubFormula faa        -- subformula A
         x = getQuantVar faa
         -- rebuild node info at (forall x)A
         newSI = setNodeInfo (mkNodeInfo [])
         -- build A[-x]
         a' = initSubStrategyInfo (nextGen g) maxVar (vars f) (consts f) True
                (fdelete x a)
         -- faa' is the new, reduced subformula
         faa' = disj [newSI faa, a']
     setWorkingFormula (setSubFormulaAtPos p faa' f)


-- Apply universal quantifier rule (UQR 3). Specification as denoted above.
-- The reduction performed is:
--   (forall x)A \/ B --> (forall y)(A[x/y] \/ B)
-- NOTE: y must not occur free in A or B. It is safe to call `applyARuleKind1'
--       to ensure this and perform this rule (UQR 3) though.

applyARuleKind3 :: Position -> ForListPartitioner -> Variable -> IOState ()
applyARuleKind3 p disjSel y =
  do (f,maxVar,g,_) <- getWorkingState
     let posList = posToList p       		-- p is the position of forall
         disjPos = listToPos (init posList)	-- disjPos is the position of \/
         i = last posList			-- forall index in disjunction
         disjFs = getSubFormulae (f@@disjPos)	-- all subformulae of the disj.
         disjRest = take i disjFs ++ drop (i+1) disjFs	-- `` without forall
         faa = disjFs!!i			-- subformula (forall x)A
         a = getSubFormula faa			-- subformula A
         x = getQuantVar faa			-- bound variable x
         ty = termFromVar y			-- variable y as term
     (b,nb) <- disjSel disjRest  -- select part B of disj (B is a list of fors!)
     if null b then applyARuleKind2 p -- no part B due to sel. func., switch to kind 2
      else do
       let -- build A[x/y]
           a' = subst x ty a
           -- rebuild node info at (forall x)A
           newSI = setNodeInfo (mkNodeInfo [])
           -- build (forall y)(A[x/y] \/ B) and set up strat. info
           r = newSI (forall y
                 (initSubStrategyInfo (nextGen g) maxVar (vars f) (consts f) True
                  (disj (a':b))))
           -- faa' is new, reduced subformula
           faa' = disj (r:nb)
       setWorkingFormula (setSubFormulaAtPos disjPos faa' f)
