----------------------------------------------------------------------
-- Module Scanner.hs						    --
--                                                                  --
-- This module implements scanners for variables, constants,        --
-- terms, formulae, etc.        				    --
-- Moreover, it delivers functions for symbol table management.     --
--                                                                  --
-- 15.01.97  initial revision                                       --
-- 12.01.00  RA support added  					    --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Scanner (
  -- data types
  FTokenPL (		-- PL1 formula token
    TokVar, TokCon, TokPred, TokTrue, TokFalse,
    TokOr, TokAnd, TokNot, TokImplies, TokEquiv,
    TokExists, TokForAll, TokOpenPar, TokClosePar,
    TokComma, TokSemicolon, TokEOF, TokError
  ),
  FTokenRA (		-- RA formula token
    RATokPred,
    RATokRelProd, RATokAbsProd, RATokRelSum, RATokAbsSum,
    RATokRelOne, RATokAbsOne, RATokAbsZero, RATokConverse,
    RATokNegation, RATokEquals, RATokIncludes, RATokTurnstile,
    RATokSemicolon, RATokComma, RATokOpenPar, RATokClosePar,
    RATokEOF, RATokError
  ),
  ScanPos,		-- position in current input file (line, column)
  Token,
  TokenPL,		-- token consisting of PL1 formula token and scan pos.
  TokenRA,		-- token consist. of RA formula token and scan pos.
  Scanner,		-- scanner data type
  IdentKind,		-- kind of an identifier: constant, variable or predicate
  IdentValue,		-- value of an identifier: usually id-index
  SymbolTable,  	-- the symbol table

  -- operations on symbol tables
  emptySymbolTable,	-- constructor
  initialRASymbolTable, -- contains 3 variables (x,y,z), and equality predicate i
  stLookupByName,	-- look-up functions
  stLookupVarByName,
  stLookupConByName,
  stLookupTermByName,
  stLookupByPred,
  stLookupByVar,
  stLookupByCon,
  stLookupPredRank,
  stSetPredRank,       	-- set rank of a predicate

  -- PL formula scanner
  forScannerPL,		-- the PL1 formula scanner
  
  -- RA formula scanner 
  forScannerRA          -- the RA formula scanner
) where


----------------------------
----- imported modules -----
----------------------------

import Formula
import Set
import Char
import List


----------------------
----- data types -----
----------------------

-- formula tokens

data FTokenPL = TokVar Variable | TokCon Constant -- PL tokens
              | TokPred Predicate
              | TokTrue         | TokFalse
              | TokOr           | TokAnd
              | TokNot
              | TokImplies      | TokEquiv
              | TokExists       | TokForAll
              | TokOpenPar      | TokClosePar
              | TokComma        | TokSemicolon
              | TokEOF          | TokError
  deriving Eq

data FTokenRA = RATokPred Predicate		  -- RA tokens
              | RATokRelProd	| RATokAbsProd
              | RATokRelSum	| RATokAbsSum
              | RATokRelOne	| RATokAbsOne	| RATokAbsZero
              | RATokConverse	| RATokNegation
              | RATokEquals	| RATokIncludes
              | RATokTurnstile	| RATokSemicolon
	      | RATokComma
              | RATokOpenPar    | RATokClosePar
              | RATokEOF	| RATokError
  deriving Eq

-- scan position

type ScanPos = (Int,Int) -- line, column


-- The tokens of the formula scanners consist of FTokens and position
-- information:

type Token ftok = (ftok, ScanPos)
type TokenPL = Token FTokenPL
type TokenRA = Token FTokenRA


-- A scanner of type a is a state transformer, where the 
-- state is a triple consisting of a symbol table, a scan position
-- and a string, which represents the characters to be read.
-- A scanner uses and updates the state and returns a value of type a.

type Scanner a = SymbolTable -> ScanPos -> String
  -> (a, SymbolTable, ScanPos, String)


-- identifier kind: variable or constant or predicate identifier

data IdentKind = VarIdent | ConIdent | PredIdent
  deriving Eq

-- value of an identifier: which variable or constant or predicate

data IdentValue = IVVar Variable | IVCon Constant | IVPred Predicate
  deriving Eq

-- Symbol table entries consist of the name (string) of the identifier,
-- the kind (variable, constant or predicate) and the value.

type SymTabEntry = (String, IdentKind, IdentValue)

-- The symbol table contains a list of symbol table entries, and, for
-- more convenient access, the values stored in the symbol table.
-- I.e. all known identifiers are additionally stored in the set of strings,
-- all variables in the set of variables, etc.

type SymbolTable = (Set Variable, Set Constant, Set Predicate,
                    Set String, [SymTabEntry])


------------------------------
----- exported functions -----
------------------------------

-- scanner functions

-- The PL1 and RA formula scanners take a symbol table and a string to parse,
-- and return a string of tokens together with the updated symbol table.
-- Input is read until an error occurs or EOF is reached. In the first
-- case, the remaining input is discarded.

forScannerPL :: SymbolTable -> String -> ([TokenPL], SymbolTable)
forScannerPL symTab input = scanItPL symTab (1,1) input
  where scanItPL st sp is =
          let (tok,st',sp',is') = scanPL st sp is
              (t,_) = tok
              (toks,st'') = scanItPL st' sp' is'
          in if t == TokEOF || t == TokError then ([tok], st')
             else (tok:toks, st'')

forScannerRA :: SymbolTable -> String -> ([TokenRA], SymbolTable)
forScannerRA symTab input = scanItRA symTab (1,1) input
  where scanItRA st sp is =
          let (tok,st',sp',is') = scanRA st sp is
              (t,_) = tok
              (toks,st'') = scanItRA st' sp' is'
          in if t == RATokEOF || t == RATokError then ([tok], st')
             else (tok:toks, st'')


-- symbol table management

-- generate new, empty symbol table

emptySymbolTable :: SymbolTable
emptySymbolTable = (empty, empty, empty, empty, [])

-- initial RA symbol table: 3 variables (x,y,z) and eq. predicate i

initialRASymbolTable :: SymbolTable
initialRASymbolTable =
  let varSet = listToSet [mkVar i | i <- [0..2]]
      predSet = singleton (mkPred 0 2)
      stringSet = listToSet ["x", "y", "z", "i"]
      stEntries = [("x", VarIdent, IVVar (mkVar 0)),
                   ("y", VarIdent, IVVar (mkVar 1)),
                   ("z", VarIdent, IVVar (mkVar 2)),
 		   ("i", PredIdent, IVPred (mkPred  0 2))]
  in (varSet, empty, predSet, stringSet, stEntries)

-- search for the value of an identifier in the symbol table

stLookupByName :: String -> SymbolTable -> Maybe IdentValue
stLookupByName str (_,_,_,ss,l) =
  let lookup str [] = error "lookup failed"
      lookup str ((s,k,v):r) = if str == s then Just v else lookup str r
  in if str `isElement` ss then lookup str l else Nothing

-- search for a variable in the symbol table, when the name is given

stLookupVarByName :: String -> SymbolTable -> Maybe Variable
stLookupVarByName str (_,_,_,_,l) =
  let lookup str [] = Nothing
      lookup str ((s,_,IVVar v):r) = if str == s then Just v else lookup str r
      lookup str (_:r) = lookup str r
  in lookup str l

-- search for a constant in the symbol table, when the name is given

stLookupConByName :: String -> SymbolTable -> Maybe Constant
stLookupConByName str (_,_,_,_,l) =
  let lookup str [] = Nothing
      lookup str ((s,_,IVCon c):r) = if str == s then Just c else lookup str r
      lookup str (_:r) = lookup str r
  in lookup str l

-- search for a term in the symbol table, when the name is given

stLookupTermByName :: String -> SymbolTable -> Maybe Term
stLookupTermByName str (_,_,_,_,l) =
  let lookup str [] = Nothing
      lookup str ((s,_,IVVar v):r) = if str == s then Just (termFromVar v)
                                     else lookup str r
      lookup str ((s,_,IVCon c):r) = if str == s then Just (termFromCon c)
                                     else lookup str r
      lookup str (_:r) = lookup str r
  in lookup str l

-- search for a predicate name when the predicate is given

stLookupByPred :: Predicate -> SymbolTable -> Maybe String
stLookupByPred pred (_,_,ps,_,l) =
  if pred `isElement` ps then Just (findPred l pred)
  else Nothing
 where findPred [] p = error "lookup failed in stLookupByPred"
       findPred ((s,_,IVPred p'):r) p = if p == p' then s
         else findPred r p
       findPred (_:r) p = findPred r p

-- search for a variable name when the variable is given

stLookupByVar :: Variable -> SymbolTable -> Maybe String
stLookupByVar var (vs,_,_,_,l) =
  if var `isElement` vs then Just (findVar l var)
  else Nothing
 where findVar [] v = error "lookup failed in stLookupByVar"
       findVar ((s,_,IVVar v'):r) v = if v == v' then s
         else findVar r v
       findVar (_:r) v = findVar r v

-- search for a constant name when the constant is given

stLookupByCon :: Constant -> SymbolTable -> Maybe String
stLookupByCon con (_,cs,_,_,l) =
  if con `isElement` cs then Just (findCon l con)
  else Nothing
 where findCon [] c = error "lookup failed in stLookupByPred"
       findCon ((s,_,IVCon c'):r) c = if c == c' then s
         else findCon r c
       findCon (_:r) c = findCon r c

-- get the rank of a given predicate

stLookupPredRank :: SymbolTable -> Predicate -> Int
stLookupPredRank (_,_,_,_,l) p = stlr l p
  where stlr [] p = -1
        stlr ((_,_, IVPred p'):r) p =
         if p == p' then getPredRank p'
                    else stlr r p
        stlr (_:r) p = stlr r p

-- "st' = stSetPredRank st p r" changes the rank of a given predicate
-- in the symbol table. p is the predicate for which the rank is to be
-- changed, r is the new rank for predicate p, and st' is the symbol table
-- st with the rank of predicate p changed.

stSetPredRank :: SymbolTable -> Predicate -> Int -> SymbolTable
stSetPredRank (vs,cs,ps,ss,l) p i = (vs,cs,ps,ss, upd l p i)
  where upd [] _ _ = []
        upd (h@(s,k,IVPred p'):r) p i =
          if p == p' then let x = getPredIndex p
                          in (s,k,IVPred (mkPred x i)):r 
          else h : upd r p i
        upd (h:r) p i = h : upd r p i


---------------------------------
----- instance declarations -----
---------------------------------

instance Show FTokenPL where
  showsPrec _ tok = case tok of
    TokVar _ -> showString "<variable>"
    TokCon _ -> showString "<constant>"
    TokPred _ -> showString "<predicate>"
    TokTrue -> showString "true"
    TokFalse -> showString "false"
    TokOr -> showString "\\/"
    TokAnd -> showString "/\\"
    TokNot -> showChar '~'
    TokImplies -> showString "->"
    TokEquiv -> showString "<=>"
    TokExists -> showString "exists"
    TokForAll -> showString "forall"
    TokOpenPar -> showChar '('
    TokClosePar -> showChar ')'
    TokComma -> showChar ','
    TokSemicolon -> showChar ';'
    TokEOF -> showString "<EOF>"
    TokError -> showString "<illegal identifier>"

instance Show FTokenRA where
  showsPrec _ tok = case tok of
    RATokPred _ -> showString "<predicate>"
    RATokRelProd -> showChar '@'
    RATokAbsProd -> showChar '*'
    RATokRelSum -> showChar '#'
    RATokAbsSum -> showChar '+'
    RATokRelOne -> showChar 'i'
    RATokAbsOne -> showChar '1'
    RATokAbsZero -> showChar '0'
    RATokConverse -> showChar '^'
    RATokNegation -> showChar '~'
    RATokEquals -> showChar '='
    RATokIncludes -> showChar '<'
    RATokTurnstile -> showString "|-"
    RATokSemicolon -> showChar ';'
    RATokComma -> showChar ','
    RATokOpenPar -> showChar '('
    RATokClosePar -> showChar ')'
    RATokEOF -> showString "<EOF>"
    RATokError -> showString "<illegal identifier>"


------------------------------
----- internal functions -----
------------------------------

-- character classes

isSpace' c = c `elem` " \t\r\f\v" 		-- space, but not newline
isVarChar c = c >= 'm' && c <= 'z'		-- variable characters
isConChar c = (c >= 'a' && c <= 'l') || isDigit c -- constant characters
isSpecial c = c `elem` "!#=<>[]&%+*-/\\"	-- special predicate characters
isPredChar c = isUpper c || isSpecial c		-- all predicate characters
isRAPredChar c = isUpper c			-- all RA predicate characters

-- scanner procedure for PL1 formulae:
-- read a token and update the state of the scanner

scanPL :: Scanner TokenPL
scanPL st (l,c) "" = ((TokEOF,(l,c)), st, (l,c), "")
scanPL st (l,c) ('\n':r) = scanPL st (l+1,1) r
scanPL st (l,c) (h:r) | isSpace' h = let (space,r') = span isSpace' r
                                    in scanPL st (l,c + length space + 1) r'
scanPL st (l,c) ('-':'-':r) = scanPL st (l,1) (dropWhile (/= '\n') r)
scanPL st (l,c) ('(':r) = ((TokOpenPar, (l,c)), st, (l,c+1), r)
scanPL st (l,c) (')':r) = ((TokClosePar, (l,c)), st, (l,c+1), r)
scanPL st (l,c) (',':r) = ((TokComma, (l,c)), st, (l,c+1), r)
scanPL st (l,c) (';':r) = ((TokSemicolon, (l,c)), st, (l,c+1), r)
scanPL st (l,c) ('~':r) = ((TokNot, (l,c)), st, (l,c+1), r)
scanPL st (l,c) ('t':'r':'u':'e':r) = ((TokTrue, (l,c)), st, (l,c+4), r)
scanPL st (l,c) ('f':'a':'l':'s':'e':r) = ((TokFalse, (l,c)), st, (l,c+5), r)
scanPL st (l,c) ('\\':'/':r) = ((TokOr, (l,c)), st, (l,c+2), r)
scanPL st (l,c) ('/':'\\':r) = ((TokAnd, (l,c)), st, (l,c+2), r)
scanPL st (l,c) ('-':'>':r) = ((TokImplies, (l,c)), st, (l,c+2), r)
scanPL st (l,c) ('<':'=':'>':r) = ((TokEquiv, (l,c)), st, (l,c+3), r)
scanPL st (l,c) ('e':'x':'i':'s':'t':'s':r) = ((TokExists, (l,c)), st, (l,c+6), r)
scanPL st (l,c) ('f':'o':'r':'a':'l':'l':r) = ((TokForAll, (l,c)), st, (l,c+6), r)
scanPL st (l,c) (h:r) | isVarChar h =
  let (varChars,r') = span isVarChar r
      varString = h:varChars
      (iv,st') = symbolTableLookup VarIdent varString st
      IVVar v = iv
  in ((TokVar v, (l,c)), st', (l,c + length varString), r')
scanPL st (l,c) (h:r) | isConChar h =
  let (conChars,r') = span isConChar r
      conString = h:conChars
      (iv,st') = symbolTableLookup ConIdent conString st
      IVCon k = iv
  in ((TokCon k, (l,c)), st', (l,c + length conString), r')
scanPL st (l,c) (h:r) | isPredChar h =
  let (predChars,r') = span isPredChar r
      predString = h:predChars
      (iv,st') = symbolTableLookup PredIdent predString st
      IVPred p = iv
  in ((TokPred p, (l,c)), st', (l,c + length predString), r')
scanPL st (l,c) (h:r) = ((TokError, (l,c)), st, (l,c), r)


-- scanner procedure for RA formulae:
-- read a token and update the state of the scanner

scanRA :: Scanner TokenRA
scanRA st (l,c) "" = ((RATokEOF,(l,c)), st, (l,c), "")
scanRA st (l,c) ('\n':r) = scanRA st (l+1,1) r
scanRA st (l,c) (h:r) | isSpace' h = let (space,r') = span isSpace' r
                                    in scanRA st (l,c + length space + 1) r'
scanRA st (l,c) ('-':'-':r) = scanRA st (l,1) (dropWhile (/= '\n') r)
scanRA st (l,c) ('(':r) = ((RATokOpenPar, (l,c)), st, (l,c+1), r)
scanRA st (l,c) (')':r) = ((RATokClosePar, (l,c)), st, (l,c+1), r)
scanRA st (l,c) (',':r) = ((RATokComma, (l,c)), st, (l,c+1), r)
scanRA st (l,c) (';':r) = ((RATokSemicolon, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('~':r) = ((RATokNegation, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('^':r) = ((RATokConverse, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('i':r) = ((RATokRelOne, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('1':r) = ((RATokAbsOne, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('0':r) = ((RATokAbsZero, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('+':r) = ((RATokAbsSum, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('*':r) = ((RATokAbsProd, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('@':r) = ((RATokRelProd, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('#':r) = ((RATokRelSum, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('=':r) = ((RATokEquals, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('<':r) = ((RATokIncludes, (l,c)), st, (l,c+1), r)
scanRA st (l,c) ('|':'-':r) = ((RATokTurnstile, (l,c)), st, (l,c+1), r)
scanRA st (l,c) (h:r) | isRAPredChar h =
  let (predChars,r') = span isRAPredChar r
      predString = h:predChars
      (iv,st') = symbolTableLookup PredIdent predString st
      IVPred p = iv
  in ((RATokPred p, (l,c)), st', (l,c + length predString), r')
scanRA st (l,c) (h:r) = ((RATokError, (l,c)), st, (l,c), r)


-- symbol table handling:

-- "(val, st') = symbolTableLookup ik n st" searches for an
-- identifier of kind ik (variable, predicate, constant) with name
-- n in symbol table st. If it exists, the value val of the
-- identifier is returned and st' = st, otherwise the identifier
-- is inserted in symbol table st, resulting in st', and a fresh
-- value for that identifier with appropriate kind is returned.

symbolTableLookup :: IdentKind -> String -> SymbolTable
  -> (IdentValue, SymbolTable)
symbolTableLookup identKind idString symTab =
  let mv = stLookupByName idString symTab; Just v = mv
      (lv,st') = insertIdent identKind idString symTab
  in if mv == Nothing then (lv, st')
     else (v, symTab)

-- "(val, st') = insertIdent ik n st" inserts an identifier into a
-- symbol table. The identifier to be inserted must not be stored
-- in the symbol table already. ik is the kind, n the name of the
-- identifier to be inserted. st' is the augmented symbol table,
-- and val the newly generated value for the inserted identifier.

insertIdent :: IdentKind -> String -> SymbolTable -> (IdentValue, SymbolTable)
insertIdent VarIdent s (vs,cs,ps,ss,l) =
  let v = head ([mkVar 0..] \\ setToOrdList vs)  -- generate new variable
      vs' = add v vs
      ss' = add s ss
      l' = (s, VarIdent, IVVar v):l	-- adapt symbol table
  in (IVVar v, (vs',cs,ps,ss',l'))
insertIdent ConIdent s (vs,cs,ps,ss,l) =
  let c = head ([mkCon 0..] \\ setToOrdList cs)	-- generate new constant
      cs' = add c cs
      ss' = add s ss
      l' = (s, ConIdent, IVCon c):l	-- adapt symbol table
  in (IVCon c, (vs,cs',ps,ss',l'))
insertIdent PredIdent s (vs,cs,ps,ss,l) =
  let p = head ([mkPred 0 (-1)..] \\ setToOrdList ps) -- generate new predicate
      ps' = add p ps
      ss' = add s ss
      l' = (s, PredIdent, IVPred p):l	-- adapt symbol table
  in (IVPred p, (vs,cs,ps',ss',l'))
