--------------------------------------------------------------------------
-- Module Set.hs                                                	--
--                                                              	--
-- Implementation of datatype `Set' and corresponding methods.  	--
-- NOTE: An ordering relation on the elements of the set is assumed	--
--       to be given.                					--
--                                                              	--
-- 26.09.96  initial revision                                   	--
--------------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Set(
  -- data type
  Set,
  -- constructors
  empty,           -- empty set
  singleton,       -- set with one element
  add,             -- add some element to a set
  -- operations
  isElement,       -- test on element relationship
  isEmpty,         -- is this set empty?
  isSubset,        -- subset relation
  cardinality,     -- number of elements of the set
  minus,           -- set difference
  sunion,          -- set union
  intersection,    -- set intersection
  setToOrdList,    -- conversion from set to ordered list
  listToSet        -- conversion from ordered list to set
) where


----------------------------
----- imported modules -----
----------------------------

import List( (\\) )


----------------------
----- data types -----
----------------------

newtype Set e = Set [e]   -- e is element type


------------------------------
----- exported functions -----
------------------------------

empty :: Set e
empty = Set []

singleton :: e -> Set e
singleton x = Set [x]

add :: Ord e => e -> Set e -> Set e
add x (Set s) = Set (insert x s)
  where insert x [] = [x]
        insert x s@(e:es) | x < e     = x:s
                          | x == e    = s
                          | otherwise = e : insert x es

isElement :: Ord e => e -> Set e -> Bool
isElement x (Set s) = x `element` s
  where x `element` [] = False
        x `element` (e:es) | x < e     = False
                           | x == e    = True
                           | otherwise = x `element` es

isEmpty :: Set e -> Bool
isEmpty (Set s) = null s

cardinality :: Set e -> Int
cardinality (Set s) = length s

isSubset :: Ord e => Set e -> Set e -> Bool
isSubset (Set s) (Set s') = s `subset` s'
  where [] `subset` _ = True
        _ `subset` [] = False
        s1@(x:xs) `subset` s2@(y:ys) | x < y     = False
                                     | x == y    = xs `subset` ys
                                     | otherwise = s1 `subset` ys

minus :: Ord e => Set e -> Set e -> Set e
minus (Set s) (Set s') = Set (s `setMinus` s')
  where [] `setMinus` _ = []
        s1 `setMinus` [] = s1
        s1@(x:xs) `setMinus` s2@(y:ys) | x < y     = x : xs `setMinus` s2
                                       | x == y    = xs `setMinus` ys
                                       | otherwise = s1 `setMinus` ys

sunion :: Ord e => Set e -> Set e -> Set e
sunion (Set s) (Set s') = Set (setUnion s s')
  where setUnion xs [] = xs
        setUnion [] ys = ys
        setUnion s1@(x:xs) s2@(y:ys) | x < y     = x : setUnion xs s2
                                     | x == y    = x : setUnion xs ys
                                     | otherwise = y : setUnion s1 ys

intersection :: Ord e => Set e -> Set e -> Set e
intersection (Set s) (Set s') = Set (intersect s s')
  where intersect _ [] = []
        intersect [] _ = []
        intersect s1@(x:xs) s2@(y:ys) | x < y     = intersect xs s2
                                      | x == y    = x : intersect xs ys
                                      | otherwise = intersect s1 ys

setToOrdList :: Set e -> [e]
setToOrdList (Set s) = s

listToSet :: Ord e => [e] -> Set e
listToSet xs = foldr add empty xs


---------------------------------
----- instance declarations -----
---------------------------------

instance Ord e => Eq (Set e) where
  s == s' = isSubset s s' && isSubset s' s

instance Show e => Show (Set e) where
  showsPrec _ (Set s) = showAsSet s
     where showAsSet []     = showString "{}"
           showAsSet (e:es) = showChar '{' . shows e . showAsSetl es
           showAsSetl []     = showChar '}'
           showAsSetl (e:es) = showString ", " . shows e . showAsSetl es
