----------------------------------------------------------------------
-- Module Simplify.hs                                               --
--                                                                  --
-- This module contains routines to perform some minor		    --
-- simplifications on formulae.					    --
--                                                                  --
-- 12.11.96  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Simplify (
  -- exported functions
  nnf,                             -- compute negation normal form
  flatten, flattenTop,		   -- flatten formula (complete/top position only)
  simplify,			   -- perform several simple reductions
  minimizeQuantifierScopes,	   -- reduce quantifier domains
  deleteOuterUniversalQuantifiers, -- delete outermost universal quantifiers
  dropOuterUniversalQuantifiers,   -- recursively delete univ. quant. at top position
  partialSkolemization,            -- perform a partial skolemization
  split, topSplit,		   -- split formula at conjunctions (only top pos.)
  minimizeVarAndConstDomains,	   -- rename variables and constants -> minimal index
  collectLiterals,      	   -- collects literals, ordered by contained predicates
  deleteBadLiterals		   -- removes unnecessary literals, see spec. below
) where


-------------------
----- imports -----
-------------------

import Set
import List (partition, intersect, nub, (\\))
import Formula
import Auxiliary


------------------------
-- exported functions --
------------------------

-- `nnf' computes the negation normal form of an arbitrary formula f.
-- I.e. the output formula is normalized with respect to the following rules:
-- ~true        --> false,         ~false       --> true,
-- ~(A \/ B)    --> ~A /\ ~B,      ~(A /\ B)    --> ~A \/ ~B,
-- ~(forall x)A --> (exists x)~A,  ~(exists x)A --> (forall x)~A,
-- ~(~A)        --> A
-- Node information may be reset!

nnf  :: Formula -> Formula
nnf f = if isNeg f then negFor (nnf (getSubFormula f))
        else setSubFormulae (map nnf (getSubFormulae f)) f
   where negFor f
          | isVerum f     = falsum
          | isFalsum f    = verum
          | isPosLit f    = neg f
          | isNeg f       = getSubFormula f
          | isConj f      = disj (map negFor (getSubFormulae f))
          | isDisj f      = conj (map negFor (getSubFormulae f))
          | isExists f    = forall (getQuantVar f) (negFor (getSubFormula f))
          | isForAll f    = exists (getQuantVar f) (negFor (getSubFormula f))


-- `flatten' converts an arbitrary formula f to flattened form, i.e.
-- it performs the following reductions:
-- \/(f_1,....f_{i-1},\/(g_1,...,g_m),f_{i+1},...,f_n)
--    --> \/(f_1,...,f_{i-1},g_1,...,g_m,f_{i+1},...,f_n)  and
-- /\(f_1,....f_{i-1},/\(g_1,...,g_m),f_{i+1},...,f_n)
--    --> /\(f_1,...,f_{i-1},g_1,...,g_m,f_{i+1},...,f_n)
-- The order of subformulae is not conserved; disjunctions and conjunctions
-- with m,n < 2 must not occur in the input formula!
-- When node information of conjunctions is to be merged, the mark of the
-- outermost conjunction is used.

flatten :: Formula -> Formula
flatten f = flattenTop (setSubFormulae (map flatten (getSubFormulae f)) f)


-- `flattenTop' flattens a formula f at top-position.
-- The specification of flattening is the same as for function `flatten'
-- above.
-- The order of subformulae is not conserved; disjunctions and conjunctions
-- with m,n < 2 (see specification above) must not occur in the input formula!
-- When node information of conjunctions is to be merged, the mark of the
-- outermost conjunction is used.

flattenTop :: Formula -> Formula
flattenTop f
  | isDisj f  = setSubFormulae (flattenDisjTop (getSubFormulae f)) f
  | isConj f =  setSubFormulae (flattenConjTop (getSubFormulae f)) f
  | otherwise = f


-- `simplify' performs several simple reductions on a formula f.
-- These reductions are either part of the reduction calculus RPC_n,
-- or are compatible with it.
-- The input formula f must be in negation normal form; the output formula
-- f' is computed by normalising f w.r.t the following rules:
-- A \/ true   --> true,  A /\ true  --> A
-- A \/ false  --> A,     A /\ false --> false
-- L \/ ~L     --> true,  L /\ ~L    --> false
-- L \/ L      --> L,     L /\ L     --> L
-- (exists x)A --> A, if x is not free in A
-- (forall x)A --> A, if x is not free in A
-- (A is an arbitrary formula, L a literal.)
-- The output formula f' is flattened and in negation normal form.
-- Literal node information is preserved.

simplify :: Formula -> Formula
simplify g
  | isDisj g = simplifyDisj g
  | isConj g = simplifyConj g
  | isExists g || isForAll g = simplifyQuant g
  | otherwise = g


-- `minimizeQuantifierScopes' reduces the domains of all quantifiers of
-- an input formula f in negation normal form.
-- The output formula f' is normalized with respect to the following rules:
-- (exists x)(A \/ B) --> (exists x)A \/ (exists x)B
-- (exists x)(A /\ B) --> (exists x)A /\ B, if x is not free in B
-- (forall x)(A \/ B) --> (forall x)A \/ B, if x is not free in B
-- (forall x)(A /\ B) --> (forall x)A /\ (forall x)B
-- f' is flattened and in negation normal form, but not necessarily
-- simplified, i.e. it may contain literal-pairs L \/ ~L, L /\ ~L,
-- L \/ L resp. L /\ L, even though the input formula f was simplified.
-- Node information may be destroyed!

minimizeQuantifierScopes :: Formula -> Formula
minimizeQuantifierScopes f
  | isDisj f = disj (flattenDisjTop sfs)
  | isConj f = conj (flattenConjTop sfs)
  | isExists f = if isDisj sf then disj (map (exists qv) vf ++ vnf)
      else if isConj sf then combine conj (exists qv) vf vnf
      else setSubFormula sf f
  | isForAll f = if isConj sf then conj (map (forall qv) vf ++ vnf)
      else if isDisj sf then combine disj (forall qv) vf vnf
      else setSubFormula sf f
  | otherwise  = setSubFormulae sfs f
  where sfs = map minimizeQuantifierScopes (getSubFormulae f)
        sf:_ = sfs
        ssfs = getSubFormulae sf   -- length of ssfs is >=2
        qv = getQuantVar f
        (vf,vnf) = partition (\f -> qv `isElement` free f) ssfs
        vfh:vfr = vf; vnfh:vnfr = vnf
        combine c qc [] gs = c gs
        combine c qc (f:[]) [] = qc f
        combine c qc (f:[]) gs = c (qc f : gs)
        combine c qc fs [] = qc (c fs)
        combine c qc fs gs = c (qc (c fs) : gs)


-- `deleteOuterUniversalQuantifiers' takes an input formula f in negation
-- normal form and removes all universal quantifiers which are not in
-- the scope of an existential quantifier. This is achieved by first applying
-- the following rules to the input formula (in combination with renaming of
-- bound variables, where necessary), and then stripping off the leading
-- universal quantifiers.
-- (forall x)A /\ B  --> (forall x)(A /\ B), if x is not free in B
-- (forall x)A \/ B  --> (forall x)(A \/ B), if x is not free in B
-- Thus some kind of universal prenex form is computed.
-- The output formula f' is in negation normal form and flattened, but not
-- necessarily simplified, i.e. it may contain literal-pairs L \/ ~L, L /\ ~L,
-- L \/ L resp. L /\ L, even though the input formula f is simplified.
-- Node information may be reset!

deleteOuterUniversalQuantifiers :: Formula -> Formula
deleteOuterUniversalQuantifiers f = let (f',_,_) = delOUQuants f in f'


-- variation: drop only outermost universal quantifers, no further changes

dropOuterUniversalQuantifiers :: Formula -> Formula
dropOuterUniversalQuantifiers f
  | isForAll f = dropOuterUniversalQuantifiers (getSubFormula f)
  | otherwise = f


-- `partialSkolemization' replaces all variables occurring free in a formula
-- f by new distinct constants. Properties like negation normal form, flat
-- or simplified are preserved.
-- Node information may be destroyed!

partialSkolemization :: Formula -> Formula
partialSkolemization f = repl (setToOrdList fv) ncs f
  where fv = free f; n = cardinality fv
        ncs = take n ([mkCon 0..] \\ setToOrdList (consts f))
        repl [] _ f = f
        repl (v:vs) (c:cs) f = repl vs cs (subst v (termFromCon c) f)


-- `split' takes an input formula f, which must be flat and in negation
-- normal form, and converts it into a list (f_1,...,f_n)
-- of formulae, such that |= f iff |= f_1 and |= f_2 and ... and |= f_n.
-- All output formulae f_i are flat and in negation normal form.
-- Output formulae f_i are not necessarily simplified.
-- Note: 1. n may become quite large.
--       2. The input formula f should not contain any subformula of the form
--          (exists x)A, where x is not free in A; otherwise the result may
--          not be optimal (though correct).
-- Node information may be reset!

split :: Formula -> [Formula]
split f
  | isVerum f = [f]
  | isFalsum f = [f]
  | isLiteral f = [f]
  | isDisj f = map disj (takeOneOfEach (map split (getSubFormulae f)))
  | isConj f = foldr ((++) . split) [] (getSubFormulae f)
  | isExists f = [f]
  | isForAll f = map (flip setSubFormula f) (split (getSubFormula f))


-- `topSplit' is similar to `split', but works only at top position.
-- Node information is preserved.

topSplit :: Formula -> [Formula]
topSplit f
  | isConj f = getSubFormulae f
  | otherwise = [f]


-- `minimizeVarAndConstDomains' renames all variables and constants in an
-- input formula f, such that the set of variables resp. constants occurring
-- in the output formula f' is {v_0,...,v_m} resp. {c_0,...,c_n} with minimal
-- m and n, and |= f iff |= f'. (More specifically, the universal closures of
-- f and f' are equivalent modulo renaming of bound variables.)
-- Properties like negation normal form, flat or simplified are preserved.

minimizeVarAndConstDomains :: Formula -> Formula
minimizeVarAndConstDomains f = minimize' vmap f
  where vs = setToOrdList (free f)
        cs = setToOrdList (consts f)
        vmap = [(v, mkVar v') | v' <- [0..length vs - 1], let v = vs!!v']
        cmap = [(c, mkCon c') | c' <- [0..length cs - 1], let c = cs!!c']
        minimize' vm f
          | isPosLit f = literal (renameLitTerms (getLiteral f) vm cmap)
          | isExists f = exists qv' (minimize' vm' sf)
          | isForAll f = forall qv' (minimize' vm' sf)
          | otherwise = setSubFormulae (map (minimize' vm) (getSubFormulae f)) f
          where qv = getQuantVar f
                sf = getSubFormula f
                qv' = head ([mkVar 0..] \\ map (assoc vm)
                                               (setToOrdList (free f)))
                vm' = (qv, qv') : vm


-- `f' = deleteBadLiterals f' replaces all literals L of a set of literals M
-- by falsum. The set M consists of those literals L occurring in f, with no
-- complementary literal ~L occurring in f connected to L by a disjunction.
-- I.e. for a literal L \in M there is no literal ~L such that A(L) \/ B(~L)
-- is a subformula of f. (A(L) should be read as an arbitrary formula A containing
-- literal L.)

deleteBadLiterals :: Formula -> Formula
deleteBadLiterals f = foldr (\p f -> setSubFormulaAtPos p falsum f) f bps
  where bps = foldr ((++) . removableLiteralPositions f) []
                    (collectLiterals f)


-- `l = collectLiterals f' collects all literal positions in a formula f ordered
-- by the predicate of the literal. Therefore it returns a list l of triples
-- (pr, pp, np) where pr is a predicate, pp and np are lists of positions, such
-- that a position p is in pp iff the literal at position p in f is positive
-- and contains predicate pr. The same holds for np, where `positive literal'
-- is to be replaced by `negative literal' in the last sentence.

collectLiterals :: Formula -> [(Predicate, [Position], [Position])]
collectLiterals f = colLits lambda f
  where colLits pos f
         | isPosLit f = [(getLitPred (getLiteral f), [pos], [])]
         | isNegLit f = [(getLitPred (getLiteral sf), [], [pos])]
         | otherwise = foldr col [] cs
         where sf = getSubFormula f
               sfs = getSubFormulae f; l = length sfs
               cs = [c | i <- [0..l-1],
                     c <- colLits (mkPos pos i) (sfs!!i)]
               col c [] = [c]
               col c@(pi, ps, nps) (c'@(pi', ps', nps'):r) = if pi == pi' then
                 (pi, ps ++ ps', nps ++ nps') : r else c' : col c r


------------------------
-- internal functions --
------------------------

-- "fs' = flattenDisjTop fs" takes a list of formulae, which are assumed to
-- be direct subformuae of a disjunction, and flattens this assumed disjunction
-- at top position.

flattenDisjTop :: [Formula] -> [Formula]
flattenDisjTop fs = let (d,nd) = partition isDisj fs
  in foldr ((++) . getSubFormulae) nd d


-- "fs' = flattenConjTop fs" is the dual version of flattenDisjTop.

flattenConjTop :: [Formula] -> [Formula]
flattenConjTop fs = let (c,nc) = partition isConj fs
  in foldr ((++) . getSubFormulae) nc c


-- "f' = simplifyDisj f" simplifies a formula with a disjunction at top position.
-- The simplifications performed are documented in the specs of function "simplify"
-- above.

simplifyDisj :: Formula -> Formula
simplifyDisj g =
  let -- decompose subformulae according to type: veri, falsi, positive literals,
      -- negative literals, complex formulae
      (v,f,p,n,c) = partitionFormulaList (getSubFormulae g)
      -- remove duplicate literals (while preserving node information)
      np = remEqLit p; nn = remEqLit n
      -- recursion, simpl. of complex formulae with flattening at top position
      (v',f',p',n',c') = partitionFormulaList (flattenDisjTop (map simplify c))
      -- build new literal lists
      np' = remEqLit (np ++ p'); nn' = remEqLit (nn ++ n')
      -- ... and the whole new disjunction
      df = np' ++ nn' ++ c'
  in -- trivial cases: verum in disjunction, L \/ ~L occurrence
     if not (null v) || litPair np nn || not (null v') || litPair np' nn' then verum
     else if null df then falsum
     -- standard result
     else disj df


-- "simplifyConj" is the dual version of "simplifyDisj" above.

simplifyConj :: Formula -> Formula
simplifyConj g =
  let (v,f,p,n,c) = partitionFormulaList (getSubFormulae g)
      np = remEqLit p; nn = remEqLit n
      (v',f',p',n',c') = partitionFormulaList (flattenConjTop (map simplify c))
      np' = remEqLit (np ++ p'); nn' = remEqLit (nn ++ n')
      cf = np' ++ nn' ++ c'; nf:nfs = cf
      ni = getNodeInfo g
  in -- trivial cases: falsum in conjunction, L /\ ~L occurrence
     if not (null f) || litPair np nn || not (null f') || litPair np' nn' then falsum
     -- conjunction collapsed: no subformula
     else if null cf then verum
     -- conjunction collapsed: only one subformula
     else if null nfs then nf
     -- normal case: reset node information to previous value
     else setNodeInfo ni (conj cf)


-- "f' = simplifyQuant f" simplifies a formula with a quantifier at top position,
-- according to the appropriate sub-specification of algorithm "simplify" above.

simplifyQuant :: Formula -> Formula
simplifyQuant f =
  let v = getQuantVar f
      g = simplify (getSubFormula f)
      f' = setSubFormula g f
  in -- bound variable occurs free in subformula
     if v `isElement` free g then f'
     -- no variable bound by quantifier -> remove it
     else g


-- Partition a list of formulae according to top operator:
-- veri, falsi, positive literals, negative literals, other formulae
-- (disjunctions, conjunctions, quantified formulae).

partitionFormulaList :: [Formula]
   -> ([Formula],[Formula],[Formula],[Formula],[Formula])
partitionFormulaList [] = ([],[],[],[],[])
partitionFormulaList (g:gs)
  | isVerum g     = (g:v,f,p,n,c)
  | isFalsum g    = (v,g:f,p,n,c)
  | isPosLit g    = (v,f,g:p,n,c)
  | isNegLit g    = (v,f,p,g:n,c)
  | otherwise     = (v,f,p,n,g:c)
  where (v,f,p,n,c) = partitionFormulaList gs


-- "t = litPair pls nls" takes two formulae lists, where pls contains only
-- positive literals and nls only negative literals. Then t is true iff there
-- is a literal pair (L,L'), such that L \in pls and L' \in nls and ~L = L'.

litPair :: [Formula] -> [Formula] -> Bool
litPair posL negL = not (null (intersect posL negL'))
  where negL' = map getSubFormula negL


-- Remove equal literals from a list of literals preserving node information.
-- (Similar to function nub in the standard prelude.)

remEqLit :: [Formula] -> [Formula]
remEqLit [] = []
remEqLit (x:xs) =
  let xs' = remEqLit xs
      (l,r) = partition (== x) xs'; h:_ = l
      ni = mergeNodeInfo (getNodeInfo x) (getNodeInfo h)
  in if null l then x : xs'
     else setNodeInfo ni x : r


-- `ps = removavleLiteralPositions f l' returns a list of literal positions,
-- at which falsum may be substitued without changing the validity of the
-- formula f. This is required for algorithm "deleteBadLiterals".
-- The list l is the output of algorithm collect literals applied to f.

removableLiteralPositions :: Formula -> (Predicate, [Position], [Position])
  -> [Position]
removableLiteralPositions f (_,ps,nps) = (ps \\ pok) ++ (nps \\ nok)
  where lps = takeOneOfEach [ps,nps]
        lok = filter (\[a,b] -> isDisj (f@@(commonPrefix a b))) lps
        (pok,nok) = unzip (map (\[a,b] -> (a,b)) lok)


-- "(f',vs,vs') = delOUQuants f" removes outermost universal quantifiers
-- after having performed several reduction steps of kind
--  (forall x)A /\ B --> (forall x)(A /\ B), if x is not free in B, or
--  (forall x)A \/ B --> (forall x)(A \/ B), if x is not free in B.
-- f is a formula in negation normal form, f' is a flat formula in negation normal
-- form, vs and vs' are sets of variables.
-- vs is the set of free variables of f' which were bound by
-- a universal quantifier in f (probably renamed).
-- vs' is the set of free variables of f' without the variables of the set vs.
-- Thus vs \cup vs' = free f and vs \cap vs' = \emptyset.
-- The aforementioned reductions are not applied within the scope of an existential
-- quantifier.

delOUQuants :: Formula -> (Formula, Set Variable, Set Variable)
delOUQuants f
  | isVerum f = (f, empty, empty)
  | isFalsum f = (f, empty, empty)
  | isLiteral f = (f, empty, free f)
  | isExists f = (f, empty, free f)
  | isForAll f = let (f', vs, fvs) = delOUQuants (getSubFormula f)
                     qv = getQuantVar f
                 in (f', add qv vs, fvs `minus` singleton qv)
  | isConj f = let (sfs, bvss, fvss) = unzip3 (map delOUQuants (getSubFormulae f))
                   fvs = foldr sunion empty fvss
                   mbv = maximum (map cardinality bvss)
                   nbvs = take mbv ([mkVar 0..] \\ setToOrdList fvs)
                   sfs' = map (changeFreeVarsC nbvs) (zip sfs bvss)
               in (conj (flattenConjTop sfs'), listToSet nbvs, fvs)
  | isDisj f = let (sfs, bvss, fvss) = unzip3 (map delOUQuants (getSubFormulae f))
                   fvs = foldr sunion empty fvss
                   mbvs = foldr ((+) . cardinality) 0 bvss
                   nbvs = take mbvs ([mkVar 0..] \\ setToOrdList fvs)
                   sfs' = changeFreeVarsD nbvs (zip sfs bvss)
               in (disj (flattenDisjTop sfs'), listToSet nbvs, fvs)


-- "f' = changeFreeVarsC nv fv" takes a list of variables nv = (nv_0,...,nv_l)
-- and a formula - variable set pair fv = (f,vs), and returns a formula f'.
-- All variables v of set vs must occur free in f, the length l of the variable
-- list must be equal to the cardinality of vs, all variables of list nv
-- must be distinct, and no variable of nv is an element of (free f \ vs).
-- f' is obtained from f_i by replacing all variables of set vs by
-- new variables of set nv.

changeFreeVarsC :: [Variable] -> (Formula, Set Variable) -> Formula
changeFreeVarsC nbv (f,ibv) = renv nbv' f ibv'
  where ibvl = setToOrdList ibv; vis  = intersect nbv ibvl
        nbv' = nbv \\ vis
        ibv' = ibvl \\ vis
        renv _ f [] = f
        renv (v':vs') f (v:vs) = renv vs' (subst v (termFromVar v') f) vs


-- "fs' = changeFreeVarsD nv fv" takes a list of variables nv = (nv_0,...,nv_l)
-- and a list of formula - variable set pairs fv = ((f_0,vs_0),...,(f_k,vs_k)),
-- and returns a list of formulae fs' = (f'_0,...,f'_k).
-- All variables v of variable set vs_i must occur free in formula f_i,
-- the length of the new variable list l must be equal to the cummulated
-- cardinalities of the variable sets vs_i, i.e. l = |vs_0| + ... + |vs_k|,
-- and no variable of nv is an element of the set M where M is the union
-- of (free f_i \ vs_i) for all i.
-- f'_i is obtained from f_i by replacing all variables v of set vs_i
-- by new variables v' of set nv, such that no variable v' of nv occurs free
-- in more than one formula f'_i.

changeFreeVarsD :: [Variable] -> [(Formula, Set Variable)] -> [Formula]
changeFreeVarsD _ [] = []
changeFreeVarsD nv ((f,vs):r) =
  let -- convert variable set to list
      vl = setToOrdList vs
      -- variable renamer
      renVars f [] _ = f
      renVars f (v:vs) (v':vs') = renVars (subst v (termFromVar v') f) vs vs'
  in renVars f vl nv : changeFreeVarsD (drop (cardinality vs) nv) r


-- assoc: find an (existing) entry in a lookup-table, and fail if it is
-- not there.

assoc :: Eq a => [(a,b)] -> a -> b
assoc [] _ = undefined
assoc ((x,y):ps) a = if x == a then y else assoc ps a


-- "l' = renameLitTerms vm cm l" takes a finite variable map, a finite constant
-- map (represented as association lists) and renames terms in literal l
-- according to these maps, thus providing l'.

renameLitTerms :: Literal -> [(Variable,Variable)]
                -> [(Constant,Constant)] -> Literal
renameLitTerms l vm cm = mkLit (getLitPred l) (map mp (getLitTerms l))
  where mp t | termIsVar t = (\(Just x) -> termFromVar x) (lookup v vm)
             | termIsCon t = (\(Just x) -> termFromCon x) (lookup c cm)
             where v = getTermVar t
                   c = getTermCon t
