----------------------------------------------------------------------
-- Module Strategy.hs                                               --
--                                                                  --
-- Implements different reduction strategies                        --
--                                                                  --
-- 11.12.96  initial revision                                       --
----------------------------------------------------------------------

module Strategy where

import Set
import Formula
import Auxiliary
import Reduction
import Simplify
import List (nub, intersect, (\\), partition)
import FormulaIO
import LiteralPair
import IOStateMonad
import Unification
import LiteralPair
import GlobalState
import Monad

type ReductionStrategy = IOState ()

-- reducers mustn't be applied to irreducible formulae!

-- oldest reduction possibility reducer
type ORPReducer = ForListPartitioner -> ForListPartitioner
  -> ForListPartitioner -> ReductionSelector -> TermSelector
  -> ReductionStrategy

-- literal pair reducer
type LPReducer = LitPairSelector -> DisjSelectType -> TermSelector
  -> ReductionStrategy



-----------------------------------------
-- interactive initialization routines --
-----------------------------------------

initORPReducer :: IOState ReductionStrategy
initORPReducer = do
  putStrLn' "Initialization of oldest reduction possibility reduction strategy:"
  putStrLn' "Choose a reduction selector:"
  redSel <- queryRedSel
  putStrLn' "Choose a term selector:"
  termSel <- queryTermSel
  putStrLn' "Choose a disjunction selector for universal quantifier rule:"
  disjSelA <- queryDisjSel
  putStrLn' "Choose a disjunction selector for distribution rule:"
  disjSelD <- queryDisjSel
  putStrLn' "Choose a conjunction selector for distribution rule:"
  conjSelD <- queryConjSel
  return (theORPReducer disjSelA disjSelD conjSelD redSel termSel)

initLPReducer :: IOState ReductionStrategy
initLPReducer = do
  putStrLn' "Initialization of literal pair reduction strategy:"
  putStrLn' "Choose a literal pair selector:"
  litPairSel <- queryLitPairSel
  putStrLn' "Choose a disjunction selector type:"
  disjSelType <- queryDisjSelType
  putStrLn' "Choose a term selector:"
  termSel <- queryTermSel
  return (theLPReducer litPairSel disjSelType termSel)

initAK3Reducer :: IOState ReductionStrategy
initAK3Reducer = do
  putStrLn' "Initialization of literal pair reduction strategy:"
  putStrLn' "Choose a literal pair selector:"
  litPairSel <- queryLitPairSel
  putStrLn' "Choose a disjunction selector type:"
  disjSelType <- queryDisjSelType
  putStrLn' "Choose a term selector:"
  termSel <- queryTermSel
  return (theAK3Reducer litPairSel disjSelType termSel)

queryRedSel :: IOState ReductionSelector
queryRedSel = do
  putStrLn' "  (d) default reduction selector"
  putStrLn' "  (i) interactive reduction selector"
  c <- query "Selection (d/i)? " "di"
  case c of 'd' -> return defaultRedSel
            'i' -> return interactiveRedSel

queryTermSel :: IOState TermSelector
queryTermSel = do
  putStrLn' "  (d) default term selector"
  putStrLn' "  (i) interactive term selector"
  putStrLn' "  (s) super term selector"
  c <- query "Selection (d/i/s)? " "dis"
  case c of 'd' -> return defaultTermSelector
            'i' -> return interactiveTermSelector
            's' -> return superTermSelector


queryDisjSel :: IOState ForListPartitioner
queryDisjSel = do
  putStrLn' "  (n) minimal disjunction"
  putStrLn' "  (a) maximal disjunction"
  putStrLn' "  (i) interactive disjunction selector"
  c <- query "Selection (n/a/i)? " "nia"
  case c of 'n' -> return minDisjSel
            'a' -> return maxDisjSel
            'i' -> return interactiveDisjSel

queryConjSel :: IOState ForListPartitioner
queryConjSel = do
  putStrLn' "  (d) default conjunction selector"
  putStrLn' "  (i) interactive conjunction selector"
  c <- query "Selection (d/i)? " "di"
  case c of 'd' -> return defaultConjSel
            'i' -> return interactiveConjSel

queryLitPairSel :: IOState LitPairSelector
queryLitPairSel = do
  putStrLn' "  (d) minimal position distance"
  putStrLn' "  (l) least recently used literals"
  putStrLn' "  (i) interactive literal selection"
  putStrLn' "  (s) super literal selector"
  putStrLn' "  (f) first literal"
  c <- query "Selection (d/l/i/s/f)? " "dlifs"
  case c of 'd' -> return minDistLitPairSel
            'l' -> return lruLitPairSel
            'i' -> return interactiveLitPairSel
            's' -> return superLitPairSel
            'f' -> return firstLitPairSel

queryDisjSelType :: IOState DisjSelectType
queryDisjSelType = do
  putStrLn' "  (i) minimal disjunction"
  putStrLn' "  (l) matching literal selector"
  putStrLn' "  (a) maximal disjunction"
  c <- query "Selection (i/l/a)? " "ila"
  case c of 'i' -> return MinDSel
            'l' -> return NormDSel
            'a' -> return MaxDSel


----------------------
-- generic reducers --
----------------------

theORPReducer :: ORPReducer
theORPReducer disjSelA disjSelD conjSelD redSel termSel =
  do f <- getWorkingFormula
     (redPos, ruleType, redTerms) <- redSel
     mt <- if ruleType == DistributionRule then return Nothing
            else do t <- if null redTerms then termSel redPos AnyVar
                      else termSel redPos (AnyOf redTerms)
                    return (Just t)
     opts <- getProofOptions
     when (printing opts >= OMStandard) $ do
       let Just t = mt
       putStr' ("Reducing at position " ++ show redPos)
       when (mt /= Nothing) $ putStr' (" using term " ++ show t)
       putStrLn' "."
     applyRule ruleType redPos (Just disjSelA) (Just disjSelD) (Just conjSelD) mt
     nextGeneration
     nextStep
   
theLPReducer :: LPReducer
theLPReducer litPairSel disjSelType termSel =
  do opt <- getProofOptions
     (f,_,g,s) <- getWorkingState
     let omode = printing opt; debug = debugging opt
     when (omode == OMVerbose) $
       putStrLn' ("\nstarted literal unification step #" ++
                  show (s+1))
     (p,p') <- litPairSel
     setLiteralPair p p'
--------------
{-
     putStrLn' "before slpTest"
     slpTest f p p'
-}
--------------
     when (omode == OMVerbose) $ do
       putStr' "  selected literal pair: "
       showFor (f@@p); putStr' ", "; showFor (f@@p')
       newLine'
       when debug $ do
         putStrLn' "DEBUG> literal pair subformula:"
         putStr' "DEBUG> "
         showFor (closure (extractLitPairSubFor (p,p') f))
         newLine'
       printWith putStrLn ("    @ positions: " ++ show p ++ ", " ++ show p')
     let (ur, umap) = unifyX f (p,p')
     setUnification umap ur
     when (omode == OMVerbose && ur /= Fail) $ do
       putStrLn' "  literal unification equation system solved"
       when debug $ do
         putStrLn' "DEBUG> solved unification equation system:"
         putStr' "DEBUG> "
         showUResult; newLine'
-------
{-
         putStrLn' ("before solve: " ++ show (unificationEqSystem (uniqueBoundVars (closure (extractLitPairSubFor (p,p') f)))))
         putStrLn' ("after solve: " ++ show (solve (unificationEqSystem (uniqueBoundVars (closure (extractLitPairSubFor (p,p') f))))))
         putStrLn' (show (ur,umap))
-}
-------
     lpReducer disjSelType termSel
     g' <- getGeneration
     when (omode >= OMStandard) $
       putStrLn' ("reduction step #" ++ show (s+1) ++ " completed, " ++
                  show (getGen g'- getGen g) ++ " RPC-reduction(s), " ++
		  show (getGen g') ++ " in total.")
     nextStep
 where lpReducer :: DisjSelectType -> TermSelector -> IOState ()
       lpReducer disjSelType termSel =
         do opt <- getProofOptions
            let omode = printing opt; debug = debugging opt
            when (omode == OMVerbose) $ putStr' "  preprocessing... "
            preProcessRootQuant termSel
            when (omode == OMVerbose) $
              putStr' "\n  main unification process... "
            processLitPath termSel disjSelType
            when (omode == OMVerbose) $
              printWith putStr "\n  postprocessing... "
            postProcessRootQuant
            when (omode == OMVerbose) newLine'
-----
            f <- getWorkingFormula
            (p,p') <- getLiteralPair
            if neg (f@@p) /= f@@p' then error ("lit.unif. failed: " ++ show (f@@p, f@@p'))
             else return ()
-----

-----------------------
-- combined reducers --
-----------------------

-- literal pair reducer with priority to non-expanding universal rule applications

theAK3Reducer :: LPReducer
theAK3Reducer litPairSel disjSelType termSel = do
  f <- getWorkingFormula
  putStrLn' "\nSearching for non-expanding universal quantifier rule..."
  let mak3 = searchAK3 f; Just (rp,vt) = mak3
  lastAK3 <- getLastAK3Info
  if mak3 == Nothing || mak3 == lastAK3 then do
    putStrLn' "No such rule found, performing normal LP reduction."
    theLPReducer litPairSel disjSelType termSel
   else do
    putStrLn' "Found, performing non-expanding universal quantifier rule reduction."
    -- store position and red. term
    setLastAK3Info mak3
    theORPReducer maxDisjSel maxDisjSel defaultConjSel (return (rp, UniversalRule, [vt]))
                  (\_ _ -> return vt)

searchAK3 :: Formula -> Maybe (Position, Term)
searchAK3 f =
  let aps = findAPositions f lambda
      okaps = [(p,vt) | p <- aps, let { mvt = freeVarAPos f p; Just vt = mvt },
                        mvt /= Nothing]
  in if null okaps then Nothing else Just (head okaps)

findAPositions :: Formula -> Position -> [Position]
findAPositions f p
  | isForAll f = [p]
  | otherwise = concat lps
  where sfs = getSubFormulae f; l = length sfs
        lps = [ps | i <- [0..l-1], let ps = findAPositions (sfs!!i) (mkPos p i)]

-- Nothing indicates position is not ok. for application of AK3

freeVarAPos :: Formula -> Position -> Maybe Term
freeVarAPos f p =
  let pl = posToList p in if null pl then Nothing else
    let rootp = listToPos (init pl); i = last pl
        dis = getSubFormulae (f@@rootp)
        fa = dis!!i
        fb = disj (take i dis ++ drop (i+1) dis)
        freeA = free fa; freeB = free fb
        freeAB = freeA `sunion` freeB
        sits = getSITerms (getStrategyInfo (getNodeInfo fa))
        oksits = sits \\ map termFromVar (setToOrdList freeAB)
    in if not (isDisj (f@@rootp)) || null oksits then Nothing
       else Just (head oksits)

-------------------------
-- reduction selectors --
-------------------------

defaultRedSel :: ReductionSelector
defaultRedSel = do
  f <- getWorkingFormula
  let reds = findReductions f
  oldestRed f reds
 where oldestRed :: Formula -> [ReductionInfo] -> IOState ReductionInfo
       oldestRed _ [] = failS (iosError 1 "no more reductions")
       oldestRed _ [ri] = return ri
       oldestRed f (ri@(rp,rt,ts):ri'@(rp',rt',ts'):r) =
         if olderRP f rp rp' then oldestRed f (ri:r) else oldestRed f (ri':r)
       olderRP f p p' =
         let fp = f@@p; fp' = f@@p'
             si = getStrategyInfo (getNodeInfo fp)
             si' = getStrategyInfo (getNodeInfo fp')
             g = map fst si; g' = map fst si'
             mg = minimum g; mg' = minimum g'
         in mg < mg'

interactiveRedSel :: ReductionSelector
interactiveRedSel = do
  f <- getWorkingFormula
  let reds = findReductions f
      l = length reds
  if null reds then failS (iosError 1 "no more literal pairs")
   else do drawWithNodeInfoAndLitPositions f
           newLine'
           printReds reds
           putStr' ("Select a reduction [1.." ++ show l ++ "]: ")
           i <- rReadCond rReadLn (\x -> x > 0 && x <= l)
           return (reds!!(i-1))

printReds :: [ReductionInfo] -> IOState ()
printReds = printReds' 1
 where printReds' _ [] = done
       printReds' n (r:rs) = do
         let (p,rt,_) = r
         putStrLn' (show n ++ ") " ++ show rt ++ " at position " ++ show p)
         printReds' (n+1) rs


--------------------
-- term selectors --
--------------------

defaultTermSelector :: TermSelector
defaultTermSelector p AnyVar =
  do f <- getWorkingFormula
     let ni = getNodeInfo (f@@p)
         si = getStrategyInfo ni
         mt = selectSIOldest si; Just t = mt
     return (if nodeInfoIsEmpty ni || null si || mt == Nothing
               then termFromVar (mkVar 0)
             else t)
defaultTermSelector p AnyTerm =
  do f <- getWorkingFormula
     let ni = getNodeInfo (f@@p)
         si = getStrategyInfo ni
         csi = filter (\(_,mt) -> let Just t = mt in mt /= Nothing && termIsCon t)
                 si
         mc = selectSIOldest csi; Just ct = mc
         mt = selectSIOldest si; Just t = mt
     return (if nodeInfoIsEmpty ni || null si then termFromVar (mkVar 0)
             else if null csi then t else ct)
defaultTermSelector p (AnyOf ts) =
  do f <- getWorkingFormula
     let ni = getNodeInfo (f@@p)
         si = getStrategyInfo ni
         siFilter (_,Nothing) = False
         siFilter (_,Just t) = t `elem` ts
         si' = filter siFilter si
         Just t = selectSIOldest si'
     return (if nodeInfoIsEmpty ni || null si' then head ts
             else t)

superTermSelector :: TermSelector
superTermSelector p AnyVar = defaultTermSelector p AnyVar
superTermSelector p (AnyOf fs) = defaultTermSelector p (AnyOf fs)
superTermSelector p AnyTerm = do
  wf <- getWorkingFormula
  (lp,lp') <- getLiteralPair
  let qf = wf@@p; x = getQuantVar qf
      sf = getSubFormula qf
      sits = getSITerms (getStrategyInfo (getNodeInfo qf))
      -- subformulae under quantifier with SITerms substituted
      sbfs = map (\t -> subst x t sf) sits
      -- formulae with term substituted
      fs = map (\sf -> setSubFormulaAtPos p sf wf) sbfs
      -- new positions
      posLitUQ = lp < p
      l = length (posToList p)
      p1 = concPos p (listToPos (drop (l+1) (posToList lp)))
      p2 = concPos p (listToPos (drop (l+1) (posToList lp')))
      (np,np') = if posLitUQ then (p1,lp') else (lp,p2)
      -- compute quality indices
      qis = map (\f -> let (o,n) = qualityIndex' f np np' in o - n) fs
      -- find best term according to qis
      qvl = zip qis sits
      selt [(_,t)] = t
      selt (qt@(q,t):qt'@(q',t'):r) = if q > q' then selt (qt:r) else selt (qt':r)
{-
  putStrLn' ("sits = " ++ (show sits))
  putStrLn' ("sbfs = " ++ (show sbfs))
  putStrLn' ("fs = " ++ (show (map (@@p) fs)))
  putStrLn' ("lp < p = " ++ show (lp<p) ++ ", posLitUQ = " ++ show posLitUQ)
  putStrLn' ("np/np' = " ++ show (np,np'))
  putStrLn' ("qif = " ++ show (map (\f -> extractSLPFor f np np') fs))
  let rp = commonPrefix np np'; rootPath = getPath (head fs) lambda rp
      rvs = foldr (add . getQuantVar) empty (filter isExists (pathToList rootPath))
  putStrLn' ("rvs = " ++ show rvs)
  putStrLn' ("cs = " ++ show (cnf' (extractSLPFor (head fs) np np')))
  putStrLn' ("pdcs = " ++ show (processDisjunctions (cnf' (extractSLPFor (head fs) np np'))))
  putStrLn' ("qis = " ++ show qis)
  putStrLn' ("qvl = " ++ show qvl)
-}
  return (if null qvl then termFromVar (mkVar 0) else selt qvl)

  

selectSIOldest :: StrategyInfo -> Maybe Term
selectSIOldest [] = Nothing
selectSIOldest [(_,mt)] = mt
selectSIOldest (s@(g,mt):s'@(g',mt'):r) =
 if g <= g' then selectSIOldest (s:r)
 else selectSIOldest (s':r)


interactiveTermSelector :: TermSelector
interactiveTermSelector p AnyVar =
  do putStr' "Enter any variable: "
     v <- rReadVar
     return (termFromVar v)
interactiveTermSelector p AnyTerm =
  do putStr' "Enter any term: "
     rReadTerm
interactiveTermSelector p (AnyOf ts) =
  do putStr' ("Enter a term, one of " ++ show ts ++ ": ")
     rReadCond rReadTerm (flip elem ts)


---------------------------
-- disjunction selectors --
---------------------------

maxDisjSel :: ForListPartitioner
maxDisjSel fs = return (fs,[])

minDisjSel :: ForListPartitioner
minDisjSel fs = return ([head fs], tail fs)

interactiveDisjSel :: ForListPartitioner
interactiveDisjSel fs = do
  let l = length fs
  putStrLn' ("Enter a list of subformulae [0.." ++ show (l-1) ++
             "]: ")
  sl <- rReadLn
  let sfs = [f | i <- sl, let f = fs!!i]
      nsfs = [f | i <- [0..(l-1)] \\ sl, let f = fs!!i]
  return (sfs, nsfs)

---------------------------
-- conjunction selectors --
---------------------------

defaultConjSel :: ForListPartitioner
defaultConjSel fs = return ([head fs], tail fs)

interactiveConjSel :: ForListPartitioner
interactiveConjSel = interactiveDisjSel

----------------------------
-- literal pair selectors --
----------------------------

lruLitPairSel :: LitPairSelector
lruLitPairSel = do
  f <- getWorkingFormula
  let lp = unifiableLitPairs f
      notSameMark (p,p') =
          null (intersect (getStrategyInfo (getNodeInfo (f@@p)))
                          (getStrategyInfo (getNodeInfo (f@@p'))))
      lps = filter notSameMark lp
      dl = [(p,p',dist f p p') | (p,p') <- lps]
      mymin x1@(_,_,d1) x2@(_,_,d2) = if d1 < d2 then x1 else x2
      (rp,rp',_) = foldl1 mymin dl
  if null lps then failS (iosError 1 "no more literal pairs") else return (rp,rp')
 where dist f p p' =
         let si1 = getStrategyInfo (getNodeInfo (f@@p))
             si2 = getStrategyInfo (getNodeInfo (f@@p'))
         in length si1 + length si2

firstLitPairSel :: LitPairSelector
firstLitPairSel = do
  f <- getWorkingFormula
  let litPositions = collectLiterals f
      pairs pl nl = takeOneOfEach [pl,nl]
      litPairs = foldr ((++) . (\(_,pls,nls) -> pairs pls nls)) [] litPositions
      sameMark (p,p') = not (
          null (intersect (getStrategyInfo (getNodeInfo (f@@p)))
                          (getStrategyInfo (getNodeInfo (f@@p')))))
      dl = dropWhile (\[p,p'] -> sameMark (p,p') ||
                           not (litPairIsUnifiable f (p,p'))) litPairs
      [p,p']:_ = dl
  if null dl then failS (iosError 1 "no more literal pairs") else return (p,p')

minDistLitPairSel :: LitPairSelector
minDistLitPairSel = do
  f <- getWorkingFormula
  let -- get literal pairs with different marks
      lps = goodLitPairs f
      -- sort by estimated literal distance
      slps = sortLitPairsBy (estimatedLitDist f) lps
--      slps = sortLitPairsBy posLitDist lps
      -- throw away not unifiable literal pairs
      ulps = filter (litPairIsUnifiable f) slps
      -- get unifiable literal pair with minimal distance
      lp:_ = ulps
  if null ulps then failS (iosError 1 "no more literal pairs")
               else return lp


superLitPairSel :: LitPairSelector
superLitPairSel = do
  f <- getWorkingFormula
  let lps = goodLitPairs f
      -- sort by estimated literal distance
      slps = sortLitPairsBy posLitDist lps
      -- get unifiable literal pair with minimal distance
      lps' = take 20 (filter (litPairIsUnifiable f) slps)
      lp:_ = sortLitPairsBy (superSorter f) lps'
  if null lps' then failS (iosError 1 "no more literal pairs")
               else return lp
 where superSorter :: Formula -> Position -> Position -> Int
       superSorter f p p' =
         let d = estimatedLitDist f p p'
             (o,n) = qualityIndex' f p p'
             q o | o < 10 = o
                 | o < 100 = 15
                 | otherwise = 20
         in 8*d + 5*n + 2 * (q o)

{-
superLitPairSel :: LitPairSelector
superLitPairSel = do
  f <- getWorkingFormula
  let lps = goodUnifiableLitPairs f
      -- sort by estimated literal distance & qi
      slps = sortLitPairsBy (superSorter f) lps
      -- get unifiable literal pair with minimal distance
      lp:_ = slps
  if null slps then failS (iosError 1 "no more literal pairs")
               else return lp
 where superSorter :: Formula -> Position -> Position -> Int
       superSorter f p p' =
         let d = estimatedLitDist f p p'
             (o,n) = qualityIndex f p p'
             q o | o < 5 = 0
                 | o < 100 = 1
                 | otherwise = 2
         in 10*d + 5*n + q o
-}
interactiveLitPairSel :: LitPairSelector
interactiveLitPairSel = do
  f <- getWorkingFormula
  let lps = goodLitPairs f
      slps = sortLitPairsBy (estimatedLitDist f) lps
      ulps = filter (litPairIsUnifiable f) slps
      l = length ulps
  if null ulps then failS (iosError 1 "no more literal pairs")
   else do drawWithNodeInfoAndLitPositions f
           newLine'
           printLitPairs f ulps
           putStr' ("Select a literal pair [1.." ++ show l ++ "]: ")
           i <- rReadCond rReadLn (\x -> x >0 && x <= l)
           return (ulps!!(i-1))

-------------------------------------------------
-- literal pair selectors, auxiliary functions --
-------------------------------------------------

-- find all literal pairs in formula f where the positive
-- and negative literals do not share same marks.

goodUnifiableLitPairs :: Formula -> [(Position, Position)]
goodUnifiableLitPairs f =
  let lps = unifiableLitPairs f
      notSameMark (p,p') =
       let sip = getStrategyInfo (getNodeInfo (f@@p))
           sip' = getStrategyInfo (getNodeInfo (f@@p'))
       in null (intersect sip sip')
  in filter notSameMark lps

goodLitPairs :: Formula -> [(Position, Position)]
goodLitPairs f =
  let lps = collectLiteralPairs f
      notSameMark (p,p') =
       let sip = getStrategyInfo (getNodeInfo (f@@p))
           sip' = getStrategyInfo (getNodeInfo (f@@p'))
       in null (intersect sip sip')
  in filter notSameMark lps      

markCheck :: Formula -> (Position, Position) -> Bool
markCheck f (p,p') =
  let  sip = getStrategyInfo (getNodeInfo (f@@p))
       sip' = getStrategyInfo (getNodeInfo (f@@p'))
  in null (intersect sip sip')

-- sort literal pairs by distance function df such that smaller distances
-- occur before greater distances in the resulting list

sortLitPairsBy :: (Position -> Position -> Int) -> [(Position, Position)]
  -> [(Position, Position)]
sortLitPairsBy df lps =
  let distComp (p1,p1') (p2,p2') = compare (df p1 p1') (df p2 p2')
  in quickSortBy distComp lps


estimatedLitDist :: Formula -> Position -> Position -> Int
estimatedLitDist f p p' =
  let root = commonPrefix p p'
      quantPosLitPath = filter isQuant (pathToList (getPath f root p))
      quantNegLitPath = filter isQuant (pathToList (getPath f root p'))
      quantPosLitVars = foldr (add . getQuantVar) empty quantPosLitPath
      quantNegLitVars = foldr (add . getQuantVar) empty quantNegLitPath
      posLitVars = vars (f@@p); negLitVars = vars (f@@p')
      critPosVars = posLitVars `minus` quantPosLitVars
      critNegVars = negLitVars `minus` quantNegLitVars
      critVars = critPosVars `sunion` critNegVars
      rootEQs = filter (isExists . fst) (getFormulaeAndPositions f lambda root)
      eQuantVars = map (\(f,p) -> (getQuantVar f,p)) rootEQs
      qPos [] = Nothing
      qPos ((v,p):r) = if v `isElement` critVars then Just p
                       else qPos r
      mqp = qPos eQuantVars; Just qp = mqp
      lr = length (posToList root); lq = length (posToList qp)
      lp = length (posToList p); lp' = length (posToList p')
  in if mqp == Nothing then (lp-lr)+(lp'-lr)
     else (lp-lq)+(lp'-lq)

posLitDist :: Position -> Position -> Int
posLitDist p p' =
  let rl = length (posToList (commonPrefix p p'))
      pl = length (posToList p); pl' = length (posToList p')
  in (pl-rl)+(pl'-rl)


printLitPairs :: Formula -> [(Position,Position)] -> IOState ()
printLitPairs = printLitPairs' 0

printLitPairs' :: Int -> Formula -> [(Position, Position)] -> IOState ()
printLitPairs' _ _ [] = done
printLitPairs' n f ((p,p'):r) = do
  putStr' (show (n+1) ++ ") ")
  putStr' (show p ++ " ("); showFor (f@@p); putStr' ") and "
  let root = commonPrefix p p'; rl = length (posToList root)
      pl = length (posToList p); pl' = length (posToList p')
      d = (pl-rl)+(pl'-rl)
  putStr' (show p' ++ " ("); showFor (f@@p')
  putStrLn' ("), distance " ++ show d)
  putStr' ("     est.dist.: " ++ show (estimatedLitDist f p p'))
  putStr' (", qual.ind.: " ++ show (qualityIndex f p p'))
  putStrLn' (", new qual.ind.: " ++ show (qualityIndex' f p p'))
  if n `mod` 10 == 9 then do
    putStr' "<RETURN> to continue, any other character to abort. "
    str <- readWith getLine
    if null str then printLitPairs' (n+1) f r
     else putStrLn' "Aborted, wait..."
   else  printLitPairs' (n+1) f r



----------- E N D -----------------------


mixedStrategy :: Int -> DisjSelectType -> LitPairSelector -> ReductionStrategy
mixedStrategy n dst lps =
  do s <- getStep
     if s `mod` (n+1) == n then
       -- call ORP reducer
       theORPReducer maxDisjSel maxDisjSel defaultConjSel defaultRedSel
                     defaultTermSelector
      else
       -- call LP reducer
       theLPReducer lps dst defaultTermSelector


--------------------------------------------------

collectLitPos :: Formula -> [(Position, Position)]
collectLitPos f =
  let litPos = collectLiterals f
      pairs pl nl = map (\[a,b] -> (a,b)) (takeOneOfEach [pl,nl])
   in foldr ((++) . (\(_,pls,nls) -> pairs pls nls)) [] litPos


-- findLitPairs f ps:
-- compute all literal pairs where one member is a literal at position p,
-- and p is an element of ps
findLitPairs :: Formula -> [Position] -> [(Position, Position)]
findLitPairs _ [] = []
findLitPairs f (p:ps) =
  let findMatchLitPos :: Position -> Formula -> Predicate -> Bool -> [Position]
      findMatchLitPos p f pred posLit
        | isPosLit f = if getLitPred (getLiteral f) == pred && posLit then [p] else []
        | isNegLit f = if getLitPred (getLiteral f) == pred && not posLit then [p] else []
        | otherwise = concat [px | i <- [0..l-1],
                              let px = findMatchLitPos (mkPos p i) (sfs!!i) pred posLit]
        where sfs = getSubFormulae f; l = length sfs
      pred = getLitPred (getLiteral (f@@p))
      posLit = isNegLit (f@@p)  -- search for positive literals if f@@p is negative
      mkPair pos = if posLit then (pos,p) else (p,pos)
      newPairs = map mkPair (findMatchLitPos lambda f pred posLit)
      otherPairs = findLitPairs f ps
  in newPairs ++ otherPairs

-----------------


-----------------

collectLitPred :: Predicate -> Formula -> ([Position], [Position])
collectLitPred pr f = colLits lambda pr f
  where colLits :: Position -> Predicate -> Formula -> ([Position], [Position])
        colLits pos pr f
         | isPosLit f = if predOk then ([pos], []) else ([],[])
         | isNegLit f = if predOk then ([], [pos]) else ([],[])
         | otherwise = foldr (\(a,b) (c,d) -> (a++c,b++d)) ([],[]) cs
         where predOk = getLitPred (getLiteral f) == pr
               sf = getSubFormula f
               sfs = getSubFormulae f; l = length sfs
               cs :: [([Position],[Position])]
               cs = [colLits (mkPos pos i) pr (sfs!!i) | i <- [0..l-1]]

-----------------------------

 


---------- what's this ?
{-
minSelect [] = error "wrong argument in minSelect"
minSelect [(p,rt,si)] = (p, rt, [minSelSI si])
minSelect ((p,rt,si):(p',rt',si'):ris) =
  let msi = minSelSI si; msi' = minSelSI si'
  in if fst msi < fst msi' then minSelect ((p,rt,[msi]):ris)
     else minSelect ((p',rt',[msi']):ris)

minSelSI :: StrategyInfo -> (Int, Maybe Term)
minSelSI [] = error "wrong arg in minSelSI"
minSelSI sis = foldl1 mini sis
  where mini (a,b) (c,d) = if a <=c then (a,b) else (c,d)

------------------------------------------
----- interactive reduction strategy -----
------------------------------------------

iSelectTerm :: RuleType -> StrategyInfo -> IO (Maybe Term)
iSelectTerm _ [] = error "wrong args in iSelectTerm"
iSelectTerm _ [(_,mt)] = return mt
iSelectTerm rt s@((_,mt):_) = case rt of
  DistributionRule -> return mt
  ExistentialRule -> do putStr "Enter a term: (one of "
                        putStrLn (show (map (\(g, Just t) -> t) s) ++ ")")
                        t <- readLn
                        return (Just t)
  UniversalRule -> do putStr "Enter a variable: (one of "
                      putStrLn (show (map (\(g, Just t) -> t) s) ++ ")")
                      v <- readLn
                      return (Just (termFromVar v))
                         
printReds :: [ReductionInfo] -> IO ()
printReds = printReds' 0

printReds' :: Int -> [ReductionInfo] -> IO ()
printReds' _ [] = return ()
printReds' n (r:rs) = do
  putStr (show n ++ ") ")
  let (p,rt,_) = r
  putStrLn (show rt ++ " at position " ++ show p)
  printReds' (n+1) rs

interactiveReducer :: Reducer
interactiveReducer mv dsA dsD csD =
  do nextGeneration
     (f,g) <- getForGenStep
     let rs = findPReductions f
     printWith printReds rs
     printWith putStr "Select a reduction: "
     i <- rReadLn
     let (p, rt, si) = rs!!i
     mt <- doIO (iSelectTerm rt si)
     f' <- doIO (applyRule g mv rt p (Just dsA) (Just dsD) (Just csD) mt f)
     setFormula (simplify f')
-}


-------------------

-- "collectLiteralPositions f p" returns a list of all those literal positions
-- in formula f which occur at a position p' which is a subposition of p.

collectLiteralPositions :: Formula -> Position -> [Position]
collectLiteralPositions f p =
  let colLP f p
       | isPosLit f = [p]
       | isNegLit f = [p]
       | otherwise = concat lps
       where sfs = getSubFormulae f; l = length sfs
             lps = [ps | i <- [0..l-1], let ps = colLP (sfs!!i) (mkPos p i)]
  in map (concPos p) (colLP (f@@p) lambda)

-- "collectPosLitPositions f p" returns  a list of all those positive literal positions
-- in formula f which occur at a position p', a subposition of p.

collectPosLitPositions :: Formula -> Position -> [Position]
collectPosLitPositions f p =
  let colLP f p
       | isPosLit f = [p]
       | otherwise = concat lps
       where sfs = getSubFormulae f; l = length sfs
             lps = [ps | i <- [0..l-1], let ps = colLP (sfs!!i) (mkPos p i)]
  in map (concPos p) (colLP (f@@p) lambda)

-- "findMatchingLiteralPositons f p lp" returns all those literal positions p':
-- in formula f such that:
-- 1. p' < p (p' is subposition of p) and
-- 2. the literal at position lp has the same, but negated, predicate as the literal
--    at positon p'.

findMatchingLiteralPositions :: Formula -> Position -> Position -> [Position]
findMatchingLiteralPositions f p lp =
  let flp = f@@lp
      searchPosLits = isNegLit flp
      pred = getLitPred (getLiteral flp)
      predok f = getLitPred (getLiteral f) == pred
      colMLP f p ltp
       | isPosLit f = if ltp && predok f then [p] else []
       | isNegLit f = if ltp || not (predok f) then [] else [p]
       | otherwise = concat lps
       where sfs = getSubFormulae f; l = length sfs
             lps = [ps | i <- [0..l-1], let ps = colMLP (sfs!!i) (mkPos p i) ltp]
  in map (concPos p) (colMLP (f@@p) lambda searchPosLits)

-- "collectLiteralPairs f p" returns a list of literal pairs (pp,np) where pp is the
-- position of a positive, np of a negative literal, the predicates of both literals are
-- the same, and both positions are subpositions of p.

collectLiteralPairs' :: Formula -> Position -> [(Position,Position)]
collectLiteralPairs' f p =
  [(pl,nl) | pl <- collectPosLitPositions f p,
             nl <- findMatchingLiteralPositions f p pl]


collectLiteralPairs :: Formula -> [(Position, Position)]
collectLiteralPairs f =
  let litPositions = collectLiterals f
      pairs pl nl = takeOneOfEach [pl,nl]
      ps = foldr ((++) . (\(_,pls,nls) -> pairs pls nls)) [] litPositions
  in map (\[a,b] -> (a,b)) ps


------------------------ TEST AREA ------------------------


qualityIndex :: Formula -> Position -> Position -> (Int,Int)
qualityIndex wf p p' =
  let f = extractSLPFor wf p p'
      -- compute prenex normal form
      f' = pnf f
      -- get existential quantifiers in quantifier prefix
      eqpvs = eqvsInQuantPrefix f'
      -- get existential quantifiers on root path
      rp = commonPrefix p p'; rootPath = getPath wf lambda rp
      rvs = foldr (add . getQuantVar) empty (filter isExists (pathToList rootPath))
      eqrvs = rvs `intersection` free f'
      eqvs = eqpvs `sunion` eqrvs
      -- compute cunjunctive normal form of matrix
      cs = cnf' (dropQuantPrefix f')
      cs' = processDisjunctions cs
      -- test disjunctions
      ts = testDisjunctions eqvs cs'
      -- compute quality index for literal pair
      (l,m) = partition id ts
  in (length l, length m)

qualityIndex' :: Formula -> Position -> Position -> (Int,Int)
qualityIndex' wf p p' =
  let f = extractSLPFor wf p p'
      -- compute prenex normal form
      f' = pnf f
      -- get existential quantifiers in quantifier prefix
      eqpvs = eqvsInQuantPrefix f'
      -- get existential quantifiers on root path
      rp = commonPrefix p p'; rootPath = getPath wf lambda rp
      rvs = foldr (add . getQuantVar) empty (filter isExists (pathToList rootPath))
      eqrvs = rvs `intersection` free f'
      eqvs = eqpvs `sunion` eqrvs
      -- compute cunjunctive normal form of matrix
      cs = cnf' (dropQuantPrefix f')
      cs' = processDisjunctions cs
      -- test disjunctions
      ts = testDisj eqvs cs'
      -- compute quality index for literal pair
      (l,m) = partition id ts
  in (length l, length m)

extractSLPFor :: Formula -> Position -> Position -> Formula
extractSLPFor f p p' =
  let rp = commonPrefix p p'; rl = length (posToList rp)
      i = head (drop rl (posToList p)); j = head (drop rl (posToList p'))
      sfs = getSubFormulae (f@@rp)
      lsfs = filter isLiteral sfs
      fp = if isLiteral (sfs!!i) then [] else [sfs!!i]
      fp' = if isLiteral (sfs!!j) then [] else [sfs!!j]
  in disj (lsfs ++ fp ++ fp')

slpTest :: Formula -> Position -> Position -> IOState ()
slpTest wf p p' = do
  let f = extractSLPFor wf p p'
      -- compute prenex normal form
      f' = pnf f
      -- get existential quantifiers in quantifier prefix
      eqpvs = eqvsInQuantPrefix f'
      -- get existential quantifiers on root path
      rp = commonPrefix p p'; rootPath = getPath wf lambda rp
      rvs = foldr (add . getQuantVar) empty (filter isExists (pathToList rootPath))
      eqrvs = rvs `intersection` free f'
      eqvs = eqpvs `sunion` eqrvs
      -- compute cunjunctive normal form of matrix
      cs = cnf' (dropQuantPrefix f')
      cs' = processDisjunctions cs
      -- test disjunctions
      ts = testDisjunctions eqvs cs'
      -- compute quality index for literal pair
      qi = (\(l,m) -> (length l, length m)) (partition id ts)
  putStrLn' ("extracted for.: " ++ show f)
  putStrLn' ("E-quant. variables: " ++ show eqvs)
  putStrLn' ("cs: " ++ show cs)
  putStrLn' ("cs': " ++ show cs')
  putStrLn' ("ts: " ++ show ts)
  putStrLn' ("qi: " ++ show qi)

extractSLPSubFor :: IOState Formula
extractSLPSubFor = do
  f <- getWorkingFormula
  (p,p') <- getLiteralPair
  let rp = commonPrefix p p'; rl = length (posToList rp)
      i = head (drop rl (posToList p)); j = head (drop rl (posToList p'))
      sfs = getSubFormulae (f@@rp)
      lsfs = filter isLiteral sfs
      fp = if isLiteral (sfs!!i) then [] else [sfs!!i]
      fp' = if isLiteral (sfs!!j) then [] else [sfs!!j]
  return (disj (lsfs ++ fp ++ fp'))

-- input is cnf representation of boolean formula

processDisjunctions :: [[Formula]] -> [[Formula]]
processDisjunctions cs = map procDisj cs
  where procDisj :: [Formula] -> [Formula]
        procDisj ds =
          let ds' = filter isLiteral ds
              (pls,nls) = partition isPosLit ds'
              ppreds = foldr (add . getLitPred . getLiteral) empty pls
              npreds = foldr (add . getLitPred . getLiteral) empty nls
              pfilter f = getLitPred (getLiteral f) `isElement` npreds
              nfilter f = getLitPred (getLiteral f) `isElement` ppreds
              fpls = filter pfilter pls; fnls = filter nfilter nls
          in if null fpls || null fnls then [] else fpls ++ fnls

testDisj :: Set Variable -> [[Formula]] -> [Bool]
testDisj _ [] = []
testDisj vs (c:cs) =
  let (vs',ok,cs') = td vs c cs
      oks = testDisj vs' cs'
  in ok:oks

td :: Set Variable -> [Formula] -> [[Formula]]
  -> (Set Variable, Bool, [[Formula]])
td vs [] cs = (vs, False, cs)
td vs lits cs =
  let (pls, nls) = partition isPosLit lits
      litPairs = [(lp,ln) | lp <- pls, ln <- nls]
      oklps = filter fst3 (map (tlp vs) litPairs)
      (ok,sbs,vs'):_ = oklps
      cs' = foldr (\(v,t) -> map (map (subst v t))) cs sbs
  in if null oklps || not ok then (vs,False,cs)
     else (vs',True,cs')
      
tlp :: Set Variable -> (Formula, Formula)
  -> (Bool, [(Variable,Term)], Set Variable)
tlp vs (pl,nl) =
  let pts = getLitTerms (getLiteral pl)
      nts = getLitTerms (getLiteral nl)
      tps = zip pts nts
      cf :: Set Variable -> [(Term, Term)] -> (Bool, [(Variable,Term)], Set Variable)
      cf vs [] = (True, [], vs)
      cf vs (tp:r) =
        let (ok, sb, vs') = tts vs tp
            (okr, sbr, vsr) = cf vs' (map (appSubTP sb) r)
        in if not ok || not okr then (False, [], vs)
           else (True, sb ++ sbr, vsr)
      appSubTP [] tp = tp
      appSubTP [(v,t)] (t1,t2) = (appSubT v t t1, appSubT v t t2)
      appSubT v t t' | termIsVar t' = if v == v' then t else termFromVar v'
                     | otherwise    = t'
        where v' = getTermVar t'
  in cf vs tps
      
tts :: Set Variable -> (Term, Term) -> (Bool, [(Variable, Term)], Set Variable)
tts vs (t,t')
  | termIsVar t && termIsVar t' = if v == v' then (True, [(v,t')], nvs)
                                  else if vinvs then (True, [(v,t')], nvs)
                                  else if vinvs' then (True, [(v',t)], nvs')
                                  else (False, [], vs)
  | termIsVar t && termIsCon t' = if vinvs then (True, [(v,t')], nvs) else (False, [], vs)
  | termIsCon t && termIsVar t' = if vinvs' then (True, [(v',t)], nvs') else (False, [], vs)
  | termIsCon t && termIsCon t' = (c == c', [], vs)
  where v = getTermVar t; v' = getTermVar t'
        c = getTermCon t; c' = getTermCon t'
        vinvs = v `isElement` vs; vinvs' = v' `isElement` vs
        nvs = vs `minus` singleton v; nvs' = vs `minus` singleton v'


testDisjunctions :: Set Variable -> [[Formula]] -> [Bool]
testDisjunctions vs cs = map (testDisj vs) cs
  where testDisj :: Set Variable -> [Formula] -> Bool
        testDisj _ [] = False
        testDisj vs lits =
          let (pls, nls) = partition isPosLit lits
              litPairs = [(lp,ln) | p <- pls, n <- nls,
                          let { lp = getLiteral p; ln = getLiteral n }]
          in any (testLitPair vs) litPairs
              
-- positive and negative literals with same predicate

testLitPair :: Set Variable -> (Literal, Literal) -> Bool
testLitPair vs (pl,nl) =
  let pts = getLitTerms pl; nts = getLitTerms nl
      termOK (t,t') =
        let v = getTermVar t; v' = getTermVar t'
            c = getTermCon t; c' = getTermCon t'
            tv = termIsVar t; t'v = termIsVar t'
        in tv && v `isElement` vs || t'v && v' `isElement` vs ||
           tv && t'v && v == v' || not tv && not t'v && c == c'
  in all termOK (zip pts nts)

-- prenex normal form:
-- input formula must be in nnf
-- output: not flat, not simplified

pnf :: Formula -> Formula
pnf f | isDisj f = pnfx pnsf disj
      | isConj f = pnfx pnsf conj
      | otherwise = setSubFormulae pnsf f
      where sf = getSubFormulae f
            pnsf = map pnf sf

-- input fs in pnf

pnfx :: [Formula] -> ([Formula] -> Formula) -> Formula
pnfx fs c =
  if null q then c nq
  else if x `isElement` badVars then
    setQuantVar y (setSubFormula (pnfx nfs' c) h)
  else setSubFormula (pnfx nfs c) h
  where (q,nq) = partition isQuant fs
        (aq,eq) = partition isForAll q
        h:r = aq ++ eq; x = getQuantVar h
        badVars = foldr (sunion . free) empty fs
        y = head ([mkVar 0..] \\ setToOrdList badVars)
        y' = termFromVar y
        h' = getSubFormula h; h'' = subst x y' h'
        nfs = h':r++nq; nfs' = h'':r++nq

eqvsInQuantPrefix :: Formula -> Set Variable
eqvsInQuantPrefix f
  | isExists f = add (getQuantVar f) (eqvsInQuantPrefix sf)
  | isForAll f = eqvsInQuantPrefix sf
  | otherwise = empty
  where sf = getSubFormula f

-- cnf: conjunctive normal form of boolean part

cnf :: Formula -> Formula
cnf f = conj cs'
  where cs = cnf' f
        cs' = map disj cs

cnf' :: Formula -> [[Formula]]
cnf' f | isDisj f =  [cs | cs <- takeOneOfEachAndConcat cnfs]
       | isConj f = concat cnfs
       | otherwise = [[f]]
 where sfs = getSubFormulae f
       cnfs = map cnf' sfs 

takeOneOfEachAndConcat :: [[[a]]] -> [[a]]
takeOneOfEachAndConcat [] = []
takeOneOfEachAndConcat (x:[]) = x
takeOneOfEachAndConcat (x:xs) = [(i ++ r) | i <- x, r <- takeOneOfEachAndConcat xs]

dropQuantPrefix :: Formula -> Formula
dropQuantPrefix f
  | isQuant f = dropQuantPrefix (getSubFormula f)
  | otherwise = f
