module Subsumption where

import Auxiliary
import Simplify
import Formula
import List
import Unification
import Reduction
import Strategy
import IOStateMonad
import GlobalState

perms :: [a] -> [[a]]
perms [] = [[]]
perms (x:xs) = [zs | ys <- perms xs, zs <- interleave x ys ]
  where interleave :: a -> [a] -> [[a]]
        interleave x [] = [[x]]
        interleave x (y:ys) = [x:y:ys] ++ map (y:) (interleave x ys)

-- input formulae must be flat and in nnf

embv f g = equalModBoundVars (f,g)

equalModBoundVars :: (Formula,Formula) -> Bool
equalModBoundVars (f, g)
  | isVerum f && isVerum g = True
  | isFalsum f && isFalsum g = True
  | isLiteral f && isLiteral g = f == g
  | isDisj f && isDisj g =
      vf == vg && ff == fg && pf' == pg' && nf' == ng' && opListEq df dg &&
      opListEq cf cg && opListEq ef eg && opListEq af ag
  | isConj f && isConj g =
      vf == vg && ff == fg && pf' == pg' && nf' == ng' && opListEq df dg &&
      opListEq cf cg && opListEq ef eg && opListEq af ag
  | isExists f && isExists g = equalModBoundVars (f', (subst y (termFromVar x) g'))
  | isForAll f && isForAll g = equalModBoundVars (f', (subst y (termFromVar x) g'))
  | otherwise = False
 where
  sff@(f':_) = getSubFormulae f; sfg@(g':_) = getSubFormulae g
  -- split depending on top operator
  (vf,ff,pf,nf,df,cf,ef,af) = partitionFL sff
  (vg,fg,pg,ng,dg,cg,eg,ag) = partitionFL sfg
  x = getQuantVar f; y = getQuantVar g
  pf' = quickSort (map getLiteral pf); pg' = quickSort (map getLiteral pg)
  nf' = quickSort (map getLiteral nf); ng' = quickSort (map getLiteral ng)
  

-- equality modulo bound variables on formula lists
equalModBoundVarsL fs gs = all equalModBoundVars (zip fs gs)

-- formula lists with same top operator
opListEq :: [Formula] -> [Formula] -> Bool
opListEq fs gs =
  let ordTupByFst (x,_) (y,_) = compare x y
      ofs = quickSortBy ordTupByFst (splitBy size fs)
      ogs = quickSortBy ordTupByFst (splitBy size gs)
      pair :: [(Int, [Formula])] -> [(Int, [Formula])]
               -> Maybe [([Formula], [Formula])]
      pair [] [] = Just []
      pair [] _ = Nothing
      pair _ [] = Nothing
      pair ((k,fs):r) ((k',fs'):r') =
        let mp = pair r r'; Just p = mp
        in if mp == Nothing then Nothing
           else if k == k' && length fs == length fs' then Just ((fs,fs'):p)
           else Nothing
      mps = pair ofs ogs; Just ps = mps
      equ :: ([Formula], [Formula]) -> Bool
      equ (fs,gs) = let fs' = perms fs in any (equalModBoundVarsL gs) fs'
  in if length fs /= length gs || mps == Nothing then False
     else all equ ps


splitBy :: Eq k => (a -> k) -> [a] -> [(k,[a])]
splitBy _ [] = []
splitBy f (x:xs) =
  let k = f x
      r = splitBy f xs
  in addItem k x r
 where addItem k x [] = [(k,[x])]
       addItem k x ((k',l'):r) =
         if k == k' then (k,x:l'):r else (k',l'):addItem k x r

partitionFL :: [Formula] -> ([Formula],[Formula],[Formula],[Formula]
  ,[Formula],[Formula],[Formula],[Formula])
partitionFL [] = ([],[],[],[],[],[],[],[])
partitionFL (g:gs)
  | isVerum g   = (g:v,f,p,n,d,c,e,a)
  | isFalsum g  = (v,g:f,p,n,d,c,e,a)
  | isPosLit g  = (v,f,g:p,n,d,c,e,a)
  | isNegLit g  = (v,f,p,g:n,d,c,e,a)
  | isDisj g    = (v,f,p,n,g:d,c,e,a)
  | isConj g    = (v,f,p,n,d,g:c,e,a)
  | isExists g  = (v,f,p,n,d,c,g:e,a)
  | isForAll g  = (v,f,p,n,d,c,e,g:a)
  where (v,f,p,n,d,c,e,a) = partitionFL gs

-- remove equal subformulae: input formula must be flat and in nnf

remEqSubFors :: Formula -> Formula
remEqSubFors f
  | isDisj f = if sfsCollapsed then head sfs' else setSubFormulae sfs' f
  | isConj f = if sfsCollapsed then head sfs' else setSubFormulae sfs' f
  | isExists f = setSubFormula sf' f
  | isForAll f = setSubFormula sf' f
  | otherwise = f
 where sfs = map remEqSubFors (getSubFormulae f)
       sf' = remEqSubFors (getSubFormula f)
       sfs' = nubBy embv sfs
       sfsCollapsed = null (tail sfs')


--------------

-- delete literals which are not unifiable

delNotUnifiableLiterals :: Formula -> Formula
delNotUnifiableLiterals f =
  let ulp = unifiableLitPairs f
      lits = collectLiterals f
      litps = foldr ((++) . (\(_,p,np) -> p ++ np)) [] lits
      (up,up') = unzip ulp
      ulits = nub (up ++ up')
      notUnif = litps \\ ulits
  in foldr (\p f -> setSubFormulaAtPos p falsum f) f notUnif

----------------

-- perform all distribution rules which can result in formula splitting
-- input formula must be simplified. output formula is simplified

distribOut :: IOState ()
distribOut = do
  f <- getWorkingFormula
  let -- find reduction position
      reds = findReductions f
      dreds = filter (\(_,r,_) -> r == DistributionRule) reds
      dfilter (p,_,_) = let path = filterPath (not . isDisj) (getPath f lambda p)
                        in pathIsEmpty (pathRest path)
      d2reds = filter dfilter dreds
      p = fst3 (head d2reds)
  if null d2reds then return () -- no more distribution rules applicable
   else do
     -- reduce
     applyDRule p maxDisjSel defaultConjSel
     updateWorkingFormula simplify
     distribOut
