----------------------------------------------------------------------
-- Module Unification.hs                                            --
--                                                                  --
-- Implementation of literal pair unification algorithms.	    --
--                                                                  --
-- 06.12.96  initial revision                                       --
----------------------------------------------------------------------


--------------------
----- comments -----
--------------------

-- The literal pair reduction strategy requires the computation of
-- a most general unifier for a literal pair.
-- A literal pair of a formula f is a pair (l,l') of literals where l is
-- positive, l' negative and the literals are connected by a disjunction
-- in formula f.
-- The unification result can be used to compute reduction steps to
-- delete a literal pair from a formula.
-- This is possible whenever the unification of the literal pair provides
-- a most general unifier.
-- 
-- A unification mapping delivers the link between terms of
-- formula f and unification terms.
--
-- To compute a unification a literal pair subformula is extracted
-- consisting only of the two literals l and l' in question, the
-- disjunction connecting them and all quantifiers binding a
-- variable of any of the two literals:
--
-- l = P(s_1,...,s_n)    l' = ~P(t_1,...t_n)
--              \           /
--               Qx        Qy
--                \       /   <---- neg. lit.path
--   pos. ---->    Qz    Qu
--   lit.           \   /
--   path            \ / <---- disjunction
--                    |
--                    Qv
--                    |    <---- root path
--                    Qw
--


----------------------------
----- module interface -----
----------------------------

module Unification (
  -- data types
  UTerm (		-- unification term: variable, constant or function
    UVar, UCon, UOp
  ),
  UEquation,		-- unification equation: pair of unification terms
  UResult (		-- unification result: failure or list of equations
    Fail, Solution
  ),
  PathInfo (		-- origin of variable in extracted subformula
    RootPath,
    PosLitPath,
    NegLitPath,
    Unknown             -- see def. of data type for comments
  ),
  UMap,			-- unification mapping (for variables and constants)

  -- unification map & result lookup operations
  uMapLookupByUTerm,		-- search in unification map for term and path info
  uMapLookupByTermAndPathInfo,	-- search in unification map for unification term
  uResultLookup,		-- search for complementary unification term
  computeERedTerm,		-- compute term for existential quantifier reductions

  -- unification
  unifiableLitPairs,	-- collect all unifiable literal pairs of a formula
  litPairIsUnifiable,	-- is this literal pair unifiable?
  unify,		-- (try to) unify a literal pair
  unifyX,		-- '' and compute unification mapping
  extractLitPairSubFor,	-- extract subformula containing literal pair
  closure,		-- universal closure of a formula
  uniqueBoundVars,	-- rename bound variable to make them unique
  unificationEqSystem,	-- build unification equation system

  solve,		-- solve unification equation system
  solvable,		-- has this unification euation system a solution?
  occurCheck,		-- occur check for unification terms

  -- auxiliary functions
  equalsModSubTerms,	-- equality on unification terms (mod. subterms)
  isUVar,		-- unification term is variable
  absUTerm,		-- absolute value of unification term (var. or func.)
  negUTerm,		-- build complementary unification term
  getUOpVars		-- get variables occurring in unification function
) where


----------------------------
----- imported modules -----
----------------------------

import Set
import Formula
import Simplify
import List ((\\), nub, nubBy)
import Auxiliary


----------------------
----- data types -----
----------------------

-- Unification terms represent variables or constants occurring in a
-- literal pair subformula f. Variables are encoded differently
-- depending on whether they are free or bound by a universal quantifier
-- or if they are bound by an existential quantifier.
-- The encoding works as follows:
-- o Constants:
--     c_i  ->  UCon i       for a constant c_i of formula f
-- o Free variables or variables bound by a universal quantifier:
--     v_i  ->  UOp  (i+1)   if v_i occurs in the positive literal
--     v_i  ->  UOp (-i-1)   if v_i occurs in the negative literal
--   For a variable v_i which is present in both the positive and the negative
--   literal both representations are used, respectively.
--   Subterms of the unification operators UOp depend on existential
--   quantifiers binding variables of the same literal.
-- o Variables bound by an existential quantifier:
--     v_i  ->  UVar  (i+1)  if v_i occurs in the positive literal
--     v_i  ->  UVar (-i-1)  if v_i occurs in the negative literal
-- Bound variables in the extracted literal pair subformula f must
-- be unique.

data UTerm = UVar Int
           | UCon Int
           | UOp Int [UTerm]
  deriving (Eq, Ord, Show)

-- unification equation, represented as pair of unification terms

type UEquation = (UTerm, UTerm)

-- Unifcation solution data type:
-- A unification result is either a solution consisting of
-- unification equations of the form v == t where v is a unification
-- variable and t a unification term, or Fail if there is no
-- solution.
-- The solution may be interpreted as a substitution delivering
-- the most general common instance (mgu).
-- Non-failing unification results used within this module are
-- always inter-reduced.

data UResult = Fail
             | Solution [UEquation]  -- (UVar, UTerm) pairs
  deriving (Eq, Show)

-- A unification map delivers a connection between unification terms
-- and terms occurring in the formula for which the unification
-- was computed.
-- Any unification term can be mapped to a term - path information pair,
-- where the mapping is one-one for all terms but variables occurring in both
-- literals, which are either both free or both bound by the same quantifier
-- in the extracted literal pair subformula.
-- The path information indicates the origin of the term where `Unknown'
-- is used for all constants, and the other three possibilities for
-- variables, as indicated at the beginning of this module.

type UMap = [(UTerm, Term, PathInfo)]

data PathInfo = RootPath | PosLitPath | NegLitPath | Unknown
  deriving (Eq, Show)


-- A variable map is a finite mapping from variables to variables.
-- It is used internally to keep track of bound variable renamings.

type VMap = [(Variable, Variable)]


------------------------------
----- exported functions -----
------------------------------

--------------------------------------
-- unification map lookup functions --
--------------------------------------


-- "r = uMapLookupByUTerm ut um" searches for the term - path info pair
-- corresponding to unification term ut in the unification map um.
-- If the unification term ut can be found in the unification map the
-- term - path info pair r, otherwise Nothing is returned.

uMapLookupByUTerm :: UTerm -> UMap -> Maybe (Term, PathInfo)
uMapLookupByUTerm _ [] = Nothing
uMapLookupByUTerm lut ((ut, t, i):r) = case (lut,ut) of
  (UVar x, UVar y) -> if x == y then Just (t,i) else uMapLookupByUTerm lut r
  (UCon x, UCon y) -> if x == y then Just (t,i) else uMapLookupByUTerm lut r
  (UOp x _, UOp y _) -> if x == y then Just (t,i) else uMapLookupByUTerm lut r
  otherwise -> uMapLookupByUTerm lut r

-- "ut = uMapLookupByTermAndPathInfo t i um" searches for the unification
-- term ut corresponding to the term t with path info i in unification map um.
-- On success the unification term ut is returned, on failure Nothing.
-- NOTE: A lookup for a variable on the root path does not produce unique
--       results. The returned unification term may be positive or negative
--       and has to be adapted afterwards depending on which literal it
--       belongs to.

uMapLookupByTermAndPathInfo :: Term -> PathInfo -> UMap -> Maybe UTerm
uMapLookupByTermAndPathInfo _ _ [] = Nothing
uMapLookupByTermAndPathInfo lt li ((ut,t,i):r) =
  if lt == t && li == i then Just ut else uMapLookupByTermAndPathInfo lt li r


-- "t = computeERedTerm v i a ur um" computes a reduction term for an existential
-- quantifier reduction depending on a previosly computed unification.
-- ur and um are the unification result and unification mapping of a literal
-- pair unification for literal pair (l,l'). v is a variable bound by an
-- existential quantifier on either the root path or positive or negative literal
-- path. i is the corresponding path information.
-- If the reduction term is arbitrary, Nothing is returned.
-- When variable v has to be unified with another existentially quantified
-- variable, the reduction term is in principle arbitrary. When parameter a
-- is set to True, an arbitrary reduction term is returned (indicated by
-- Nothing), otherwise the variable of the other existentially quantified
-- variable is returned, thus making the reduction term unique.

computeERedTerm :: Variable -> PathInfo -> Bool -> UResult -> UMap -> Maybe Term
computeERedTerm _ _ _ Fail _ = error "unification result is failure in computeERedTerm"
computeERedTerm _ _ _ _ [] = Nothing
computeERedTerm v i anyT ures umap =
  let v' = termFromVar v
      ut = uMapLookupByTermAndPathInfo v' i umap; Just uTerm = ut
      rt = uResultLookup uTerm ures; Just redTerm = rt
      t = uMapLookupByUTerm redTerm umap; Just (term,_) = t
  in if ut == Nothing || rt == Nothing || t == Nothing then Nothing
     else if isUVar redTerm && anyT then Nothing else Just term

-- "ut' = uResultLookup ut ur" searches for the corresponding unification term
-- ut' of a unification term ut in unification result ur; i.e. it returns
-- "the other side" of an equation for ut.
-- When no equation with term ut exists in the unification result, Nothing is
-- returned.
-- The unifcation result should be interreduced to avoid variable chains.

uResultLookup :: UTerm -> UResult -> Maybe UTerm
uResultLookup _ Fail = Nothing
uResultLookup t (Solution []) = Nothing
uResultLookup t (Solution ((l,r):es)) =
  if t == l then Just r
  else if t == r then Just l
  else uResultLookup t (Solution es)


-----------------------------------------
-- literal pair subformula unification --
-----------------------------------------

-- "l = unifiableLitPairs f" determines all unifiable literal pairs
-- in a formula f. l is a list of literal pairs represented by
-- the positions (p,p') of the literals in f. p is the position of a positive,
-- p' the position of a negative literal.
-- Unifiability of such a literal pair implies that the formula can be reduced
-- to a formula, where the literal pair initially at positions (p,p') has vanished.

unifiableLitPairs :: Formula -> [(Position, Position)]
unifiableLitPairs f =
  let -- find all literal occurrences in f
      litPositions = collectLiterals f
      -- build all (p,p') pairs where ~p = p'
      pairs pl nl = map (\[x,y] -> (x,y)) (takeOneOfEach [pl,nl])
      litPairs = foldr ((++) . (\(_,pls,nls) -> pairs pls nls)) [] litPositions
  in filter (litPairIsUnifiable f) litPairs


-- "t = litPairIsUnifiable f (p,p')" computes whether or not the literal
-- pair at positions p and p' is unifiable.

litPairIsUnifiable :: Formula -> (Position, Position) -> Bool
litPairIsUnifiable f (p, p') =
  let -- perform disjunction test
      root = commonPrefix p p'
      disjTest = isDisj (f@@root)
      -- compute literal pair subformula and closure
      f' = closure (extractLitPairSubFor (p,p') f)
      -- rename bound variables
      f'' = uniqueBoundVars f'
      -- build unification equation system
      es = unificationEqSystem f''
  in disjTest && solvable es

-- "ur = unify f (p,p')" computes the unification result (mgu, substitution) 
-- for a unifiable literal pair (l,l') occurring at positions (p,p') in formula
-- f.

unify :: Formula -> (Position, Position) -> UResult
unify f (p,p') =
  let -- compute closure of literal pair subformula
      f' = closure (extractLitPairSubFor (p,p') f)
      -- rename bound variables
      f'' = uniqueBoundVars f'
      -- build unification equation system
      es = unificationEqSystem f''
  in solve es

-- "(ur,um) = unifyX f (p,p')" is an extended unification algorithm:
-- It also returns the unification map for the literal pair at positions
-- (p,p') in formula f.
-- In other respects the specification is as for algorithm `unify' (see above).

unifyX :: Formula -> (Position, Position) -> (UResult, UMap)
unifyX f (p,p') =
  let -- compute closure of literal pair subformula
      f' = closure (extractLitPairSubFor (p,p') f)
      -- rename bound variables and get extended variable map
      (f'', vmap) = uniqueBoundVarsAndVMap f'
      -- build unification equation system and unification map
      (es, umap) = unificationEqSystemAndUMap f''
      -- solve unification equation system
      uresult = solve es
      -- apply vmap to umap and throw away unused entries
      umap' = applyVMapToUMap vmap umap
      inSolution (Solution []) _ = False
      inSolution (Solution ((l,r):eqs)) e@(ut,_,_) =
        if ut `equalsModSubTerms`l || ut `equalsModSubTerms` r then True
        else inSolution (Solution eqs) e
      umap'' = filter (inSolution uresult) umap'
  in (uresult, umap'')


----------------------------------------
-- literal pair subformula extraction --
----------------------------------------

-- extractLitPairSubFor (p,p') f = f'
-- p is a pos. Lit. Position, p' is a neg. Lit. Position in f,
-- connected by disjunction,
-- f' is the Subformula of f consisting of p, p' and the 
-- corresponding quantifier subtree.

extractLitPairSubFor :: (Position, Position) -> Formula -> Formula
extractLitPairSubFor (p, p') f =
  let root = commonPrefix p p'
      pQuantPath = filterPath isQuantOrLit (getPath f root p)
      nQuantPath = filterPath isQuantOrLit (getPath f root p')
      rQuantPath = filterPath isQuant (getPath f lambda root)
      df = disj [connectPath pQuantPath, connectPath nQuantPath]
      rpl = pathLength rQuantPath
      dpos = listToPos (replicate rpl 0)
  -- removeUQ: remove unused quantifiers
  in if pathIsEmpty rQuantPath then removeUQ df
     else removeUQ (setSubFormulaAtPos dpos df (connectPath rQuantPath))


-- "f' = closure f" computes the universal closure f' of a formula f.
-- I.e. it apends universal quantifiers for all free variables of
-- formula f.

closure :: Formula -> Formula
closure f = foldr forall f (setToOrdList (free f))


-----------------------
-- variable renaming --
-----------------------

-- "f' = uniqueBoundVars f" renames bound variables in formula f
-- such that each variable bound by some quantifier is not bound
-- by another quantifier.

uniqueBoundVars :: Formula -> Formula
uniqueBoundVars = fst . uniqueBoundVarsAndVMap
  

----------------------------------------------
-- unification equation system construction --
----------------------------------------------

-- "es = unificationEqSystem f" builds the unification equation
-- system of a literal pair subformula f. I.e. formula f must be of the form
-- described at the beginning of this module: It contains a literal pair
-- (positive and negative literal), a disjunction and arbitrary quantifiers.
-- Ther must not contain any free variable in formula f and all bound variables
-- have to be distinct.
-- A formula of the required type is produced by e.g.:
--   `uniqueBoundVars . closure . extractLitPairSubFor'
-- Then the equation system needed for unification is computed.

unificationEqSystem :: Formula -> [UEquation]
unificationEqSystem = fst . unificationEqSystemAndUMap


------------------------------------------
-- solving unification equation systems --
------------------------------------------

-- "r = solve es" tries to solve a system of unification equation.
-- On success, a substitution (mgu) is returned, otherwise Fail.
-- The solution produced on success is interreduced.
-- Variables with different signs are assumed to be equal, as
-- a unification variable occurring both positive and negative must
-- lie on the root path.

solve :: [UEquation] -> UResult
solve [] = Solution []
solve ((e@(l,r)):es) = case (l,r) of
-- changed on 24.03.: added x == -y
  (UVar x, UVar y) -> if x == y || x == -y then solve es
                      else addEq e (solve (map (applySubs [e]) es))
  (UVar _, UCon _) -> addEq e (solve (map (applySubs [e]) es))
  (UVar _, UOp _ ts) -> if occurCheck l ts then Fail
                        else addEq e (solve (map (applySubs [e]) es))
  (UCon _, UVar _) -> addEq (r,l) (solve (map (applySubs [(r,l)]) es))
  (UCon c, UCon d) -> if c == d then solve es else Fail
  (UCon _, UOp _ _) -> Fail
  (UOp _ ts, UVar _) -> if occurCheck r ts then Fail
                        else addEq (r,l) (solve (map (applySubs [(r,l)]) es))
  (UOp _ _, UCon _) -> Fail
-- added i=-j on 24.02.
  (UOp i ts, UOp j us) -> if i == j || i == -j then solve (es ++ zip ts us)
                          else Fail
  otherwise -> error "unexpected case in solve"
 where -- insert equation (substitution) to solution
       addEq :: UEquation -> UResult -> UResult
       addEq _ Fail = Fail
       addEq e (Solution es) =
         let s@(l,r) = applySubs es e
         in case r of UOp _ ts -> if occurCheck l ts then Fail
                                  else Solution (s:es)
                      otherwise -> Solution (s:es)
       -- apply substitution to a equation
       applySubs :: [UEquation] -> UEquation -> UEquation
       applySubs es (l,r) = (appTSub es l, appTSub es r)
       -- apply substitution to a term
       appTSub :: [UEquation] -> UTerm -> UTerm
       appTSub es t = foldr app t es
         where app (UVar v, r) uv@(UVar v') = if v == v' then r else uv
               app _ c@(UCon _) = c
               app e (UOp x ts) = UOp x ts

-- test, if equation system is solvable

solvable :: [UEquation] -> Bool
solvable es = solve es /= Fail

-- "t = occurCheck v ts" tests whether or not variable v occurs in any
-- of the terms ts.

occurCheck :: UTerm -> [UTerm] -> Bool
occurCheck _ [] = False
occurCheck uv@(UVar v) ((UVar v'):r) =
  if v == v' then True else occurCheck uv r
occurCheck uv ((UCon _):r) = occurCheck uv r
occurCheck uv ((UOp _ ts):r) = occurCheck uv (ts ++ r)


-------------------------
-- auxiliary functions --
-------------------------

-- Equality relation on unification terms, treating unification
-- operators with different subterms as equal.

equalsModSubTerms :: UTerm -> UTerm -> Bool
ut `equalsModSubTerms` ut' = case (ut,ut') of
  (UVar x, UVar y) -> x == y
  (UCon x, UCon y) -> x == y
  (UOp x _, UOp y _) -> x == y
  otherwise -> False

-- Is this unification term a unification variable?

isUVar :: UTerm -> Bool
isUVar (UVar _) = True
isUVar _ = False

-- Compute the absolute value of a unification term, i.e.
-- no negative indices are used for unification variables and operators.

absUTerm :: UTerm -> UTerm
absUTerm (UVar x) = UVar (abs x)
absUTerm (UCon c) = UCon c
absUTerm (UOp x ts) = UOp (abs x) (map absUTerm ts)

-- "ut' = negUTerm ut" negates all unification variables and operators
-- un unification term ut.

negUTerm :: UTerm -> UTerm
negUTerm (UVar v) = UVar (-v)
negUTerm (UCon c) = UCon c
negUTerm (UOp x ts) = UOp (-x) (map negUTerm ts)

-- "uts = getUOpVars ut" extracts all unification variables occurring
-- as subterms of a unification operator. This procedure is applied
-- recursively, sucht that nested variables are considered, too.

getUOpVars :: UTerm -> [UTerm]
getUOpVars (UOp _ ts) = nub (foldr ((++) . g) [] ts)
  where g :: UTerm -> [UTerm]
        g (UVar x) = [UVar x]
        g (UCon _) = []
        g (UOp _ ts) = nub (foldr ((++) . g) [] ts)
getUOpVars _ = []


------------------------------
----- internal functions -----
------------------------------

-- "f' = removeUQ f" removes unused quantifiers from formula f.
-- Universal and existential quantifiers binding no free variable occurrence
-- are removed.

removeUQ :: Formula -> Formula
removeUQ f
  | isExists f = if x `isElement` vs then f' else sf'
  | isForAll f = if x `isElement` vs then f' else sf'
  | otherwise = setSubFormulae (map removeUQ (getSubFormulae f)) f
  where sf = getSubFormula f
        sf' = removeUQ sf
        f' = setSubFormula sf' f
        x = getQuantVar f
        vs = free sf'
       

-- "t = isQuantOrLit f" returns True, if the top of formula f is a quantifier
-- or a (positive or negative) literal.

isQuantOrLit :: Formula -> Bool
isQuantOrLit f = isQuant f || isLiteral f



-- "f = connectPath p" converts a path (approx.: list of formulae)
-- to a formula. The path has to consist solely of quantifier and literal
-- formula.

connectPath :: Path -> Formula
connectPath p
  | pathIsEmpty p = error "wrong argument in connectPath"
  | otherwise = let (h,r) = decomposePath p
                    fr = connectPath r
                in if pathIsEmpty r then h else setSubFormula fr h

-- "(f,vm) = uniqueBoundVarsAndVMap" renames bound variables in formula
-- f, such that each variable bound by some quantifier is not bound
-- by another quantifier. The renaming of bound variables is captured
-- in the variable mapping vm, which consists of variable pairs (n,o)
-- where o is the old variable and n the new, unique variable.

uniqueBoundVarsAndVMap :: Formula -> (Formula, VMap)
uniqueBoundVarsAndVMap f = (f',vmap)
  where (f',_,vmap) = ubv f empty
        -- second parameter is set of already bound variables
        ubv :: Formula -> Set Variable -> (Formula, Set Variable, VMap)
        ubv f bv
          | isQuant f =
              let -- decompose f
                  x = getQuantVar f
                  sf = getSubFormula f
                  -- get new variable y, not already bound and not free in f
                  y = head ([mkVar 0..] \\ setToOrdList (bv `sunion` free f))
                  yt = termFromVar y
                  -- new bound variable set; naming convention:
                  -- x': no renaming necessary, x'': renaming necessary
                  bv' = add x bv; bv'' = add y bv
                  -- new subformula and variable map
                  (nsf',nbv',vmap') = ubv sf bv'
                  (nsf'',nbv'',vmap'') = ubv (subst x yt sf) bv''
                  -- update vmap
                  nvmap' = vmap'
                  nvmap'' = (y,x) : updVMap y x vmap''
              in if x `isElement` bv then
                (setSubFormula nsf'' (setQuantVar y f), nbv'', nvmap'')
              else
                (setSubFormula nsf' f, nbv', nvmap')
          | otherwise =
              let sfs = getSubFormulae f
                  listubv :: [Formula] -> Set Variable ->
		    ([Formula], Set Variable, VMap)
                  listubv [] bv = ([], bv, [])
                  listubv (f:fs) bv =
                    let (f',bv',vmap') = ubv f bv
                        (fs',bv'',vmap'') = listubv fs bv'
                    in (f':fs', bv'', vmap' ++ vmap'')
                  (sfs',bv',vmap') = listubv sfs bv
              in (setSubFormulae sfs' f, bv', vmap')
        -- update variable map
        updVMap :: Variable -> Variable -> VMap -> VMap
        updVMap _ _ [] = []
        updVMap y x (p@(n,o):r) = if o == x then (n,y):r else p : updVMap y x r


-- "um' = applyVMapToUMap vm um" applies a variable mapping to a unification
-- mapping. All variable entries in the unification map must have distinct
-- variables as second (term) arguments.
-- These second arguments are then adapted according to the variable mapping.

applyVMapToUMap :: VMap -> UMap -> UMap
applyVMapToUMap vmap umap = map (applyVMap vmap) umap
  where applyVMap [] ume = ume
        applyVMap ((nv,ov):r) ume@(ut,t,i) =
          if termIsVar t && getTermVar t == nv then (ut, termFromVar ov, i)
          else applyVMap r ume

-- "(es,um) = unificationEqSystemAndUMap f" builds the  unification equation
-- system of a literal pair subformula f. I.e. formula f must be of the form
-- described at the beginning of this module: It contains a literal pair
-- (positive and negative literal), a disjunction and arbitrary quantifiers.
-- Ther must not contain any free variable in formula f and all bound variables
-- have to be distinct.
-- A formula of the required type is produced by e.g.:
--   `uniqueBoundVars . closure . extractLitPairSubFor'
-- Then the equation system needed for unification is computed as well as a
-- unification mapping containing the origins of the unification terms
-- stored as term - path info pairs.

unificationEqSystemAndUMap :: Formula -> ([UEquation], UMap)
unificationEqSystemAndUMap f =
  let -- get root path, positive and negative literals and corr. quantifier paths
      (rootp, pl, pqp, nl, nqp) = decomposeLitPairSubFor f
      -- get literal terms
      plts = getLitTerms pl; nlts = getLitTerms nl
      -- build unification terms
      (puts, pumap) = buildPosUTermsAndUMap plts rootp pqp
      (nuts, numap) = buildNegUTermsAndUMap nlts rootp nqp
    in (zip puts nuts, nubBy uTermEq (pumap ++ numap))
 where uTermEq (UVar x,_,_) (UVar y,_,_) = x == y
       uTermEq (UCon x,_,_) (UCon y,_,_) = x == y
       uTermEq (UOp x _,_,_) (UOp y _,_,_) = x == y
       uTermEq _ _ = False

-- "(rp,l,p,l',n) = decomposeLitPairSubFor f" decomposes a literal pair
-- subformula (as described above) into its various parts:
-- rp is the root path, l the positive literal, p the positive literal
-- path, l' the negative literal, and n the negative literal path.

decomposeLitPairSubFor :: Formula -> (Path, Literal, Path, Literal, Path)
decomposeLitPairSubFor f
  | isQuant f =    -- root path
      let sf = getSubFormula f
          (rp, pl, pqp, nl, nqp) = decomposeLitPairSubFor sf
          rp' = mkPath f rp
      in (rp', pl, pqp, nl, nqp)
  | otherwise =    -- disjunction 
      -- this must be a disjunction!
      let f1:f2:_ = getSubFormulae f
          (pl, pqp) = decompLitFor f1  -- positive literal path
          (nl, nqp) = decompLitFor f2  -- negative literal path
      in (emptyPath, pl, pqp, nl, nqp)
  where decompLitFor :: Formula -> (Literal, Path)
        decompLitFor f
         | isQuant f =   -- pos. or neg. literal path
             let sf = getSubFormula f
                 (l, qp) = decompLitFor sf
                 qp' = mkPath f qp
             in (l, qp')
         | otherwise = (getLiteral f, emptyPath)   -- literals



-- build unification terms
-- positive: positive indices for UVars and UOps
-- negative: negative indices for UVars and UOps

-- "(uts,um) = buildNegUTermsAndUMap ts rp np" computes unification terms for
-- a literal pair's negative literal terms ts. rp is the root path, np
-- the negative literal path of the extracted literal pair subformula.
-- For each veriable there must not be more than one quantifier on the
-- root and negative literal path.
-- Then uts is the set of unification terms corresponding to ts,
-- umap is the unification mapping assigning each unification term
-- its origin as a term - path info pair.

buildNegUTermsAndUMap :: [Term] -> Path -> Path -> ([UTerm], UMap)
buildNegUTermsAndUMap ts rp np =
  let -- use pos. lit. routine to compute preliminary result
      (uts, umap) = buildPosUTermsAndUMap ts rp np
      -- negate variables and operators in unification terms and map
      negComponent (ut,t,PosLitPath) = (negUTerm ut,t,NegLitPath)
      negComponent (ut,t,pi) = (negUTerm ut,t,pi)
  in (map negUTerm uts, map negComponent umap)

-- "(uts,um) = buildPosUTermsAndUMap ts rp pp" computes unification terms for
-- a literal pair's positive literal terms ts.
-- Specification is the same as for `buildNegUTermsAndUMap' with `negative'
-- changed to `positive' where needed.

buildPosUTermsAndUMap :: [Term] -> Path -> Path -> ([UTerm], UMap)
buildPosUTermsAndUMap ts rp p
  = let -- variable conversion map
        varConvMap :: Path -> [UTerm] -> [(UTerm,Term)]
        varConvMap pa exv = if pathIsEmpty pa then [] else
          let (h,r) = decomposePath pa
              x = getQuantVar h; xt = termFromVar x
              xi = getVarIndex x; ux = xi + 1
          in if isExists h then (UVar ux,xt) : varConvMap r (UVar ux:exv)
             else (UOp ux exv, xt) : varConvMap r exv
        vmap' = map (\(ut,t) -> (ut,t,RootPath)) (varConvMap rp [])
        rts = filter isUVar (map fst3 vmap')
        vmap = vmap' ++ map (\(ut,t) -> (ut,t,PosLitPath)) (varConvMap p rts)
        -- constant conversion map
        cs = filter termIsCon ts
        con2UCon c = UCon (getConIndex (getTermCon c))
        cmap = map (\c -> (con2UCon c, c, Unknown)) cs
        -- build umap
        umap = vmap ++ cmap
        -- convert terms to UTerms
        uMapLookup [] _ = error "internal error in builPosUTermsAndUMap"
        uMapLookup ((ut,t,_):r) lt =
          if t == lt then ut else uMapLookup r lt
        uts = map (uMapLookup umap) ts
    in (uts, umap)
