----------------------------------------------------------------------
-- Module IOStateMonad.hs                                           --
--                                                                  --
-- This module implements a combined state/IO monad.                --
--                                                                  --
-- 24.01.97  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module GFIOStateMonad where


-------------------
----- imports -----
-------------------

import Formula


----------------------
----- data types -----
----------------------

newtype IOStateMonad s a = SM (s -> IO (a,s))
type ForGenIOSt a = IOStateMonad (Formula, Generation)


---------------------------------
----- instance declarations -----
---------------------------------

instance Functor (IOStateMonad s) where
  map f (IOSt m) = IOSt (\s -> do (x,s') <- m s
                                  return (f x, s'))

instance Monad (IOStateMonad s) where
  return x   = IOSt (\s -> return (x, s))
  IOSt m >>= f = IOSt (\s -> do (x,s') <- m s
                                let IOSt f' = f x
                                f' s')

------------------------------
----- exported functions -----
------------------------------

-- functions for generic IOStateMonad

startingWith :: IOStateMonad s a -> s -> IO a
(IOSt m) `startingWith` v = do (x,s') <- m v
                             return x

fetch :: IOStateMonad s s
fetch = IOSt (\s -> return (s,s))

fetchWith :: (s -> a) -> IOStateMonad s a
fetchWith f = IOSt (\s -> return (f s, s))

update :: (s -> s) -> IOStateMonad s ()
update f = IOSt (\s -> return ((), f s))

set :: s -> IOStateMonad s ()
set s' = IOSt (\s -> return ((), s'))

-- functions for ForGenIOSt monad

initForGenIOSt :: ForGenIOSt a -> Formula -> Generation -> IO a
initForGenIOSt (IOSt m) f g = do (x,s) <- m (f,g)
                           return x

getGeneration :: ForGenIOSt Generation
getGeneration = fetchWith snd

getFormula :: ForGenIOSt Formula
getFormula = fetchWith fst

getForGen :: ForGenIOSt GFormula
getForGen = fetch

nextGeneration :: ForGenIOSt ()
nextGeneration = update (\(f,g) -> (f, g+1))

setGeneration :: Generation -> ForGenIOSt ()
setGeneration g = update (\(f',g') -> (f',g))

setFormula :: Formula -> ForGenIOSt ()
setFormula f = update (\(f',g') -> (f,g'))

printWith :: Show a => (a -> IO ()) -> a -> ForGenIOSt ()
printWith pf x = IOSt (\s -> do pf x
                              return ((),s))

readWith :: Read a => IO a -> ForGenIOSt a
readWith rf = IOSt (\s -> do x <- rf
                           return (x,s))


------------------------------
----- internal functions -----
------------------------------

printForGenIOSt :: (Show a) => ForGenIOSt a -> IO ()
printForGenIOSt (IOSt m) = let vsIO = m (falsum,0) in
  do (v,s) <- vsIO
     putStrLn ("Value: " ++ show v)
     putStrLn ("State: " ++ show s)

