----------------------------------------------------------------------
-- Module Formula.hs                                                --
--                                                                  --
-- Implementation of data type `Formula' and corresponding methods. --
--                                                                  --
-- 26.09.96  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Formula (
  -- exported data types
  Variable, Constant,
  StandardPredicate(MkSP), PredIdent, Term(Var, Con),
  StandardFormula, ReductionFormula,
  NodeInfo(NIEmpty, NIStrategyInfo), StrategyInfo(MkSI),
  Position,
  ID(IDVar, IDCon, IDPred), IDContext,
  -- exported classes
  Predicate(..),
  Formula(..),
  -- exported functions
  standardIDContext,
  showReductionFormula,
  convertToReductionFormula,
  lambda, mkPos, concPos, posToList, listToPos
) where


-------------------
----- imports -----
-------------------

import Set     -- for variable sets


--------------
-- notation --
--------------

-- To get the subformula at a specified position use this infix operator:
infix 8 @@

----------------------
----- data types -----
----------------------

-- variables, constants, terms, predicates

type Variable = Int
type Constant = Int

type PredIdent = Int

data Term = Var Variable | Con Constant
  deriving (Eq, Ord)

newtype StandardPredicate = MkSP (PredIdent,[Term])
  deriving (Eq, Ord)

-- A formula is a tree, where the nodes consist of formula constructors
-- and additional information used by the reduction strategy, therefore:

data Tree a = Node a (Forest a)
  deriving (Eq, Show)
type Forest a = [Tree a]

-- Formula constructor datatype: The parameter p allows different
-- representations of predicates:

data Predicate p => ForConstructor p = 
  T | F |             -- true and false
  Pred p |            -- predicate
  Not | Or | And |    -- negation, disjunction, conjunction
  Exists Variable |   -- existential quantification
  ForAll Variable     -- universal quantification
 deriving (Eq, Show)

-- A generic formula has arbitrary predicate representation and node
-- information.

type GenericFormulaNode p ni = (ForConstructor p, Maybe ni)
type GenericFormula p ni = Tree (GenericFormulaNode p ni)

-- A standard formula is a special generic formula with empty nodeinformation.
-- It is primarily used for I/O-operations.

type StandardFormulaTree = GenericFormula StandardPredicate ()
newtype StandardFormula = MkSF StandardFormulaTree

-- A reduction formula stores reduction information in each node.

type ReductionFormulaTree = GenericFormula StandardPredicate StrategyInfo
newtype ReductionFormula = MkRF ReductionFormulaTree

-- Node information data structure

newtype StrategyInfo = MkSI [(Generation, Maybe Term)]

data NodeInfo = NIEmpty
              | NIStrategyInfo StrategyInfo
  deriving Show

-- Generation is part of NodeInfo's NIStrategyInfo

type Generation = Int

-- Position data structure

newtype Position = Pos [Int]
  deriving (Eq, Show)

-- A context assigns names to variables, constants and predicates

data ID = IDVar Variable | IDCon Constant | IDPred PredIdent
type IDContext = [(String, ID)]

------------------------------
----- class declarations -----
------------------------------

-- Functions on predicates are gathered within this class:

class Predicate p where
  -- constructors
  mkPred :: PredIdent -> [Term] -> p
  -- selectors
  getPredIdent :: p -> PredIdent
  getTermList  :: p -> [Term]
  -- operations
  getPredVars :: p -> Set Variable
  getPredConsts :: p -> Set Constant


-- Common operations valid for all generic formulae are gathered
-- in this class:

class Formula f where
  -- constructors
  verum, falsum :: f
  predicate :: StandardPredicate -> f
  neg :: f -> f
  disj, conj :: [f] -> f
  exists, forall :: Variable -> f -> f
  -- selectors
  isVerum, isFalsum :: f -> Bool
  isPredicate :: f -> Bool
  isNeg, isConj, isDisj :: f -> Bool
  isExists, isForAll :: f -> Bool
  getPredicate :: f -> StandardPredicate
  getQuantVar :: f -> Variable
  getSubFormula :: f -> f
  getSubFormulae :: f -> [f]
  getNodeInfo :: f -> NodeInfo
  (@@) :: f -> Position -> f
  -- modifiers
  setQuantVar :: Variable -> f -> f
  setSubFormula :: f -> f -> f
  setSubFormulae :: [f] -> f -> f
  setNodeInfo :: NodeInfo -> f -> f
  addSubFormulae :: [f] -> f -> f
  setSubFormulaAtPos :: Position -> f -> f -> f
  -- operations
  vars :: f -> Set Variable
  free :: f -> Set Variable
  bound :: f -> Set Variable
  consts :: f -> Set Constant

---------------------------------
----- instance declarations -----
---------------------------------

-- instances of class Eq:

instance Eq StandardFormula where
  MkSF f == MkSF g = equalFT (f,g)

instance Eq ReductionFormula where
  MkRF f == MkRF g = equalFT (f,g)


-- instances of class Show:

instance Show Term where
  showsPrec _ (Var v) = showString (getVarName standardIDContext v)
  showsPrec _ (Con c) = showString (getConName standardIDContext c)

instance Show StandardPredicate where
  showsPrec _ (MkSP (pi, ts)) =
    showString (getPredName standardIDContext pi) . showTermList ts
      where showTermList = showTL id '('
            showTL e _ []     = e
            showTL e c (t:ts) = showChar c . shows t .
                                showTL (showChar ')') ',' ts

instance Show StandardFormula where
  showsPrec _ (MkSF f) = showFormulaTree 0 f

instance Show ReductionFormula where
  showsPrec p (MkRF f) = showFormulaTree p f

--instance Show Void where
--  showsPrec _ _ = id

instance Show StrategyInfo where
  showsPrec _ (MkSI l) = showSIList "" l
    where showSIList _ [] = id
          showSIList s ((g,mt):is) = showString s . showChar '(' . shows g .
            showChar ',' . case mt of
            { Nothing -> showString "||"; Just t  -> shows t } .
            showChar ')' . showSIList "," is

-- instances of class Formula:

instance Formula StandardFormula where
  -- constructors
  verum             = MkSF (Node (T, Nothing) [])
  falsum            = MkSF (Node (F, Nothing) [])
  predicate p       = MkSF (Node (Pred p, Nothing) [])
  neg (MkSF f)      = MkSF (Node (Not, Nothing) [f])
  disj fs           = MkSF (Node (Or, Nothing) (map stripSF fs))
  conj fs           = MkSF (Node (And, Nothing) (map stripSF fs))
  exists v (MkSF f) = MkSF (Node (Exists v, Nothing) [f])
  forall v (MkSF f) = MkSF (Node (ForAll v, Nothing) [f])
  -- selectors
  isVerum (MkSF f)       = getConstructor f == T
  isFalsum (MkSF f)      = getConstructor f == F
  isPredicate (MkSF f)   = case getConstructor f of
    Pred _    -> True
    otherwise -> False
  isNeg (MkSF f)         = getConstructor f == Not
  isDisj (MkSF f)        = getConstructor f == Or
  isConj (MkSF f)        = getConstructor f == And
  isExists (MkSF f)      = case getConstructor f of
    Exists _  -> True
    otherwise -> False
  isForAll (MkSF f)      = case getConstructor f of
    ForAll _  -> True
    otherwise -> False
  getPredicate (MkSF f)  = case getConstructor f of
    Pred p -> p
    otherwise -> error "wrong argument in getPredicate"
  getQuantVar (MkSF f)   = case getConstructor f of
    Exists v  -> v
    ForAll v  -> v
    otherwise -> error "wrong argument in getQuantVar"
  getSubFormula (MkSF f) = case getConstructor f of
    Not       -> sf
    Exists _  -> sf
    ForAll _  -> sf
    otherwise -> error "wrong argument in getSubFormula"
   where (sf:_) = getSubFormulae (MkSF f)
  getSubFormulae (MkSF (Node (_,Nothing) sfs)) = map wrapSF sfs
  getNodeInfo   = const NIEmpty
  (MkSF f)@@p   = MkSF (getSubForAtPos p f)
  -- modifiers
  setQuantVar v (MkSF (Node (op, Nothing) sfs)) = case op of
    Exists _  -> MkSF (Node (Exists v, Nothing) sfs)
    ForAll _  -> MkSF (Node (ForAll v, Nothing) sfs)
    otherwise -> error "wrong argument in setQuantVar"
  setSubFormula (MkSF sf) (MkSF (Node (op, Nothing) _)) = case op of
    Not       -> MkSF (Node (op, Nothing) [sf])
    Exists _  -> MkSF (Node (op, Nothing) [sf])
    ForAll _  -> MkSF (Node (op, Nothing) [sf])
    otherwise -> error "wrong argument in setSubFormulae"
  setSubFormulae sfs (MkSF (Node (op, Nothing) _)) =
    MkSF (Node (op, Nothing) (map stripSF sfs))
  setNodeInfo ni f = case ni of
    NIEmpty   -> f
    otherwise -> error "wrong argument in setNodeInfo"
  addSubFormulae fs (MkSF (Node (op, Nothing) sfs)) = case op of
    Or        -> MkSF (Node (op, Nothing) (sfs ++ map stripSF fs))
    And       -> MkSF (Node (op, Nothing) (sfs ++ map stripSF fs))
    otherwise -> error "wrong argument in addSubFormulae"
  setSubFormulaAtPos p (MkSF sf) (MkSF f) = MkSF (setSubForAtPos p sf f)
  -- operations
  vars (MkSF f)   = getVars f
  free (MkSF f)   = freeVars f
  bound (MkSF f)  = boundVars f
  consts (MkSF f) = getConsts f


instance Formula ReductionFormula where
  -- constructors
  verum              = MkRF (Node (T, Nothing) [])
  falsum             = MkRF (Node (F, Nothing) [])
  predicate p        = MkRF (Node (Pred p, Nothing) [])
  neg (MkRF f)       = MkRF (Node (Not, Nothing) [f])
  disj fs            = MkRF (Node (Or, Nothing) (map stripRF fs))
  conj fs            = MkRF (Node (And, Just (MkSI [])) (map stripRF fs))
  exists v (MkRF f)  = MkRF (Node (Exists v, Just (MkSI [])) [f])
  forall v (MkRF f)  = MkRF (Node (ForAll v, Just (MkSI [])) [f])
  -- selectors
  isVerum (MkRF f)       = getConstructor f == T
  isFalsum (MkRF f)      = getConstructor f == F
  isPredicate (MkRF f)   = case getConstructor f of
    Pred _    -> True
    otherwise -> False
  isNeg (MkRF f)         = getConstructor f == Not
  isDisj (MkRF f)        = getConstructor f == Or
  isConj (MkRF f)        = getConstructor f == And
  isExists (MkRF f)      = case getConstructor f of
    Exists _  -> True
    otherwise -> False
  isForAll (MkRF f)      = case getConstructor f of
    ForAll _  -> True
    otherwise -> False
  getPredicate (MkRF f)  = case getConstructor f of
    Pred p -> p
    otherwise -> error "wrong argument in getPredicate"
  getQuantVar (MkRF f)   = case getConstructor f of
    Exists v  -> v
    ForAll v  -> v
    otherwise -> error "wrong argument in getQuantVar"
  getSubFormula (MkRF f) = case getConstructor f of
    Not       -> sf
    Exists _  -> sf
    ForAll _  -> sf
    otherwise -> error "wrong argument in getSubFormula"
   where (sf:_) = getSubFormulae (MkRF f)
  getSubFormulae (MkRF (Node (_,_) sfs)) = map wrapRF sfs
  getNodeInfo (MkRF (Node (_, ni) _)) = case ni of
    Nothing   -> NIEmpty
    Just n -> NIStrategyInfo n
  (MkRF f)@@p   = MkRF (getSubForAtPos p f)
  -- modifiers
  setQuantVar v (MkRF (Node (op, ni) sfs)) = case op of
    Exists _  -> MkRF (Node (Exists v, ni) sfs)
    ForAll _  -> MkRF (Node (ForAll v, ni) sfs)
    otherwise -> error "wrong argument in setQuantVar"
  setSubFormula (MkRF sf) (MkRF (Node (op, ni) _)) = case op of
    Not       -> MkRF (Node (op, ni) [sf])
    Exists _  -> MkRF (Node (op, ni) [sf])
    ForAll _  -> MkRF (Node (op, ni) [sf])
    otherwise -> error "wrong argument in setSubFormula"
  setSubFormulae sfs (MkRF (Node (op, ni) _)) =
    MkRF (Node (op, ni) (map stripRF sfs))
  setNodeInfo ni (MkRF (Node (op, _) sfs)) = case ni of
    NIEmpty           -> MkRF (Node (op, Nothing) sfs)
    NIStrategyInfo si -> MkRF (Node (op, Just si) sfs)
  addSubFormulae fs (MkRF (Node (op, ni) sfs)) = case op of
    Or        -> MkRF (Node (op, ni) (sfs ++ map stripRF fs))
    And       -> MkRF (Node (op, ni) (sfs ++ map stripRF fs))
    otherwise -> error "wrong argument in addSubFormulae"
  setSubFormulaAtPos p (MkRF sf) (MkRF f) = MkRF (setSubForAtPos p sf f)
  -- other functions
  vars (MkRF f)   = getVars f
  free (MkRF f)   = freeVars f
  bound (MkRF f)  = boundVars f
  consts (MkRF f) = getConsts f


-- instances of class Predicate:

instance Predicate StandardPredicate where
  -- constructors
  mkPred i ts = MkSP (i, ts)
  -- selectors
  getPredIdent (MkSP (i, _)) = i
  getTermList (MkSP (_, ts)) = ts
  -- operations
  getPredVars (MkSP (_,ts)) = getPVars ts
  getPredConsts (MkSP (_,ts)) = getPConsts ts


------------------------------
----- exported functions -----
------------------------------

-- Identifier context to make formula I/O more readable:

standardIDContext :: IDContext
standardIDContext =
  [-- variables
   ("x", IDVar 0), ("y", IDVar 1), ("z", IDVar 2),
   ("u", IDVar 3), ("v", IDVar 4), ("w", IDVar 5),
   -- constants
   ("a", IDCon 0), ("b", IDCon 1), ("c", IDCon 2),
   ("0", IDCon 3), ("1", IDCon 4), ("2", IDCon 5),
   -- predicates
   ("P", IDPred 0), ("Q", IDPred 1), ("R", IDPred 2),
   ("S", IDPred 3), ("T", IDPred 4), ("U", IDPred 5)
  ]

-- `showReductionFormula' shows a reduction formula as well as strategy
-- information, when available.

showReductionFormula :: ReductionFormula -> String
showReductionFormula f = showsPrec 1 f ""

-- Conversion from standard formula to reduction formula.
-- Strategy information is not computed.

convertToReductionFormula :: StandardFormula -> ReductionFormula
convertToReductionFormula (MkSF f) = MkRF (mapNodeInfo nodeMap f)
  where nodeMap op ni = case op of
         And       -> Just (MkSI [])
         Exists _  -> Just (MkSI [])
         ForAll _  -> Just (MkSI [])
         otherwise -> Nothing

-- functions on positions

lambda :: Position
lambda = Pos []

mkPos :: Position -> Int -> Position
mkPos (Pos p) i = Pos (p ++ [i])

concPos :: Position -> Position -> Position
concPos (Pos p) (Pos p') = Pos (p ++ p')

posToList :: Position -> [Int]
posToList (Pos p) = p

listToPos :: [Int] -> Position
listToPos p = (Pos p)


------------------------------
----- internal functions -----
------------------------------

stripSF :: StandardFormula -> StandardFormulaTree
stripSF (MkSF f) = f

wrapSF :: StandardFormulaTree -> StandardFormula
wrapSF f = MkSF f

stripRF :: ReductionFormula -> ReductionFormulaTree
stripRF (MkRF f) = f

wrapRF :: ReductionFormulaTree -> ReductionFormula
wrapRF f = MkRF f

getVars :: Predicate p => GenericFormula p ni -> Set Variable
getVars (Node (op,_) sfs) = case op of
  Pred p -> getPredVars p
  otherwise -> foldr (sunion . getVars) empty sfs

freeVars :: Predicate p => GenericFormula p ni -> Set Variable
freeVars (Node (op,_) sfs) = case op of
  Pred p -> getPredVars p
  Exists v -> freeVars (head sfs) `minus` singleton v
  ForAll v -> freeVars (head sfs) `minus` singleton v
  otherwise -> foldr (sunion . freeVars) empty sfs

boundVars :: Predicate p => GenericFormula p ni -> Set Variable
boundVars (Node (op,_) sfs) = case op of
  Exists v -> foldr (sunion . boundVars) (singleton v) sfs
  ForAll v -> foldr (sunion . boundVars) (singleton v) sfs
  otherwise -> foldr (sunion . boundVars) empty sfs

getConsts :: Predicate p => GenericFormula p ni -> Set Constant
getConsts (Node (op,_) sfs) = case op of
  Pred p -> getPredConsts p
  otherwise -> foldr (sunion . getConsts) empty sfs

getConstructor :: Predicate p => GenericFormula p ni -> ForConstructor p
getConstructor (Node (c,_) _) = c

--------------

mapNodeInfo :: Predicate p => (ForConstructor p -> Maybe ni -> Maybe ni')
               -> GenericFormula p ni -> GenericFormula p ni'
mapNodeInfo c (Node (op,ni) sfs) =
  Node (op, c op ni) (map (mapNodeInfo c) sfs)


mapForConst :: (Predicate p, Predicate p') => (ForConstructor p ->
               ForConstructor p') -> GenericFormula p ni -> GenericFormula p' ni
mapForConst c (Node (op,ni) sfs) = Node (c op, ni) (map (mapForConst c) sfs)

-------------

getPVars :: [Term] -> Set Variable
getPVars []     = empty
getPVars ((Var v):ts) = add v (getPVars ts)
getPVars ((Con _):ts) = getPVars ts

getPConsts :: [Term] -> Set Constant
getPConsts [] = empty
getPConsts ((Var _):ts) = getPConsts ts
getPConsts ((Con c):ts) = add c (getPConsts ts)

getVarName :: IDContext -> Variable -> String
getVarName []                  v = "v_" ++ show v
getVarName ((vn, IDVar v'):pc) v = if v == v' then vn else getVarName pc v
getVarName (_:pc)              v = getVarName pc v

getConName :: IDContext -> Constant -> String
getConName []                  c = "c_" ++ show c
getConName ((cn, IDCon c'):pc) c = if c == c' then cn else getConName pc c
getConName (_:pc)              c = getConName pc c

getPredName :: IDContext -> PredIdent -> String
getPredName []                   p = "P_" ++ show p
getPredName ((pn, IDPred p'):pc) p = if p == p' then pn else getPredName pc p
getPredName (_:pc)               p = getPredName pc p

showFormulaTree :: (Predicate p, Show p, Show ni) =>
                   Int -> GenericFormula p ni -> ShowS
showFormulaTree pr f@(Node (op,ni) sfs) =
  let sf:_   = sfs
      pc f'  = precedence sf < precedence f'
  in case op of
    T -> showString "true"
    F -> showString "false"
    Pred p -> shows p
    Not -> showChar '~' . showParen (pc f) (showFormulaTree pr sf)
    Or -> showForTreeList pr sfs " \\/ " (precedence f)
    And -> showParen (pr > 0) (showForTreeList pr sfs " /\\ " (precedence f)) .
           showNI (pr >0) ni
    Exists v -> showString "(exists " .
                showString (getVarName standardIDContext v) .
                showNI (pr > 0) ni . showChar ')' .
                showParen (pc f) (showFormulaTree pr sf)
    ForAll v -> showString "(forall " .
                showString (getVarName standardIDContext v) .
                showNI (pr > 0) ni . showChar ')' .
                showParen (pc f) (showFormulaTree pr sf)

showForTreeList :: (Predicate p, Show p, Show ni) =>
                   Int -> [GenericFormula p ni] -> String -> Int -> ShowS
showForTreeList pr []     s p = error "in showForList"
showForTreeList pr (f:[]) s p = 
  showParen (precedence f < p) (showFormulaTree pr f)
showForTreeList pr (f:fs) s p =
  showParen (precedence f < p) (showFormulaTree pr f) . showString s .
  showForTreeList pr fs s p

showNI :: (Show ni) => Bool -> Maybe ni -> ShowS
showNI _ Nothing  = id
showNI o (Just ni) = if o then showChar '{' . shows ni . showChar '}' else id

precedence :: Predicate p => GenericFormula p ni -> Int
precedence (Node (op,_) _) = case op of
  T        -> 3
  F        -> 3
  Pred p   -> 3
  Exists v -> 2
  ForAll v -> 2
  Not      -> 2
  And      -> 1
  Or       -> 0

equalFT :: (Predicate p, Eq p) =>
           (GenericFormula p ni, GenericFormula p ni') -> Bool
equalFT ((Node (op,_) sfs), (Node (op',_) sfs')) =
  op == op' &&  all equalFT (zip sfs sfs')

---------

getSubForAtPos :: Predicate p => Position -> GenericFormula p ni
                  -> GenericFormula p ni
getSubForAtPos (Pos []) f = f
getSubForAtPos (Pos (i:is)) (Node _ sfs) = getSubForAtPos (Pos is) (sfs!!i)

setSubForAtPos :: Predicate p => Position -> GenericFormula p ni
  -> GenericFormula p ni -> GenericFormula p ni
setSubForAtPos (Pos []) sf f = sf
setSubForAtPos (Pos (i:is)) sf (Node (op,ni) sfs) = Node (op,ni)
    (fs ++ [setSubForAtPos (Pos is) sf (sfs!!i)] ++ gs)
  where fs = take i sfs
        gs = drop (i+1) sfs