module LitPairs where

unifiableLitPairs :: Formula -> [(Position, Position)]
unifiableLitPairs f =
  let litsAndPos = collectLiterals f
      pairs pl nl = map (\[x,y] -> (x,y)) (takeOneOfEach [pl,nl])
      okps (_,pl,nl) = filter (isUnifiable f) (pairs pl nl)
  in foldr ((++) . okps) [] litsAndPos

isUnifiable :: Formula -> (Position, Position) -> Bool
isUnifiable f pnp@(p,p') =
  let root = comPrefix p p'
      disjTest = isDisj (f@@root)
      lf = extractLitPairSubFor pnp f
      f' = closure (pnf lf)
      es = litEqSys f'
  in disjTest && solvable es

unify :: Formula -> (Position, Position) -> SResult
unify f pnp =
  let lf = extractLitPairSubFor pnp f
      f' = closure (pnf lf)
      es = litEqSys f'
  in solve es

--------------------------

type UMap = [(UTerm, Term)]

uMapLookupByTerm :: Term -> UMap -> Maybe UTerm
uMapLookupByTerm _ [] = Nothing
uMapLookupByTerm x ((ut,t):r) = if x == t then Just ut
    else uMapLookupByTerm x r

uMapLookupByUTerm :: UTerm -> UMap -> Maybe Term
uMapLookupByUTerm = lookup

unifyM :: Formula -> (Position, Position) -> (SResult, UMap)
unifyM f pnp =
  let lf = extractLitPairSubFor pnp f
      (f',vm) = pnfM lf
      f'' = closure f'
      es = litEqSys f''
      r = solve es
  in (r, buildUMap r vm)

buildUMap :: SResult -> VMap -> UMap
buildUMap s vm =
  let uts = getUTerms s
      -- lookup in VMap
      lkp :: Variable -> VMap -> Variable
      lkp x [] = x
      lkp x ((n,o):r) = if x == n then o else lkp x r
      -- convert UVar to Var
      uv2v x = lkp (mkVar (abs x - 1)) vm
      -- build UMap entry
      bld t@(UVar x) = (t, Var (uv2v x))
      bld t@(UCon c) = (t, Con (mkCon c))
      bld t@(UOp o []) = (t, Var (uv2v o))
  in map bld uts

             
getUTerms :: SResult -> [UTerm]
getUTerms Fail = []
getUTerms (Solution []) = []
getUTerms (Solution ((l,r):es)) =
  let ins :: Eq a => a -> [a] -> [a]
      ins x [] = [x]
      ins x l@(y:ys) = if x == y then l else y:ins x ys
      rt = case r of UVar _ -> r
                     UCon _ -> r
                     UOp o _ -> UOp o []
      g = getUTerms (Solution es)
  in ins l (ins rt g)


-------------------------------------
-- process quantifiers on root path
-- f is a formula, p and p' are unifiable literal positions.
-- Then in (f*, p*,p'*) = processRootQuant(f,p,p') the formulae
-- f and f* are equivalent, p* and p'* are unifiable literal positions
-- corresponding with p and p'.

-- first parameter: maxVar
processRootQuant :: Int -> UMap
   -> IOStateMonad (Formula, Generation, Position, Position) ()
processRootQuant mv um =
  do (f,g,p,p') <- fetch
     let root = comPrefix p p'
         s = unify f (p,p')
         rrp = rootReductionPosition f root s um
     printWith putStrLn ("rrp = " ++ show rrp)
     if rrp == Nothing then return () else do
       let Just pos = rrp
           x = getQuantVar (f@@pos)
           mt = Just (termFromVar x)
       f' <- doIO (applyRule g mv ExistentialRule pos (Just defaultDisjSel)
                             (Just defaultDisjSel) (Just defaultConjSel) mt f)
       let r1 = posToList pos
           rpLength = length r1
           pContainsX = x `isElement` vars (f@@p)
           (p1,p2) = if pContainsX then (p,p') else (p',p)
           p1' = drop rpLength (posToList p1)
           p2' = drop rpLength (posToList p2)
           np = listToPos (r1 ++ [0] ++ p1')
           np' = listToPos (r1 ++ [1] ++ tail p2')
       update (\(f,g,p,p') -> (f',g+1,np,np'))
     
-- compute, if a reduction on a root quantifier is necessary
-- input: f: complete formula
--        p: root position of literal pair, i.e. common pos-prefix
--        s: unification result of literal pair
-- output: the position, where reduction must take place, if any

rootReductionPosition :: Formula -> Position -> SResult -> UMap -> Maybe Position
rootReductionPosition f p r um =
  let ps = necyRQPos f p r um
  in if null ps then Nothing else Just (minimum ps)

-- output must be ordered, s.th. superposition occurs before subposition
necyRQPos :: Formula -> Position -> SResult -> UMap -> [Position]
necyRQPos _ _ (Solution []) _ = []
necyRQPos f root (Solution (eq:eqs)) um =
  let (UVar v,r) = eq
      Just (Var var) = uMapLookupByUTerm (UVar v) um
      crit = occurCheck (UVar (-v)) [r]
      qp = findEQonRootPath f var root
      poss = necyRQPos f root (Solution eqs) um
  in if crit then qp : poss else poss

-- compute bound variables of all E-Quantifiers on root path
-- input: f: formula
--        p: root position of lit.pair.
-- output: variable set bound by ex. quant. on root path

existQuantVarsOnRootPath :: Formula -> Position -> Set Variable
existQuantVarsOnRootPath f r =
  let rp = getPath f lambda r
      eq = filterPath isExists rp
      foldVar :: Path -> Set Variable
      foldVar p | pathIsEmpty p = empty
                | otherwise = let (h,r) = (pathHead p, pathRest p)
                              in add (getQuantVar h) (foldVar r)
  in foldVar eq

    
findEQonRootPath :: Formula -> Variable -> Position -> Position
findEQonRootPath _ _ (Pos []) = error "error in findEQonRootPath"
findEQonRootPath f v pos@(Pos (i:p))
  | isExists f = let x = getQuantVar f
                     Pos p' = findEQonRootPath (getSubFormulae f!!i) v (Pos p)
                 in if x == v then lambda else Pos (i:p')
  | otherwise = let Pos p' = findEQonRootPath (getSubFormulae f!!i) v (Pos p)
                in Pos (i:p')

      
      
  


-------------------------------------

-- original positions: associate with each (some) sub-formula position
-- the position in the original formula
type OPosL = [(Position, Position)]

buildFaI :: Position -> Formula -> (Formula -> Formula, OPosL, Position)
buildFaI (Pos []) _ = (id ,[], lambda)
buildFaI (Pos (i:p)) f =
  let sf = getSubFormulae f !! i
      (sf',pl', Pos ep') = buildFaI (Pos p) sf
  in if isQuant f then
      (\f' -> setSubFormula (sf' f') f, (lambda, lambda): map (posmap i) pl',
       Pos (0:ep'))
     else (\f' -> sf' f', map (posmap' i) pl', Pos ep')
 where posmap i (Pos op, Pos sp) = (Pos (i:op), Pos (0:sp))
       posmap' i (Pos op, Pos sp) = (Pos (i:op), Pos sp)

{-
remUQ :: (Formula, OPosL) -> (Formula, OPosL)
remUQ i (f,pl)
  | isExists f = if x `isElement` vs then (f',pl) else (sf',pl')
  | isForAll f = if x `isElement` vs then (f',pl) else (sf',pl')
  | otherwise = setSubFormulae (map (remUQ (i+1)) (getSubFormulae f)) f
  where sf = getSubFormula f
        sf' = remUQ (i+1) sf
        f' = setSubFormula sf' f
        x = getQuantVar f
        vs = free sf'
        pl' = map pmap pl
        pmap (Pos p, Pos p') = 
-}

extLPSF :: (Position, Position) -> Formula -> (Formula, OPosL)
extLPSF (p,p') f =
  let root = comPrefix p p'; rl = length (posToList root)
      pp = listToPos (drop rl (posToList p))
      p'p = listToPos (drop rl (posToList p'))
      (rf,rpl,ep) = buildFaI root f
      (pf,ppl,_) = buildFaI pp (f@@root)
      (p'f, p'pl,_) = buildFaI p'p (f@@root)
  in (rf (disj [pf (f@@p), p'f (f@@p')]),
      rpl ++ map (pmap root ep) ppl ++ map (pmap' root ep) p'pl)
 where pmap (Pos p) (Pos ep) (Pos op, Pos sp) = (Pos (p++op), Pos (ep++[0]++sp))
       pmap' (Pos p) (Pos ep) (Pos op, Pos sp) = (Pos (p++op), Pos (ep++[1]++sp))

-------------------------------------
-- extractLitPairSubFor (p,p') f = f'
-- p is a pos. Lit. Position, p' is a neg. Lit. Position in f,
-- connected by disjunction,
-- f' is the Subformula of f consisting of p, p' and the 
-- corresponding quantifier subtree.

extractLitPairSubFor :: (Position, Position) -> Formula -> Formula
extractLitPairSubFor (p, p') f =
  let root = comPrefix p p'
      pQuantPath = filterPath isQuantOrLit (getPath f root p)
      nQuantPath = filterPath isQuantOrLit (getPath f root p')
      rQuantPath = filterPath isQuant (getPath f lambda root)
      df = disj [connectPath pQuantPath, connectPath nQuantPath]
      rpl = pathLength rQuantPath
      dpos = listToPos (replicate rpl 0)
  -- removeUQ: remove unused quantifiers
  in removeUQ (setSubFormulaAtPos dpos df (connectPath rQuantPath))

removeUQ :: Formula -> Formula
removeUQ f
  | isExists f = if x `isElement` vs then f' else sf'
  | isForAll f = if x `isElement` vs then f' else sf'
  | otherwise = setSubFormulae (map removeUQ (getSubFormulae f)) f
  where sf = getSubFormula f
        sf' = removeUQ sf
        f' = setSubFormula sf' f
        x = getQuantVar f
        vs = free sf'
       

isQuantOrLit f = isExists f || isForAll f || isLiteral f
isQuant f = isExists f || isForAll f

-- convert a path to a formula

connectPath :: Path -> Formula
connectPath p
  | pathIsEmpty p = error "wrong argument in connectPath"
  | otherwise = let (h,r) = (pathHead p, pathRest p)
                    fr = connectPath r
                in if pathIsEmpty r then h else setSubFormula fr h

------------------------------

type VMap = [(Variable, Variable)]

pnfM :: Formula -> (Formula, VMap)
pnfM f | isDisj f = pnfxM pnsfvm disj
       | isConj f = pnfxM pnsfvm conj
       | otherwise = (setSubFormulae pnsf f, concat renmap)
       where sf = getSubFormulae f
             pnsfvm = map pnfM sf
             (pnsf, renmap) = unzip pnsfvm


pnfxM :: [(Formula,VMap)] -> ([Formula] -> Formula) -> (Formula, VMap)
pnfxM fms c =
  let (q, nq) = partition (isQuant . fst) fms
      (aq,eq) = partition (isForAll . fst) q
      (h,vm):r = aq ++ eq; x = getQuantVar h
      badVars = foldr (sunion . free . fst) empty fms
      y = head ([mkVar 0..] \\ setToOrdList badVars)
      y' = termFromVar y
      h' = getSubFormula h; h'' = subst x y' h'
      vm' = if lookup y vm == Nothing then (x,y):vm
            else let Just z = lookup y vm in (x,z):vm
      nfms = (h',vm):r++nq; nfms' = (h'',vm'):r++nq
      (pnfx1, vm1) = pnfxM nfms c
      (pnfx2, vm2) = pnfxM nfms' c
      vm2' = if lookup x vm == Nothing then (y,x):vm2
             else let Just z = lookup x vm in (y,z):vm2
  in if null q then (c (map fst nq), [])
     else if x `isElement` badVars then
       (setQuantVar y (setSubFormula pnfx2 h), vm2')
     else (setSubFormula pnfx1 h, vm1)

------------------------------

-- prenex normal form:
-- input formula must be in nnf
-- output: not flat, not simplified

pnf :: Formula -> Formula
pnf f | isDisj f = pnfx pnsf disj
      | isConj f = pnfx pnsf conj
      | otherwise = setSubFormulae pnsf f
      where sf = getSubFormulae f
            pnsf = map pnf sf

-- input fs in pnf

pnfx :: [Formula] -> ([Formula] -> Formula) -> Formula
pnfx fs c =
  if null q then c nq
  else if x `isElement` badVars then
    setQuantVar y (setSubFormula (pnfx nfs' c) h)
  else setSubFormula (pnfx nfs c) h
  where (q,nq) = partition isQuant fs
        (aq,eq) = partition isForAll q
        h:r = aq ++ eq; x = getQuantVar h
        badVars = foldr (sunion . free) empty fs
        y = head ([mkVar 0..] \\ setToOrdList badVars)
        y' = termFromVar y
        h' = getSubFormula h; h'' = subst x y' h'
        nfs = h':r++nq; nfs' = h'':r++nq

quantPrefixAndRest :: Formula -> (Path, Formula)
quantPrefixAndRest f
  | isExists f = (mkPath f p', f')
  | isForAll f = (mkPath f p', f')
  | otherwise  = (emptyPath, f)
  where sf = getSubFormula f
        (p', f') = quantPrefixAndRest sf

closure :: Formula -> Formula
closure f = foldr forall f (setToOrdList (free f))

-- litEqSys computes the 'Unification ES' of formula f
-- f is the output of the closure of extractLitPairSubFor in pnf

-- variables in neg. lit: negative indices, e.g. UVar -4,
-- variables in pos. lit: positive indices, e.g. UVar 4,

{-
litEqSys :: Formula -> [UEquation]
litEqSys f = zip utp utn
  where (qp,f') = quantPrefixAndRest f
        p1:p2:_ = getSubFormulae f'
        (pp,np) = if isNeg p1 then (p2,p1) else (p1,p2)
        tsp = getLitTerms (getLiteral pp)
        tsn = getLitTerms (getLiteral np)
        ucs = [(termFromCon c, UCon (getConIndex c)) |
                      c <- setToOrdList (consts f')]
        utsp = conv qp [] tsp ++ ucs
        utsn = conv qp [] tsn ++ ucs
        utp = map (assoc utsp) tsp
        utn = map (negUTerm . assoc utsn) tsn


type UEquation = (UTerm, UTerm)

data UTerm = UVar Int
           | UCon Int
           | UOp Int [UTerm]
  deriving (Eq, Show)

negUTerm :: UTerm -> UTerm
negUTerm (UVar v) = UVar (-v)
negUTerm (UCon c) = UCon c
negUTerm (UOp x uts) = UOp x (map negUTerm uts)

-- take a quantifier prefix and compute for all 
-- terms the corr. unification terms
-- conv p ct nt : converts terms ct, nt are neccessary terms
conv :: Path -> [UTerm] -> [Term] -> [(Term, UTerm)]
conv p uv nt | pathIsEmpty p = []
             | otherwise = let (h,r) = (pathHead p, pathRest p)
                               x = getQuantVar h; xt = termFromVar x
                               xi = getVarIndex x + 1; ux = UVar xi
                           in if isExists h then
                                if xt `elem` nt then (xt,ux) : conv r (ux:uv) nt
                                else (xt,ux) : conv r uv nt
                              else (xt, UOp xi uv): conv r uv nt
-}

litEqSys :: Formula -> [UEquation]
litEqSys f = zip utp utn
  where (qp,f') = quantPrefixAndRest f
        p1:p2:_ = getSubFormulae f'
        (pp,np) = if isNeg p1 then (p2,p1) else (p1,p2)
        tsp = getLitTerms (getLiteral pp)
        tsn = getLitTerms (getLiteral np)
        critv = nub (filter termIsVar tsp) `intersect` nub (filter termIsVar tsn)
        ucs = [(termFromCon c, UCon (getConIndex c)) |
                      c <- setToOrdList (consts f')]
        uts = conv qp critv [] ++ ucs
        utp = map (assoc uts) tsp
        utn = map (negUTerm . assoc uts) tsn

type UEquation = (UTerm, UTerm)

data UTerm = UVar Int
           | UCon Int
           | UOp Int [UTerm]
  deriving (Eq, Show)

negUTerm :: UTerm -> UTerm
negUTerm (UVar v) = UVar (-v)
negUTerm (UCon c) = UCon c
negUTerm (UOp x uts) = UOp (-x) (map negUTerm uts)

-- take a quantifier prefix and compute for all 
-- terms the corr. unification terms
conv :: Path -> [Term] -> [UTerm] -> [(Term, UTerm)]
conv p cv uv | pathIsEmpty p = []
             | otherwise = let (h,r) = (pathHead p, pathRest p)
                               x = getQuantVar h; xt = termFromVar x
                               xi = getVarIndex x + 1; ux = UVar xi
                           in if isExists h then
                                if xt `elem` cv then (xt,ux) : conv r cv (ux:uv)
                                else (xt,ux):conv r cv uv
                              else (xt, UOp xi uv): conv r cv uv

--------------------

-- unification, solve

data SResult = Fail
             | Solution [UEquation]  -- (Var, Term) pairs
  deriving (Eq, Show)

solvable :: [UEquation] -> Bool
solvable es = solve es /= Fail

addSEq :: UEquation -> SResult -> SResult
addSEq _ Fail = Fail
addSEq e (Solution es) = Solution (e:es)

-- solve unification-ES

solve :: [UEquation] -> SResult
solve [] = Solution []
solve ((e@(l,r)):es) = case (l,r) of
  (UVar x, UVar y) -> if x == y then solve es
                      else addSEq e (solve (map (enterESubs [e]) es))
  (UVar _, UCon _) -> addSEq e (solve (map (enterESubs [e]) es))
  (UVar _, UOp _ ts) -> if occurCheck l ts then Fail
                        else addSEq e (solve (map (enterESubs [e]) es))
  (UCon _, UVar _) -> addSEq (r,l) (solve (map (enterESubs [(r,l)]) es))
  (UCon c, UCon d) -> if c == d then solve es else Fail
  (UCon _, UOp _ _) -> Fail
  (UOp _ ts, UVar _) -> if occurCheck r ts then Fail
                        else addSEq (r,l) (solve (map (enterESubs [(r,l)]) es))
  (UOp _ _, UCon _) -> Fail
  (UOp i ts, UOp j us) -> if i == j then solve (es ++ zip ts us) else Fail
  otherwise -> error "unexpected case in solve"


-- occurcheck returns true, if x = f(...x...)
occurCheck :: UTerm -> [UTerm] -> Bool
occurCheck _ [] = False
occurCheck vv@(UVar v) ((UVar w):us) = if v == w then True
                                       else occurCheck vv us
occurCheck v ((UCon _):us) = occurCheck v us
occurCheck v ((UOp _ ts):us) = occurCheck v (ts ++ us)

enterESubs :: [UEquation] -> UEquation -> UEquation
enterESubs es (l,r) = (enterSubs es l, enterSubs es r)

enterSubs :: [UEquation] -> UTerm -> UTerm
enterSubs es t = foldr enterSub t es
  where enterSub (UVar v, r) (UVar w) = if v == w then r else UVar w
        enterSub e c@(UCon _) = c
        enterSub e (UOp i ts) = UOp i (map (enterSub e) ts)

        