----------------------------------------------------------------------
-- Module Parser.hs                                                 --
--                                                                  --
-- Routines for parsing formulae.                                   --
--                                                                  --
-- 22.10.96  initial revision                                       --
----------------------------------------------------------------------


----------------------------
----- module interface -----
----------------------------

module Parser where


-------------------
----- imports -----
-------------------

import Char
import Formula
import Auxiliary
import Scanner
import IOStateMonad
import GlobalState
import Unification

--------------
-- notation --
--------------

-- alternatives in parser rules are separated by `+++'.
infixr 8 +++

----------------------
----- data types -----
----------------------

-- A parser takes a list of tokens of type t and returns a list of
-- (v,r)-pairs, where v is the value of the parsed token stream,
-- and r is the rest of tokens not parsed.

data PFail t = PFail Int String [t]
  deriving Show
data ParseResult v t = Parse [(v,Int,[t])] (PFail t)
                     | None (PFail t)
  deriving Show

newtype Parser t v = Parser ([t] -> Int -> ParseResult v t)
  deriving Show


---------------------------
-- instance declarations --
---------------------------

instance Functor (Parser t) where
  -- map :: (v -> w) -> Parser t v -> Parser t w
  map f (Parser p) =
    Parser (\ts n -> let res = p ts n
                   in case res of
                     Parse l fa -> Parse [(f v,k, r) | (v,k,r) <- l] fa
                     None fa -> None fa)

instance Monad (Parser t) where
  -- return :: v -> Parser t v
  return v = Parser (\ts n -> Parse [(v,n,ts)] noFail)
  -- >>= :: Parser t v -> (v -> Parser t w) -> Parser t w
  (Parser p) >>= f = Parser (\ts n ->
    case p ts n of
      Parse l fa -> processAlts fa [] [fv r k | (v,k,r) <- l, let Parser fv = f v]
      None fa -> None fa)

processAlts :: PFail t -> [(v,Int,[t])] -> [ParseResult v t]
  -> ParseResult v t
processAlts f [] [] = None f
processAlts f rs [] = Parse rs f
processAlts f rs (None f':rws) = processAlts (bestFail f f') rs rws
processAlts f rs (Parse rs' f':rws) = processAlts (bestFail f f') (rs++rs') rws

noFail :: PFail t
noFail = PFail (-1) "" []

updateFail :: PFail t -> ParseResult v t -> ParseResult v t
updateFail f (Parse cas f') = let r = bestFail f f' in Parse cas r
updateFail f (None f') = None (bestFail f f')

bestFail :: PFail t -> PFail t -> PFail t
bestFail f@(PFail n _ _) f'@(PFail n' _ _) =
  if n > n' then f
  else if n' > n then f'
  else if n == -1 then noFail
  else f

instance MonadZero (Parser t) where
  -- zero :: Parser t v
  zero = Parser (\ts n -> None (PFail n "error in last token." ts))

instance MonadPlus (Parser t) where
  -- (++) :: Parser t v -> Parser t v -> Parser t v
  (Parser p) ++ (Parser q) = Parser (\ts n ->
    let rp = p ts n; rq = q ts n
    in case (rp,rq) of
      (Parse l fa, Parse l' fa') -> Parse (l ++ l') (bestFail fa fa')
      (Parse l fa, None fa') -> Parse l (bestFail fa fa')
      (None fa, _) -> updateFail fa rq)


------------------------
-- exported functions --
------------------------

-- none

------------------------------
-- formula parser functions --
------------------------------

-- read, but no symbol table support
instance Read Formula where
  readsPrec _ s = if null toks then [] 
                  else either (const []) (\((f,_,_):_) -> [(f,[])]) parse
    where (toks,_) = forScanner emptySymbolTable s
          parse = papply mainForParser toks

readFormula :: String -> IOState Formula
readFormula s = do
  let (toks,symTab) = forScanner emptySymbolTable s
      parse = papply mainForParser toks
  setSymbolTable symTab
  case parse of
    Left errMsg -> failS (iosError 5 errMsg)
    Right ((f,_,_):_) -> return f

-- read, but no symbol table support
-- input following term is discarded
instance Read Term where
  readsPrec _ s = readTerm' (dropWhile isSpace s)
    where readTerm' ('v':'_':r) = let n = read r
                                  in [(termFromVar (mkVar n), [])]
          readTerm' ('c':'_':r) = let n = read r
                                  in [(termFromCon (mkCon n), [])]
          readVar' _ = []

readTerm :: String -> IOState Term
readTerm s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mt = stLookupTermByName name symTab; Just t = mt
      -- default format
      rdf ('v':'_':is) = Just (termFromVar (mkVar (read is)))
      rdf ('c':'_':is) = Just (termFromCon (mkCon (read is)))
      rdf _ = Nothing
      mdt = rdf name; Just dt = mdt
  if mt == Nothing then
    if mdt == Nothing then failS (iosError 9 "no term parse")
     else return dt
   else return t
  
-- read variable, no symbol table support
-- input following variable is discarded
instance Read Variable where
  readsPrec _ s = readVar' (dropWhile isSpace s)
    where readVar' ('v':'_':r) = let n = read r in [(mkVar n, [])]
          readVar' _ = []

-- read variable
readVar :: String -> IOState Variable
readVar s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mv = stLookupVarByName name symTab; Just v = mv
      -- default format
      rdf ('v':'_':is) = Just (mkVar (read is))
      rdf _ = Nothing
      mdv = rdf name; Just dv = mdv
  if mv == Nothing then
    if mdv == Nothing then failS (iosError 7 "no variable parse")
     else return dv
   else return v

-- read constant, no symbol table support
-- input following constant is discarded
instance Read Constant where
  readsPrec _ s = readCon' (dropWhile isSpace s)
    where readCon' ('c':'_':r) = let n = read r in [(mkCon n, [])]
          readCon' _ = []

-- read constant
readCon :: String -> IOState Constant
readCon s = do
  symTab <- getSymbolTable
  let name = takeWhile (not . isSpace) (dropWhile isSpace s)
      mc = stLookupConByName name symTab; Just c = mc
      -- default format
      rdf ('c':'_':is) = Just (mkCon (read is))
      rdf _ = Nothing
      mdc = rdf name; Just dc = mdc
  if mc == Nothing then
    if mdc == Nothing then failS (iosError 8 "no constant parse")
     else return dc
   else return c

------------------------
-- internal functions --
------------------------

-- Parser routines: At first some common combinators

-- item parser: returns the next token as value

item :: Parser t t
item = Parser (\ts n -> case ts of
   [] -> None (PFail n "unexpected end of file." [])
   (t:r) -> Parse [(t,n+1,r)] noFail)

-- formula token parser with error messages

ftoken :: FToken -> Parser Token Token
ftoken ftok = Parser (\ts n -> 
  case ts of 
    [] -> None (PFail n "unexpected end of file" [])
    (t@(ft,(l,c)):r) -> if ft == ftok then Parse [(t,n+1,r)] noFail
                        else  None (PFail n ("unexpected `" ++ show ft ++
                                            "'.") ts))

ftokenp :: (FToken -> Bool) -> String -> Parser Token Token
ftokenp p expStr = Parser (\ts n -> 
  case ts of [] -> None (PFail 0 "unexpected end of file" [])
             (t@(ft,(l,c)):r) -> if p ft then Parse [(t,n+1,r)] noFail
                                 else  None (PFail n ("expected `" ++ expStr ++
                                                   "', found `" ++ show ft ++
                                                   "'.") ts))


isTokVar :: FToken -> Bool
isTokVar (TokVar _) = True
isTokVar _ = False

isTokCon :: FToken -> Bool
isTokCon (TokCon _) = True
isTokCon _ = False

isTokPred :: FToken -> Bool
isTokPred (TokPred _) = True
isTokPred _ = False

-- first returns the first alternative of a parse result

first :: Parser t v -> Parser t v
first (Parser p) = Parser (\ts n -> case p ts n of
                                      None err -> None err
                                      Parse (x:xs) fa -> Parse [x] fa)


-- force parser application

force :: Parser t v -> Parser t v
force (Parser p)        =
  Parser (\ts n -> case p ts n of
    Parse xs fa -> Parse ((fst3 (head xs), snd3 (head xs), thd3 (head xs)) : tail xs) fa
    None fa -> None fa)

-- alternative parser, list (with separator) parsers

(+++) :: Parser t v -> Parser t v -> Parser t v
p +++ q = (p ++ q)

many :: Parser t v -> Parser t [v]
many p = force (many1 p +++ return [])

many1 :: Parser t v -> Parser t [v]
many1 p = do { x <- p; xs <- many p; return (x:xs) }

sepby :: Parser t v -> Parser t w -> Parser t [v]
p `sepby` sep = (p `sepby1` sep) +++ return []

sepby1 :: Parser t v -> Parser t w -> Parser t [v]
p `sepby1` sep = do {x <- p; xs <- many (do {sep; p}); return (x:xs)}


-- The following parsers represent a grammar for formulae with precedences
-- as usually defined.

-- formula parsers, default entry point mainForParser

mainForParser :: Parser Token Formula
mainForParser = do
  f <- forParser
  ftoken TokSemicolon
  return f

-- formula parser

forParser :: Parser Token Formula
forParser = do
  f <- impParser
  g' <- forParser'
  case g' of
    Nothing -> return f
    Just g  -> return (conj [disj [neg f,g], disj [neg g, f]])

forParser' :: Parser Token (Maybe Formula)
forParser' = do { ftoken TokEquiv; f <- forParser; return (Just f) }
  +++ return Nothing

-- implication parsers

impParser :: Parser Token Formula
impParser = do
  f <- disjParser
  g' <- impParser'
  case g' of
    Nothing -> return f
    Just g  -> return (disj [neg f, g])

impParser' :: Parser Token (Maybe Formula)
impParser' = do { ftoken TokImplies; f <- impParser; return (Just f) }
  +++ return Nothing

-- disjunction parsers

disjParser :: Parser Token Formula
disjParser = do
  f <- conjParser
  g' <- disjParser'
  case g' of
    Nothing -> return f
    Just g  -> return (disj [f, g])

disjParser' :: Parser Token (Maybe Formula)
disjParser' = do { ftoken TokOr; f <- disjParser; return (Just f) }
  +++ return Nothing

-- conjunction parsers

conjParser :: Parser Token Formula
conjParser = do
  f <- quantParser
  g' <- conjParser'
  case g' of
    Nothing -> return f
    Just g  -> return (conj [f, g])

conjParser' :: Parser Token (Maybe Formula)
conjParser' = do { ftoken TokAnd; f <-conjParser; return (Just f) }
  +++ return Nothing

-- quantified formula parsers

quantParser :: Parser Token Formula
quantParser =
  do { ftoken TokOpenPar; ftoken TokExists;
       vs <- varParser `sepby1` commaParser;
       ftoken TokClosePar; f <- quantParser; return (foldr exists f vs) }
  +++ do ftoken TokOpenPar; ftoken TokForAll
         vs <- varParser `sepby1` commaParser
         ftoken TokClosePar; f <- quantParser; return (foldr forall f vs)
  +++ negParser

-- negation parser

negParser :: Parser Token Formula
negParser =
  do { ftoken TokNot; f <- quantParser; return (neg f) }
  +++ primForParser

-- primitive formula parser
-- A primitive formula is either an atomic formula or any formula
-- within parenthesis.

primForParser :: Parser Token Formula
primForParser =
  do { ftoken TokOpenPar; f <- forParser; ftoken TokClosePar; return f }
  +++ atomParser

-- atomic formula parser

atomParser :: Parser Token Formula
atomParser = do { ftoken TokTrue; return verum }
  +++ do { ftoken TokFalse; return falsum }
  +++ do (TokPred p,_) <- ftokenp isTokPred "predicate"
         ftoken TokOpenPar
         ts <- termParser `sepby` commaParser
         ftoken TokClosePar
         return (literal (mkLit p ts))
  +++ do (TokPred p,_) <- ftokenp isTokPred "predicate"
         ts <- many termParser
         return (literal (mkLit p ts))
  +++ do t1 <- termParser
         (TokPred p,_) <- ftokenp isTokPred "predicate"
         t2 <- termParser
         return (literal (mkLit p [t1,t2]))
  +++ do t1 <- termParser
         ftoken TokNot
         (TokPred p,_) <- ftokenp isTokPred "predicate"
         t2 <- termParser
         return (neg (literal (mkLit p [t1,t2])))

-- auxiliary parsers

commaParser :: Parser Token Token
commaParser = do { t <- ftoken TokComma; return t }

termParser :: Parser Token Term
termParser = do { (TokVar v,_) <- ftokenp isTokVar "term"; return (Var v) }
  +++ do { (TokCon c,_) <- ftokenp isTokCon "term"; return (Con c) }

varParser :: Parser Token Variable
varParser = do (TokVar v,_) <- ftokenp isTokVar "variable"
               return v

-- apply a parser to a token stream

papply :: Parser Token v -> [Token] -> Either String [(v,Int,[Token])]
papply (Parser p) ts =
  let res = p ts 0
  in case res of
    None (PFail _ msg ts) ->
      Left ("Error at line " ++ show l ++ ", column " ++ show c ++ 
            ": " ++ msg)
         where ((_,(l,c)):_) = ts
    Parse l _ -> Right l


-----------------------
-- semantical checks --
-----------------------

-- raises exception, if literal rank check fails

checkLitRank :: Formula -> IOState Formula
checkLitRank f = do
  st <- getSymbolTable
  if isLiteral f then do
    let l = getLiteral f; p = getLitPred l; ts = getLitTerms l
        i = getPredIndex p; r' = length ts; r = stLookupLitRank st p
    if r == -1 then do
      setSymbolTable (stSetLitRank st p r')
      return (if isPosLit f then literal (mkLit (mkPred i r') ts)
              else neg (literal (mkLit (mkPred i r') ts)))
     else if r == r' then return f
     else failS (iosError 6 ("Semantic error: different ranks " ++ show r' ++ 
      " resp. " ++ show r ++ " assigned to predicate " ++ showPred' st p ++ "."))
   else do
    let fs = getSubFormulae f
        checkLitRank' fs = foldR (\f ff -> do f' <- checkLitRank f
                                              return (f':ff)) [] fs
    fs' <- checkLitRank' fs
    return (setSubFormulae fs' f)


---------------------
-- output routines --
---------------------

showFor :: Formula -> IOState ()
showFor f = do
  st <- getSymbolTable
  putStr' (showFormula st 0 f "")

draw :: Formula -> IOState ()
draw f = do
  st <- getSymbolTable
  putStr' (drawWO st f)

drawWO :: SymbolTable -> Formula -> String
drawWO st f = draw' st 0 f ""
  where draw' st i f = showString (replicate i ' ') . draw'' st i f
        draw'' st i f
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf
          | isDisj f = showString "\\/\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isConj f = showString "/\\\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isExists f = showString "exists " . showVar st qv . showChar '\n' .
                         draw' st (i+3) sf
          | isForAll f = showString "forall " . showVar st qv . showChar '\n' .
                         draw' st (i+3) sf
          where qv = getQuantVar f
                l  = getLiteral f
                sfs = getSubFormulae f
                sf:_ = sfs

drawWithNodeInfo :: Formula -> IOState ()
drawWithNodeInfo f = do
  st <- getSymbolTable
  putStr' (drawWNI st f)

drawWNI :: SymbolTable -> Formula -> String
drawWNI st f = draw' st 0 f ""
  where draw' st i f = showString (replicate i ' ') . draw'' st i f
        draw'' st i f
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showString sni . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showString sni . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf
          | isDisj f = showString "\\/\n" . foldr ((.) . draw' st (i+3)) id sfs
          | isConj f = showString "/\\" . showString sni . showChar '\n' .
              foldr ((.) . draw' st (i+3)) id sfs
          | isExists f = showString "exists " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf
          | isForAll f = showString "forall " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf
          where qv = getQuantVar f
                l  = getLiteral f
                ni = getNodeInfo f
                sni= if nodeInfoIsEmpty ni then ""
                     else "   <" ++ showNodeInfo st ni "" ++ ">"
                sfs = getSubFormulae f
                sf:_ = sfs

drawWithNodeInfoAndLitPositions :: Formula -> IOState ()
drawWithNodeInfoAndLitPositions f = do
  st <- getSymbolTable
  putStr' (drawWNILP st f)

drawWNILP :: SymbolTable -> Formula -> String
drawWNILP st f = draw' st 0 f lambda ""
  where draw' st i f p = showString (replicate i ' ') . draw'' st i f p
        draw'' st i f p
          | isVerum f = showString "true\n"
          | isFalsum f = showString "false\n"
          | isPosLit f = showLit st l . showString sp . showString sni . showChar '\n'
          | isNegLit f = showChar '~' . showLit st l . showString sp .
              showString sni . showChar '\n'
          | isNeg f = showChar '~' . draw'' st i sf (mkPos p 0)
          | isDisj f = showString "\\/\n" .
              foldr (\(f,j) s -> draw' st (i+3) f (mkPos p j) . s) id sfs'
          | isConj f = showString "/\\" . showString sni . showChar '\n' .
              foldr (\(f,j) s -> draw' st (i+3) f (mkPos p j) . s) id sfs'
          | isExists f = showString "exists " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf (mkPos p 0)
          | isForAll f = showString "forall " . showVar st qv . showString sni .
              showChar '\n' . draw' st (i+3) sf (mkPos p 0)
          where qv = getQuantVar f
                l  = getLiteral f
                ni = getNodeInfo f
                sni= if nodeInfoIsEmpty ni then ""
                     else "   <" ++ showNodeInfo st ni "" ++ ">"
                sp = "  " ++ show p
                sfs = getSubFormulae f
                sfs' = zip sfs ([0..length sfs-1])
                sf:_ = sfs

showLit :: SymbolTable -> Literal -> ShowS
showLit st l =
  let p = getLitPred l; ts = getLitTerms l
      showTerms = showTL id '('
      showTL e _ [] = e
      showTL e c (t:ts) = showChar c . showTerm st t . 
                          showTL (showChar ')') ',' ts
  in showPred st p . showTerms ts

showPred :: SymbolTable -> Predicate -> ShowS
showPred st p =
  let mps = stLookupByPred p st; Just ps = mps
      pi = getPredIndex p
  in if mps == Nothing then showString "P_" . shows pi
     else showString ps

showPred' :: SymbolTable -> Predicate -> String
showPred' st p = showPred st p ""

showCon :: SymbolTable -> Constant -> ShowS
showCon st c =
  let mcs = stLookupByCon c st; Just cs = mcs
      ci = getConIndex c
  in if mcs == Nothing then showString "c_" . shows ci
     else showString cs

showVar :: SymbolTable -> Variable -> ShowS
showVar st v =
  let mvs = stLookupByVar v st; Just vs = mvs
      vi = getVarIndex v
  in if mvs == Nothing then showString "v_" . shows vi
     else showString vs

showTerm :: SymbolTable -> Term -> ShowS
showTerm st t
  | termIsVar t = let v = getTermVar t in showVar st v
  | termIsCon t = let c = getTermCon t in showCon st c

showFormula :: SymbolTable -> Int -> Formula -> ShowS
showFormula st p f
  | isVerum f = showString "true"
  | isFalsum f = showString "false"
  | isPosLit f = showLit st (getLiteral f)
  | isNeg f = showChar '~' . showFormula st 2 sf
  | isDisj f = showParen (p > 0) (showForList st 0 sfs " \\/ ")
  | isConj f = showParen (p > 1) (showForList st 1 sfs " /\\ ")
  | isExists f = showString "(exists " . showVar st x . showChar ')' .
                 showFormula st 2 sf
  | isForAll f = showString "(forall " . showVar st x . showChar ')' .
                 showFormula st 2 sf
 where sfs@(sf:_) = getSubFormulae f
       x = getQuantVar f

showForList :: SymbolTable -> Int -> [Formula] -> String -> ShowS
showForList _ _ [] _ = error "empty formula list in showForList"
showForList st p (f:[]) _ = showFormula st p f
showForList st p (f:fs) s = showFormula st p f . showString s .
                            showForList st p fs s

showNodeInfo :: SymbolTable -> NodeInfo -> ShowS
showNodeInfo st ni
  | nodeInfoIsEmpty ni = id
  | otherwise = let gts = getStrategyInfo ni in showNIList "" gts
  where showNIList _ [] = id
        showNIList s ((g,mt):gts) = showString s . showChar '(' . shows g .
         showChar ',' . case mt of
          { Nothing -> showString "||"; Just t -> showTerm st t } .
          showChar ')' . showNIList "," gts

-- with position output:

draw2 f = putStr (draw2F f)

draw2F :: Formula -> String
draw2F f = draw2' 0 f lambda ""
  where draw2' :: Int -> Formula -> Position -> ShowS
        draw2' i f p = showString (replicate i ' ') . draw2'' i f p
        draw2'' :: Int -> Formula -> Position -> ShowS
        draw2'' i f p
          | isVerum f = showString "true" . shows p . showChar '\n'
          | isFalsum f = showString "false" . shows p . showChar '\n'
          | isPosLit f = shows l . shows p . showString sni . showChar '\n'
          | isNeg f = showChar '~' . shows l . shows p . showString sni . showChar '\n'
          | isDisj f = showString "\\/\n" .
              foldr (\ (f,j) s -> (draw2' (i+3) f (mkPos p j)) . s) id sfs' 
          | isConj f = showString "/\\" . showString sni . showChar '\n' .
              foldr (\ (f,j) s -> (draw2' (i+3) f (mkPos p j)) . s) id sfs'
          | isExists f = showString "exists " . shows qv . showString sni .
              showChar '\n' . draw2' (i+3) sf (mkPos p 0)
          | isForAll f = showString "forall " . shows qv . showString sni .
              showChar '\n' . draw2' (i+3) sf (mkPos p 0)
          where qv = getQuantVar f
                l  = getLiteral f
                ni = getNodeInfo f
                sni= if nodeInfoIsEmpty ni then "" else "   <" ++ show ni ++ ">"
                sfs = getSubFormulae f
                sfs' = zip sfs ([0..length sfs-1])
                sf:_ = sfs


-----------------------------
-- instances of class Show --
-----------------------------

instance Show Variable where
  showsPrec _ v = showVar emptySymbolTable v

instance Show Constant where
  showsPrec _ c = showCon emptySymbolTable c

instance Show Term where
  showsPrec _ t = showTerm emptySymbolTable t

instance Show Predicate where
  showsPrec _ p = showPred emptySymbolTable p

instance Show Literal where
  showsPrec _ l = showLit emptySymbolTable l

instance Show Formula where
  showsPrec p f = showFormula emptySymbolTable p f

instance Show NodeInfo where
  showsPrec _ ni
   | nodeInfoIsEmpty ni = id
   | otherwise = let gts = getStrategyInfo ni in showNIList "" gts
   where showNIList _ [] = id
         showNIList s ((g,mt):gts) = showString s . showChar '(' . shows g .
          showChar ',' . case mt of
           { Nothing -> showString "||"; Just t -> shows t } .
           showChar ')' . showNIList "," gts

instance Show Path where
  showsPrec _ pa
    | pathIsEmpty pa = showString "<< >>"
    | otherwise = let fs = pathToList pa
                  in showString "<< " . showPath fs . showString " >>"
    where showPath [] = id
          showPath (f:[]) = showPathE f
          showPath (f:fs) = showPathE f . showString ", " . showPath fs
          showPathE f
            | isVerum f = showString "true"
            | isFalsum f = showString "false"
            | isPosLit f = shows (getLiteral f)
            | isNegLit f = showChar '~' . shows (getLiteral (getSubFormula f))
            | isConj f = showString "/\\"
            | isDisj f = showString "\\/"
            | isExists f = showString "(exists " . shows (getQuantVar f) .
                showChar ')'
            | isForAll f = showString "(forall " . shows (getQuantVar f) .
                showChar ')'


---------------------------
-- robust read functions --
---------------------------

rReadLn :: Read a => IOState a
rReadLn = rReadWith (readWith . readIO)

rReadVar :: IOState Variable
rReadVar = rReadWith readVar

rReadCon :: IOState Constant
rReadCon = rReadWith readCon

rReadTerm :: IOState Term
rReadTerm = rReadWith readTerm

rReadWith :: (String -> IOState a) -> IOState a
rReadWith rf = catchS $$ do s <- readWith getLine
                            rf s
                      $$ \_ -> do putStrLn' "Error in input."
                                  rReadWith rf

rReadCond :: IOState a -> (a -> Bool) -> IOState a
rReadCond reader cond = do
  x <- reader
  if cond x then return x
   else do putStrLn' "Input not admitted."
           rReadCond reader cond

----------------------------
-- generic query function --
----------------------------

query :: String -> [Char] -> IOState Char
query msg cs = do
  printWith putStr msg
  doIO (readAnswer cs)
 where readAnswer :: [Char] -> IO Char
       readAnswer cs = do
         cl <- getLine
         let cl' = dropWhile (\c -> c `notElem` cs) cl
         if null cl' then do putStrLn "Input error."
                             readAnswer cs
                     else return (head cl')

-------------------------------
-- unification result output --
-------------------------------

showUResult :: IOState ()
showUResult = do
  (um,ur) <- getUnification
  showUR ur um

showUR :: UResult -> UMap -> IOState ()
showUR Fail _ = putStrLn' "<no solution>"
showUR (Solution es) um = do
  putStr' "{ "
  showEQList "" es um
  putStr' " }"

showEQList _ [] _ = done
showEQList s (e:es) um = do
  putStr' s
  showEQ e um
  showEQList ", " es um

showEQ (l,r) um = do
  showUTerm l um
  putStr' " -> "
  showUTerm r um

showUTerm ut um =
  let mti = uMapLookupByUTerm ut um; Just (t,pi) = mti
  in if mti == Nothing then putStr' "<undefd.>" else do
    st <- getSymbolTable
    putStr' (showTerm st t "")
    case pi of RootPath -> putStr' "#R"
               PosLitPath -> putStr' "#P"
               NegLitPath -> putStr' "#N"
               Unknown -> done
