----------------------------------------------------------------------
-- Module Parser.hs                                                 --
--                                                                  --
-- Routines for parsing formulae.                                   --
--                                                                  --
-- 22.10.96  initial revision                                       --
-- 05.03.97  completely rewritten, error handling                   --
----------------------------------------------------------------------

import Formula

----------------------
----- data types -----
----------------------

-- formula tokens

data IdType = Short | Long
  deriving (Eq, Show)

data FToken = TokSVar Variable  | TokLVar Variable
            | TokSCon Constant  | TokLCon Constant
            | TokLPred Predicate
            | TokSPred Predicate
            | TokTrue
            | TokFalse          | TokNot
            | TokOr             | TokAnd
            | TokImplies        | TokEquiv
            | TokExists         | TokForAll
            | TokOpenPar        | TokClosePar
            | TokComma          | TokPoint
            | TokEOF
  deriving (Show, Eq)

-- A parser takes a list of tokens of type t and returns a list of
-- (v,r)-pairs, where v is the value of the parsed token stream,
-- and r is the rest of tokens not parsed.
-- v may be a parse exception.

-- 0: EOF
type PErrCode = Int
data ParseError = ParseError ScanPos PErrCode String
data ParseResult v ts = Parse [(v, ts)]
                      | Fail ParseError
newtype Parser t v = Parser ((ScanPos, [t]) -> (ScanPos, ParseResult v [t]))

type ScanPos = (Int,Int)

data ScanResult a = One a
                  | Many [a]
                  | ScanError String ScanPos
  deriving (Eq, Show)

---------------------------
-- instance declarations --
---------------------------

instance Functor (Parser t) where
  -- map :: (v -> w) -> Parser t v -> Parser t w
  map f (Parser p) = Parser $ 
    \(o,ts) -> let (o',res) = p (o,ts)
               in case res of
                    Parse ps -> (o', Parse [(f v, r) | (v,r) <- ps])
                    Fail e   -> (o', Fail e)




-----
isSpace' c =  c == ' ' || c == '\t' || c == '\r' || c == '\f' || c == '\v'
isSeparator c = isSpace c || c == '(' || c == ')' || c == ',' || c == '.'
                          || c == '/' || c == '\\' || c == '-' || c == '<'
                          || c == '>' || c == '=' || c == '~'
isSpecial c = c == '<' || c == '>' || c == '&' || c == '!' || c == '$'
           || c == '%' || c == '/' || c == '\\' || c == '|' || c == '='
           || c == '?' || c == '-' || c == '+' || c == '*' || c == '#'

isVarStart c = c >= 'm' && c <= 'z'
isVarChar c = isDigit c || isLower c
isConStart c = isDigit c || (c >= 'a' && c <= 'l')
isConChar c = isDigit c || isLower c
isPredStart c = isUpper c || isSpecial c

scan :: IdContext -> ScanPos -> String
  -> ScanResult (FToken,String,ScanPos,IdContext)
-- end of file, new lines, white spaces, comments
scan ct p "" = One (TokEOF, "", p, ct)
scan ct (l,c) ('\n':r) = scan ct (l+1,0) r
scan ct (l,c) (h:r) | isSpace h = let (space,r') = span isSpace' r
                               in scan ct (l,c+length space+1) r'
scan ct (l,c) ('-':'-':r) = scan ct (l+1,0) (dropWhile (/= '\n') r)
-- auxiliary symbols
scan ct (l,c) ('(':r) = One (TokOpenPar, r, (l,c+1), ct)
scan ct (l,c) (')':r) = One (TokClosePar, r, (l,c+1), ct)
scan ct (l,c) ('~':r) = One (TokNot, r, (l,c+1), ct)
scan ct (l,c) (',':r) = One (TokComma, r, (l,c+1), ct)
scan ct (l,c) ('.':r) = One (TokPoint, r, (l,c+1), ct)
-- formula tokens
scan ct (l,c) ('t':'r':'u':'e':r) = One (TokTrue, r, (l,c+4), ct)
scan ct (l,c) ('f':'a':'l':'s':'e':r) = One (TokFalse, r, (l,c+5), ct)
scan ct (l,c) ('e':'x':'i':'s':'t':'s':r) = One (TokExists, r, (l,c+6), ct)
scan ct (l,c) ('f':'o':'r':'a':'l':'l':r) = One (TokForAll, r, (l,c+6), ct)
scan ct (l,c) ('/':'\\':r) = One (TokAnd, r, (l,c+2), ct)
scan ct (l,c) ('\\':'/':r) = One (TokOr, r, (l,c+2), ct)
scan ct (l,c) ('-':'>':r) = One (TokImplies, r, (l, c+2), ct)
scan ct (l,c) ('<':'=':'>':r) = One (TokEquiv, r, (l,c+3), ct)
-- identifiers:
scan ct (l,c) s@(h:r) | isVarStart h = scanVar ct (l,c) s
                      | isConStart h = scanCon ct (l,c) s
                      | isPredStart h = scanPred ct (l,c) s
-- error
scan _ (l,c) s =
  let id = takeWhile (not . isSeparator) s; idl = length id
      ids = if idl > 10 then take 10 id ++ "..." else id
  in ScanError ("Unknown identifier at line " ++ show l ++
                 ", column " ++ show c ++ ": " ++ ids) (l,c)


scanVar :: IdContext -> ScanPos -> String
  -> ScanResult (FToken,String,ScanPos,IdContext)
scanVar ct (l,c) s@(h:r) =
  -- short variable name
  let mide = getEntryByName ct [h]; Just ide = mide
  in if mide == Nothing then
       -- insert variable into IdContext
       
       One (TokSVar 
 One (TokSVar (mkVar 0), r, (l,c+1), ct)

scanCon :: IdContext -> ScanPos -> String
  -> ScanResult (FToken,String,ScanPos,IdContext)
scanCon ct (l,c) s@(h:r) = One (TokSCon (mkCon 0), r, (l,c+1), ct)

scanPred :: IdContext -> ScanPos -> String
  -> ScanResult (FToken,String,ScanPos,IdContext)
scanPred ct (l,c) s@(h:r) = One (TokSPred (mkPred 0 0), r, (l,c+1), ct)



