----------------------------------------------------------------------
-- Module Strategy.hs                                               --
--                                                                  --
-- Implements different reduction strategies                        --
--                                                                  --
-- 11.12.96  initial revision                                       --
----------------------------------------------------------------------

module Strategy where

import Set
import Formula
import Auxiliary
import Reduction
import Simplify
import List (nub)
import Parser
import IOStateMonad

-- A Reduction Strategy consists of the following components:
--   1) the maximal number of variables to be used
--   2) a reducer, which keeps track of the generation
--   3) two disjunction selectors (for universal- and disjunction rule)
--   4) a conjunction selector (for disjunction rule)

type Reducer = Int -> ForListPartitioner -> ForListPartitioner
  -> ForListPartitioner -> ForGenIOSt ()

type ReductionStrategy = ForGenIOSt ()

defaultRS :: Int -> ReductionStrategy
defaultRS n = defaultReducer n defaultDisjSel defaultDisjSel defaultConjSel

interactiveRS :: Int -> ReductionStrategy
interactiveRS n = interactiveReducer n defaultDisjSel defaultDisjSel
                    defaultConjSel

defaultDisjSel :: ForListPartitioner
defaultDisjSel fs = return (fs,[])

maxDisjSel :: ForListPartitioner
maxDisjSel fs = return (fs,[])

defaultConjSel :: ForListPartitioner
defaultConjSel fs = return ([head fs], tail fs)


-- f must be reducible!
defaultReducer :: Reducer
defaultReducer mv dsA dsD csD =
  do nextGeneration
     (f,g) <- getForGen
     let rs = findPReductions f
         (p,rt,[(_,mt)]) = minSelect rs
     printWith putStrLn ("Reducing at position " ++ show p)
     f' <- doIO (applyRule g mv rt p (Just dsA) (Just dsD) (Just csD) mt f)
     setFormula (simplify f')
    
minSelect :: [ReductionInfo] -> ReductionInfo
minSelect [] = error "wrong argument in minSelect"
minSelect [(p,rt,si)] = (p, rt, [minSelSI si])
minSelect ((p,rt,si):(p',rt',si'):ris) =
  let msi = minSelSI si; msi' = minSelSI si'
  in if fst msi < fst msi' then minSelect ((p,rt,[msi]):ris)
     else minSelect ((p',rt',[msi']):ris)

minSelSI :: StrategyInfo -> (Int, Maybe Term)
minSelSI [] = error "wrong arg in minSelSI"
minSelSI sis = foldl1 mini sis
  where mini (a,b) (c,d) = if a <=c then (a,b) else (c,d)

------------------------------------------
----- interactive reduction strategy -----
------------------------------------------

iSelectTerm :: RuleType -> StrategyInfo -> IO (Maybe Term)
iSelectTerm _ [] = error "wrong args in iSelectTerm"
iSelectTerm _ [(_,mt)] = return mt
iSelectTerm rt s@((_,mt):_) = case rt of
  DistributionRule -> return mt
  ExistentialRule -> do putStr "Enter a term: (one of "
                        putStrLn (show (map (\(g, Just t) -> t) s) ++ ")")
                        t <- readLn
                        return (Just t)
  UniversalRule -> do putStr "Enter a variable: (one of "
                      putStrLn (show (map (\(g, Just t) -> t) s) ++ ")")
                      v <- readLn
                      return (Just (termFromVar v))
                         
printReds :: [ReductionInfo] -> IO ()
printReds = printReds' 0

printReds' :: Int -> [ReductionInfo] -> IO ()
printReds' _ [] = return ()
printReds' n (r:rs) = do
  putStr (show n ++ ") ")
  let (p,rt,_) = r
  putStrLn (show rt ++ " at position " ++ show p)
  printReds' (n+1) rs

interactiveReducer :: Reducer
interactiveReducer mv dsA dsD csD =
  do nextGeneration
     (f,g) <- getForGen
     let rs = findPReductions f
     printWith printReds rs
     printWith putStr "Select a reduction: "
     i <- readWith readLn
     let (p, rt, si) = rs!!i
     mt <- doIO (iSelectTerm rt si)
     f' <- doIO (applyRule g mv rt p (Just dsA) (Just dsD) (Just csD) mt f)
     setFormula (simplify f')


------------------------------------
----- preform a reduction step -----
------------------------------------

applyRS :: ReductionStrategy -> ForGenIOSt ()
applyRS rs = rs

multiReduce :: ReductionStrategy -> ForGenIOSt ()
multiReduce rs =
  do (f,g) <- getForGen
     printWith putStrLn ("formula before reduction step #" ++ show (g+1))
     printWith putStrLn ("depth: " ++ show (depth f) ++ ", size: " ++ show (size f))
--     printWith draw f
     f' <- applyRS rs
     multiReduce rs
