module UnificationMapping where

-- A variable map is a finite mapping form variables to variables
-- represented as an association list.

type VMap = [(Variable, Variable)]

-- extended variable map 

data Origin = RootPath | PosLitPath | NegLitPath | Unknown
  deriving (Eq, Show)

type XVMap = [(Variable, Variable, Origin)]

-- unification equation

type UEquation = (UTerm, UTerm)

-- unification terms
-- representation of constants: c_i -> UCon i
-- representation of variables:
--  either: v_i -> UVar (i+1) or UVar (-i-1)
--  or:     v_i -> UOp i
-- Negative indices in UVar are used in the negative literal, 
-- positive indices for positive literal variables (bound by an existential
-- quantifier).
-- UOp i represents variable v_i bound by a universal quantifier.

data UTerm = UVar Int
           | UCon Int
           | UOp Int [UTerm]
  deriving (Eq, Ord, Show)

-- unifcation solution data type

data UResult = Fail
             | Solution [UEquation]  -- (UVar, UTerm) pairs
  deriving (Eq, Show)

type MUResult = (UResult, UMap, VMap)

type XUResult = (UResult, XUMap, XVMap)


-- mapping from unification terms to terms (in original formula)

type UMap = [(UTerm, Term)]

-- extended unifcation map

type XUMap = [(UTerm, Term, Origin)]


----------------------
-- lookup functions --
----------------------

-- find term and origin by UTerm in an XUMap

xuMapLookup :: UTerm -> XUMap -> Maybe (Term, Origin)
xuMapLookup _ [] = Nothing
xuMapLookup uo@(UOp op _) ((UOp op' _,t,o):r) =
  if op == op' then Just (t,o) else xuMapLookup uo r
xuMapLookup uo@(UOp op _) (_:r) = xuMapLookup uo r
xuMapLookup lt ((ut,t,o):r) =
  if lt == ut then Just (t,o) else xuMapLookup lt r

-- find UTerm by variable and origin in an XUMap

xuMapLookupByVar :: Variable -> Origin -> XUMap -> Maybe UTerm
xuMapLookupByVar _ _ [] = Nothing
xuMapLookupByVar lv lo (h@(ut, t, o):r) =
  let vt = termFromVar lv
  in if vt == t && lo == o then Just ut
     else xuMapLookupByVar lv lo r

-- find original variable by variable and origin in an XVMap

xvMapLookup :: Variable -> Origin -> XVMap -> Maybe Variable
xvMapLookup _ _ [] = Nothing
xvMapLookup lv lo ((nv,ov,o):r) = if lv == nv && lo == o then Just ov
  else xvMapLookup lv lo r

-- find new variable by original variable and origin

xvMapLookupByOrigVar :: Variable -> Origin -> XVMap -> Maybe Variable
xvMapLookupByOrigVar _ _ [] = Nothing
xvMapLookupByOrigVar lv lo ((nv,ov,o):r) = if lv == ov && lo == o then Just nv
  else xvMapLookupByOrigVar lv lo r


-- find term and origin by UTerm in a combination of XUMap and XVMap

-- Suppose the function `unificationEqSystem . uniqueBoundVars' was
-- applied to a formula. The function `originalTerm' computes the
-- term of the original formula, given a UTerm.
-- The mappings used therefore are delivered by the aforementioned functions.

originalTerm :: UTerm -> XUMap -> XVMap -> Term
originalTerm uc@(UCon _) xumap _ =
  let origCon = xuMapLookup uc xumap; Just (oc,_) = origCon
  in if origCon == Nothing then
       error "reconstruction of constant failed in originalTerm"
     else oc
originalTerm uo@(UOp _ _) xumap xvmap =
  let renVar = xuMapLookup uo xumap; Just (rv,o) = renVar; v = getTermVar rv
      ovl = xvMapLookup v o xvmap; Just ov = ovl; ovt = termFromVar ov
  in if renVar == Nothing || ovl == Nothing then
       error "reconstruction of op-var failed in originalTerm"
     else ovt
originalTerm uv@(UVar _) xumap xvmap =
  let renVar = xuMapLookup uv xumap; Just (rv,o) = renVar; v = getTermVar rv
      ovl = xvMapLookup v o xvmap; Just ov = ovl; ovt = termFromVar ov
  in if renVar == Nothing || ovl == Nothing then
       error "reconstruction of variable failed in originalTerm"
     else ovt

-- find UTerm by variable and origin in a combination of XUMap and XVMap

findUTerm :: Variable -> Origin -> XUMap -> XVMap -> UTerm
findUTerm v o xumap xvmap =
  let nVar = xvMapLookupByOrigVar v o xvmap; Just nv = nVar
      uTerm = xuMapLookupByVar nv o xumap; Just ut = uTerm
  in if nVar == Nothing || uTerm == Nothing then
       error "UTerm lookup for variable failed in findUTerm"
     else ut


-- find subst. in UResult
findSubst :: UTerm -> UResult -> UTerm
findSubst _ Fail = error "in findSubst"
findSubst t (Solution []) = error "in findSubst"
findSubst t (Solution ((l,r):es)) =
  if t == l then r else findSubst t (Solution es)

-- find unification reduction term

findRedTerm :: Variable -> Origin -> XUResult -> Term
findRedTerm v o (eqs, xum, xvm) =
  let ut = findUTerm v o xum xvm
      ut' = findSubst ut eqs
  in originalTerm ut' xum xvm
